@@ -1,6 +1,6 @@
 #![no_main]
 
-use cranelift_assembler_x64::{fuzz, Inst};
+use cranelift_assembler_x64::{Inst, fuzz};
 use libfuzzer_sys::fuzz_target;
 
 fuzz_target!(|inst: Inst<fuzz::FuzzRegs>| {
