@@ -17,25 +17,27 @@
 //! See the main module comment in `mod.rs` for more details on the VCode-based
 //! backend pipeline.
 
+use crate::CodegenError;
 use crate::ir::pcc::*;
-use crate::ir::{self, types, Constant, ConstantData, ValueLabel};
-use crate::machinst::*;
+use crate::ir::{self, Constant, ConstantData, ValueLabel, types};
 use crate::ranges::Ranges;
 use crate::timing;
 use crate::trace;
-use crate::CodegenError;
 use crate::{LabelValueLoc, ValueLocRange};
+use crate::{machinst::*, trace_log_enabled};
 use regalloc2::{
-    Edit, Function as RegallocFunction, InstOrEdit, InstRange, MachineEnv, Operand,
-    OperandConstraint, OperandKind, PRegSet, RegClass,
+    Edit, Function as RegallocFunction, InstOrEdit, InstPosition, InstRange, Operand,
+    OperandConstraint, OperandKind, PRegSet, ProgPoint, RegClass,
 };
 use rustc_hash::FxHashMap;
 
+use core::cmp::Ordering;
+use core::fmt::{self, Write};
 use core::mem::take;
-use cranelift_entity::{entity_impl, Keys};
-use std::collections::hash_map::Entry;
+use core::ops::Range;
+use cranelift_entity::{Keys, entity_impl};
 use std::collections::HashMap;
-use std::fmt;
+use std::collections::hash_map::Entry;
 
 /// Index referring to an instruction in VCode.
 pub type InsnIndex = regalloc2::Inst;
@@ -102,6 +104,15 @@ pub struct VCode<I: VCodeInst> {
     /// user stack map.
     user_stack_maps: FxHashMap<BackwardsInsnIndex, ir::UserStackMap>,
 
+    /// A map from backwards instruction index to the debug tags for
+    /// that instruction. Each entry indexes a range in the
+    /// `debug_tag_pool`.
+    debug_tags: FxHashMap<BackwardsInsnIndex, Range<u32>>,
+
+    /// Pooled storage for sequences of debug tags; indexed by entries
+    /// in `debug_tags`.
+    debug_tag_pool: Vec<ir::DebugTag>,
+
     /// Operands: pre-regalloc references to virtual registers with
     /// constraints, in one flattened array. This allows the regalloc
     /// to efficiently access all operands without requiring expensive
@@ -195,6 +206,8 @@ pub struct VCode<I: VCodeInst> {
 
     /// Facts on VRegs, for proof-carrying code verification.
     facts: Vec<Option<Fact>>,
+
+    log2_min_function_alignment: u8,
 }
 
 /// The result of `VCode::emit`. Contains all information computed
@@ -205,33 +218,21 @@ pub struct EmitResult {
     pub buffer: MachBufferFinalized<Stencil>,
 
     /// Offset of each basic block, recorded during emission. Computed
-    /// only if `debug_value_labels` is non-empty.
+    /// only if `machine_code_cfg_info` is enabled.
     pub bb_offsets: Vec<CodeOffset>,
 
     /// Final basic-block edges, in terms of code offsets of
-    /// bb-starts. Computed only if `debug_value_labels` is non-empty.
+    /// bb-starts. Computed only if `machine_code_cfg_info` is enabled.
     pub bb_edges: Vec<(CodeOffset, CodeOffset)>,
 
-    /// Final length of function body.
-    pub func_body_len: CodeOffset,
-
     /// The pretty-printed disassembly, if any. This uses the same
     /// pretty-printing for MachInsts as the pre-regalloc VCode Debug
     /// implementation, but additionally includes the prologue and
     /// epilogue(s), and makes use of the regalloc results.
     pub disasm: Option<String>,
 
-    /// Offsets of sized stackslots.
-    pub sized_stackslot_offsets: PrimaryMap<StackSlot, u32>,
-
-    /// Offsets of dynamic stackslots.
-    pub dynamic_stackslot_offsets: PrimaryMap<DynamicStackSlot, u32>,
-
     /// Value-labels information (debug metadata).
     pub value_labels_ranges: ValueLabelsRanges,
-
-    /// Stack frame size.
-    pub frame_size: u32,
 }
 
 /// A builder for a VCode function body.
@@ -281,8 +282,16 @@ impl<I: VCodeInst> VCodeBuilder<I> {
         block_order: BlockLoweringOrder,
         constants: VCodeConstants,
         direction: VCodeBuildDirection,
+        log2_min_function_alignment: u8,
     ) -> Self {
-        let vcode = VCode::new(sigs, abi, emit_info, block_order, constants);
+        let vcode = VCode::new(
+            sigs,
+            abi,
+            emit_info,
+            block_order,
+            constants,
+            log2_min_function_alignment,
+        );
 
         VCodeBuilder {
             vcode,
@@ -370,18 +379,37 @@ impl<I: VCodeInst> VCodeBuilder<I> {
 
     /// Add a debug value label to a register.
     pub fn add_value_label(&mut self, reg: Reg, label: ValueLabel) {
-        // We'll fix up labels in reverse(). Because we're generating
-        // code bottom-to-top, the liverange of the label goes *from*
-        // the last index at which was defined (or 0, which is the end
-        // of the eventual function) *to* just this instruction, and
-        // no further.
-        let inst = InsnIndex::new(self.vcode.insts.len());
+        // 1) In the reversed order, we consider the instructions
+        //    that define ranges in the "debug_info" array to refer
+        //    to the IP **after** them (when reversed):
+        //      IP[2]__| Inst 3 |
+        //      IP[1]__| Inst 2 |
+        //      IP[0]__| Inst 1 |
+        //             | Inst 0 |
+        //    This is so that we can represent IP[<function start>],
+        //    done at the cost of not being to represent IP[<function end>],
+        //    which is OK since no values will be live at that point.
+        // 2) The live range for "reg" begins at the current IP
+        //    and continues until the next, in execution order,
+        //    VReg that defines "label". Since the ranges are open
+        //    at the end, the subtraction of 1 cancels out:
+        //      [last..current IP] <=>
+        //      [last..last emitted inst index] <=>
+        //      [last..next_inst_index - 1] <=>
+        //      [last..next_inst_index)
+        //
+        let next_inst_index = self.vcode.insts.len();
+        if next_inst_index == 0 {
+            // This would produce a defective [0..0) range.
+            return;
+        }
+        let next_inst = InsnIndex::new(next_inst_index);
         let labels = self.debug_info.entry(label).or_insert_with(|| vec![]);
         let last = labels
             .last()
             .map(|(_start, end, _vreg)| *end)
             .unwrap_or(InsnIndex::new(0));
-        labels.push((last, inst, reg.into()));
+        labels.push((last, next_inst, reg.into()));
     }
 
     /// Access the constants.
@@ -487,7 +515,7 @@ impl<I: VCodeInst> VCodeBuilder<I> {
     }
 
     fn collect_operands(&mut self, vregs: &VRegAllocator<I>) {
-        let allocatable = PRegSet::from(self.vcode.machine_env());
+        let allocatable = PRegSet::from(self.vcode.abi.machine_env());
         for (i, insn) in self.vcode.insts.iter_mut().enumerate() {
             // Push operands from the instruction onto the operand list.
             //
@@ -590,6 +618,14 @@ impl<I: VCodeInst> VCodeBuilder<I> {
         let old_entry = self.vcode.user_stack_maps.insert(inst, stack_map);
         debug_assert!(old_entry.is_none());
     }
+
+    /// Add debug tags for the associated instruction.
+    pub fn add_debug_tags(&mut self, inst: BackwardsInsnIndex, entries: &[ir::DebugTag]) {
+        let start = u32::try_from(self.vcode.debug_tag_pool.len()).unwrap();
+        self.vcode.debug_tag_pool.extend(entries.iter().cloned());
+        let end = u32::try_from(self.vcode.debug_tag_pool.len()).unwrap();
+        self.vcode.debug_tags.insert(inst, start..end);
+    }
 }
 
 const NO_INST_OFFSET: CodeOffset = u32::MAX;
@@ -602,6 +638,7 @@ impl<I: VCodeInst> VCode<I> {
         emit_info: I::Info,
         block_order: BlockLoweringOrder,
         constants: VCodeConstants,
+        log2_min_function_alignment: u8,
     ) -> Self {
         let n_blocks = block_order.lowered_order().len();
         VCode {
@@ -609,6 +646,8 @@ impl<I: VCodeInst> VCode<I> {
             vreg_types: vec![],
             insts: Vec::with_capacity(10 * n_blocks),
             user_stack_maps: FxHashMap::default(),
+            debug_tags: FxHashMap::default(),
+            debug_tag_pool: vec![],
             operands: Vec::with_capacity(30 * n_blocks),
             operand_ranges: Ranges::with_capacity(10 * n_blocks),
             clobbers: FxHashMap::default(),
@@ -630,14 +669,10 @@ impl<I: VCodeInst> VCode<I> {
             constants,
             debug_value_labels: vec![],
             facts: vec![],
+            log2_min_function_alignment,
         }
     }
 
-    /// Get the ABI-dependent MachineEnv for managing register allocation.
-    pub fn machine_env(&self) -> &MachineEnv {
-        self.abi.machine_env(&self.sigs)
-    }
-
     /// Get the number of blocks. Block indices will be in the range `0 ..
     /// (self.num_blocks() - 1)`.
     pub fn num_blocks(&self) -> usize {
@@ -649,8 +684,12 @@ impl<I: VCodeInst> VCode<I> {
         self.insts.len()
     }
 
-    fn compute_clobbers(&self, regalloc: &regalloc2::Output) -> Vec<Writable<RealReg>> {
+    fn compute_clobbers_and_function_calls(
+        &self,
+        regalloc: &regalloc2::Output,
+    ) -> (Vec<Writable<RealReg>>, FunctionCalls) {
         let mut clobbered = PRegSet::default();
+        let mut function_calls = FunctionCalls::None;
 
         // All moves are included in clobbers.
         for (_, Edit::Move { to, .. }) in &regalloc.edits {
@@ -660,16 +699,6 @@ impl<I: VCodeInst> VCode<I> {
         }
 
         for (i, range) in self.operand_ranges.iter() {
-            // Skip this instruction if not "included in clobbers" as
-            // per the MachInst. (Some backends use this to implement
-            // ABI specifics; e.g., excluding calls of the same ABI as
-            // the current function from clobbers, because by
-            // definition everything clobbered by the call can be
-            // clobbered by this function without saving as well.)
-            if !self.insts[i].is_included_in_clobbers() {
-                continue;
-            }
-
             let operands = &self.operands[range.clone()];
             let allocs = &regalloc.allocs[range];
             for (operand, alloc) in operands.iter().zip(allocs.iter()) {
@@ -680,16 +709,40 @@ impl<I: VCodeInst> VCode<I> {
                 }
             }
 
+            function_calls.update(self.insts[i].call_type());
+
             // Also add explicitly-clobbered registers.
-            if let Some(&inst_clobbered) = self.clobbers.get(&InsnIndex::new(i)) {
-                clobbered.union_from(inst_clobbered);
+            //
+            // Skip merging this instruction's clobber list if not
+            // "included in clobbers" as per the MachInst. (Some
+            // backends use this to implement ABI specifics; e.g.,
+            // excluding calls of the same ABI as the current function
+            // from clobbers, because by definition everything
+            // clobbered by the call can be clobbered by this function
+            // without saving as well.
+            //
+            // This is important for a particular optimization: when
+            // some registers are "half-clobbered", e.g. vector/float
+            // registers on aarch64, we want them to be seen as
+            // clobbered by regalloc so it avoids carrying values
+            // across calls in these registers but not seen as
+            // clobbered by prologue generation here (because the
+            // actual half-clobber implied by the clobber list fits
+            // within the clobbers that we allow without
+            // clobber-saves).
+            if self.insts[i].is_included_in_clobbers() {
+                if let Some(&inst_clobbered) = self.clobbers.get(&InsnIndex::new(i)) {
+                    clobbered.union_from(inst_clobbered);
+                }
             }
         }
 
-        clobbered
+        let clobbered_regs = clobbered
             .into_iter()
             .map(|preg| Writable::from_reg(RealReg::from(preg)))
-            .collect()
+            .collect();
+
+        (clobbered_regs, function_calls)
     }
 
     /// Emit the instructions to a `MachBuffer`, containing fixed-up
@@ -709,11 +762,9 @@ impl<I: VCodeInst> VCode<I> {
     where
         I: VCodeInst,
     {
-        // To write into disasm string.
-        use core::fmt::Write;
-
         let _tt = timing::vcode_emit();
         let mut buffer = MachBuffer::new();
+        buffer.set_log2_min_function_alignment(self.log2_min_function_alignment);
         let mut bb_starts: Vec<Option<CodeOffset>> = vec![];
 
         // The first M MachLabels are reserved for block indices.
@@ -745,9 +796,13 @@ impl<I: VCodeInst> VCode<I> {
         // mutate `VCode`. The info it usually carries prior to
         // setting clobbers is fairly minimal so this should be
         // relatively cheap.
-        let clobbers = self.compute_clobbers(regalloc);
-        self.abi
-            .compute_frame_layout(&self.sigs, regalloc.num_spillslots, clobbers);
+        let (clobbers, function_calls) = self.compute_clobbers_and_function_calls(regalloc);
+        self.abi.compute_frame_layout(
+            &self.sigs,
+            regalloc.num_spillslots,
+            clobbers,
+            function_calls,
+        );
 
         // Emit blocks.
         let mut cur_srcloc = None;
@@ -894,8 +949,11 @@ impl<I: VCodeInst> VCode<I> {
                                 // function.
                                 let index = iix.to_backwards_insn_index(self.num_insts());
                                 let user_stack_map = self.user_stack_maps.remove(&index);
-                                let user_stack_map_disasm =
-                                    user_stack_map.as_ref().map(|m| format!("  ; {m:?}"));
+                                let user_stack_map_disasm = if want_disasm {
+                                    user_stack_map.as_ref().map(|m| format!("  ; {m:?}"))
+                                } else {
+                                    None
+                                };
                                 (user_stack_map, user_stack_map_disasm)
                             };
 
@@ -906,11 +964,50 @@ impl<I: VCodeInst> VCode<I> {
                             None
                         };
 
+                        // Place debug tags in the emission buffer
+                        // either at the offset prior to the
+                        // instruction or after the instruction,
+                        // depending on whether this is a call. See
+                        // the documentation on [`MachDebugTagPos`]
+                        // for details on why.
+                        let mut debug_tag_disasm = None;
+                        let mut place_debug_tags =
+                            |this: &VCode<I>, pos: MachDebugTagPos, buffer: &mut MachBuffer<I>| {
+                                // As above, translate the forward instruction
+                                // index to a backward index for the lookup.
+                                let debug_tag_range = {
+                                    let index = iix.to_backwards_insn_index(this.num_insts());
+                                    this.debug_tags.get(&index)
+                                };
+                                if let Some(range) = debug_tag_range {
+                                    let start = usize::try_from(range.start).unwrap();
+                                    let end = usize::try_from(range.end).unwrap();
+                                    let tags = &this.debug_tag_pool[start..end];
+
+                                    if want_disasm {
+                                        debug_tag_disasm =
+                                            Some(format!("  ; ^-- debug @ {pos:?}: {tags:?}"));
+                                    }
+                                    buffer.push_debug_tags(pos, tags);
+                                }
+                            };
+                        let debug_tag_pos =
+                            if self.insts[iix.index()].call_type() == CallType::Regular {
+                                MachDebugTagPos::Post
+                            } else {
+                                MachDebugTagPos::Pre
+                            };
+
+                        if debug_tag_pos == MachDebugTagPos::Pre {
+                            place_debug_tags(&self, debug_tag_pos, &mut buffer);
+                        }
+
                         // If the instruction we are about to emit is
                         // a return, place an epilogue at this point
                         // (and don't emit the return; the actual
                         // epilogue will contain it).
                         if self.insts[iix.index()].is_term() == MachTerminator::Ret {
+                            log::trace!("emitting epilogue");
                             for inst in self.abi.gen_epilogue() {
                                 do_emit(&inst, &mut disasm, &mut buffer, &mut state);
                             }
@@ -920,21 +1017,25 @@ impl<I: VCodeInst> VCode<I> {
                             let mut allocs = regalloc.inst_allocs(iix).iter();
                             self.insts[iix.index()].get_operands(
                                 &mut |reg: &mut Reg, constraint, _kind, _pos| {
-                                    let alloc = allocs
-                                        .next()
-                                        .expect("enough allocations for all operands")
-                                        .as_reg()
-                                        .expect("only register allocations, not stack allocations")
-                                        .into();
+                                    let alloc =
+                                        allocs.next().expect("enough allocations for all operands");
 
-                                    if let OperandConstraint::FixedReg(rreg) = constraint {
-                                        debug_assert_eq!(Reg::from(rreg), alloc);
+                                    if let Some(alloc) = alloc.as_reg() {
+                                        let alloc: Reg = alloc.into();
+                                        if let OperandConstraint::FixedReg(rreg) = constraint {
+                                            debug_assert_eq!(Reg::from(rreg), alloc);
+                                        }
+                                        *reg = alloc;
+                                    } else if let Some(alloc) = alloc.as_stack() {
+                                        let alloc: Reg = alloc.into();
+                                        *reg = alloc;
                                     }
-                                    *reg = alloc;
                                 },
                             );
                             debug_assert!(allocs.next().is_none());
 
+                            log::trace!("emitting: {:?}", self.insts[iix.index()]);
+
                             // Emit the instruction!
                             do_emit(
                                 &self.insts[iix.index()],
@@ -942,10 +1043,19 @@ impl<I: VCodeInst> VCode<I> {
                                 &mut buffer,
                                 &mut state,
                             );
+
+                            if debug_tag_pos == MachDebugTagPos::Post {
+                                place_debug_tags(&self, debug_tag_pos, &mut buffer);
+                            }
+
                             if let Some(stack_map_disasm) = stack_map_disasm {
                                 disasm.push_str(&stack_map_disasm);
                                 disasm.push('\n');
                             }
+                            if let Some(debug_tag_disasm) = debug_tag_disasm {
+                                disasm.push_str(&debug_tag_disasm);
+                                disasm.push('\n');
+                            }
                         }
                     }
 
@@ -1067,18 +1177,16 @@ impl<I: VCodeInst> VCode<I> {
         self.monotonize_inst_offsets(&mut inst_offsets[..], func_body_len);
         let value_labels_ranges =
             self.compute_value_labels_ranges(regalloc, &inst_offsets[..], func_body_len);
-        let frame_size = self.abi.frame_size();
+
+        // Store metadata about frame layout in the MachBuffer.
+        buffer.set_frame_layout(self.abi.frame_slot_metadata());
 
         EmitResult {
             buffer: buffer.finish(&self.constants, ctrl_plane),
             bb_offsets,
             bb_edges,
-            func_body_len,
             disasm: if want_disasm { Some(disasm) } else { None },
-            sized_stackslot_offsets: self.abi.sized_stackslot_offsets().clone(),
-            dynamic_stackslot_offsets: self.abi.dynamic_stackslot_offsets().clone(),
             value_labels_ranges,
-            frame_size,
         }
     }
 
@@ -1110,9 +1218,7 @@ impl<I: VCodeInst> VCode<I> {
             if inst_offset > next_offset {
                 trace!(
                     "Fixing code offset of the removed Inst {}: {} -> {}",
-                    inst_index,
-                    inst_offset,
-                    next_offset
+                    inst_index, inst_offset, next_offset
                 );
                 inst_offsets[inst_index] = next_offset;
                 continue;
@@ -1132,16 +1238,28 @@ impl<I: VCodeInst> VCode<I> {
             return ValueLabelsRanges::default();
         }
 
+        if trace_log_enabled!() {
+            self.log_value_labels_ranges(regalloc, inst_offsets);
+        }
+
         let mut value_labels_ranges: ValueLabelsRanges = HashMap::new();
         for &(label, from, to, alloc) in &regalloc.debug_locations {
-            let ranges = value_labels_ranges
-                .entry(ValueLabel::from_u32(label))
-                .or_insert_with(|| vec![]);
-            let from_offset = inst_offsets[from.inst().index()];
-            let to_offset = if to.inst().index() == inst_offsets.len() {
+            let label = ValueLabel::from_u32(label);
+            let ranges = value_labels_ranges.entry(label).or_insert_with(|| vec![]);
+            let prog_point_to_inst = |prog_point: ProgPoint| {
+                let mut inst = prog_point.inst();
+                if prog_point.pos() == InstPosition::After {
+                    inst = inst.next();
+                }
+                inst.index()
+            };
+            let from_inst_index = prog_point_to_inst(from);
+            let to_inst_index = prog_point_to_inst(to);
+            let from_offset = inst_offsets[from_inst_index];
+            let to_offset = if to_inst_index == inst_offsets.len() {
                 func_body_len
             } else {
-                inst_offsets[to.inst().index()]
+                inst_offsets[to_inst_index]
             };
 
             // Empty ranges or unavailable offsets can happen
@@ -1167,54 +1285,236 @@ impl<I: VCodeInst> VCode<I> {
                 LabelValueLoc::CFAOffset(cfa_to_sp_offset + slot_offset)
             };
 
-            // ValueLocRanges are recorded by *instruction-end
-            // offset*. `from_offset` is the *start* of the
-            // instruction; that is the same as the end of another
-            // instruction, so we only want to begin coverage once
-            // we are past the previous instruction's end.
-            let start = from_offset + 1;
-
-            // Likewise, `end` is exclusive, but we want to
-            // *include* the end of the last
-            // instruction. `to_offset` is the start of the
-            // `to`-instruction, which is the exclusive end, i.e.,
-            // the first instruction not covered. That
-            // instruction's start is the same as the end of the
-            // last instruction that is included, so we go one
-            // byte further to be sure to include it.
-            let end = to_offset + 1;
-
             // Coalesce adjacent ranges that for the same location
             // to minimize output size here and for the consumers.
             if let Some(last_loc_range) = ranges.last_mut() {
-                if last_loc_range.loc == loc && last_loc_range.end == start {
+                if last_loc_range.loc == loc && last_loc_range.end == from_offset {
                     trace!(
-                        "Extending debug range for VL{} in {:?} to {}",
-                        label,
-                        loc,
-                        end
+                        "Extending debug range for {:?} in {:?} to Inst {} ({})",
+                        label, loc, to_inst_index, to_offset
                     );
-                    last_loc_range.end = end;
+                    last_loc_range.end = to_offset;
                     continue;
                 }
             }
 
             trace!(
-                "Recording debug range for VL{} in {:?}: [Inst {}..Inst {}) [{}..{})",
-                label,
-                loc,
-                from.inst().index(),
-                to.inst().index(),
-                start,
-                end
+                "Recording debug range for {:?} in {:?}: [Inst {}..Inst {}) [{}..{})",
+                label, loc, from_inst_index, to_inst_index, from_offset, to_offset
             );
 
-            ranges.push(ValueLocRange { loc, start, end });
+            ranges.push(ValueLocRange {
+                loc,
+                start: from_offset,
+                end: to_offset,
+            });
         }
 
         value_labels_ranges
     }
 
+    fn log_value_labels_ranges(&self, regalloc: &regalloc2::Output, inst_offsets: &[CodeOffset]) {
+        debug_assert!(trace_log_enabled!());
+
+        // What debug labels do we have? Note we'll skip those that have not been
+        // allocated any location at all. They will show up as numeric gaps in the table.
+        let mut labels = vec![];
+        for &(label, _, _, _) in &regalloc.debug_locations {
+            if Some(&label) == labels.last() {
+                continue;
+            }
+            labels.push(label);
+        }
+
+        // Reformat the data on what VRegs were the VLs assigned to by lowering, since
+        // the array we have is sorted by VReg, and we want it sorted by VL for easy
+        // access in the loop below.
+        let mut vregs = vec![];
+        for &(vreg, start, end, label) in &self.debug_value_labels {
+            if matches!(labels.binary_search(&label), Ok(_)) {
+                vregs.push((label, start, end, vreg));
+            }
+        }
+        vregs.sort_unstable_by(
+            |(l_label, l_start, _, _), (r_label, r_start, _, _)| match l_label.cmp(r_label) {
+                Ordering::Equal => l_start.cmp(r_start),
+                cmp => cmp,
+            },
+        );
+
+        #[derive(PartialEq)]
+        enum Mode {
+            Measure,
+            Emit,
+        }
+        #[derive(PartialEq)]
+        enum Row {
+            Head,
+            Line,
+            Inst(usize, usize),
+        }
+
+        let mut widths = vec![0; 3 + 2 * labels.len()];
+        let mut row = String::new();
+        let mut output_row = |row_kind: Row, mode: Mode| {
+            let mut column_index = 0;
+            row.clear();
+
+            macro_rules! output_cell_impl {
+                ($fill:literal, $span:literal, $($cell_fmt:tt)*) => {
+                    let column_start = row.len();
+                    {
+                        row.push('|');
+                        write!(row, $($cell_fmt)*).unwrap();
+                    }
+
+                    let next_column_index = column_index + $span;
+                    let expected_width: usize = widths[column_index..next_column_index].iter().sum();
+                    if mode == Mode::Measure {
+                        let actual_width = row.len() - column_start;
+                        if actual_width > expected_width {
+                            widths[next_column_index - 1] += actual_width - expected_width;
+                        }
+                    } else {
+                        let column_end = column_start + expected_width;
+                        while row.len() != column_end {
+                            row.push($fill);
+                        }
+                    }
+                    column_index = next_column_index;
+                };
+            }
+            macro_rules! output_cell {
+                ($($cell_fmt:tt)*) => {
+                    output_cell_impl!(' ', 1, $($cell_fmt)*);
+                };
+            }
+
+            match row_kind {
+                Row::Head => {
+                    output_cell!("BB");
+                    output_cell!("Inst");
+                    output_cell!("IP");
+                    for label in &labels {
+                        output_cell_impl!(' ', 2, "{:?}", ValueLabel::from_u32(*label));
+                    }
+                }
+                Row::Line => {
+                    debug_assert!(mode == Mode::Emit);
+                    for _ in 0..3 {
+                        output_cell_impl!('-', 1, "");
+                    }
+                    for _ in &labels {
+                        output_cell_impl!('-', 2, "");
+                    }
+                }
+                Row::Inst(block_index, inst_index) => {
+                    debug_assert!(inst_index < self.num_insts());
+                    if self.block_ranges.get(block_index).start == inst_index {
+                        output_cell!("B{}", block_index);
+                    } else {
+                        output_cell!("");
+                    }
+                    output_cell!("Inst {inst_index} ");
+                    output_cell!("{} ", inst_offsets[inst_index]);
+
+                    for label in &labels {
+                        // First, the VReg.
+                        use regalloc2::Inst;
+                        let vreg_cmp = |inst: usize,
+                                        vreg_label: &u32,
+                                        range_start: &Inst,
+                                        range_end: &Inst| {
+                            match vreg_label.cmp(&label) {
+                                Ordering::Equal => {
+                                    if range_end.index() <= inst {
+                                        Ordering::Less
+                                    } else if range_start.index() > inst {
+                                        Ordering::Greater
+                                    } else {
+                                        Ordering::Equal
+                                    }
+                                }
+                                cmp => cmp,
+                            }
+                        };
+                        let vreg_index =
+                            vregs.binary_search_by(|(l, s, e, _)| vreg_cmp(inst_index, l, s, e));
+                        if let Ok(vreg_index) = vreg_index {
+                            let mut prev_vreg = None;
+                            if inst_index > 0 {
+                                let prev_vreg_index = vregs.binary_search_by(|(l, s, e, _)| {
+                                    vreg_cmp(inst_index - 1, l, s, e)
+                                });
+                                if let Ok(prev_vreg_index) = prev_vreg_index {
+                                    prev_vreg = Some(vregs[prev_vreg_index].3);
+                                }
+                            }
+
+                            let vreg = vregs[vreg_index].3;
+                            if Some(vreg) == prev_vreg {
+                                output_cell!("*");
+                            } else {
+                                output_cell!("{}", vreg);
+                            }
+                        } else {
+                            output_cell!("");
+                        }
+
+                        // Second, the allocated location.
+                        let inst_prog_point = ProgPoint::before(Inst::new(inst_index));
+                        let range_index = regalloc.debug_locations.binary_search_by(
+                            |(range_label, range_start, range_end, _)| match range_label.cmp(label)
+                            {
+                                Ordering::Equal => {
+                                    if *range_end <= inst_prog_point {
+                                        Ordering::Less
+                                    } else if *range_start > inst_prog_point {
+                                        Ordering::Greater
+                                    } else {
+                                        Ordering::Equal
+                                    }
+                                }
+                                cmp => cmp,
+                            },
+                        );
+                        if let Ok(range_index) = range_index {
+                            // Live at this instruction, print the location.
+                            if let Some(reg) = regalloc.debug_locations[range_index].3.as_reg() {
+                                output_cell!("{:?}", Reg::from(reg));
+                            } else {
+                                output_cell!("Stk");
+                            }
+                        } else {
+                            // Not live at this instruction.
+                            output_cell!("");
+                        }
+                    }
+                }
+            }
+            row.push('|');
+
+            if mode == Mode::Emit {
+                trace!("{}", row.as_str());
+            }
+        };
+
+        for block_index in 0..self.num_blocks() {
+            for inst_index in self.block_ranges.get(block_index) {
+                output_row(Row::Inst(block_index, inst_index), Mode::Measure);
+            }
+        }
+        output_row(Row::Head, Mode::Measure);
+
+        output_row(Row::Head, Mode::Emit);
+        output_row(Row::Line, Mode::Emit);
+        for block_index in 0..self.num_blocks() {
+            for inst_index in self.block_ranges.get(block_index) {
+                output_row(Row::Inst(block_index, inst_index), Mode::Emit);
+            }
+        }
+    }
+
     /// Get the IR block for a BlockIndex, if one exists.
     pub fn bindex_to_bb(&self, block: BlockIndex) -> Option<ir::Block> {
         self.block_order.lowered_order()[block.index()].orig_block()
@@ -1310,13 +1610,13 @@ impl<I: VCodeInst> RegallocFunction for VCode<I> {
             // We treat blocks terminated by an unconditional trap like a return for regalloc.
             MachTerminator::None => self.insts[insn.index()].is_trap(),
             MachTerminator::Ret | MachTerminator::RetCall => true,
-            MachTerminator::Uncond | MachTerminator::Cond | MachTerminator::Indirect => false,
+            MachTerminator::Branch => false,
         }
     }
 
     fn is_branch(&self, insn: InsnIndex) -> bool {
         match self.insts[insn.index()].is_term() {
-            MachTerminator::Cond | MachTerminator::Uncond | MachTerminator::Indirect => true,
+            MachTerminator::Branch => true,
             _ => false,
         }
     }
@@ -1733,11 +2033,7 @@ impl VCodeConstantData {
 
     /// Calculate the alignment of the constant data.
     pub fn alignment(&self) -> u32 {
-        if self.as_slice().len() <= 8 {
-            8
-        } else {
-            16
-        }
+        if self.as_slice().len() <= 8 { 8 } else { 16 }
     }
 }
 
