@@ -9,12 +9,12 @@ use super::HashMap;
 use crate::data_context::DataDescription;
 use core::fmt::Display;
 use cranelift_codegen::binemit::{CodeOffset, Reloc};
-use cranelift_codegen::entity::{entity_impl, PrimaryMap};
-use cranelift_codegen::ir::function::{Function, VersionMarker};
+use cranelift_codegen::entity::{PrimaryMap, entity_impl};
 use cranelift_codegen::ir::ExternalName;
+use cranelift_codegen::ir::function::{Function, VersionMarker};
 use cranelift_codegen::settings::SetError;
 use cranelift_codegen::{
-    ir, isa, CodegenError, CompileError, Context, FinalizedMachReloc, FinalizedRelocTarget,
+    CodegenError, CompileError, Context, FinalizedMachReloc, FinalizedRelocTarget, ir, isa,
 };
 use cranelift_control::ControlPlane;
 use std::borrow::{Cow, ToOwned};
@@ -182,6 +182,14 @@ impl Linkage {
         }
     }
 
+    /// Test whether this linkage must have a definition.
+    pub fn requires_definition(self) -> bool {
+        match self {
+            Self::Import | Self::Preemptible => false,
+            Self::Local | Self::Hidden | Self::Export => true,
+        }
+    }
+
     /// Test whether this linkage will have a definition that cannot be preempted.
     pub fn is_final(self) -> bool {
         match self {
@@ -979,7 +987,7 @@ pub trait Module {
         relocs: &[ModuleReloc],
     ) -> ModuleResult<()>;
 
-    /// Define a data object, producing the data contents from the given `DataContext`.
+    /// Define a data object, producing the data contents from the given `DataDescription`.
     fn define_data(&mut self, data_id: DataId, data: &DataDescription) -> ModuleResult<()>;
 }
 
