@@ -4,17 +4,17 @@ use crate::utils::read_to_string;
 use anyhow::{Context as _, Result};
 use clap::Parser;
 use cranelift::prelude::Value;
+use cranelift_codegen::Context;
 use cranelift_codegen::cursor::{Cursor, FuncCursor};
 use cranelift_codegen::flowgraph::ControlFlowGraph;
-use cranelift_codegen::ir::types::{F32, F64, I128, I64};
+use cranelift_codegen::ir::types::{F32, F64, I64, I128};
 use cranelift_codegen::ir::{
     self, Block, FuncRef, Function, GlobalValueData, Inst, InstBuilder, InstructionData,
     StackSlots, TrapCode,
 };
 use cranelift_codegen::isa::TargetIsa;
-use cranelift_codegen::Context;
 use cranelift_entity::PrimaryMap;
-use cranelift_reader::{parse_sets_and_triple, parse_test, ParseOptions};
+use cranelift_reader::{ParseOptions, parse_sets_and_triple, parse_test};
 use indicatif::{ProgressBar, ProgressDrawTarget, ProgressStyle};
 use std::collections::HashMap;
 use std::path::PathBuf;
@@ -420,7 +420,9 @@ impl Mutator for ReplaceBlockParamWithConst {
         // Remove parameters in branching instructions that point to this block
         for pred in cfg.pred_iter(self.block) {
             let dfg = &mut func.dfg;
-            for branch in dfg.insts[pred.inst].branch_destination_mut(&mut dfg.jump_tables) {
+            for branch in dfg.insts[pred.inst]
+                .branch_destination_mut(&mut dfg.jump_tables, &mut dfg.exception_tables)
+            {
                 if branch.block(&dfg.value_lists) == self.block {
                     branch.remove(param_index, &mut dfg.value_lists);
                 }
@@ -708,7 +710,8 @@ impl Mutator for MergeBlocks {
 
         // If the branch instruction that lead us to this block wasn't an unconditional jump, then
         // we have a conditional jump sequence that we should not break.
-        let branch_dests = func.dfg.insts[pred.inst].branch_destination(&func.dfg.jump_tables);
+        let branch_dests = func.dfg.insts[pred.inst]
+            .branch_destination(&func.dfg.jump_tables, &func.dfg.exception_tables);
         if branch_dests.len() != 1 {
             return Some((
                 func,
@@ -717,7 +720,9 @@ impl Mutator for MergeBlocks {
             ));
         }
 
-        let branch_args = branch_dests[0].args_slice(&func.dfg.value_lists).to_vec();
+        let branch_args = branch_dests[0]
+            .args(&func.dfg.value_lists)
+            .collect::<Vec<_>>();
 
         // TODO: should we free the entity list associated with the block params?
         let block_params = func
@@ -731,8 +736,10 @@ impl Mutator for MergeBlocks {
         // If there were any block parameters in block, then the last instruction in pred will
         // fill these parameters. Make the block params aliases of the terminator arguments.
         for (block_param, arg) in block_params.into_iter().zip(branch_args) {
-            if block_param != arg {
-                func.dfg.change_to_alias(block_param, arg);
+            if let Some(arg) = arg.as_value() {
+                if block_param != arg {
+                    func.dfg.change_to_alias(block_param, arg);
+                }
             }
         }
 
@@ -989,7 +996,6 @@ impl<'a> CrashCheckContext<'a> {
         }
     }
 
-    #[cfg_attr(test, expect(unreachable_code, reason = "test-specific code"))]
     fn check_for_crash(&mut self, func: &Function) -> CheckResult {
         self.context.clear();
 
@@ -1010,7 +1016,7 @@ impl<'a> CrashCheckContext<'a> {
         }
 
         #[cfg(test)]
-        {
+        if true {
             // For testing purposes we emulate a panic caused by the existence of
             // a `call` instruction.
             let contains_call = func.layout.blocks().any(|block| {
