@@ -22,6 +22,9 @@ const FEATURES: &[&str] = &[
     "DEBUG_BUILTINS",
     "WAT",
     "POOLING_ALLOCATOR",
+    "COMPONENT_MODEL",
+    "PULLEY",
+    "ALL_ARCH",
 ];
 // ... if you add a line above this be sure to change the other locations
 // marked WASMTIME_FEATURE_LIST
