@@ -97,7 +97,7 @@ wasmtime_option_group! {
         /// optimize the size of memory slots.
         #[serde(default)]
         #[serde(deserialize_with = "crate::opt::cli_parse_wrapper")]
-        pub pooling_memory_protection_keys: Option<wasmtime::MpkEnabled>,
+        pub pooling_memory_protection_keys: Option<wasmtime::Enabled>,
 
         /// Sets an upper limit on how many memory protection keys (MPK) Wasmtime
         /// will use. (default: 16)
@@ -193,6 +193,12 @@ wasmtime_option_group! {
 
         /// DEPRECATED: Use `-Cmemory-reservation-for-growth=N` instead.
         pub dynamic_memory_reserved_for_growth: Option<u64>,
+
+        /// Whether or not `PAGEMAP_SCAN` ioctls are used to reset linear
+        /// memory.
+        #[serde(default)]
+        #[serde(deserialize_with = "crate::opt::cli_parse_wrapper")]
+        pub pooling_pagemap_scan: Option<wasmtime::Enabled>,
     }
 
     enum Optimize {
@@ -237,6 +243,9 @@ wasmtime_option_group! {
         /// object files.
         pub native_unwind_info: Option<bool>,
 
+        /// Whether to perform function inlining during compilation.
+        pub inlining: Option<bool>,
+
         #[prefixed = "cranelift"]
         #[serde(default)]
         /// Set a cranelift-specific option. Use `wasmtime settings` to see
@@ -255,6 +264,8 @@ wasmtime_option_group! {
     pub struct DebugOptions {
         /// Enable generation of DWARF debug information in compiled code.
         pub debug_info: Option<bool>,
+        /// Enable guest debugging insrumentation.
+        pub guest_debug: Option<bool>,
         /// Configure whether compiled code can map native addresses to wasm.
         pub address_map: Option<bool>,
         /// Configure whether logging is enabled.
@@ -358,6 +369,8 @@ wasmtime_option_group! {
         pub tail_call: Option<bool>,
         /// Configure support for the threads proposal.
         pub threads: Option<bool>,
+        /// Configure support for the shared-everything-threads proposal.
+        pub shared_everything_threads: Option<bool>,
         /// Configure support for the memory64 proposal.
         pub memory64: Option<bool>,
         /// Configure support for the component-model proposal.
@@ -370,8 +383,19 @@ wasmtime_option_group! {
         /// Component model support for async lifting/lowering: this corresponds
         /// to the 🚟 emoji in the component model specification.
         pub component_model_async_stackful: Option<bool>,
+        /// Component model support for threading: this corresponds
+        /// to the 🧵 emoji in the component model specification.
+        pub component_model_threading: Option<bool>,
+        /// Component model support for `error-context`: this corresponds
+        /// to the 📝 emoji in the component model specification.
+        pub component_model_error_context: Option<bool>,
+        /// GC support in the component model: this corresponds to the 🛸 emoji
+        /// in the component model specification.
+        pub component_model_gc: Option<bool>,
         /// Configure support for the function-references proposal.
         pub function_references: Option<bool>,
+        /// Configure support for the stack-switching proposal.
+        pub stack_switching: Option<bool>,
         /// Configure support for the GC proposal.
         pub gc: Option<bool>,
         /// Configure support for the custom-page-sizes proposal.
@@ -380,6 +404,10 @@ wasmtime_option_group! {
         pub wide_arithmetic: Option<bool>,
         /// Configure support for the extended-const proposal.
         pub extended_const: Option<bool>,
+        /// Configure support for the exceptions proposal.
+        pub exceptions: Option<bool>,
+        /// Whether or not any GC infrastructure in Wasmtime is enabled or not.
+        pub gc_support: Option<bool>,
     }
 
     enum Wasm {
@@ -415,9 +443,11 @@ wasmtime_option_group! {
         /// Enable support for WASI key-value imports (experimental)
         pub keyvalue: Option<bool>,
         /// Inherit environment variables and file descriptors following the
-        /// systemd listen fd specification (UNIX only)
+        /// systemd listen fd specification (UNIX only) (legacy wasip1
+        /// implementation only)
         pub listenfd: Option<bool>,
-        /// Grant access to the given TCP listen socket
+        /// Grant access to the given TCP listen socket (experimental, legacy
+        /// wasip1 implementation only)
         #[serde(default)]
         pub tcplisten: Vec<String>,
         /// Enable support for WASI TLS (Transport Layer Security) imports (experimental)
@@ -458,6 +488,8 @@ wasmtime_option_group! {
         /// Preset data for the In-Memory provider of WASI key-value API.
         #[serde(skip)]
         pub keyvalue_in_memory_data: Vec<KeyValuePair>,
+        /// Enable support for WASIp3 APIs.
+        pub p3: Option<bool>,
     }
 
     enum Wasi {
@@ -677,6 +709,11 @@ impl CommonOptions {
         if let Some(enable) = self.debug.debug_info {
             config.debug_info(enable);
         }
+        match_feature! {
+            ["debug" : self.debug.guest_debug]
+            enable => config.guest_debug(enable),
+            _ => err,
+        }
         if self.debug.coredump.is_some() {
             #[cfg(feature = "coredump")]
             config.coredump_on_trap(true);
@@ -727,14 +764,12 @@ impl CommonOptions {
 
         #[cfg(feature = "cache")]
         if self.codegen.cache != Some(false) {
-            match &self.codegen.cache_config {
-                Some(path) => {
-                    config.cache_config_load(path)?;
-                }
-                None => {
-                    config.cache_config_load_default()?;
-                }
-            }
+            use wasmtime::Cache;
+            let cache = match &self.codegen.cache_config {
+                Some(path) => Cache::from_file(Some(Path::new(path)))?,
+                None => Cache::from_file(None)?,
+            };
+            config.cache(Some(cache));
         }
         #[cfg(not(feature = "cache"))]
         if self.codegen.cache == Some(true) {
@@ -808,6 +843,26 @@ impl CommonOptions {
         if let Some(enable) = self.codegen.native_unwind_info {
             config.native_unwind_info(enable);
         }
+        if let Some(enable) = self.codegen.inlining {
+            config.compiler_inlining(enable);
+        }
+
+        // async_stack_size enabled by either async or stack-switching, so
+        // cannot directly use match_feature!
+        #[cfg(any(feature = "async", feature = "stack-switching"))]
+        {
+            if let Some(size) = self.wasm.async_stack_size {
+                config.async_stack_size(size);
+            }
+        }
+        #[cfg(not(any(feature = "async", feature = "stack-switching")))]
+        {
+            if let Some(_size) = self.wasm.async_stack_size {
+                anyhow::bail!(concat!(
+                    "support for async/stack-switching disabled at compile time"
+                ));
+            }
+        }
 
         match_feature! {
             ["pooling-allocator" : self.opts.pooling_allocator.or(pooling_allocator_default)]
@@ -894,6 +949,9 @@ impl CommonOptions {
                         max => cfg.total_gc_heaps(max),
                         _ => err,
                     }
+                    if let Some(enabled) = self.opts.pooling_pagemap_scan {
+                        cfg.pagemap_scan(enabled);
+                    }
                     config.allocation_strategy(wasmtime::InstanceAllocationStrategy::Pooling(cfg));
                 }
             },
@@ -914,11 +972,6 @@ impl CommonOptions {
             );
         }
 
-        match_feature! {
-            ["async" : self.wasm.async_stack_size]
-            size => config.async_stack_size(size),
-            _ => err,
-        }
         match_feature! {
             ["async" : self.wasm.async_stack_zeroing]
             enable => config.async_stack_zeroing(enable),
@@ -931,7 +984,7 @@ impl CommonOptions {
             // If `-Wasync-stack-size` isn't passed then automatically adjust it
             // to the wasm stack size provided here too. That prevents the need
             // to pass both when one can generally be inferred from the other.
-            #[cfg(feature = "async")]
+            #[cfg(any(feature = "async", feature = "stack-switching"))]
             if self.wasm.async_stack_size.is_none() {
                 const DEFAULT_HOST_STACK: usize = 512 << 10;
                 config.async_stack_size(max + DEFAULT_HOST_STACK);
@@ -947,6 +1000,10 @@ impl CommonOptions {
             true => err,
         }
 
+        if let Some(enable) = self.wasm.gc_support {
+            config.gc_support(enable);
+        }
+
         Ok(config)
     }
 
@@ -974,6 +1031,9 @@ impl CommonOptions {
         if let Some(enable) = self.wasm.memory64.or(all) {
             config.wasm_memory64(enable);
         }
+        if let Some(enable) = self.wasm.stack_switching {
+            config.wasm_stack_switching(enable);
+        }
         if let Some(enable) = self.wasm.custom_page_sizes.or(all) {
             config.wasm_custom_page_sizes(enable);
         }
@@ -1002,11 +1062,25 @@ impl CommonOptions {
             ("component-model-async", component_model_async, wasm_component_model_async)
             ("component-model-async", component_model_async_builtins, wasm_component_model_async_builtins)
             ("component-model-async", component_model_async_stackful, wasm_component_model_async_stackful)
+            ("component-model-async", component_model_threading, wasm_component_model_threading)
+            ("component-model", component_model_error_context, wasm_component_model_error_context)
             ("threads", threads, wasm_threads)
             ("gc", gc, wasm_gc)
             ("gc", reference_types, wasm_reference_types)
             ("gc", function_references, wasm_function_references)
+            ("gc", exceptions, wasm_exceptions)
+            ("stack-switching", stack_switching, wasm_stack_switching)
         }
+
+        if let Some(enable) = self.wasm.component_model_gc {
+            #[cfg(all(feature = "component-model", feature = "gc"))]
+            config.wasm_component_model_gc(enable);
+            #[cfg(not(all(feature = "component-model", feature = "gc")))]
+            if enable && all.is_none() {
+                anyhow::bail!("support for `component-model-gc` was disabled at compile time")
+            }
+        }
+
         Ok(())
     }
 
