@@ -1,18 +1,21 @@
 use super::{ArrayInit, GcCompiler};
+use crate::bounds_checks::BoundsCheck;
 use crate::func_environ::{Extension, FuncEnvironment};
-use crate::translate::{StructFieldsVec, TargetEnvironment};
-use crate::TRAP_INTERNAL_ASSERT;
+use crate::translate::{Heap, HeapData, StructFieldsVec, TargetEnvironment};
+use crate::{Reachability, TRAP_INTERNAL_ASSERT};
+use cranelift_codegen::ir::immediates::Offset32;
+use cranelift_codegen::ir::{BlockArg, ExceptionTableData, ExceptionTableItem};
 use cranelift_codegen::{
     cursor::FuncCursor,
-    ir::{self, condcodes::IntCC, InstBuilder},
+    ir::{self, InstBuilder, condcodes::IntCC},
 };
 use cranelift_entity::packed_option::ReservedValue;
 use cranelift_frontend::FunctionBuilder;
-use smallvec::SmallVec;
+use smallvec::{SmallVec, smallvec};
 use wasmtime_environ::{
-    wasm_unsupported, Collector, GcArrayLayout, GcLayout, GcStructLayout, ModuleInternedTypeIndex,
-    PtrSize, TypeIndex, VMGcKind, WasmHeapTopType, WasmHeapType, WasmRefType, WasmResult,
-    WasmStorageType, WasmValType, I31_DISCRIMINANT,
+    Collector, GcArrayLayout, GcLayout, GcStructLayout, I31_DISCRIMINANT, ModuleInternedTypeIndex,
+    PtrSize, TagIndex, TypeIndex, VMGcKind, WasmCompositeInnerType, WasmHeapTopType, WasmHeapType,
+    WasmRefType, WasmResult, WasmStorageType, WasmValType, wasm_unsupported,
 };
 
 #[cfg(feature = "gc-drc")]
@@ -56,7 +59,10 @@ pub fn gc_compiler(func_env: &mut FuncEnvironment<'_>) -> WasmResult<Box<dyn GcC
     }
 }
 
-#[cfg_attr(not(feature = "gc-drc"), allow(dead_code))]
+#[cfg_attr(
+    not(feature = "gc-drc"),
+    expect(dead_code, reason = "easier to define")
+)]
 fn unbarriered_load_gc_ref(
     builder: &mut FunctionBuilder,
     ty: WasmHeapType,
@@ -71,7 +77,10 @@ fn unbarriered_load_gc_ref(
     Ok(gc_ref)
 }
 
-#[cfg_attr(not(any(feature = "gc-drc", feature = "gc-null")), allow(dead_code))]
+#[cfg_attr(
+    not(any(feature = "gc-drc", feature = "gc-null")),
+    expect(dead_code, reason = "easier to define")
+)]
 fn unbarriered_store_gc_ref(
     builder: &mut FunctionBuilder,
     ty: WasmHeapType,
@@ -115,8 +124,10 @@ fn read_field_at_addr(
             WasmValType::F64 => builder.ins().load(ir::types::F64, flags, addr, 0),
             WasmValType::V128 => builder.ins().load(ir::types::I8X16, flags, addr, 0),
             WasmValType::Ref(r) => match r.heap_type.top() {
-                WasmHeapTopType::Any | WasmHeapTopType::Extern => gc_compiler(func_env)?
-                    .translate_read_gc_reference(func_env, builder, r, addr, flags)?,
+                WasmHeapTopType::Any | WasmHeapTopType::Extern | WasmHeapTopType::Exn => {
+                    gc_compiler(func_env)?
+                        .translate_read_gc_reference(func_env, builder, r, addr, flags)?
+                }
                 WasmHeapTopType::Func => {
                     let expected_ty = match r.heap_type {
                         WasmHeapType::Func => ModuleInternedTypeIndex::reserved_value(),
@@ -151,7 +162,12 @@ fn read_field_at_addr(
                         .call(get_interned_func_ref, &[vmctx, func_ref_id, expected_ty]);
                     builder.func.dfg.first_result(call_inst)
                 }
-                WasmHeapTopType::Cont => todo!(), // FIXME: #10248 stack switching support.
+                WasmHeapTopType::Cont => {
+                    // TODO(#10248) GC integration for stack switching
+                    return Err(wasmtime_environ::WasmError::Unsupported(
+                        "Stack switching feature not compatible with GC, yet".to_string(),
+                    ));
+                }
             },
         },
     };
@@ -305,7 +321,9 @@ pub fn translate_struct_get(
     struct_ref: ir::Value,
     extension: Option<Extension>,
 ) -> WasmResult<ir::Value> {
-    log::trace!("translate_struct_get({struct_type_index:?}, {field_index:?}, {struct_ref:?}, {extension:?})");
+    log::trace!(
+        "translate_struct_get({struct_type_index:?}, {field_index:?}, {struct_ref:?}, {extension:?})"
+    );
 
     // TODO: If we know we have a `(ref $my_struct)` here, instead of maybe a
     // `(ref null $my_struct)`, we could omit the `trapz`. But plumbing that
@@ -315,9 +333,8 @@ pub fn translate_struct_get(
     let field_index = usize::try_from(field_index).unwrap();
     let interned_type_index = func_env.module.types[struct_type_index].unwrap_module_type_index();
 
-    let struct_layout = func_env.struct_layout(interned_type_index);
+    let struct_layout = func_env.struct_or_exn_layout(interned_type_index);
     let struct_size = struct_layout.size;
-    let struct_size_val = builder.ins().iconst(ir::types::I32, i64::from(struct_size));
 
     let field_offset = struct_layout.fields[field_index].offset;
     let field_ty = &func_env.types.unwrap_struct(interned_type_index)?.fields[field_index];
@@ -327,8 +344,11 @@ pub fn translate_struct_get(
     let field_addr = func_env.prepare_gc_ref_access(
         builder,
         struct_ref,
-        Offset::Static(field_offset),
-        BoundsCheck::Object(struct_size_val),
+        BoundsCheck::StaticObjectField {
+            offset: field_offset,
+            access_size: u8::try_from(field_size).unwrap(),
+            object_size: struct_size,
+        },
     );
 
     let result = read_field_at_addr(
@@ -360,9 +380,8 @@ pub fn translate_struct_set(
     let field_index = usize::try_from(field_index).unwrap();
     let interned_type_index = func_env.module.types[struct_type_index].unwrap_module_type_index();
 
-    let struct_layout = func_env.struct_layout(interned_type_index);
+    let struct_layout = func_env.struct_or_exn_layout(interned_type_index);
     let struct_size = struct_layout.size;
-    let struct_size_val = builder.ins().iconst(ir::types::I32, i64::from(struct_size));
 
     let field_offset = struct_layout.fields[field_index].offset;
     let field_ty = &func_env.types.unwrap_struct(interned_type_index)?.fields[field_index];
@@ -372,8 +391,11 @@ pub fn translate_struct_set(
     let field_addr = func_env.prepare_gc_ref_access(
         builder,
         struct_ref,
-        Offset::Static(field_offset),
-        BoundsCheck::Object(struct_size_val),
+        BoundsCheck::StaticObjectField {
+            offset: field_offset,
+            access_size: u8::try_from(field_size).unwrap(),
+            object_size: struct_size,
+        },
     );
 
     write_field_at_addr(
@@ -388,6 +410,117 @@ pub fn translate_struct_set(
     Ok(())
 }
 
+pub fn translate_exn_unbox(
+    func_env: &mut FuncEnvironment<'_>,
+    builder: &mut FunctionBuilder<'_>,
+    tag_index: TagIndex,
+    exn_ref: ir::Value,
+) -> WasmResult<SmallVec<[ir::Value; 4]>> {
+    log::trace!("translate_exn_unbox({tag_index:?}, {exn_ref:?})");
+
+    // We know that the `exn_ref` is not null because we reach this
+    // operation only in catch blocks, and throws are initiated from
+    // runtime code that checks for nulls first.
+
+    // Get the GcExceptionLayout associated with this tag's
+    // function type, and generate loads for each field.
+    let exception_ty_idx = func_env
+        .exception_type_from_tag(tag_index)
+        .unwrap_module_type_index();
+    let exception_ty = func_env.types.unwrap_exn(exception_ty_idx)?;
+    let exn_layout = func_env.struct_or_exn_layout(exception_ty_idx);
+    let exn_size = exn_layout.size;
+
+    // Gather accesses first because these require a borrow on
+    // `func_env`, which we later mutate below via
+    // `prepare_gc_ref_access()`.
+    let mut accesses: SmallVec<[_; 4]> = smallvec![];
+    for (field_ty, field_layout) in exception_ty.fields.iter().zip(exn_layout.fields.iter()) {
+        accesses.push((field_layout.offset, field_ty.element_type));
+    }
+
+    let mut result = smallvec![];
+    for (field_offset, field_ty) in accesses {
+        let field_size = wasmtime_environ::byte_size_of_wasm_ty_in_gc_heap(&field_ty);
+        assert!(field_offset + field_size <= exn_size);
+        let field_addr = func_env.prepare_gc_ref_access(
+            builder,
+            exn_ref,
+            BoundsCheck::StaticObjectField {
+                offset: field_offset,
+                access_size: u8::try_from(field_size).unwrap(),
+                object_size: exn_size,
+            },
+        );
+
+        let value = read_field_at_addr(func_env, builder, field_ty, field_addr, None)?;
+        result.push(value);
+    }
+
+    log::trace!("translate_exn_unbox(..) -> {result:?}");
+    Ok(result)
+}
+
+pub fn translate_exn_throw(
+    func_env: &mut FuncEnvironment<'_>,
+    builder: &mut FunctionBuilder<'_>,
+    tag_index: TagIndex,
+    args: &[ir::Value],
+) -> WasmResult<()> {
+    let (instance_id, defined_tag_id) = func_env.get_instance_and_tag(builder, tag_index);
+    let exnref = gc_compiler(func_env)?.alloc_exn(
+        func_env,
+        builder,
+        tag_index,
+        args,
+        instance_id,
+        defined_tag_id,
+    )?;
+    translate_exn_throw_ref(func_env, builder, exnref)
+}
+
+pub fn translate_exn_throw_ref(
+    func_env: &mut FuncEnvironment<'_>,
+    builder: &mut FunctionBuilder<'_>,
+    exnref: ir::Value,
+) -> WasmResult<()> {
+    let builtin = func_env.builtin_functions.throw_ref(builder.func);
+    let sig = builder.func.dfg.ext_funcs[builtin].signature;
+    let vmctx = func_env.vmctx_val(&mut builder.cursor());
+
+    // Generate a `try_call` with handlers from the current
+    // stack. This libcall is unique among libcall implementations of
+    // opcodes: we know the others will not throw, but `throw_ref`'s
+    // entire purpose is to throw. So if there are any handlers in the
+    // local function body, we need to attach them to this callsite
+    // like any other.
+    let continuation = builder.create_block();
+    let current_block = builder.current_block().unwrap();
+    builder.insert_block_after(continuation, current_block);
+    let continuation_call = builder.func.dfg.block_call(continuation, &[]);
+    let mut table_items = vec![ExceptionTableItem::Context(vmctx)];
+    for (tag, block) in func_env.stacks.handlers.handlers() {
+        let block_call = builder
+            .func
+            .dfg
+            .block_call(block, &[BlockArg::TryCallExn(0)]);
+        table_items.push(match tag {
+            Some(tag) => ExceptionTableItem::Tag(tag, block_call),
+            None => ExceptionTableItem::Default(block_call),
+        });
+    }
+    let etd = ExceptionTableData::new(sig, continuation_call, table_items);
+    let et = builder.func.dfg.exception_tables.push(etd);
+
+    builder.ins().try_call(builtin, &[vmctx, exnref], et);
+
+    builder.switch_to_block(continuation);
+    builder.seal_block(continuation);
+    func_env.trap(builder, crate::TRAP_UNREACHABLE);
+
+    Ok(())
+}
+
 pub fn translate_array_new(
     func_env: &mut FuncEnvironment<'_>,
     builder: &mut FunctionBuilder,
@@ -446,7 +579,10 @@ pub fn translate_array_new_fixed(
 
 impl ArrayInit<'_> {
     /// Get the length (as an `i32`-typed `ir::Value`) of these array elements.
-    #[cfg_attr(not(any(feature = "gc-drc", feature = "gc-null")), allow(dead_code))]
+    #[cfg_attr(
+        not(any(feature = "gc-drc", feature = "gc-null")),
+        expect(dead_code, reason = "easier to define")
+    )]
     fn len(self, pos: &mut FuncCursor) -> ir::Value {
         match self {
             ArrayInit::Fill { len, .. } => len,
@@ -458,7 +594,10 @@ impl ArrayInit<'_> {
     }
 
     /// Initialize a newly-allocated array's elements.
-    #[cfg_attr(not(any(feature = "gc-drc", feature = "gc-null")), allow(dead_code))]
+    #[cfg_attr(
+        not(any(feature = "gc-drc", feature = "gc-null")),
+        expect(dead_code, reason = "easier to define")
+    )]
     fn initialize(
         self,
         func_env: &mut FuncEnvironment<'_>,
@@ -532,7 +671,9 @@ fn emit_array_fill_impl(
         ir::Value,
     ) -> WasmResult<()>,
 ) -> WasmResult<()> {
-    log::trace!("emit_array_fill_impl(elem_addr: {elem_addr:?}, elem_size: {elem_size:?}, fill_end: {fill_end:?})");
+    log::trace!(
+        "emit_array_fill_impl(elem_addr: {elem_addr:?}, elem_size: {elem_size:?}, fill_end: {fill_end:?})"
+    );
 
     let pointer_ty = func_env.pointer_type();
 
@@ -571,13 +712,14 @@ fn emit_array_fill_impl(
 
     // Current block: jump to the loop header block with the first element's
     // address.
-    builder.ins().jump(loop_header_block, &[elem_addr]);
+    builder.ins().jump(loop_header_block, &[elem_addr.into()]);
 
     // Loop header block: check if we're done, then jump to either the continue
     // block or the loop body block.
     builder.switch_to_block(loop_header_block);
     builder.append_block_param(loop_header_block, pointer_ty);
     log::trace!("emit_array_fill_impl: loop header");
+    func_env.translate_loop_header(builder)?;
     let elem_addr = builder.block_params(loop_header_block)[0];
     let done = builder.ins().icmp(IntCC::Equal, elem_addr, fill_end);
     builder
@@ -590,7 +732,9 @@ fn emit_array_fill_impl(
     log::trace!("emit_array_fill_impl: loop body");
     emit_elem_write(func_env, builder, elem_addr)?;
     let next_elem_addr = builder.ins().iadd(elem_addr, elem_size);
-    builder.ins().jump(loop_header_block, &[next_elem_addr]);
+    builder
+        .ins()
+        .jump(loop_header_block, &[next_elem_addr.into()]);
 
     // Continue...
     builder.switch_to_block(continue_block);
@@ -635,8 +779,10 @@ pub fn translate_array_fill(
     let elem_addr = func_env.prepare_gc_ref_access(
         builder,
         array_ref,
-        Offset::Dynamic(obj_offset),
-        BoundsCheck::Object(obj_size),
+        BoundsCheck::DynamicObjectField {
+            offset: obj_offset,
+            object_size: obj_size,
+        },
     );
 
     // Calculate the end address, just after the filled region.
@@ -679,10 +825,12 @@ pub fn translate_array_len(
     let len_field = func_env.prepare_gc_ref_access(
         builder,
         array_ref,
-        Offset::Static(len_offset),
         // Note: We can't bounds check the whole array object's size because we
         // don't know its length yet. Chicken and egg problem.
-        BoundsCheck::Access(ir::types::I32.bytes()),
+        BoundsCheck::StaticOffset {
+            offset: len_offset,
+            access_size: u8::try_from(ir::types::I32.bytes()).unwrap(),
+        },
     );
     let result = builder.ins().load(
         ir::types::I32,
@@ -813,8 +961,10 @@ fn array_elem_addr(
     func_env.prepare_gc_ref_access(
         builder,
         array_ref,
-        Offset::Dynamic(offset_in_array),
-        BoundsCheck::Object(obj_size),
+        BoundsCheck::DynamicObjectField {
+            offset: offset_in_array,
+            object_size: obj_size,
+        },
     )
 }
 
@@ -864,17 +1014,18 @@ pub fn translate_array_set(
 pub fn translate_ref_test(
     func_env: &mut FuncEnvironment<'_>,
     builder: &mut FunctionBuilder<'_>,
-    ref_ty: WasmRefType,
+    test_ty: WasmRefType,
     val: ir::Value,
+    val_ty: WasmRefType,
 ) -> WasmResult<ir::Value> {
-    log::trace!("translate_ref_test({ref_ty:?}, {val:?})");
+    log::trace!("translate_ref_test({test_ty:?}, {val:?})");
 
     // First special case: testing for references to bottom types.
-    if ref_ty.heap_type.is_bottom() {
-        let result = if ref_ty.nullable {
+    if test_ty.heap_type.is_bottom() {
+        let result = if test_ty.nullable {
             // All null references (within the same type hierarchy) match null
             // references to the bottom type.
-            func_env.translate_ref_is_null(builder.cursor(), val)?
+            func_env.translate_ref_is_null(builder.cursor(), val, val_ty)?
         } else {
             // `ref.test` is always false for non-nullable bottom types, as the
             // bottom types are uninhabited.
@@ -888,11 +1039,11 @@ pub fn translate_ref_test(
     // the same type hierarchy as `heap_ty`, if `heap_ty` is its hierarchy's top
     // type, we only need to worry about whether we are testing for nullability
     // or not.
-    if ref_ty.heap_type.is_top() {
-        let result = if ref_ty.nullable {
+    if test_ty.heap_type.is_top() {
+        let result = if test_ty.nullable {
             builder.ins().iconst(ir::types::I32, 1)
         } else {
-            let is_null = func_env.translate_ref_is_null(builder.cursor(), val)?;
+            let is_null = func_env.translate_ref_is_null(builder.cursor(), val, val_ty)?;
             let zero = builder.ins().iconst(ir::types::I32, 0);
             let one = builder.ins().iconst(ir::types::I32, 1);
             builder.ins().select(is_null, zero, one)
@@ -903,14 +1054,14 @@ pub fn translate_ref_test(
 
     // `i31ref`s are a little interesting because they don't point to GC
     // objects; we test the bit pattern of the reference itself.
-    if ref_ty.heap_type == WasmHeapType::I31 {
+    if test_ty.heap_type == WasmHeapType::I31 {
         let i31_mask = builder.ins().iconst(
             ir::types::I32,
             i64::from(wasmtime_environ::I31_DISCRIMINANT),
         );
         let is_i31 = builder.ins().band(val, i31_mask);
-        let result = if ref_ty.nullable {
-            let is_null = func_env.translate_ref_is_null(builder.cursor(), val)?;
+        let result = if test_ty.nullable {
+            let is_null = func_env.translate_ref_is_null(builder.cursor(), val, val_ty)?;
             builder.ins().bor(is_null, is_i31)
         } else {
             is_i31
@@ -922,19 +1073,21 @@ pub fn translate_ref_test(
     // Otherwise, in the general case, we need to inspect our given object's
     // actual type, which also requires null-checking and i31-checking it.
 
-    let is_any_hierarchy = ref_ty.heap_type.top() == WasmHeapTopType::Any;
+    let is_any_hierarchy = test_ty.heap_type.top() == WasmHeapTopType::Any;
 
     let non_null_block = builder.create_block();
     let non_null_non_i31_block = builder.create_block();
     let continue_block = builder.create_block();
 
     // Current block: check if the reference is null and branch appropriately.
-    let is_null = func_env.translate_ref_is_null(builder.cursor(), val)?;
-    let result_when_is_null = builder.ins().iconst(ir::types::I32, ref_ty.nullable as i64);
+    let is_null = func_env.translate_ref_is_null(builder.cursor(), val, val_ty)?;
+    let result_when_is_null = builder
+        .ins()
+        .iconst(ir::types::I32, test_ty.nullable as i64);
     builder.ins().brif(
         is_null,
         continue_block,
-        &[result_when_is_null],
+        &[result_when_is_null.into()],
         non_null_block,
         &[],
     );
@@ -954,14 +1107,14 @@ pub fn translate_ref_test(
         let result_when_is_i31 = builder.ins().iconst(
             ir::types::I32,
             matches!(
-                ref_ty.heap_type,
+                test_ty.heap_type,
                 WasmHeapType::Any | WasmHeapType::Eq | WasmHeapType::I31
             ) as i64,
         );
         builder.ins().brif(
             is_i31,
             continue_block,
-            &[result_when_is_i31],
+            &[result_when_is_i31.into()],
             non_null_non_i31_block,
             &[],
         );
@@ -982,15 +1135,14 @@ pub fn translate_ref_test(
                              val: ir::Value,
                              expected_kind: VMGcKind|
      -> ir::Value {
-        let header_size = builder.ins().iconst(
-            ir::types::I32,
-            i64::from(wasmtime_environ::VM_GC_HEADER_SIZE),
-        );
         let kind_addr = func_env.prepare_gc_ref_access(
             builder,
             val,
-            Offset::Static(wasmtime_environ::VM_GC_HEADER_KIND_OFFSET),
-            BoundsCheck::Object(header_size),
+            BoundsCheck::StaticObjectField {
+                offset: wasmtime_environ::VM_GC_HEADER_KIND_OFFSET,
+                access_size: wasmtime_environ::VM_GC_KIND_SIZE,
+                object_size: wasmtime_environ::VM_GC_HEADER_SIZE,
+            },
         );
         let actual_kind = builder.ins().load(
             ir::types::I32,
@@ -1008,13 +1160,17 @@ pub fn translate_ref_test(
             .icmp(ir::condcodes::IntCC::Equal, and, expected_kind);
         builder.ins().uextend(ir::types::I32, kind_matches)
     };
-    let result = match ref_ty.heap_type {
+    let result = match test_ty.heap_type {
         WasmHeapType::Any
         | WasmHeapType::None
         | WasmHeapType::Extern
         | WasmHeapType::NoExtern
         | WasmHeapType::Func
         | WasmHeapType::NoFunc
+        | WasmHeapType::Cont
+        | WasmHeapType::NoCont
+        | WasmHeapType::Exn
+        | WasmHeapType::NoExn
         | WasmHeapType::I31 => unreachable!("handled top, bottom, and i31 types above"),
 
         // For these abstract but non-top and non-bottom types, we check the
@@ -1030,7 +1186,9 @@ pub fn translate_ref_test(
         // TODO: This check should ideally be done inline, but we don't have a
         // good way to access the `TypeRegistry`'s supertypes arrays from Wasm
         // code at the moment.
-        WasmHeapType::ConcreteArray(ty) | WasmHeapType::ConcreteStruct(ty) => {
+        WasmHeapType::ConcreteArray(ty)
+        | WasmHeapType::ConcreteStruct(ty)
+        | WasmHeapType::ConcreteExn(ty) => {
             let expected_interned_ty = ty.unwrap_module_type_index();
             let expected_shared_ty =
                 func_env.module_interned_to_shared_ty(&mut builder.cursor(), expected_interned_ty);
@@ -1038,8 +1196,10 @@ pub fn translate_ref_test(
             let ty_addr = func_env.prepare_gc_ref_access(
                 builder,
                 val,
-                Offset::Static(wasmtime_environ::VM_GC_HEADER_TYPE_INDEX_OFFSET),
-                BoundsCheck::Access(func_env.offsets.size_of_vmshared_type_index().into()),
+                BoundsCheck::StaticOffset {
+                    offset: wasmtime_environ::VM_GC_HEADER_TYPE_INDEX_OFFSET,
+                    access_size: func_env.offsets.size_of_vmshared_type_index(),
+                },
             );
             let actual_shared_ty = builder.ins().load(
                 ir::types::I32,
@@ -1067,10 +1227,14 @@ pub fn translate_ref_test(
 
             func_env.is_subtype(builder, actual_shared_ty, expected_shared_ty)
         }
-
-        WasmHeapType::Cont | WasmHeapType::ConcreteCont(_) | WasmHeapType::NoCont => todo!(), // FIXME: #10248 stack switching support.
+        WasmHeapType::ConcreteCont(_) => {
+            // TODO(#10248) GC integration for stack switching
+            return Err(wasmtime_environ::WasmError::Unsupported(
+                "Stack switching feature not compatible with GC, yet".to_string(),
+            ));
+        }
     };
-    builder.ins().jump(continue_block, &[result]);
+    builder.ins().jump(continue_block, &[result.into()]);
 
     // Control flow join point with the result.
     builder.switch_to_block(continue_block);
@@ -1084,41 +1248,6 @@ pub fn translate_ref_test(
     Ok(result)
 }
 
-/// A static or dynamic offset from a GC reference.
-#[derive(Debug)]
-enum Offset {
-    /// A static offset from a GC reference.
-    Static(u32),
-
-    /// A dynamic `i32` offset from a GC reference.
-    Dynamic(ir::Value),
-}
-
-/// The kind of bounds check to perform when accessing a GC object's fields and
-/// elements.
-#[derive(Debug)]
-enum BoundsCheck {
-    /// Check that this whole object is inside the GC heap:
-    ///
-    /// ```ignore
-    /// gc_ref + size <= gc_heap_bound
-    /// ```
-    ///
-    /// The object size must be an `i32` value.
-    Object(ir::Value),
-
-    /// Check that this one access in particular is inside the GC heap:
-    ///
-    /// ```ignore
-    /// gc_ref + offset + access_size <= gc_heap_bound
-    /// ```
-    ///
-    /// Prefer `Bound::Object` over `Bound::Access` when possible, as that
-    /// approach allows the mid-end to deduplicate bounds checks across multiple
-    /// accesses to the same object.
-    Access(u32),
-}
-
 fn uextend_i32_to_pointer_type(
     builder: &mut FunctionBuilder,
     pointer_type: ir::Type,
@@ -1136,7 +1265,10 @@ fn uextend_i32_to_pointer_type(
 /// in its initialization.
 ///
 /// Traps if the size overflows.
-#[cfg_attr(not(any(feature = "gc-drc", feature = "gc-null")), allow(dead_code))]
+#[cfg_attr(
+    not(any(feature = "gc-drc", feature = "gc-null")),
+    expect(dead_code, reason = "easier to define")
+)]
 fn emit_array_size(
     func_env: &mut FuncEnvironment<'_>,
     builder: &mut FunctionBuilder<'_>,
@@ -1181,7 +1313,10 @@ fn emit_array_size(
 
 /// Common helper for struct-field initialization that can be reused across
 /// collectors.
-#[cfg_attr(not(any(feature = "gc-drc", feature = "gc-null")), allow(dead_code))]
+#[cfg_attr(
+    not(any(feature = "gc-drc", feature = "gc-null")),
+    expect(dead_code, reason = "easier to define")
+)]
 fn initialize_struct_fields(
     func_env: &mut FuncEnvironment<'_>,
     builder: &mut FunctionBuilder<'_>,
@@ -1196,17 +1331,19 @@ fn initialize_struct_fields(
         ir::Value,
     ) -> WasmResult<()>,
 ) -> WasmResult<()> {
-    let struct_layout = func_env.struct_layout(struct_ty);
+    let struct_layout = func_env.struct_or_exn_layout(struct_ty);
     let struct_size = struct_layout.size;
     let field_offsets: SmallVec<[_; 8]> = struct_layout.fields.iter().map(|f| f.offset).collect();
     assert_eq!(field_offsets.len(), field_values.len());
 
     assert!(!func_env.types[struct_ty].composite_type.shared);
-    let struct_ty = func_env.types[struct_ty]
-        .composite_type
-        .inner
-        .unwrap_struct();
-    let field_types: SmallVec<[_; 8]> = struct_ty.fields.iter().cloned().collect();
+    let fields = match &func_env.types[struct_ty].composite_type.inner {
+        WasmCompositeInnerType::Struct(s) => &s.fields,
+        WasmCompositeInnerType::Exn(e) => &e.fields,
+        _ => panic!("Not a struct or exception type"),
+    };
+
+    let field_types: SmallVec<[_; 8]> = fields.iter().cloned().collect();
     assert_eq!(field_types.len(), field_values.len());
 
     for ((ty, val), offset) in field_types.into_iter().zip(field_values).zip(field_offsets) {
@@ -1240,44 +1377,89 @@ impl FuncEnvironment<'_> {
         self.gc_layout(type_index).unwrap_array()
     }
 
-    /// Get the `GcStructLayout` for the struct type at the given `type_index`.
-    fn struct_layout(&mut self, type_index: ModuleInternedTypeIndex) -> &GcStructLayout {
-        self.gc_layout(type_index).unwrap_struct()
+    /// Get the `GcStructLayout` for the struct or exception type at the given `type_index`.
+    fn struct_or_exn_layout(&mut self, type_index: ModuleInternedTypeIndex) -> &GcStructLayout {
+        let result = self.gc_layout(type_index).unwrap_struct();
+        result
     }
 
-    /// Get the GC heap's base pointer.
+    /// Get or create the global for our GC heap's base pointer.
+    fn get_gc_heap_base_global(&mut self, func: &mut ir::Function) -> ir::GlobalValue {
+        if let Some(base) = self.gc_heap_base {
+            return base;
+        }
+
+        let store_context_ptr = self.get_vmstore_context_ptr_global(func);
+        let offset = self.offsets.ptr.vmstore_context_gc_heap_base();
+
+        let mut flags = ir::MemFlags::trusted();
+        if !self
+            .tunables
+            .gc_heap_memory_type()
+            .memory_may_move(self.tunables)
+        {
+            flags.set_readonly();
+            flags.set_can_move();
+        }
+
+        let base = func.create_global_value(ir::GlobalValueData::Load {
+            base: store_context_ptr,
+            offset: Offset32::new(offset.into()),
+            global_type: self.pointer_type(),
+            flags,
+        });
+
+        self.gc_heap_base = Some(base);
+        base
+    }
+
+    /// Get the GC heap's base.
+    #[cfg(any(feature = "gc-null", feature = "gc-drc"))]
     fn get_gc_heap_base(&mut self, builder: &mut FunctionBuilder) -> ir::Value {
-        let ptr_ty = self.pointer_type();
-        let flags = ir::MemFlags::trusted().with_readonly().with_can_move();
+        let global = self.get_gc_heap_base_global(&mut builder.func);
+        builder.ins().global_value(self.pointer_type(), global)
+    }
 
-        let vmctx = self.vmctx(builder.func);
-        let vmctx = builder.ins().global_value(ptr_ty, vmctx);
-
-        let base_offset = self.offsets.ptr.vmctx_gc_heap_base();
-        let base_offset = i32::from(base_offset);
-
-        builder.ins().load(ptr_ty, flags, vmctx, base_offset)
+    fn get_gc_heap_bound_global(&mut self, func: &mut ir::Function) -> ir::GlobalValue {
+        if let Some(bound) = self.gc_heap_bound {
+            return bound;
+        }
+        let store_context_ptr = self.get_vmstore_context_ptr_global(func);
+        let offset = self.offsets.ptr.vmstore_context_gc_heap_current_length();
+        let bound = func.create_global_value(ir::GlobalValueData::Load {
+            base: store_context_ptr,
+            offset: Offset32::new(offset.into()),
+            global_type: self.pointer_type(),
+            flags: ir::MemFlags::trusted(),
+        });
+        self.gc_heap_bound = Some(bound);
+        bound
     }
 
     /// Get the GC heap's bound.
+    #[cfg(feature = "gc-null")]
     fn get_gc_heap_bound(&mut self, builder: &mut FunctionBuilder) -> ir::Value {
-        let ptr_ty = self.pointer_type();
-        let flags = ir::MemFlags::trusted().with_readonly().with_can_move();
-
-        let vmctx = self.vmctx(builder.func);
-        let vmctx = builder.ins().global_value(ptr_ty, vmctx);
-
-        let bound_offset = self.offsets.ptr.vmctx_gc_heap_bound();
-        let bound_offset = i32::from(bound_offset);
-
-        builder.ins().load(ptr_ty, flags, vmctx, bound_offset)
+        let global = self.get_gc_heap_bound_global(&mut builder.func);
+        builder.ins().global_value(self.pointer_type(), global)
     }
 
-    /// Get the GC heap's base pointer and bound.
-    fn get_gc_heap_base_bound(&mut self, builder: &mut FunctionBuilder) -> (ir::Value, ir::Value) {
-        let base = self.get_gc_heap_base(builder);
-        let bound = self.get_gc_heap_bound(builder);
-        (base, bound)
+    /// Get or create the `Heap` for our GC heap.
+    fn get_gc_heap(&mut self, func: &mut ir::Function) -> Heap {
+        if let Some(heap) = self.gc_heap {
+            return heap;
+        }
+
+        let base = self.get_gc_heap_base_global(func);
+        let bound = self.get_gc_heap_bound_global(func);
+        let memory = self.tunables.gc_heap_memory_type();
+        let heap = self.heaps.push(HeapData {
+            base,
+            bound,
+            pcc_memory_type: None,
+            memory,
+        });
+        self.gc_heap = Some(heap);
+        heap
     }
 
     /// Get the raw pointer of `gc_ref[offset]` bounds checked for an access of
@@ -1299,57 +1481,31 @@ impl FuncEnvironment<'_> {
         &mut self,
         builder: &mut FunctionBuilder,
         gc_ref: ir::Value,
-        offset: Offset,
-        check: BoundsCheck,
+        bounds_check: BoundsCheck,
     ) -> ir::Value {
-        log::trace!("prepare_gc_ref_access({gc_ref:?}, {offset:?}, {check:?})");
+        log::trace!("prepare_gc_ref_access({gc_ref:?}, {bounds_check:?})");
         assert_eq!(builder.func.dfg.value_type(gc_ref), ir::types::I32);
 
-        let pointer_type = self.pointer_type();
-        let (base, bound) = self.get_gc_heap_base_bound(builder);
-        let index = uextend_i32_to_pointer_type(builder, pointer_type, gc_ref);
-
-        let offset = match offset {
-            Offset::Dynamic(offset) => uextend_i32_to_pointer_type(builder, pointer_type, offset),
-            Offset::Static(offset) => builder.ins().iconst(pointer_type, i64::from(offset)),
-        };
-
-        let index_and_offset =
-            builder
-                .ins()
-                .uadd_overflow_trap(index, offset, TRAP_INTERNAL_ASSERT);
-
-        let end = match check {
-            BoundsCheck::Object(object_size) => {
-                // Check that `index + object_size` is in bounds. This can be
-                // deduplicated across multiple accesses to different fields
-                // within the same object.
-                let object_size = uextend_i32_to_pointer_type(builder, pointer_type, object_size);
-                builder
-                    .ins()
-                    .uadd_overflow_trap(index, object_size, TRAP_INTERNAL_ASSERT)
-            }
-            BoundsCheck::Access(access_size) => {
-                // Check that `index + offset + access_size` is in bounds.
-                let access_size = builder.ins().iconst(pointer_type, i64::from(access_size));
-                builder.ins().uadd_overflow_trap(
-                    index_and_offset,
-                    access_size,
-                    TRAP_INTERNAL_ASSERT,
-                )
+        let gc_heap = self.get_gc_heap(&mut builder.func);
+        let gc_heap = self.heaps[gc_heap].clone();
+        let result = match crate::bounds_checks::bounds_check_and_compute_addr(
+            builder,
+            self,
+            &gc_heap,
+            gc_ref,
+            bounds_check,
+            crate::TRAP_INTERNAL_ASSERT,
+        ) {
+            Reachability::Reachable(v) => v,
+            Reachability::Unreachable => {
+                // We are now in unreachable code, but we don't want to plumb
+                // through a bunch of `Reachability` through all of our callers,
+                // so just assert we won't reach here and return `null`
+                let null = builder.ins().iconst(self.pointer_type(), 0);
+                builder.ins().trapz(null, crate::TRAP_INTERNAL_ASSERT);
+                null
             }
         };
-
-        let is_in_bounds =
-            builder
-                .ins()
-                .icmp(ir::condcodes::IntCC::UnsignedLessThanOrEqual, end, bound);
-        builder.ins().trapz(is_in_bounds, TRAP_INTERNAL_ASSERT);
-
-        // NB: No need to check for overflow here, as that would mean that the
-        // GC heap is hanging off the end of the address space, which is
-        // impossible.
-        let result = builder.ins().iadd(base, index_and_offset);
         log::trace!("prepare_gc_ref_access(..) -> {result:?}");
         result
     }
@@ -1364,7 +1520,10 @@ impl FuncEnvironment<'_> {
     /// reference is null or is an `i31ref`; otherwise, it will be zero.
     ///
     /// This method is collector-agnostic.
-    #[cfg_attr(not(feature = "gc-drc"), allow(dead_code))]
+    #[cfg_attr(
+        not(feature = "gc-drc"),
+        expect(dead_code, reason = "easier to define")
+    )]
     fn gc_ref_is_null_or_i31(
         &mut self,
         builder: &mut FunctionBuilder,
@@ -1401,14 +1560,17 @@ impl FuncEnvironment<'_> {
             // Can only ever be `null`.
             WasmHeapType::NoExtern => false,
 
+            WasmHeapType::Exn | WasmHeapType::ConcreteExn(_) | WasmHeapType::NoExn => false,
+
             // Wrong type hierarchy, and also funcrefs are not GC-managed
             // types. Should have been caught by the assertion at the start of
             // the function.
             WasmHeapType::Func | WasmHeapType::ConcreteFunc(_) | WasmHeapType::NoFunc => {
                 unreachable!()
             }
-
-            WasmHeapType::Cont | WasmHeapType::ConcreteCont(_) | WasmHeapType::NoCont => todo!(), // FIXME: #10248 stack switching support.
+            WasmHeapType::Cont | WasmHeapType::ConcreteCont(_) | WasmHeapType::NoCont => {
+                unreachable!()
+            }
         };
 
         match (ty.nullable, might_be_i31) {
@@ -1450,9 +1612,13 @@ impl FuncEnvironment<'_> {
         log::trace!("is_subtype: fast path check for exact same types");
         let same_ty = builder.ins().icmp(IntCC::Equal, a, b);
         let same_ty = builder.ins().uextend(ir::types::I32, same_ty);
-        builder
-            .ins()
-            .brif(same_ty, continue_block, &[same_ty], diff_tys_block, &[]);
+        builder.ins().brif(
+            same_ty,
+            continue_block,
+            &[same_ty.into()],
+            diff_tys_block,
+            &[],
+        );
 
         // Different types block: fall back to the `is_subtype` libcall.
         builder.switch_to_block(diff_tys_block);
@@ -1461,7 +1627,7 @@ impl FuncEnvironment<'_> {
         let vmctx = self.vmctx_val(&mut builder.cursor());
         let call_inst = builder.ins().call(is_subtype, &[vmctx, a, b]);
         let result = builder.func.dfg.first_result(call_inst);
-        builder.ins().jump(continue_block, &[result]);
+        builder.ins().jump(continue_block, &[result.into()]);
 
         // Continue block: join point for the result.
         builder.switch_to_block(continue_block);
