@@ -21,11 +21,11 @@
 use crate::component::dfg::CoreDef;
 use crate::component::{
     Adapter, AdapterOptions as AdapterOptionsDfg, ComponentTypesBuilder, FlatType, InterfaceType,
-    StringEncoding, Transcode, TypeFuncIndex,
+    RuntimeComponentInstanceIndex, StringEncoding, Transcode, TypeFuncIndex,
 };
 use crate::fact::transcode::Transcoder;
-use crate::prelude::*;
 use crate::{EntityRef, FuncIndex, GlobalIndex, MemoryIndex, PrimaryMap};
+use crate::{ModuleInternedTypeIndex, prelude::*};
 use std::borrow::Cow;
 use std::collections::HashMap;
 use wasm_encoder::*;
@@ -36,6 +36,21 @@ mod trampoline;
 mod transcode;
 mod traps;
 
+/// Fixed parameter types for the `prepare_call` built-in function.
+///
+/// Note that `prepare_call` also takes a variable number of parameters in
+/// addition to these, determined by the signature of the function for which
+/// we're generating an adapter.
+pub static PREPARE_CALL_FIXED_PARAMS: &[ValType] = &[
+    ValType::FUNCREF, // start
+    ValType::FUNCREF, // return
+    ValType::I32,     // caller_instance
+    ValType::I32,     // callee_instance
+    ValType::I32,     // task_return_type
+    ValType::I32,     // string_encoding
+    ValType::I32,     // result_count_or_max_if_async
+];
+
 /// Representation of an adapter module.
 pub struct Module<'a> {
     /// Whether or not debug code is inserted into the adapters themselves.
@@ -65,6 +80,15 @@ pub struct Module<'a> {
     imported_resource_enter_call: Option<FuncIndex>,
     imported_resource_exit_call: Option<FuncIndex>,
 
+    // Cached versions of imported trampolines for working with the async ABI.
+    imported_async_start_calls: HashMap<(Option<FuncIndex>, Option<FuncIndex>), FuncIndex>,
+
+    // Cached versions of imported trampolines for working with `stream`s,
+    // `future`s, and `error-context`s.
+    imported_future_transfer: Option<FuncIndex>,
+    imported_stream_transfer: Option<FuncIndex>,
+    imported_error_context_transfer: Option<FuncIndex>,
+
     // Current status of index spaces from the imports generated so far.
     imported_funcs: PrimaryMap<FuncIndex, Option<CoreDef>>,
     imported_memories: PrimaryMap<MemoryIndex, CoreDef>,
@@ -73,6 +97,8 @@ pub struct Module<'a> {
     funcs: PrimaryMap<FunctionId, Function>,
     helper_funcs: HashMap<Helper, FunctionId>,
     helper_worklist: Vec<(FunctionId, Helper)>,
+
+    exports: Vec<(u32, String)>,
 }
 
 struct AdapterData {
@@ -95,6 +121,7 @@ struct AdapterData {
 /// These options are typically unique per-adapter and generally aren't needed
 /// when translating recursive types within an adapter.
 struct AdapterOptions {
+    instance: RuntimeComponentInstanceIndex,
     /// The ascribed type of this adapter.
     ty: TypeFuncIndex,
     /// The global that represents the instance flags for where this adapter
@@ -106,14 +133,9 @@ struct AdapterOptions {
     options: Options,
 }
 
-/// This type is split out of `AdapterOptions` and is specifically used to
-/// deduplicate translation functions within a module. Consequently this has
-/// as few fields as possible to minimize the number of functions generated
-/// within an adapter module.
 #[derive(PartialEq, Eq, Hash, Copy, Clone)]
-struct Options {
-    /// The encoding that strings use from this adapter.
-    string_encoding: StringEncoding,
+/// Linear memory.
+struct LinearMemoryOptions {
     /// Whether or not the `memory` field, if present, is a 64-bit memory.
     memory64: bool,
     /// An optionally-specified memory where values may travel through for
@@ -122,13 +144,51 @@ struct Options {
     /// An optionally-specified function to be used to allocate space for
     /// types such as strings as they go into a module.
     realloc: Option<FuncIndex>,
-    callback: Option<FuncIndex>,
-    async_: bool,
 }
 
-enum Context {
-    Lift,
-    Lower,
+impl LinearMemoryOptions {
+    fn ptr(&self) -> ValType {
+        if self.memory64 {
+            ValType::I64
+        } else {
+            ValType::I32
+        }
+    }
+
+    fn ptr_size(&self) -> u8 {
+        if self.memory64 { 8 } else { 4 }
+    }
+}
+
+/// The data model for objects passed through an adapter.
+#[derive(PartialEq, Eq, Hash, Copy, Clone)]
+enum DataModel {
+    Gc {},
+    LinearMemory(LinearMemoryOptions),
+}
+
+impl DataModel {
+    #[track_caller]
+    fn unwrap_memory(&self) -> &LinearMemoryOptions {
+        match self {
+            DataModel::Gc {} => panic!("`unwrap_memory` on GC"),
+            DataModel::LinearMemory(opts) => opts,
+        }
+    }
+}
+
+/// This type is split out of `AdapterOptions` and is specifically used to
+/// deduplicate translation functions within a module. Consequently this has
+/// as few fields as possible to minimize the number of functions generated
+/// within an adapter module.
+#[derive(PartialEq, Eq, Hash, Copy, Clone)]
+struct Options {
+    /// The encoding that strings use from this adapter.
+    string_encoding: StringEncoding,
+    callback: Option<FuncIndex>,
+    async_: bool,
+    core_type: ModuleInternedTypeIndex,
+    data_model: DataModel,
 }
 
 /// Representation of a "helper function" which may be generated as part of
@@ -166,6 +226,12 @@ enum HelperLocation {
     Stack,
     /// Located in linear memory as configured by `opts`.
     Memory,
+    /// Located in a GC struct field.
+    #[expect(dead_code, reason = "CM+GC is still WIP")]
+    StructField,
+    /// Located in a GC array element.
+    #[expect(dead_code, reason = "CM+GC is still WIP")]
+    ArrayElement,
 }
 
 impl<'a> Module<'a> {
@@ -189,6 +255,11 @@ impl<'a> Module<'a> {
             imported_resource_transfer_borrow: None,
             imported_resource_enter_call: None,
             imported_resource_exit_call: None,
+            imported_async_start_calls: HashMap::new(),
+            imported_future_transfer: None,
+            imported_stream_transfer: None,
+            imported_error_context_transfer: None,
+            exports: Vec::new(),
         }
     }
 
@@ -209,7 +280,7 @@ impl<'a> Module<'a> {
         // Import the core wasm function which was lifted using its appropriate
         // signature since the exported function this adapter generates will
         // call the lifted function.
-        let signature = self.types.signature(&lift, Context::Lift);
+        let signature = self.types.signature(&lift);
         let ty = self
             .core_types
             .function(&signature.params, &signature.results);
@@ -246,13 +317,14 @@ impl<'a> Module<'a> {
         let AdapterOptionsDfg {
             instance,
             string_encoding,
-            memory,
-            memory64,
-            realloc,
             post_return: _, // handled above
             callback,
             async_,
+            core_type,
+            data_model,
+            cancellable,
         } = options;
+        assert!(!cancellable);
 
         let flags = self.import_global(
             "flags",
@@ -264,44 +336,54 @@ impl<'a> Module<'a> {
             },
             CoreDef::InstanceFlags(*instance),
         );
-        let memory = memory.as_ref().map(|memory| {
-            self.import_memory(
-                "memory",
-                &format!("m{}", self.imported_memories.len()),
-                MemoryType {
-                    minimum: 0,
-                    maximum: None,
-                    shared: false,
+
+        let data_model = match data_model {
+            crate::component::DataModel::Gc {} => DataModel::Gc {},
+            crate::component::DataModel::LinearMemory {
+                memory,
+                memory64,
+                realloc,
+            } => {
+                let memory = memory.as_ref().map(|memory| {
+                    self.import_memory(
+                        "memory",
+                        &format!("m{}", self.imported_memories.len()),
+                        MemoryType {
+                            minimum: 0,
+                            maximum: None,
+                            shared: false,
+                            memory64: *memory64,
+                            page_size_log2: None,
+                        },
+                        memory.clone().into(),
+                    )
+                });
+                let realloc = realloc.as_ref().map(|func| {
+                    let ptr = if *memory64 {
+                        ValType::I64
+                    } else {
+                        ValType::I32
+                    };
+                    let ty = self.core_types.function(&[ptr, ptr, ptr, ptr], &[ptr]);
+                    self.import_func(
+                        "realloc",
+                        &format!("f{}", self.imported_funcs.len()),
+                        ty,
+                        func.clone(),
+                    )
+                });
+                DataModel::LinearMemory(LinearMemoryOptions {
                     memory64: *memory64,
-                    page_size_log2: None,
-                },
-                memory.clone().into(),
-            )
-        });
-        let realloc = realloc.as_ref().map(|func| {
-            let ptr = if *memory64 {
-                ValType::I64
-            } else {
-                ValType::I32
-            };
-            let ty = self.core_types.function(&[ptr, ptr, ptr, ptr], &[ptr]);
-            self.import_func(
-                "realloc",
-                &format!("f{}", self.imported_funcs.len()),
-                ty,
-                func.clone(),
-            )
-        });
+                    memory,
+                    realloc,
+                })
+            }
+        };
+
         let callback = callback.as_ref().map(|func| {
-            let ptr = if *memory64 {
-                ValType::I64
-            } else {
-                ValType::I32
-            };
-            let ty = self.core_types.function(
-                &[ptr, ValType::I32, ValType::I32, ValType::I32],
-                &[ValType::I32],
-            );
+            let ty = self
+                .core_types
+                .function(&[ValType::I32, ValType::I32, ValType::I32], &[ValType::I32]);
             self.import_func(
                 "callback",
                 &format!("f{}", self.imported_funcs.len()),
@@ -311,16 +393,16 @@ impl<'a> Module<'a> {
         });
 
         AdapterOptions {
+            instance: *instance,
             ty,
             flags,
             post_return: None,
             options: Options {
                 string_encoding: *string_encoding,
-                memory64: *memory64,
-                memory,
-                realloc,
                 callback,
                 async_: *async_,
+                core_type: *core_type,
+                data_model,
             },
         }
     }
@@ -407,9 +489,30 @@ impl<'a> Module<'a> {
         results: &[ValType],
         import: Import,
         get: impl Fn(&mut Self) -> &mut Option<FuncIndex>,
+    ) -> FuncIndex {
+        self.import_simple_get_and_set(
+            module,
+            name,
+            params,
+            results,
+            import,
+            |me| *get(me),
+            |me, v| *get(me) = Some(v),
+        )
+    }
+
+    fn import_simple_get_and_set(
+        &mut self,
+        module: &str,
+        name: &str,
+        params: &[ValType],
+        results: &[ValType],
+        import: Import,
+        get: impl Fn(&mut Self) -> Option<FuncIndex>,
+        set: impl Fn(&mut Self, FuncIndex),
     ) -> FuncIndex {
         if let Some(idx) = get(self) {
-            return *idx;
+            return idx;
         }
         let ty = self.core_types.function(params, results);
         let ty = EntityType::Function(ty);
@@ -417,10 +520,154 @@ impl<'a> Module<'a> {
 
         self.imports.push(import);
         let idx = self.imported_funcs.push(None);
-        *get(self) = Some(idx);
+        set(self, idx);
         idx
     }
 
+    /// Import a host built-in function to set up a subtask for a sync-lowered
+    /// import call to an async-lifted export.
+    ///
+    /// Given that the callee may exert backpressure before the host can copy
+    /// the parameters, the adapter must use this function to set up the subtask
+    /// and stash the parameters as part of that subtask until any backpressure
+    /// has cleared.
+    fn import_prepare_call(
+        &mut self,
+        suffix: &str,
+        params: &[ValType],
+        memory: Option<MemoryIndex>,
+    ) -> FuncIndex {
+        let ty = self.core_types.function(
+            &PREPARE_CALL_FIXED_PARAMS
+                .iter()
+                .copied()
+                .chain(params.iter().copied())
+                .collect::<Vec<_>>(),
+            &[],
+        );
+        self.core_imports.import(
+            "sync",
+            &format!("[prepare-call]{suffix}"),
+            EntityType::Function(ty),
+        );
+        let import = Import::PrepareCall {
+            memory: memory.map(|v| self.imported_memories[v].clone()),
+        };
+        self.imports.push(import);
+        self.imported_funcs.push(None)
+    }
+
+    /// Import a host built-in function to start a subtask for a sync-lowered
+    /// import call to an async-lifted export.
+    ///
+    /// This call with block until the subtask has produced result(s) via the
+    /// `task.return` intrinsic.
+    ///
+    /// Note that this could potentially be combined with the `sync-prepare`
+    /// built-in into a single built-in function that does both jobs.  However,
+    /// we've kept them separate to allow a future optimization where the caller
+    /// calls the callee directly rather than using `sync-start` to have the host
+    /// do it.
+    fn import_sync_start_call(
+        &mut self,
+        suffix: &str,
+        callback: Option<FuncIndex>,
+        results: &[ValType],
+    ) -> FuncIndex {
+        let ty = self
+            .core_types
+            .function(&[ValType::FUNCREF, ValType::I32], results);
+        self.core_imports.import(
+            "sync",
+            &format!("[start-call]{suffix}"),
+            EntityType::Function(ty),
+        );
+        let import = Import::SyncStartCall {
+            callback: callback
+                .map(|callback| self.imported_funcs.get(callback).unwrap().clone().unwrap()),
+        };
+        self.imports.push(import);
+        self.imported_funcs.push(None)
+    }
+
+    /// Import a host built-in function to start a subtask for an async-lowered
+    /// import call to an async- or sync-lifted export.
+    ///
+    /// Note that this could potentially be combined with the `async-prepare`
+    /// built-in into a single built-in function that does both jobs.  However,
+    /// we've kept them separate to allow a future optimization where the caller
+    /// calls the callee directly rather than using `async-start` to have the
+    /// host do it.
+    fn import_async_start_call(
+        &mut self,
+        suffix: &str,
+        callback: Option<FuncIndex>,
+        post_return: Option<FuncIndex>,
+    ) -> FuncIndex {
+        self.import_simple_get_and_set(
+            "async",
+            &format!("[start-call]{suffix}"),
+            &[ValType::FUNCREF, ValType::I32, ValType::I32, ValType::I32],
+            &[ValType::I32],
+            Import::AsyncStartCall {
+                callback: callback
+                    .map(|callback| self.imported_funcs.get(callback).unwrap().clone().unwrap()),
+                post_return: post_return.map(|post_return| {
+                    self.imported_funcs
+                        .get(post_return)
+                        .unwrap()
+                        .clone()
+                        .unwrap()
+                }),
+            },
+            |me| {
+                me.imported_async_start_calls
+                    .get(&(callback, post_return))
+                    .copied()
+            },
+            |me, v| {
+                assert!(
+                    me.imported_async_start_calls
+                        .insert((callback, post_return), v)
+                        .is_none()
+                )
+            },
+        )
+    }
+
+    fn import_future_transfer(&mut self) -> FuncIndex {
+        self.import_simple(
+            "future",
+            "transfer",
+            &[ValType::I32; 3],
+            &[ValType::I32],
+            Import::FutureTransfer,
+            |me| &mut me.imported_future_transfer,
+        )
+    }
+
+    fn import_stream_transfer(&mut self) -> FuncIndex {
+        self.import_simple(
+            "stream",
+            "transfer",
+            &[ValType::I32; 3],
+            &[ValType::I32],
+            Import::StreamTransfer,
+            |me| &mut me.imported_stream_transfer,
+        )
+    }
+
+    fn import_error_context_transfer(&mut self) -> FuncIndex {
+        self.import_simple(
+            "error-context",
+            "transfer",
+            &[ValType::I32; 3],
+            &[ValType::I32],
+            Import::ErrorContextTransfer,
+            |me| &mut me.imported_error_context_transfer,
+        )
+    }
+
     fn import_resource_transfer_own(&mut self) -> FuncIndex {
         self.import_simple(
             "resource",
@@ -496,6 +743,9 @@ impl<'a> Module<'a> {
                 exports.export(name, ExportKind::Func, idx.as_u32());
             }
         }
+        for (idx, name) in &self.exports {
+            exports.export(name, ExportKind::Func, *idx);
+        }
 
         // With all functions numbered the fragments of the body of each
         // function can be assigned into one final adapter function.
@@ -528,6 +778,9 @@ impl<'a> Module<'a> {
                     Body::Call(id) => {
                         Instruction::Call(id_to_index[*id].as_u32()).encode(&mut body);
                     }
+                    Body::RefFunc(id) => {
+                        Instruction::RefFunc(id_to_index[*id].as_u32()).encode(&mut body);
+                    }
                 }
             }
             code.raw(&body);
@@ -585,12 +838,23 @@ pub enum Import {
     /// Tears down a previous entry and handles checking borrow-related
     /// metadata.
     ResourceExitCall,
-    /// An intrinsic used by FACT-generated modules to begin a call to an
-    /// async-lowered import function.
-    AsyncEnterCall,
-    /// An intrinsic used by FACT-generated modules to complete a call to an
-    /// async-lowered import function.
-    AsyncExitCall {
+    /// An intrinsic used by FACT-generated modules to begin a call involving
+    /// an async-lowered import and/or an async-lifted export.
+    PrepareCall {
+        /// The memory used to verify that the memory specified for the
+        /// `task.return` that is called at runtime (if any) matches the one
+        /// specified in the lifted export.
+        memory: Option<CoreDef>,
+    },
+    /// An intrinsic used by FACT-generated modules to complete a call involving
+    /// a sync-lowered import and async-lifted export.
+    SyncStartCall {
+        /// The callee's callback function, if any.
+        callback: Option<CoreDef>,
+    },
+    /// An intrinsic used by FACT-generated modules to complete a call involving
+    /// an async-lowered import function.
+    AsyncStartCall {
         /// The callee's callback function, if any.
         callback: Option<CoreDef>,
 
@@ -609,33 +873,20 @@ pub enum Import {
 }
 
 impl Options {
-    fn ptr(&self) -> ValType {
-        if self.memory64 {
-            ValType::I64
-        } else {
-            ValType::I32
-        }
-    }
-
-    fn ptr_size(&self) -> u8 {
-        if self.memory64 {
-            8
-        } else {
-            4
-        }
-    }
-
     fn flat_types<'a>(
         &self,
         ty: &InterfaceType,
         types: &'a ComponentTypesBuilder,
     ) -> Option<&'a [FlatType]> {
         let flat = types.flat_types(ty)?;
-        Some(if self.memory64 {
-            flat.memory64
-        } else {
-            flat.memory32
-        })
+        match self.data_model {
+            DataModel::Gc {} => todo!("CM+GC"),
+            DataModel::LinearMemory(mem_opts) => Some(if mem_opts.memory64 {
+                flat.memory64
+            } else {
+                flat.memory32
+            }),
+        }
     }
 }
 
@@ -704,6 +955,7 @@ struct Function {
 enum Body {
     Raw(Vec<u8>, Vec<(usize, traps::Trap)>),
     Call(FunctionId),
+    RefFunc(FunctionId),
 }
 
 impl Function {
@@ -736,7 +988,8 @@ impl Helper {
         // stack-based representation.
         match self.dst.loc {
             HelperLocation::Stack => self.dst.push_flat(&mut results, types),
-            HelperLocation::Memory => params.push(self.dst.opts.ptr()),
+            HelperLocation::Memory => params.push(self.dst.opts.data_model.unwrap_memory().ptr()),
+            HelperLocation::StructField | HelperLocation::ArrayElement => todo!("CM+GC"),
         }
 
         core_types.function(&params, &results)
@@ -752,8 +1005,9 @@ impl HelperType {
                 }
             }
             HelperLocation::Memory => {
-                dst.push(self.opts.ptr());
+                dst.push(self.opts.data_model.unwrap_memory().ptr());
             }
+            HelperLocation::StructField | HelperLocation::ArrayElement => todo!("CM+GC"),
         }
     }
 }
