@@ -1,6 +1,6 @@
-use crate::TripleExt;
-use anyhow::{anyhow, bail, Result};
-use core::fmt;
+use crate::{IndexType, Limits, Memory, TripleExt};
+use anyhow::{Error, Result, anyhow, bail};
+use core::{fmt, str::FromStr};
 use serde_derive::{Deserialize, Serialize};
 use target_lexicon::{PointerWidth, Triple};
 
@@ -74,7 +74,11 @@ define_tunables! {
         pub memory_reservation_for_growth: u64,
 
         /// Whether or not to generate native DWARF debug information.
-        pub generate_native_debuginfo: bool,
+        pub debug_native: bool,
+
+        /// Whether we are enabling precise Wasm-level debugging in
+        /// the guest.
+        pub debug_guest: bool,
 
         /// Whether or not to retain DWARF sections in compiled modules.
         pub parse_wasm_debuginfo: bool,
@@ -121,6 +125,21 @@ define_tunables! {
 
         /// Whether CoW images might be used to initialize linear memories.
         pub memory_init_cow: bool,
+
+        /// Whether to enable inlining in Wasmtime's compilation orchestration
+        /// or not.
+        pub inlining: bool,
+
+        /// Whether to inline calls within the same core Wasm module or not.
+        pub inlining_intra_module: IntraModuleInlining,
+
+        /// The size of "small callees" that can be inlined regardless of the
+        /// caller's size.
+        pub inlining_small_callee_size: u32,
+
+        /// The general size threshold for the sum of the caller's and callee's
+        /// sizes, past which we will generally not inline calls anymore.
+        pub inlining_sum_size_threshold: u32,
     }
 
     pub struct ConfigTunables {
@@ -178,7 +197,7 @@ impl Tunables {
 
             // General options which have the same defaults regardless of
             // architecture.
-            generate_native_debuginfo: false,
+            debug_native: false,
             parse_wasm_debuginfo: true,
             consume_fuel: false,
             epoch_interruption: false,
@@ -191,6 +210,11 @@ impl Tunables {
             winch_callable: false,
             signals_based_traps: false,
             memory_init_cow: true,
+            inlining: false,
+            inlining_intra_module: IntraModuleInlining::WhenUsingGc,
+            inlining_small_callee_size: 50,
+            inlining_sum_size_threshold: 2000,
+            debug_guest: false,
         }
     }
 
@@ -234,6 +258,20 @@ impl Tunables {
             ..Tunables::default_miri()
         }
     }
+
+    /// Get the GC heap's memory type, given our configured tunables.
+    pub fn gc_heap_memory_type(&self) -> Memory {
+        Memory {
+            idx_type: IndexType::I32,
+            limits: Limits { min: 0, max: None },
+            shared: false,
+            // We *could* try to match the target architecture's page size, but that
+            // would require exercising a page size for memories that we don't
+            // otherwise support for Wasm; we conservatively avoid that, and just
+            // use the default Wasm page size, for now.
+            page_size_log2: 16,
+        }
+    }
 }
 
 /// The garbage collector implementation to use.
@@ -253,3 +291,28 @@ impl fmt::Display for Collector {
         }
     }
 }
+
+/// Whether to inline function calls within the same module.
+#[derive(Clone, Copy, Hash, Serialize, Deserialize, Debug, PartialEq, Eq)]
+#[expect(missing_docs, reason = "self-describing variants")]
+pub enum IntraModuleInlining {
+    Yes,
+    No,
+    WhenUsingGc,
+}
+
+impl FromStr for IntraModuleInlining {
+    type Err = Error;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        match s {
+            "y" | "yes" | "true" => Ok(Self::Yes),
+            "n" | "no" | "false" => Ok(Self::No),
+            "gc" => Ok(Self::WhenUsingGc),
+            _ => bail!(
+                "invalid intra-module inlining option string: `{s}`, \
+                 only yes,no,gc accepted"
+            ),
+        }
+    }
+}
