@@ -1,4 +1,4 @@
-use crate::{wasm_unsupported, Tunables, WasmResult};
+use crate::{Tunables, WasmResult, wasm_unsupported};
 use alloc::borrow::Cow;
 use alloc::boxed::Box;
 use core::{fmt, ops::Range};
@@ -232,6 +232,16 @@ impl WasmValType {
             size => panic!("invalid int bits for WasmValType: {size}"),
         }
     }
+
+    /// Returns the contained reference type.
+    ///
+    /// Panics if the value type is not a vmgcref
+    pub fn unwrap_ref_type(&self) -> WasmRefType {
+        match self {
+            WasmValType::Ref(ref_type) => *ref_type,
+            _ => panic!("Called WasmValType::unwrap_ref_type on non-reference type"),
+        }
+    }
 }
 
 /// WebAssembly reference type -- equivalent of `wasmparser`'s RefType
@@ -362,6 +372,7 @@ impl EngineOrModuleTypeIndex {
     }
 
     /// Get the underlying engine-level type index, if any.
+    #[inline]
     pub fn as_engine_type_index(self) -> Option<VMSharedTypeIndex> {
         match self {
             Self::Engine(e) => Some(e),
@@ -371,6 +382,7 @@ impl EngineOrModuleTypeIndex {
 
     /// Get the underlying engine-level type index, or panic.
     #[track_caller]
+    #[inline]
     pub fn unwrap_engine_type_index(self) -> VMSharedTypeIndex {
         match self.as_engine_type_index() {
             Some(x) => x,
@@ -436,6 +448,11 @@ pub enum WasmHeapType {
     ConcreteFunc(EngineOrModuleTypeIndex),
     NoFunc,
 
+    // Exception types.
+    Exn,
+    ConcreteExn(EngineOrModuleTypeIndex),
+    NoExn,
+
     // Continuation types.
     Cont,
     ConcreteCont(EngineOrModuleTypeIndex),
@@ -460,6 +477,7 @@ impl From<WasmHeapTopType> for WasmHeapType {
             WasmHeapTopType::Any => Self::Any,
             WasmHeapTopType::Func => Self::Func,
             WasmHeapTopType::Cont => Self::Cont,
+            WasmHeapTopType::Exn => Self::Exn,
         }
     }
 }
@@ -472,6 +490,7 @@ impl From<WasmHeapBottomType> for WasmHeapType {
             WasmHeapBottomType::None => Self::None,
             WasmHeapBottomType::NoFunc => Self::NoFunc,
             WasmHeapBottomType::NoCont => Self::NoCont,
+            WasmHeapBottomType::NoExn => Self::NoExn,
         }
     }
 }
@@ -494,6 +513,9 @@ impl fmt::Display for WasmHeapType {
             Self::ConcreteArray(i) => write!(f, "array {i}"),
             Self::Struct => write!(f, "struct"),
             Self::ConcreteStruct(i) => write!(f, "struct {i}"),
+            Self::Exn => write!(f, "exn"),
+            Self::ConcreteExn(i) => write!(f, "exn {i}"),
+            Self::NoExn => write!(f, "noexn"),
             Self::None => write!(f, "none"),
         }
     }
@@ -532,9 +554,10 @@ impl WasmHeapType {
     #[inline]
     pub fn is_vmgcref_type(&self) -> bool {
         match self.top() {
-            // All `t <: (ref null any)` and `t <: (ref null extern)` are
-            // represented as `VMGcRef`s.
-            WasmHeapTopType::Any | WasmHeapTopType::Extern => true,
+            // All `t <: (ref null any)`, `t <: (ref null extern)`,
+            // and `t <: (ref null exn)` are represented as
+            // `VMGcRef`s.
+            WasmHeapTopType::Any | WasmHeapTopType::Extern | WasmHeapTopType::Exn => true,
 
             // All `t <: (ref null func)` are not.
             WasmHeapTopType::Func => false,
@@ -572,6 +595,10 @@ impl WasmHeapType {
                 WasmHeapTopType::Cont
             }
 
+            WasmHeapType::Exn | WasmHeapType::ConcreteExn(_) | WasmHeapType::NoExn => {
+                WasmHeapTopType::Exn
+            }
+
             WasmHeapType::Any
             | WasmHeapType::Eq
             | WasmHeapType::I31
@@ -603,6 +630,10 @@ impl WasmHeapType {
                 WasmHeapBottomType::NoCont
             }
 
+            WasmHeapType::Exn | WasmHeapType::ConcreteExn(_) | WasmHeapType::NoExn => {
+                WasmHeapBottomType::NoExn
+            }
+
             WasmHeapType::Any
             | WasmHeapType::Eq
             | WasmHeapType::I31
@@ -616,7 +647,7 @@ impl WasmHeapType {
 }
 
 /// A top heap type.
-#[derive(Debug, Clone, Copy, Eq, PartialEq)]
+#[derive(Debug, Clone, Copy, Eq, PartialEq, Hash, Serialize, Deserialize)]
 pub enum WasmHeapTopType {
     /// The common supertype of all external references.
     Extern,
@@ -624,6 +655,8 @@ pub enum WasmHeapTopType {
     Any,
     /// The common supertype of all function references.
     Func,
+    /// The common supertype of all exception references.
+    Exn,
     /// The common supertype of all continuation references.
     Cont,
 }
@@ -637,6 +670,8 @@ pub enum WasmHeapBottomType {
     None,
     /// The common subtype of all function references.
     NoFunc,
+    /// The common subtype of all exception references.
+    NoExn,
     /// The common subtype of all continuation references.
     NoCont,
 }
@@ -801,6 +836,15 @@ impl WasmContType {
     pub fn new(idx: EngineOrModuleTypeIndex) -> Self {
         WasmContType(idx)
     }
+
+    /// Returns the (module interned) index to the underlying function type.
+    pub fn unwrap_module_type_index(self) -> ModuleInternedTypeIndex {
+        match self.0 {
+            EngineOrModuleTypeIndex::Engine(_) => panic!("not module interned"),
+            EngineOrModuleTypeIndex::Module(idx) => idx,
+            EngineOrModuleTypeIndex::RecGroup(_) => todo!(),
+        }
+    }
 }
 
 impl TypeTrace for WasmContType {
@@ -819,6 +863,79 @@ impl TypeTrace for WasmContType {
     }
 }
 
+/// WebAssembly exception type.
+///
+/// This "exception type" is not a Wasm language-level
+/// concept. Instead, it denotes an *exception object signature* --
+/// the types of the payload values.
+///
+/// In contrast, at the Wasm language level, exception objects are
+/// associated with specific tags, and these tags refer to their
+/// signatures (function types). However, tags are *nominal*: like
+/// memories and tables, a separate instance of a tag exists for every
+/// instance of the defining module, and these tag instances can be
+/// imported and exported. At runtime we handle tags like we do
+/// memories and tables, but these runtime instances do not exist in
+/// the type system here.
+///
+/// Because the Wasm type system does not have concrete `exn` types
+/// (i.e., the heap-type lattice has only top `exn` and bottom
+/// `noexn`), we are free to decide what we mean by "concrete type"
+/// here. Thus, we define an "exception type" to refer to the
+/// type-level *signature*. When a particular *exception object* is
+/// created in a store, it can be associated with a particular *tag
+/// instance* also in that store, and the compatibility is checked
+/// (the tag's function type must match the function type in the
+/// associated WasmExnType).
+#[derive(Debug, Clone, Eq, PartialEq, Hash, Serialize, Deserialize)]
+pub struct WasmExnType {
+    /// The function type from which we get our signature. We hold
+    /// this directly so that we can efficiently derive a FuncType
+    /// without re-interning the field types.
+    pub func_ty: EngineOrModuleTypeIndex,
+    /// The fields (payload values) that make up this exception type.
+    ///
+    /// While we could obtain these by looking up the `func_ty` above,
+    /// we also need to be able to derive a GC object layout from this
+    /// type descriptor without referencing other type descriptors; so
+    /// we directly inline the information here.
+    pub fields: Box<[WasmFieldType]>,
+}
+
+impl fmt::Display for WasmExnType {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "(exn ({})", self.func_ty)?;
+        for ty in self.fields.iter() {
+            write!(f, " {ty}")?;
+        }
+        write!(f, ")")
+    }
+}
+
+impl TypeTrace for WasmExnType {
+    fn trace<F, E>(&self, func: &mut F) -> Result<(), E>
+    where
+        F: FnMut(EngineOrModuleTypeIndex) -> Result<(), E>,
+    {
+        func(self.func_ty)?;
+        for f in self.fields.iter() {
+            f.trace(func)?;
+        }
+        Ok(())
+    }
+
+    fn trace_mut<F, E>(&mut self, func: &mut F) -> Result<(), E>
+    where
+        F: FnMut(&mut EngineOrModuleTypeIndex) -> Result<(), E>,
+    {
+        func(&mut self.func_ty)?;
+        for f in self.fields.iter_mut() {
+            f.trace_mut(func)?;
+        }
+        Ok(())
+    }
+}
+
 /// Represents storage types introduced in the GC spec for array and struct fields.
 #[derive(Debug, Copy, Clone, Eq, PartialEq, Hash, Serialize, Deserialize)]
 pub enum WasmStorageType {
@@ -1008,6 +1125,7 @@ pub enum WasmCompositeInnerType {
     Func(WasmFuncType),
     Struct(WasmStructType),
     Cont(WasmContType),
+    Exn(WasmExnType),
 }
 
 impl fmt::Display for WasmCompositeInnerType {
@@ -1017,6 +1135,7 @@ impl fmt::Display for WasmCompositeInnerType {
             Self::Func(ty) => fmt::Display::fmt(ty, f),
             Self::Struct(ty) => fmt::Display::fmt(ty, f),
             Self::Cont(ty) => fmt::Display::fmt(ty, f),
+            Self::Exn(ty) => fmt::Display::fmt(ty, f),
         }
     }
 }
@@ -1094,6 +1213,24 @@ impl WasmCompositeInnerType {
     pub fn unwrap_cont(&self) -> &WasmContType {
         self.as_cont().unwrap()
     }
+
+    #[inline]
+    pub fn is_exn(&self) -> bool {
+        matches!(self, Self::Exn(_))
+    }
+
+    #[inline]
+    pub fn as_exn(&self) -> Option<&WasmExnType> {
+        match self {
+            Self::Exn(f) => Some(f),
+            _ => None,
+        }
+    }
+
+    #[inline]
+    pub fn unwrap_exn(&self) -> &WasmExnType {
+        self.as_exn().unwrap()
+    }
 }
 
 impl TypeTrace for WasmCompositeType {
@@ -1106,6 +1243,7 @@ impl TypeTrace for WasmCompositeType {
             WasmCompositeInnerType::Func(f) => f.trace(func),
             WasmCompositeInnerType::Struct(a) => a.trace(func),
             WasmCompositeInnerType::Cont(c) => c.trace(func),
+            WasmCompositeInnerType::Exn(e) => e.trace(func),
         }
     }
 
@@ -1118,6 +1256,7 @@ impl TypeTrace for WasmCompositeType {
             WasmCompositeInnerType::Func(f) => f.trace_mut(func),
             WasmCompositeInnerType::Struct(a) => a.trace_mut(func),
             WasmCompositeInnerType::Cont(c) => c.trace_mut(func),
+            WasmCompositeInnerType::Exn(e) => e.trace_mut(func),
         }
     }
 }
@@ -1237,6 +1376,26 @@ impl WasmSubType {
         assert!(!self.composite_type.shared);
         self.composite_type.inner.unwrap_cont()
     }
+
+    #[inline]
+    pub fn is_exn(&self) -> bool {
+        self.composite_type.inner.is_exn() && !self.composite_type.shared
+    }
+
+    #[inline]
+    pub fn as_exn(&self) -> Option<&WasmExnType> {
+        if self.composite_type.shared {
+            None
+        } else {
+            self.composite_type.inner.as_exn()
+        }
+    }
+
+    #[inline]
+    pub fn unwrap_exn(&self) -> &WasmExnType {
+        assert!(!self.composite_type.shared);
+        self.composite_type.inner.unwrap_exn()
+    }
 }
 
 impl TypeTrace for WasmSubType {
@@ -1643,6 +1802,7 @@ impl ConstExpr {
     /// Returns the new const expression as well as the escaping function
     /// indices that appeared in `ref.func` instructions, if any.
     pub fn from_wasmparser(
+        env: &dyn TypeConvert,
         expr: wasmparser::ConstExpr<'_>,
     ) -> WasmResult<(Self, SmallVec<[FuncIndex; 1]>)> {
         let mut iter = expr
@@ -1668,12 +1828,13 @@ impl ConstExpr {
                 escaped.push(FuncIndex::from_u32(*function_index));
             }
 
-            ops.push(ConstOp::from_wasmparser(op, offset)?);
+            ops.push(ConstOp::from_wasmparser(env, op, offset)?);
         }
         Ok((Self { ops }, escaped))
     }
 
     /// Get the opcodes that make up this const expression.
+    #[inline]
     pub fn ops(&self) -> &[ConstOp] {
         &self.ops
     }
@@ -1718,7 +1879,7 @@ pub enum ConstOp {
     V128Const(u128),
     GlobalGet(GlobalIndex),
     RefI31,
-    RefNull,
+    RefNull(WasmHeapTopType),
     RefFunc(FuncIndex),
     I32Add,
     I32Sub,
@@ -1742,11 +1903,17 @@ pub enum ConstOp {
         array_type_index: TypeIndex,
         array_size: u32,
     },
+    ExternConvertAny,
+    AnyConvertExtern,
 }
 
 impl ConstOp {
     /// Convert a `wasmparser::Operator` to a `ConstOp`.
-    pub fn from_wasmparser(op: wasmparser::Operator<'_>, offset: usize) -> WasmResult<Self> {
+    pub fn from_wasmparser(
+        env: &dyn TypeConvert,
+        op: wasmparser::Operator<'_>,
+        offset: usize,
+    ) -> WasmResult<Self> {
         use wasmparser::Operator as O;
         Ok(match op {
             O::I32Const { value } => Self::I32Const(value),
@@ -1754,7 +1921,7 @@ impl ConstOp {
             O::F32Const { value } => Self::F32Const(value.bits()),
             O::F64Const { value } => Self::F64Const(value.bits()),
             O::V128Const { value } => Self::V128Const(u128::from_le_bytes(*value.bytes())),
-            O::RefNull { hty: _ } => Self::RefNull,
+            O::RefNull { hty } => Self::RefNull(env.convert_heap_type(hty)?.top()),
             O::RefFunc { function_index } => Self::RefFunc(FuncIndex::from_u32(function_index)),
             O::GlobalGet { global_index } => Self::GlobalGet(GlobalIndex::from_u32(global_index)),
             O::RefI31 => Self::RefI31,
@@ -1783,6 +1950,8 @@ impl ConstOp {
                 array_type_index: TypeIndex::from_u32(array_type_index),
                 array_size,
             },
+            O::ExternConvertAny => Self::ExternConvertAny,
+            O::AnyConvertExtern => Self::AnyConvertExtern,
             op => {
                 return Err(wasm_unsupported!(
                     "unsupported opcode in const expression at offset {offset:#x}: {op:?}",
@@ -1983,7 +2152,7 @@ impl Memory {
     pub fn can_elide_bounds_check(&self, tunables: &Tunables, host_page_size_log2: u8) -> bool {
         self.can_use_virtual_memory(tunables, host_page_size_log2)
             && self.idx_type == IndexType::I32
-            && tunables.memory_reservation >= (1 << 32)
+            && tunables.memory_reservation + tunables.memory_guard_size >= (1 << 32)
     }
 
     /// Returns the static size of this heap in bytes at runtime, if available.
@@ -1992,14 +2161,10 @@ impl Memory {
     pub fn static_heap_size(&self) -> Option<u64> {
         let min = self.minimum_byte_size().ok()?;
         let max = self.maximum_byte_size().ok()?;
-        if min == max {
-            Some(min)
-        } else {
-            None
-        }
+        if min == max { Some(min) } else { None }
     }
 
-    /// Returs whether or not the base pointer of this memory is allowed to be
+    /// Returns whether or not the base pointer of this memory is allowed to be
     /// relocated at runtime.
     ///
     /// When this function returns `false` then it means that after the initial
@@ -2019,6 +2184,12 @@ impl Memory {
             return false;
         }
 
+        // If its minimum and maximum are the same, then the memory will never
+        // be resized, and therefore will never move.
+        if self.limits.max.is_some_and(|max| self.limits.min == max) {
+            return false;
+        }
+
         // If the maximum size of this memory is above the threshold of the
         // initial memory reservation then the memory may move.
         let max = self.maximum_byte_size().unwrap_or(u64::MAX);
@@ -2067,6 +2238,8 @@ impl From<wasmparser::MemoryType> for Memory {
 pub struct Tag {
     /// The tag signature type.
     pub signature: EngineOrModuleTypeIndex,
+    /// The corresponding exception type.
+    pub exception: EngineOrModuleTypeIndex,
 }
 
 impl TypeTrace for Tag {
@@ -2074,14 +2247,18 @@ impl TypeTrace for Tag {
     where
         F: FnMut(EngineOrModuleTypeIndex) -> Result<(), E>,
     {
-        func(self.signature)
+        func(self.signature)?;
+        func(self.exception)?;
+        Ok(())
     }
 
     fn trace_mut<F, E>(&mut self, func: &mut F) -> Result<(), E>
     where
         F: FnMut(&mut EngineOrModuleTypeIndex) -> Result<(), E>,
     {
-        func(&mut self.signature)
+        func(&mut self.signature)?;
+        func(&mut self.exception)?;
+        Ok(())
     }
 }
 
@@ -2089,11 +2266,11 @@ impl TypeTrace for Tag {
 #[expect(missing_docs, reason = "self-describing functions")]
 pub trait TypeConvert {
     /// Converts a wasmparser table type into a wasmtime type
-    fn convert_global_type(&self, ty: &wasmparser::GlobalType) -> Global {
-        Global {
-            wasm_ty: self.convert_valtype(ty.content_type),
+    fn convert_global_type(&self, ty: &wasmparser::GlobalType) -> WasmResult<Global> {
+        Ok(Global {
+            wasm_ty: self.convert_valtype(ty.content_type)?,
             mutability: ty.mutable,
-        }
+        })
     }
 
     /// Converts a wasmparser table type into a wasmtime type
@@ -2103,43 +2280,46 @@ pub trait TypeConvert {
             true => IndexType::I64,
         };
         let limits = Limits {
-            min: ty.initial.try_into().unwrap(),
-            max: ty.maximum.map(|i| i.try_into().unwrap()),
+            min: ty.initial,
+            max: ty.maximum,
         };
         Ok(Table {
             idx_type,
             limits,
-            ref_type: self.convert_ref_type(ty.element_type),
+            ref_type: self.convert_ref_type(ty.element_type)?,
         })
     }
 
-    fn convert_sub_type(&self, ty: &wasmparser::SubType) -> WasmSubType {
-        WasmSubType {
+    fn convert_sub_type(&self, ty: &wasmparser::SubType) -> WasmResult<WasmSubType> {
+        Ok(WasmSubType {
             is_final: ty.is_final,
             supertype: ty.supertype_idx.map(|i| self.lookup_type_index(i.unpack())),
-            composite_type: self.convert_composite_type(&ty.composite_type),
-        }
+            composite_type: self.convert_composite_type(&ty.composite_type)?,
+        })
     }
 
-    fn convert_composite_type(&self, ty: &wasmparser::CompositeType) -> WasmCompositeType {
+    fn convert_composite_type(
+        &self,
+        ty: &wasmparser::CompositeType,
+    ) -> WasmResult<WasmCompositeType> {
         let inner = match &ty.inner {
             wasmparser::CompositeInnerType::Func(f) => {
-                WasmCompositeInnerType::Func(self.convert_func_type(f))
+                WasmCompositeInnerType::Func(self.convert_func_type(f)?)
             }
             wasmparser::CompositeInnerType::Array(a) => {
-                WasmCompositeInnerType::Array(self.convert_array_type(a))
+                WasmCompositeInnerType::Array(self.convert_array_type(a)?)
             }
             wasmparser::CompositeInnerType::Struct(s) => {
-                WasmCompositeInnerType::Struct(self.convert_struct_type(s))
+                WasmCompositeInnerType::Struct(self.convert_struct_type(s)?)
             }
             wasmparser::CompositeInnerType::Cont(c) => {
                 WasmCompositeInnerType::Cont(self.convert_cont_type(c))
             }
         };
-        WasmCompositeType {
+        Ok(WasmCompositeType {
             inner,
             shared: ty.shared,
-        }
+        })
     }
 
     /// Converts a wasmparser continuation type to a wasmtime type
@@ -2151,73 +2331,73 @@ pub trait TypeConvert {
         }
     }
 
-    fn convert_struct_type(&self, ty: &wasmparser::StructType) -> WasmStructType {
-        WasmStructType {
+    fn convert_struct_type(&self, ty: &wasmparser::StructType) -> WasmResult<WasmStructType> {
+        Ok(WasmStructType {
             fields: ty
                 .fields
                 .iter()
                 .map(|f| self.convert_field_type(f))
-                .collect(),
-        }
+                .collect::<WasmResult<_>>()?,
+        })
     }
 
-    fn convert_array_type(&self, ty: &wasmparser::ArrayType) -> WasmArrayType {
-        WasmArrayType(self.convert_field_type(&ty.0))
+    fn convert_array_type(&self, ty: &wasmparser::ArrayType) -> WasmResult<WasmArrayType> {
+        Ok(WasmArrayType(self.convert_field_type(&ty.0)?))
     }
 
-    fn convert_field_type(&self, ty: &wasmparser::FieldType) -> WasmFieldType {
-        WasmFieldType {
-            element_type: self.convert_storage_type(&ty.element_type),
+    fn convert_field_type(&self, ty: &wasmparser::FieldType) -> WasmResult<WasmFieldType> {
+        Ok(WasmFieldType {
+            element_type: self.convert_storage_type(&ty.element_type)?,
             mutable: ty.mutable,
-        }
+        })
     }
 
-    fn convert_storage_type(&self, ty: &wasmparser::StorageType) -> WasmStorageType {
-        match ty {
+    fn convert_storage_type(&self, ty: &wasmparser::StorageType) -> WasmResult<WasmStorageType> {
+        Ok(match ty {
             wasmparser::StorageType::I8 => WasmStorageType::I8,
             wasmparser::StorageType::I16 => WasmStorageType::I16,
-            wasmparser::StorageType::Val(v) => WasmStorageType::Val(self.convert_valtype(*v)),
-        }
+            wasmparser::StorageType::Val(v) => WasmStorageType::Val(self.convert_valtype(*v)?),
+        })
     }
 
     /// Converts a wasmparser function type to a wasmtime type
-    fn convert_func_type(&self, ty: &wasmparser::FuncType) -> WasmFuncType {
+    fn convert_func_type(&self, ty: &wasmparser::FuncType) -> WasmResult<WasmFuncType> {
         let params = ty
             .params()
             .iter()
             .map(|t| self.convert_valtype(*t))
-            .collect();
+            .collect::<WasmResult<_>>()?;
         let results = ty
             .results()
             .iter()
             .map(|t| self.convert_valtype(*t))
-            .collect();
-        WasmFuncType::new(params, results)
+            .collect::<WasmResult<_>>()?;
+        Ok(WasmFuncType::new(params, results))
     }
 
     /// Converts a wasmparser value type to a wasmtime type
-    fn convert_valtype(&self, ty: wasmparser::ValType) -> WasmValType {
-        match ty {
+    fn convert_valtype(&self, ty: wasmparser::ValType) -> WasmResult<WasmValType> {
+        Ok(match ty {
             wasmparser::ValType::I32 => WasmValType::I32,
             wasmparser::ValType::I64 => WasmValType::I64,
             wasmparser::ValType::F32 => WasmValType::F32,
             wasmparser::ValType::F64 => WasmValType::F64,
             wasmparser::ValType::V128 => WasmValType::V128,
-            wasmparser::ValType::Ref(t) => WasmValType::Ref(self.convert_ref_type(t)),
-        }
+            wasmparser::ValType::Ref(t) => WasmValType::Ref(self.convert_ref_type(t)?),
+        })
     }
 
     /// Converts a wasmparser reference type to a wasmtime type
-    fn convert_ref_type(&self, ty: wasmparser::RefType) -> WasmRefType {
-        WasmRefType {
+    fn convert_ref_type(&self, ty: wasmparser::RefType) -> WasmResult<WasmRefType> {
+        Ok(WasmRefType {
             nullable: ty.is_nullable(),
-            heap_type: self.convert_heap_type(ty.heap_type()),
-        }
+            heap_type: self.convert_heap_type(ty.heap_type())?,
+        })
     }
 
     /// Converts a wasmparser heap type to a wasmtime type
-    fn convert_heap_type(&self, ty: wasmparser::HeapType) -> WasmHeapType {
-        match ty {
+    fn convert_heap_type(&self, ty: wasmparser::HeapType) -> WasmResult<WasmHeapType> {
+        Ok(match ty {
             wasmparser::HeapType::Concrete(i) => self.lookup_heap_type(i),
             wasmparser::HeapType::Abstract { ty, shared: false } => match ty {
                 wasmparser::AbstractHeapType::Extern => WasmHeapType::Extern,
@@ -2230,16 +2410,13 @@ pub trait TypeConvert {
                 wasmparser::AbstractHeapType::Array => WasmHeapType::Array,
                 wasmparser::AbstractHeapType::Struct => WasmHeapType::Struct,
                 wasmparser::AbstractHeapType::None => WasmHeapType::None,
-
-                wasmparser::AbstractHeapType::Exn
-                | wasmparser::AbstractHeapType::NoExn
-                | wasmparser::AbstractHeapType::Cont
-                | wasmparser::AbstractHeapType::NoCont => {
-                    unimplemented!("unsupported heap type {ty:?}");
-                }
+                wasmparser::AbstractHeapType::Cont => WasmHeapType::Cont,
+                wasmparser::AbstractHeapType::NoCont => WasmHeapType::NoCont,
+                wasmparser::AbstractHeapType::Exn => WasmHeapType::Exn,
+                wasmparser::AbstractHeapType::NoExn => WasmHeapType::NoExn,
             },
-            _ => unimplemented!("unsupported heap type {ty:?}"),
-        }
+            _ => return Err(wasm_unsupported!("unsupported heap type {ty:?}")),
+        })
     }
 
     /// Converts the specified type index from a heap type into a canonicalized
