@@ -1,16 +1,12 @@
 //! Generate a configuration for both Wasmtime and the Wasm module to execute.
 
-use super::{
-    AsyncConfig, CodegenSettings, InstanceAllocationStrategy, MemoryConfig, ModuleConfig,
-    NormalMemoryConfig, UnalignedMemoryCreator,
-};
+use super::{AsyncConfig, CodegenSettings, InstanceAllocationStrategy, MemoryConfig, ModuleConfig};
 use crate::oracles::{StoreLimits, Timeout};
 use anyhow::Result;
 use arbitrary::{Arbitrary, Unstructured};
-use std::sync::Arc;
 use std::time::Duration;
-use wasmtime::{Engine, Module, MpkEnabled, Store};
-use wasmtime_test_util::wast::{limits, WastConfig, WastTest};
+use wasmtime::{Enabled, Engine, Module, Store};
+use wasmtime_test_util::wast::{WastConfig, WastTest, limits};
 
 /// Configuration for `wasmtime::Config` and generated modules for a session of
 /// fuzzing.
@@ -79,7 +75,7 @@ impl Config {
             pooling.total_memories = config.max_memories as u32;
             pooling.max_memory_size = 10 << 16;
             pooling.max_memories_per_module = config.max_memories as u32;
-            if pooling.memory_protection_keys == MpkEnabled::Auto
+            if pooling.memory_protection_keys == Enabled::Auto
                 && pooling.max_memory_protection_keys > 1
             {
                 pooling.total_memories =
@@ -92,11 +88,10 @@ impl Config {
 
             pooling.core_instance_size = 1_000_000;
 
-            if let MemoryConfig::Normal(cfg) = &mut self.wasmtime.memory_config {
-                match &mut cfg.memory_reservation {
-                    Some(size) => *size = (*size).max(pooling.max_memory_size as u64),
-                    other @ None => *other = Some(pooling.max_memory_size as u64),
-                }
+            let cfg = &mut self.wasmtime.memory_config;
+            match &mut cfg.memory_reservation {
+                Some(size) => *size = (*size).max(pooling.max_memory_size as u64),
+                other @ None => *other = Some(pooling.max_memory_size as u64),
             }
         }
 
@@ -131,6 +126,7 @@ impl Config {
             custom_page_sizes,
             multi_memory,
             threads,
+            shared_everything_threads,
             gc,
             function_references,
             relaxed_simd,
@@ -141,11 +137,19 @@ impl Config {
             component_model_async,
             component_model_async_builtins,
             component_model_async_stackful,
+            component_model_threading,
+            component_model_error_context,
+            component_model_gc,
             simd,
+            exceptions,
+            legacy_exceptions: _,
+            custom_descriptors: _,
 
             hogs_memory: _,
             nan_canonicalization: _,
             gc_types: _,
+            stack_switching: _,
+            spec_test: _,
         } = test.config;
 
         // Enable/disable some proposals that aren't configurable in wasm-smith
@@ -157,6 +161,10 @@ impl Config {
             component_model_async_builtins.unwrap_or(false);
         self.module_config.component_model_async_stackful =
             component_model_async_stackful.unwrap_or(false);
+        self.module_config.component_model_threading = component_model_threading.unwrap_or(false);
+        self.module_config.component_model_error_context =
+            component_model_error_context.unwrap_or(false);
+        self.module_config.component_model_gc = component_model_gc.unwrap_or(false);
 
         // Enable/disable proposals that wasm-smith has knobs for which will be
         // read when creating `wasmtime::Config`.
@@ -170,24 +178,21 @@ impl Config {
         config.tail_call_enabled = tail_call.unwrap_or(false);
         config.custom_page_sizes_enabled = custom_page_sizes.unwrap_or(false);
         config.threads_enabled = threads.unwrap_or(false);
+        config.shared_everything_threads_enabled = shared_everything_threads.unwrap_or(false);
         config.gc_enabled = gc.unwrap_or(false);
         config.reference_types_enabled = config.gc_enabled
             || self.module_config.function_references_enabled
             || reference_types.unwrap_or(false);
         config.extended_const_enabled = extended_const.unwrap_or(false);
+        config.exceptions_enabled = exceptions.unwrap_or(false);
         if multi_memory.unwrap_or(false) {
             config.max_memories = limits::MEMORIES_PER_MODULE as usize;
         } else {
             config.max_memories = 1;
         }
 
-        match &mut self.wasmtime.memory_config {
-            MemoryConfig::Normal(config) => {
-                if let Some(n) = &mut config.memory_reservation {
-                    *n = (*n).max(limits::MEMORY_SIZE as u64);
-                }
-            }
-            MemoryConfig::CustomUnaligned => {}
+        if let Some(n) = &mut self.wasmtime.memory_config.memory_reservation {
+            *n = (*n).max(limits::MEMORY_SIZE as u64);
         }
 
         // FIXME: it might be more ideal to avoid the need for this entirely
@@ -226,6 +231,7 @@ impl Config {
             pooling.component_instance_size = pooling
                 .component_instance_size
                 .max(limits::CORE_INSTANCE_SIZE);
+            pooling.total_stacks = pooling.total_stacks.max(limits::TOTAL_STACKS);
         }
 
         // Return the test configuration that this fuzz configuration represents
@@ -262,6 +268,7 @@ impl Config {
         cfg.codegen.native_unwind_info =
             Some(cfg!(target_os = "windows") || self.wasmtime.native_unwind_info);
         cfg.codegen.parallel_compilation = Some(false);
+
         cfg.debug.address_map = Some(self.wasmtime.generate_address_map);
         cfg.opts.opt_level = Some(self.wasmtime.opt_level.to_wasmtime());
         cfg.opts.regalloc_algorithm = Some(self.wasmtime.regalloc_algorithm.to_wasmtime());
@@ -279,6 +286,10 @@ impl Config {
             Some(self.module_config.component_model_async_builtins);
         cfg.wasm.component_model_async_stackful =
             Some(self.module_config.component_model_async_stackful);
+        cfg.wasm.component_model_threading = Some(self.module_config.component_model_threading);
+        cfg.wasm.component_model_error_context =
+            Some(self.module_config.component_model_error_context);
+        cfg.wasm.component_model_gc = Some(self.module_config.component_model_gc);
         cfg.wasm.custom_page_sizes = Some(self.module_config.config.custom_page_sizes_enabled);
         cfg.wasm.epoch_interruption = Some(self.wasmtime.epoch_interruption);
         cfg.wasm.extended_const = Some(self.module_config.config.extended_const_enabled);
@@ -293,7 +304,10 @@ impl Config {
         cfg.wasm.simd = Some(self.module_config.config.simd_enabled);
         cfg.wasm.tail_call = Some(self.module_config.config.tail_call_enabled);
         cfg.wasm.threads = Some(self.module_config.config.threads_enabled);
+        cfg.wasm.shared_everything_threads =
+            Some(self.module_config.config.shared_everything_threads_enabled);
         cfg.wasm.wide_arithmetic = Some(self.module_config.config.wide_arithmetic_enabled);
+        cfg.wasm.exceptions = Some(self.module_config.config.exceptions_enabled);
         if !self.module_config.config.simd_enabled {
             cfg.wasm.relaxed_simd = Some(false);
         }
@@ -315,9 +329,32 @@ impl Config {
             && self.wasmtime.pcc
             && !self.module_config.config.memory64_enabled;
 
+        cfg.codegen.inlining = self.wasmtime.inlining;
+
         // Only set cranelift specific flags when the Cranelift strategy is
         // chosen.
         if cranelift_strategy {
+            if let Some(option) = self.wasmtime.inlining_intra_module {
+                cfg.codegen.cranelift.push((
+                    "wasmtime_inlining_intra_module".to_string(),
+                    Some(option.to_string()),
+                ));
+            }
+            if let Some(size) = self.wasmtime.inlining_small_callee_size {
+                cfg.codegen.cranelift.push((
+                    "wasmtime_inlining_small_callee_size".to_string(),
+                    // Clamp to avoid extreme code size blow up.
+                    Some(std::cmp::min(1000, size).to_string()),
+                ));
+            }
+            if let Some(size) = self.wasmtime.inlining_sum_size_threshold {
+                cfg.codegen.cranelift.push((
+                    "wasmtime_inlining_sum_size_threshold".to_string(),
+                    // Clamp to avoid extreme code size blow up.
+                    Some(std::cmp::min(1000, size).to_string()),
+                ));
+            }
+
             // If the wasm-smith-generated module use nan canonicalization then we
             // don't need to enable it, but if it doesn't enable it already then we
             // enable this codegen option.
@@ -364,11 +401,11 @@ impl Config {
         // - shared memories are required to be aligned which means that the
         //   `CustomUnaligned` variant isn't actually safe to use with a shared
         //   memory.
-        let host_memory = if !self.module_config.config.threads_enabled {
+        if !self.module_config.config.threads_enabled {
             // If PCC is enabled, force other options to be compatible: PCC is currently only
             // supported when bounds checks are elided.
             let memory_config = if pcc {
-                MemoryConfig::Normal(NormalMemoryConfig {
+                MemoryConfig {
                     memory_reservation: Some(4 << 30), // 4 GiB
                     memory_guard_size: Some(2 << 30),  // 2 GiB
                     memory_reservation_for_growth: Some(0),
@@ -376,28 +413,12 @@ impl Config {
                     memory_init_cow: true,
                     // Doesn't matter, only using virtual memory.
                     cranelift_enable_heap_access_spectre_mitigations: None,
-                })
+                }
             } else {
                 self.wasmtime.memory_config.clone()
             };
 
-            match &memory_config {
-                MemoryConfig::Normal(memory_config) => {
-                    memory_config.configure(&mut cfg);
-                    None
-                }
-                MemoryConfig::CustomUnaligned => {
-                    cfg.opts.memory_reservation = Some(0);
-                    cfg.opts.memory_guard_size = Some(0);
-                    cfg.opts.memory_reservation_for_growth = Some(0);
-                    cfg.opts.guard_before_linear_memory = Some(false);
-                    cfg.opts.memory_init_cow = Some(false);
-                    log::debug!("a custom unaligned host memory will be in use");
-                    Some(Arc::new(UnalignedMemoryCreator))
-                }
-            }
-        } else {
-            None
+            memory_config.configure(&mut cfg);
         };
 
         // If malloc-based memory is going to be used, which requires these four
@@ -422,10 +443,6 @@ impl Config {
         log::debug!("creating wasmtime config with CLI options:\n{cfg}");
         let mut cfg = cfg.config(None).expect("failed to create wasmtime::Config");
 
-        if let Some(host_memory) = host_memory {
-            cfg.with_host_memory(host_memory);
-        }
-
         if self.wasmtime.async_config != AsyncConfig::Disabled {
             log::debug!("async config in use {:?}", self.wasmtime.async_config);
             self.wasmtime.async_config.configure(&mut cfg);
@@ -446,7 +463,12 @@ impl Config {
     /// Configures a store based on this configuration.
     pub fn configure_store(&self, store: &mut Store<StoreLimits>) {
         store.limiter(|s| s as &mut dyn wasmtime::ResourceLimiter);
+        self.configure_store_epoch_and_fuel(store);
+    }
 
+    /// Configures everything unrelated to `T` in a store, such as epochs and
+    /// fuel.
+    pub fn configure_store_epoch_and_fuel<T>(&self, store: &mut Store<T>) {
         // Configure the store to never abort by default, that is it'll have
         // max fuel or otherwise trap on an epoch change but the epoch won't
         // ever change.
@@ -555,7 +577,7 @@ pub struct WasmtimeConfig {
     regalloc_algorithm: RegallocAlgorithm,
     debug_info: bool,
     canonicalize_nans: bool,
-    interruptable: bool,
+    interruptible: bool,
     pub(crate) consume_fuel: bool,
     pub(crate) epoch_interruption: bool,
     /// The Wasmtime memory configuration to use.
@@ -563,6 +585,10 @@ pub struct WasmtimeConfig {
     force_jump_veneers: bool,
     memory_init_cow: bool,
     memory_guaranteed_dense_image_size: u64,
+    inlining: Option<bool>,
+    inlining_intra_module: Option<IntraModuleInlining>,
+    inlining_small_callee_size: Option<u32>,
+    inlining_sum_size_threshold: Option<u32>,
     use_precompiled_cwasm: bool,
     async_stack_zeroing: bool,
     /// Configuration for the instance allocation strategy to use.
@@ -619,7 +645,7 @@ impl WasmtimeConfig {
     pub fn update_module_config(
         &mut self,
         config: &mut ModuleConfig,
-        u: &mut Unstructured<'_>,
+        _u: &mut Unstructured<'_>,
     ) -> arbitrary::Result<()> {
         match self.compiler_strategy {
             CompilerStrategy::CraneliftNative => {}
@@ -645,6 +671,7 @@ impl WasmtimeConfig {
                 config.config.gc_enabled = false;
                 config.config.tail_call_enabled = false;
                 config.config.reference_types_enabled = false;
+                config.config.exceptions_enabled = false;
                 config.function_references_enabled = false;
 
                 // Winch's SIMD implementations require AVX and AVX2.
@@ -670,18 +697,6 @@ impl WasmtimeConfig {
             }
         }
 
-        // Forcibly don't use the `CustomUnaligned` memory configuration when
-        // wasm threads are enabled or when the pooling allocator is used. For
-        // the pooling allocator it doesn't use custom memory creators anyway
-        // and for wasm threads that will require some refactoring of the
-        // `LinearMemory` trait to bubble up the request that the linear memory
-        // not move. Otherwise that just generates a panic right now.
-        if config.config.threads_enabled
-            || matches!(self.strategy, InstanceAllocationStrategy::Pooling(_))
-        {
-            self.avoid_custom_unaligned_memory(u)?;
-        }
-
         // If using the pooling allocator, constrain the memory and module configurations
         // to the module limits.
         if let InstanceAllocationStrategy::Pooling(pooling) = &mut self.strategy {
@@ -704,10 +719,9 @@ impl WasmtimeConfig {
                         .try_into()
                         .unwrap_or(u64::MAX),
                 );
-            let mut min = min_bytes.min(pooling.max_memory_size as u64);
-            if let MemoryConfig::Normal(cfg) = &self.memory_config {
-                min = min.min(cfg.memory_reservation.unwrap_or(0));
-            }
+            let min = min_bytes
+                .min(pooling.max_memory_size as u64)
+                .min(self.memory_config.memory_reservation.unwrap_or(0));
             pooling.max_memory_size = min as usize;
             config.config.max_memory32_bytes = min;
             config.config.max_memory64_bytes = min as u128;
@@ -720,11 +734,10 @@ impl WasmtimeConfig {
                     pooling.max_memory_size = 1 << 16;
                     config.config.max_memory32_bytes = 1 << 16;
                     config.config.max_memory64_bytes = 1 << 16;
-                    if let MemoryConfig::Normal(cfg) = &mut self.memory_config {
-                        match &mut cfg.memory_reservation {
-                            Some(size) => *size = (*size).max(pooling.max_memory_size as u64),
-                            size @ None => *size = Some(pooling.max_memory_size as u64),
-                        }
+                    let cfg = &mut self.memory_config;
+                    match &mut cfg.memory_reservation {
+                        Some(size) => *size = (*size).max(pooling.max_memory_size as u64),
+                        size @ None => *size = Some(pooling.max_memory_size as u64),
                     }
                 }
                 // .. additionally update tables
@@ -755,9 +768,8 @@ impl WasmtimeConfig {
 
             // Spectre-based heap mitigations require signal handlers so this
             // must always be disabled if signals-based traps are disabled.
-            if let MemoryConfig::Normal(cfg) = &mut self.memory_config {
-                cfg.cranelift_enable_heap_access_spectre_mitigations = None;
-            }
+            self.memory_config
+                .cranelift_enable_heap_access_spectre_mitigations = None;
         }
 
         self.make_internally_consistent();
@@ -776,15 +788,6 @@ impl WasmtimeConfig {
         })
     }
 
-    /// Helper to switch `MemoryConfig::CustomUnaligned` to
-    /// `MemoryConfig::Normal`
-    fn avoid_custom_unaligned_memory(&mut self, u: &mut Unstructured<'_>) -> arbitrary::Result<()> {
-        if let MemoryConfig::CustomUnaligned = self.memory_config {
-            self.memory_config = MemoryConfig::Normal(u.arbitrary()?);
-        }
-        Ok(())
-    }
-
     /// Helper method to handle some dependencies between various configuration
     /// options. This is intended to be called whenever a `Config` is created or
     /// modified to ensure that the final result is an instantiable `Config`.
@@ -795,25 +798,24 @@ impl WasmtimeConfig {
     /// the current state of the engine's development.
     fn make_internally_consistent(&mut self) {
         if !self.signals_based_traps {
-            if let MemoryConfig::Normal(cfg) = &mut self.memory_config {
-                // Spectre-based heap mitigations require signal handlers so
-                // this must always be disabled if signals-based traps are
-                // disabled.
-                cfg.cranelift_enable_heap_access_spectre_mitigations = None;
+            let cfg = &mut self.memory_config;
+            // Spectre-based heap mitigations require signal handlers so
+            // this must always be disabled if signals-based traps are
+            // disabled.
+            cfg.cranelift_enable_heap_access_spectre_mitigations = None;
 
-                // With configuration settings that match the use of malloc for
-                // linear memories cap the `memory_reservation_for_growth` value
-                // to something reasonable to avoid OOM in fuzzing.
-                if !cfg.memory_init_cow
-                    && cfg.memory_guard_size == Some(0)
-                    && cfg.memory_reservation == Some(0)
-                {
-                    let min = 10 << 20; // 10 MiB
-                    if let Some(val) = &mut cfg.memory_reservation_for_growth {
-                        *val = (*val).min(min);
-                    } else {
-                        cfg.memory_reservation_for_growth = Some(min);
-                    }
+            // With configuration settings that match the use of malloc for
+            // linear memories cap the `memory_reservation_for_growth` value
+            // to something reasonable to avoid OOM in fuzzing.
+            if !cfg.memory_init_cow
+                && cfg.memory_guard_size == Some(0)
+                && cfg.memory_reservation == Some(0)
+            {
+                let min = 10 << 20; // 10 MiB
+                if let Some(val) = &mut cfg.memory_reservation_for_growth {
+                    *val = (*val).min(min);
+                } else {
+                    cfg.memory_reservation_for_growth = Some(min);
                 }
             }
         }
@@ -847,7 +849,32 @@ impl RegallocAlgorithm {
     fn to_wasmtime(&self) -> wasmtime::RegallocAlgorithm {
         match self {
             RegallocAlgorithm::Backtracking => wasmtime::RegallocAlgorithm::Backtracking,
-            RegallocAlgorithm::SinglePass => wasmtime::RegallocAlgorithm::SinglePass,
+            RegallocAlgorithm::SinglePass => {
+                // FIXME(#11850)
+                const SINGLE_PASS_KNOWN_BUGGY_AT_THIS_TIME: bool = true;
+                if SINGLE_PASS_KNOWN_BUGGY_AT_THIS_TIME {
+                    wasmtime::RegallocAlgorithm::Backtracking
+                } else {
+                    wasmtime::RegallocAlgorithm::SinglePass
+                }
+            }
+        }
+    }
+}
+
+#[derive(Arbitrary, Clone, Copy, Debug, PartialEq, Eq, Hash)]
+enum IntraModuleInlining {
+    Yes,
+    No,
+    WhenUsingGc,
+}
+
+impl std::fmt::Display for IntraModuleInlining {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        match self {
+            IntraModuleInlining::Yes => write!(f, "yes"),
+            IntraModuleInlining::No => write!(f, "no"),
+            IntraModuleInlining::WhenUsingGc => write!(f, "gc"),
         }
     }
 }
