@@ -29,7 +29,7 @@ use std::future::Future;
 use std::pin::Pin;
 use std::sync::atomic::{AtomicBool, AtomicUsize, Ordering::SeqCst};
 use std::sync::{Arc, Condvar, Mutex};
-use std::task::{Context, Poll};
+use std::task::{Context, Poll, Waker};
 use std::time::{Duration, Instant};
 use wasmtime::*;
 use wasmtime_wast::WastContext;
@@ -51,7 +51,7 @@ pub fn log_wasm(wasm: &[u8]) {
     let i = CNT.fetch_add(1, SeqCst);
     let name = format!("testcase{i}.wasm");
     std::fs::write(&name, wasm).expect("failed to write wasm file");
-    log::debug!("wrote wasm file to `{}`", name);
+    log::debug!("wrote wasm file to `{name}`");
     let wat = format!("testcase{i}.wat");
     match wasmprinter::print_bytes(wasm) {
         Ok(s) => std::fs::write(&wat, s).expect("failed to write wat file"),
@@ -282,7 +282,7 @@ pub fn instantiate_many(
         match command {
             Command::Instantiate(index) => {
                 let index = *index % modules.len();
-                log::info!("instantiating {}", index);
+                log::info!("instantiating {index}");
                 let module = &modules[index];
                 let mut store = Store::new(&engine, limits.clone());
                 config.configure_store(&mut store);
@@ -299,7 +299,7 @@ pub fn instantiate_many(
                 }
                 let index = *index % stores.len();
 
-                log::info!("dropping {}", index);
+                log::info!("dropping {index}");
                 stores.swap_remove(index);
             }
         }
@@ -409,9 +409,17 @@ fn unwrap_instance(
         return None;
     }
 
-    // Allow traps which can happen normally with `unreachable` or a
-    // timeout or such
-    if e.is::<Trap>() {
+    // Allow traps which can happen normally with `unreachable` or a timeout or
+    // such.
+    if e.is::<Trap>()
+        // Also allow failures to instantiate as a result of hitting pooling
+        // limits.
+        || e.is::<wasmtime::PoolConcurrencyLimitError>()
+        // And GC heap OOMs.
+        || e.is::<wasmtime::GcHeapOutOfMemory<()>>()
+        // And thrown exceptions.
+        || e.is::<wasmtime::ThrownException>()
+    {
         return None;
     }
 
@@ -424,11 +432,6 @@ fn unwrap_instance(
         return None;
     }
 
-    // Also allow failures to instantiate as a result of hitting pooling limits.
-    if e.is::<wasmtime::PoolConcurrencyLimitError>() {
-        return None;
-    }
-
     // Everything else should be a bug in the fuzzer or a bug in wasmtime
     panic!("failed to instantiate: {e:?}");
 }
@@ -451,7 +454,7 @@ pub fn differential(
     args: &[DiffValue],
     result_tys: &[DiffValueType],
 ) -> anyhow::Result<bool> {
-    log::debug!("Evaluating: `{}` with {:?}", name, args);
+    log::debug!("Evaluating: `{name}` with {args:?}");
     let lhs_results = match lhs.evaluate(name, args, result_tys) {
         Ok(Some(results)) => Ok(results),
         Err(e) => Err(e),
@@ -616,7 +619,7 @@ pub fn make_api_calls(api: generators::api::ApiCalls) {
             }
 
             ApiCall::ModuleNew { id, wasm } => {
-                log::debug!("creating module: {}", id);
+                log::debug!("creating module: {id}");
                 log_wasm(&wasm);
                 let module = match Module::new(store.as_ref().unwrap().engine(), &wasm) {
                     Ok(m) => m,
@@ -627,12 +630,12 @@ pub fn make_api_calls(api: generators::api::ApiCalls) {
             }
 
             ApiCall::ModuleDrop { id } => {
-                log::trace!("dropping module: {}", id);
+                log::trace!("dropping module: {id}");
                 drop(modules.remove(&id));
             }
 
             ApiCall::InstanceNew { id, module } => {
-                log::trace!("instantiating module {} as {}", module, id);
+                log::trace!("instantiating module {module} as {id}");
                 let module = match modules.get(&module) {
                     Some(m) => m,
                     None => continue,
@@ -645,12 +648,12 @@ pub fn make_api_calls(api: generators::api::ApiCalls) {
             }
 
             ApiCall::InstanceDrop { id } => {
-                log::trace!("dropping instance {}", id);
+                log::trace!("dropping instance {id}");
                 instances.remove(&id);
             }
 
             ApiCall::CallExportedFunc { instance, nth } => {
-                log::trace!("calling instance export {} / {}", instance, nth);
+                log::trace!("calling instance export {instance} / {nth}");
                 let instance = match instances.get(&instance) {
                     Some(i) => i,
                     None => {
@@ -701,12 +704,13 @@ pub fn wast_test(u: &mut arbitrary::Unstructured<'_>) -> arbitrary::Result<()> {
 
     let mut fuzz_config: generators::Config = u.arbitrary()?;
     let test: generators::WastTest = u.arbitrary()?;
-    if u.arbitrary()? {
-        fuzz_config.enable_async(u)?;
-    }
 
     let test = &test.test;
 
+    if test.config.component_model_async() || u.arbitrary()? {
+        fuzz_config.enable_async(u)?;
+    }
+
     // Discard tests that allocate a lot of memory as we don't want to OOM the
     // fuzzer and we also limit memory growth which would cause the test to
     // fail.
@@ -754,7 +758,11 @@ pub fn wast_test(u: &mut arbitrary::Unstructured<'_>) -> arbitrary::Result<()> {
     } else {
         wasmtime_wast::Async::Yes
     };
-    let mut wast_context = WastContext::new(fuzz_config.to_store(), async_);
+    log::debug!("async: {async_:?}");
+    let engine = Engine::new(&fuzz_config.to_wasmtime()).unwrap();
+    let mut wast_context = WastContext::new(&engine, async_, move |store| {
+        fuzz_config.configure_store_epoch_and_fuel(store);
+    });
     wast_context
         .register_spectest(&wasmtime_wast::SpectestConfig {
             use_shared_memory: true,
@@ -762,7 +770,7 @@ pub fn wast_test(u: &mut arbitrary::Unstructured<'_>) -> arbitrary::Result<()> {
         })
         .unwrap();
     wast_context
-        .run_buffer(test.path.to_str().unwrap(), test.contents.as_bytes())
+        .run_wast(test.path.to_str().unwrap(), test.contents.as_bytes())
         .unwrap();
     Ok(())
 }
@@ -773,9 +781,9 @@ pub fn wast_test(u: &mut arbitrary::Unstructured<'_>) -> arbitrary::Result<()> {
 /// case -- used to test below that gc happens reasonably soon and eventually.
 pub fn table_ops(
     mut fuzz_config: generators::Config,
-    ops: generators::table_ops::TableOps,
+    mut ops: generators::table_ops::TableOps,
 ) -> Result<usize> {
-    let expected_drops = Arc::new(AtomicUsize::new(ops.num_params as usize));
+    let expected_drops = Arc::new(AtomicUsize::new(0));
     let num_dropped = Arc::new(AtomicUsize::new(0));
 
     let num_gcs = Arc::new(AtomicUsize::new(0));
@@ -809,16 +817,23 @@ pub fn table_ops(
             move |mut caller: Caller<'_, StoreLimits>, _params, results| {
                 log::info!("table_ops: GC");
                 if num_gcs.fetch_add(1, SeqCst) < MAX_GCS {
-                    caller.gc();
+                    caller.gc(None);
                 }
 
-                let a = ExternRef::new(&mut caller, CountDrops(num_dropped.clone()))?;
-                let b = ExternRef::new(&mut caller, CountDrops(num_dropped.clone()))?;
-                let c = ExternRef::new(&mut caller, CountDrops(num_dropped.clone()))?;
+                let a = ExternRef::new(
+                    &mut caller,
+                    CountDrops::new(&expected_drops, num_dropped.clone()),
+                )?;
+                let b = ExternRef::new(
+                    &mut caller,
+                    CountDrops::new(&expected_drops, num_dropped.clone()),
+                )?;
+                let c = ExternRef::new(
+                    &mut caller,
+                    CountDrops::new(&expected_drops, num_dropped.clone()),
+                )?;
 
-                log::info!("table_ops: gc() -> ({:?}, {:?}, {:?})", a, b, c);
-
-                expected_drops.fetch_add(3, SeqCst);
+                log::info!("table_ops: gc() -> ({a:?}, {b:?}, {c:?})");
                 results[0] = Some(a).into();
                 results[1] = Some(b).into();
                 results[2] = Some(c).into();
@@ -881,12 +896,20 @@ pub fn table_ops(
             move |mut caller, _params, results| {
                 log::info!("table_ops: make_refs");
 
-                let a = ExternRef::new(&mut caller, CountDrops(num_dropped.clone()))?;
-                let b = ExternRef::new(&mut caller, CountDrops(num_dropped.clone()))?;
-                let c = ExternRef::new(&mut caller, CountDrops(num_dropped.clone()))?;
-                expected_drops.fetch_add(3, SeqCst);
+                let a = ExternRef::new(
+                    &mut caller,
+                    CountDrops::new(&expected_drops, num_dropped.clone()),
+                )?;
+                let b = ExternRef::new(
+                    &mut caller,
+                    CountDrops::new(&expected_drops, num_dropped.clone()),
+                )?;
+                let c = ExternRef::new(
+                    &mut caller,
+                    CountDrops::new(&expected_drops, num_dropped.clone()),
+                )?;
 
-                log::info!("table_ops: make_refs() -> ({:?}, {:?}, {:?})", a, b, c);
+                log::info!("table_ops: make_refs() -> ({a:?}, {b:?}, {c:?})");
 
                 results[0] = Some(a).into();
                 results[1] = Some(b).into();
@@ -905,19 +928,19 @@ pub fn table_ops(
 
             log::info!(
                 "table_ops: begin allocating {} externref arguments",
-                ops.num_globals
+                ops.limits.num_globals
             );
-            let args: Vec<_> = (0..ops.num_params)
+            let args: Vec<_> = (0..ops.limits.num_params)
                 .map(|_| {
                     Ok(Val::ExternRef(Some(ExternRef::new(
                         &mut scope,
-                        CountDrops(num_dropped.clone()),
+                        CountDrops::new(&expected_drops, num_dropped.clone()),
                     )?)))
                 })
                 .collect::<Result<_>>()?;
             log::info!(
                 "table_ops: end allocating {} externref arguments",
-                ops.num_globals
+                ops.limits.num_globals
             );
 
             // The generated function should always return a trap. The only two
@@ -925,20 +948,23 @@ pub fn table_ops(
             // and `table.set` generated or an out-of-fuel trap. Otherwise any other
             // error is unexpected and should fail fuzzing.
             log::info!("table_ops: calling into Wasm `run` function");
-            let trap = run
-                .call(&mut scope, &args, &mut [])
-                .unwrap_err()
-                .downcast::<Trap>()
-                .unwrap();
-
-            match trap {
-                Trap::TableOutOfBounds | Trap::OutOfFuel => {}
-                _ => panic!("unexpected trap: {trap}"),
+            let err = run.call(&mut scope, &args, &mut []).unwrap_err();
+            match err.downcast::<GcHeapOutOfMemory<CountDrops>>() {
+                Ok(_oom) => {}
+                Err(err) => {
+                    let trap = err
+                        .downcast::<Trap>()
+                        .expect("if not GC oom, error should be a Wasm trap");
+                    match trap {
+                        Trap::TableOutOfBounds | Trap::OutOfFuel => {}
+                        _ => panic!("unexpected trap: {trap}"),
+                    }
+                }
             }
         }
 
         // Do a final GC after running the Wasm.
-        store.gc();
+        store.gc(None);
     }
 
     assert_eq!(num_dropped.load(SeqCst), expected_drops.load(SeqCst));
@@ -946,9 +972,21 @@ pub fn table_ops(
 
     struct CountDrops(Arc<AtomicUsize>);
 
+    impl CountDrops {
+        fn new(expected_drops: &AtomicUsize, num_dropped: Arc<AtomicUsize>) -> Self {
+            let expected = expected_drops.fetch_add(1, SeqCst);
+            log::info!(
+                "CountDrops::new: expected drops: {expected} -> {}",
+                expected + 1
+            );
+            Self(num_dropped)
+        }
+    }
+
     impl Drop for CountDrops {
         fn drop(&mut self) {
-            self.0.fetch_add(1, SeqCst);
+            let drops = self.0.fetch_add(1, SeqCst);
+            log::info!("CountDrops::drop: actual drops: {drops} -> {}", drops + 1);
         }
     }
 }
@@ -1012,7 +1050,7 @@ pub fn dynamic_component_api_target(input: &mut arbitrary::Unstructured) -> arbi
     use wasmtime::component::{Component, Linker, Val};
     use wasmtime_test_util::component::FuncExt;
     use wasmtime_test_util::component_fuzz::{
-        TestCase, Type, EXPORT_FUNCTION, IMPORT_FUNCTION, MAX_TYPE_DEPTH,
+        EXPORT_FUNCTION, IMPORT_FUNCTION, MAX_TYPE_DEPTH, TestCase, Type,
     };
 
     crate::init_fuzzing();
@@ -1053,6 +1091,7 @@ pub fn dynamic_component_api_target(input: &mut arbitrary::Unstructured) -> arbi
         .root()
         .func_new(IMPORT_FUNCTION, {
             move |mut cx: StoreContextMut<'_, (Vec<Val>, Option<Vec<Val>>)>,
+                  _,
                   params: &[Val],
                   results: &mut [Val]|
                   -> Result<()> {
@@ -1071,17 +1110,16 @@ pub fn dynamic_component_api_target(input: &mut arbitrary::Unstructured) -> arbi
 
     let instance = linker.instantiate(&mut store, &component).unwrap();
     let func = instance.get_func(&mut store, EXPORT_FUNCTION).unwrap();
-    let param_tys = func.params(&store);
-    let result_tys = func.results(&store);
+    let ty = func.ty(&store);
 
     while input.arbitrary()? {
-        let params = param_tys
-            .iter()
-            .map(|(_, ty)| component_types::arbitrary_val(ty, input))
+        let params = ty
+            .params()
+            .map(|(_, ty)| component_types::arbitrary_val(&ty, input))
             .collect::<arbitrary::Result<Vec<_>>>()?;
-        let results = result_tys
-            .iter()
-            .map(|ty| component_types::arbitrary_val(ty, input))
+        let results = ty
+            .results()
+            .map(|ty| component_types::arbitrary_val(&ty, input))
             .collect::<arbitrary::Result<Vec<_>>>()?;
 
         *store.data_mut() = (params.clone(), Some(results.clone()));
@@ -1130,7 +1168,7 @@ pub fn call_async(wasm: &[u8], config: &generators::Config, mut poll_amts: &[u32
                     let ty = ty.clone();
                     Box::new(async move {
                         caller.engine().increment_epoch();
-                        log::info!("yielding {} times in import", poll_amt);
+                        log::info!("yielding {poll_amt} times in import");
                         YieldN(poll_amt).await;
                         for (ret_ty, result) in ty.results().zip(results) {
                             *result = ret_ty.default_value().unwrap();
@@ -1141,9 +1179,9 @@ pub fn call_async(wasm: &[u8], config: &generators::Config, mut poll_amts: &[u32
                 .into()
             }
             other_ty => match other_ty.default_value(&mut store) {
-                Some(item) => item,
-                None => {
-                    log::warn!("couldn't create import for {import:?}");
+                Ok(item) => item,
+                Err(e) => {
+                    log::warn!("couldn't create import for {import:?}: {e:?}");
                     return;
                 }
             },
@@ -1198,7 +1236,7 @@ pub fn call_async(wasm: &[u8], config: &generators::Config, mut poll_amts: &[u32
             .map(|ty| ty.default_value().unwrap())
             .collect::<Vec<_>>();
 
-        log::info!("invoking export {:?}", name);
+        log::info!("invoking export {name:?}");
         let future = func.call_async(&mut store, &params, &mut results);
         match run(Timeout {
             future,
@@ -1289,7 +1327,7 @@ pub fn call_async(wasm: &[u8], config: &generators::Config, mut poll_amts: &[u32
 
     fn run<F: Future>(future: F) -> F::Output {
         let mut f = Box::pin(future);
-        let mut cx = Context::from_waker(futures::task::noop_waker_ref());
+        let mut cx = Context::from_waker(Waker::noop());
         loop {
             match f.as_mut().poll(&mut cx) {
                 Poll::Ready(val) => break val,
@@ -1385,7 +1423,8 @@ mod tests {
             | WasmFeatures::GC
             | WasmFeatures::GC_TYPES
             | WasmFeatures::CUSTOM_PAGE_SIZES
-            | WasmFeatures::EXTENDED_CONST;
+            | WasmFeatures::EXTENDED_CONST
+            | WasmFeatures::EXCEPTIONS;
 
         // All other features that wasmparser supports, which is presumably a
         // superset of the features that wasm-smith supports, are listed here as
