@@ -3,7 +3,7 @@
 
 use crate::generators::{Config, DiffValue, DiffValueType};
 use crate::oracles::engine::{DiffEngine, DiffInstance};
-use anyhow::{anyhow, Error, Result};
+use anyhow::{Error, Result, anyhow};
 use wasm_spec_interpreter::SpecValue;
 use wasmtime::Trap;
 
@@ -28,6 +28,7 @@ impl SpecInterpreter {
         config.custom_page_sizes_enabled = false;
         config.wide_arithmetic_enabled = false;
         config.extended_const_enabled = false;
+        config.exceptions_enabled = false;
 
         Self
     }
@@ -40,7 +41,7 @@ impl DiffEngine for SpecInterpreter {
 
     fn instantiate(&mut self, wasm: &[u8]) -> Result<Box<dyn DiffInstance>> {
         let instance = wasm_spec_interpreter::instantiate(wasm)
-            .map_err(|e| anyhow!("failed to instantiate in spec interpreter: {}", e))?;
+            .map_err(|e| anyhow!("failed to instantiate in spec interpreter: {e}"))?;
         Ok(Box::new(SpecInstance { instance }))
     }
 
@@ -77,7 +78,7 @@ impl DiffInstance for SpecInstance {
     }
 
     fn get_global(&mut self, name: &str, _ty: DiffValueType) -> Option<DiffValue> {
-        use wasm_spec_interpreter::{export, SpecExport::Global};
+        use wasm_spec_interpreter::{SpecExport::Global, export};
         if let Ok(Global(g)) = export(&self.instance, name) {
             Some(g.into())
         } else {
@@ -86,7 +87,7 @@ impl DiffInstance for SpecInstance {
     }
 
     fn get_memory(&mut self, name: &str, _shared: bool) -> Option<Vec<u8>> {
-        use wasm_spec_interpreter::{export, SpecExport::Memory};
+        use wasm_spec_interpreter::{SpecExport::Memory, export};
         if let Ok(Memory(m)) = export(&self.instance, name) {
             Some(m)
         } else {
@@ -103,16 +104,20 @@ impl From<&DiffValue> for SpecValue {
             DiffValue::F32(n) => SpecValue::F32(n as i32),
             DiffValue::F64(n) => SpecValue::F64(n as i64),
             DiffValue::V128(n) => SpecValue::V128(n.to_le_bytes().to_vec()),
-            DiffValue::FuncRef { .. } | DiffValue::ExternRef { .. } | DiffValue::AnyRef { .. } => {
+            DiffValue::FuncRef { .. }
+            | DiffValue::ExternRef { .. }
+            | DiffValue::AnyRef { .. }
+            | DiffValue::ExnRef { .. }
+            | DiffValue::ContRef { .. } => {
                 unimplemented!()
             }
         }
     }
 }
 
-impl Into<DiffValue> for SpecValue {
-    fn into(self) -> DiffValue {
-        match self {
+impl From<SpecValue> for DiffValue {
+    fn from(spec: SpecValue) -> DiffValue {
+        match spec {
             SpecValue::I32(n) => DiffValue::I32(n),
             SpecValue::I64(n) => DiffValue::I64(n),
             SpecValue::F32(n) => DiffValue::F32(n as u32),
