@@ -3,7 +3,7 @@
 use crate::generators::{self, CompilerStrategy, DiffValue, DiffValueType, WasmtimeConfig};
 use crate::oracles::dummy;
 use crate::oracles::engine::DiffInstance;
-use crate::oracles::{compile_module, engine::DiffEngine, StoreLimits};
+use crate::oracles::{StoreLimits, compile_module, engine::DiffEngine};
 use crate::single_module_fuzzer::KnownValid;
 use anyhow::{Context, Error, Result};
 use arbitrary::Unstructured;
@@ -223,13 +223,21 @@ impl From<&DiffValue> for Val {
                 assert!(null);
                 Val::AnyRef(None)
             }
+            DiffValue::ExnRef { null } => {
+                assert!(null);
+                Val::ExnRef(None)
+            }
+            DiffValue::ContRef { null } => {
+                assert!(null);
+                Val::ExnRef(None)
+            }
         }
     }
 }
 
-impl Into<DiffValue> for Val {
-    fn into(self) -> DiffValue {
-        match self {
+impl From<Val> for DiffValue {
+    fn from(val: Val) -> DiffValue {
+        match val {
             Val::I32(n) => DiffValue::I32(n),
             Val::I64(n) => DiffValue::I64(n),
             Val::F32(n) => DiffValue::F32(n),
@@ -238,6 +246,8 @@ impl Into<DiffValue> for Val {
             Val::ExternRef(r) => DiffValue::ExternRef { null: r.is_none() },
             Val::FuncRef(r) => DiffValue::FuncRef { null: r.is_none() },
             Val::AnyRef(r) => DiffValue::AnyRef { null: r.is_none() },
+            Val::ExnRef(e) => DiffValue::ExnRef { null: e.is_none() },
+            Val::ContRef(c) => DiffValue::ContRef { null: c.is_none() },
         }
     }
 }
