@@ -8,7 +8,7 @@
 
 use arbitrary::{Arbitrary, Unstructured};
 use proc_macro2::{Ident, TokenStream};
-use quote::{format_ident, quote, ToTokens};
+use quote::{ToTokens, format_ident, quote};
 use std::borrow::Cow;
 use std::fmt::{self, Debug, Write};
 use std::iter;
