@@ -25,18 +25,42 @@ pub mod limits {
     pub const CORE_INSTANCES: u32 = 900;
     pub const TABLE_ELEMENTS: usize = 1000;
     pub const CORE_INSTANCE_SIZE: usize = 64 * 1024;
+    pub const TOTAL_STACKS: u32 = 10;
 }
 
 /// Local all `*.wast` tests under `root` which should be the path to the root
 /// of the wasmtime repository.
 pub fn find_tests(root: &Path) -> Result<Vec<WastTest>> {
     let mut tests = Vec::new();
-    add_tests(&mut tests, &root.join("tests/spec_testsuite"), false)?;
-    add_tests(&mut tests, &root.join("tests/misc_testsuite"), true)?;
+
+    let spec_tests = root.join("tests/spec_testsuite");
+    add_tests(
+        &mut tests,
+        &spec_tests,
+        &FindConfig::Infer(spec_test_config),
+    )
+    .with_context(|| format!("failed to add tests from `{}`", spec_tests.display()))?;
+
+    let misc_tests = root.join("tests/misc_testsuite");
+    add_tests(&mut tests, &misc_tests, &FindConfig::InTest)
+        .with_context(|| format!("failed to add tests from `{}`", misc_tests.display()))?;
+
+    let cm_tests = root.join("tests/component-model/test");
+    add_tests(
+        &mut tests,
+        &cm_tests,
+        &FindConfig::Infer(component_test_config),
+    )
+    .with_context(|| format!("failed to add tests from `{}`", cm_tests.display()))?;
     Ok(tests)
 }
 
-fn add_tests(tests: &mut Vec<WastTest>, path: &Path, has_config: bool) -> Result<()> {
+enum FindConfig {
+    InTest,
+    Infer(fn(&Path) -> TestConfig),
+}
+
+fn add_tests(tests: &mut Vec<WastTest>, path: &Path, config: &FindConfig) -> Result<()> {
     for entry in path.read_dir().context("failed to read directory")? {
         let entry = entry.context("failed to read directory entry")?;
         let path = entry.path();
@@ -45,7 +69,7 @@ fn add_tests(tests: &mut Vec<WastTest>, path: &Path, has_config: bool) -> Result
             .context("failed to get file type")?
             .is_dir()
         {
-            add_tests(tests, &path, has_config).context("failed to read sub-directory")?;
+            add_tests(tests, &path, config).context("failed to read sub-directory")?;
             continue;
         }
 
@@ -55,11 +79,10 @@ fn add_tests(tests: &mut Vec<WastTest>, path: &Path, has_config: bool) -> Result
 
         let contents =
             fs::read_to_string(&path).with_context(|| format!("failed to read test: {path:?}"))?;
-        let config = if has_config {
-            parse_test_config(&contents)
-                .with_context(|| format!("failed to parse test configuration: {path:?}"))?
-        } else {
-            spec_test_config(&path)
+        let config = match config {
+            FindConfig::InTest => parse_test_config(&contents, ";;!")
+                .with_context(|| format!("failed to parse test configuration: {path:?}"))?,
+            FindConfig::Infer(f) => f(&path),
         };
         tests.push(WastTest {
             path,
@@ -72,12 +95,8 @@ fn add_tests(tests: &mut Vec<WastTest>, path: &Path, has_config: bool) -> Result
 
 fn spec_test_config(test: &Path) -> TestConfig {
     let mut ret = TestConfig::default();
+    ret.spec_test = Some(true);
     match spec_proposal_from_path(test) {
-        Some("multi-memory") => {
-            ret.multi_memory = Some(true);
-            ret.reference_types = Some(true);
-            ret.simd = Some(true);
-        }
         Some("wide-arithmetic") => {
             ret.wide_arithmetic = Some(true);
         }
@@ -85,47 +104,69 @@ fn spec_test_config(test: &Path) -> TestConfig {
             ret.threads = Some(true);
             ret.reference_types = Some(false);
         }
-        Some("tail-call") => {
-            ret.tail_call = Some(true);
-            ret.reference_types = Some(true);
-        }
-        Some("relaxed-simd") => {
-            ret.relaxed_simd = Some(true);
-        }
-        Some("memory64") => {
-            ret.memory64 = Some(true);
-            ret.tail_call = Some(true);
-            ret.gc = Some(true);
-            ret.extended_const = Some(true);
-            ret.multi_memory = Some(true);
-            ret.relaxed_simd = Some(true);
-        }
-        Some("extended-const") => {
-            ret.extended_const = Some(true);
-            ret.reference_types = Some(true);
-        }
         Some("custom-page-sizes") => {
             ret.custom_page_sizes = Some(true);
             ret.multi_memory = Some(true);
+            ret.memory64 = Some(true);
+
+            // See commentary below in `wasm-3.0` case for why these "hog
+            // memory"
+            if test.ends_with("memory_max.wast") || test.ends_with("memory_max_i64.wast") {
+                ret.hogs_memory = Some(true);
+            }
         }
-        Some("exception-handling") => {
-            ret.reference_types = Some(true);
+        Some("custom-descriptors") => {
+            ret.custom_descriptors = Some(true);
         }
-        Some("gc") => {
-            ret.gc = Some(true);
-            ret.tail_call = Some(true);
-        }
-        Some("function-references") => {
-            ret.function_references = Some(true);
-            ret.tail_call = Some(true);
-        }
-        Some("annotations") => {
-            ret.simd = Some(true);
-        }
-        Some(proposal) => panic!("unsuported proposal {proposal:?}"),
+        Some(proposal) => panic!("unsupported proposal {proposal:?}"),
         None => {
             ret.reference_types = Some(true);
             ret.simd = Some(true);
+            ret.simd = Some(true);
+            ret.relaxed_simd = Some(true);
+            ret.multi_memory = Some(true);
+            ret.gc = Some(true);
+            ret.reference_types = Some(true);
+            ret.memory64 = Some(true);
+            ret.tail_call = Some(true);
+            ret.extended_const = Some(true);
+            ret.exceptions = Some(true);
+
+            if test.parent().unwrap().ends_with("legacy") {
+                ret.legacy_exceptions = Some(true);
+            }
+
+            // These tests technically don't actually hog any memory but they
+            // do have a module definition with a table/memory that is the
+            // maximum size. These modules fail to compile in the pooling
+            // allocator which has limits on the minimum size of
+            // memories/tables by default.
+            //
+            // Pretend that these hog memory to avoid running the tests in the
+            // pooling allocator.
+            if test.ends_with("memory.wast")
+                || test.ends_with("table.wast")
+                || test.ends_with("memory64.wast")
+                || test.ends_with("table64.wast")
+            {
+                ret.hogs_memory = Some(true);
+            }
+        }
+    }
+
+    ret
+}
+
+fn component_test_config(test: &Path) -> TestConfig {
+    let mut ret = TestConfig::default();
+    ret.spec_test = Some(true);
+    ret.reference_types = Some(true);
+    ret.multi_memory = Some(true);
+
+    if let Some(parent) = test.parent() {
+        if parent.ends_with("async") {
+            ret.component_model_async = Some(true);
+            ret.component_model_async_builtins = Some(true);
         }
     }
 
@@ -134,7 +175,7 @@ fn spec_test_config(test: &Path) -> TestConfig {
 
 /// Parse test configuration from the specified test, comments starting with
 /// `;;!`.
-pub fn parse_test_config<T>(wat: &str) -> Result<T>
+pub fn parse_test_config<T>(wat: &str, comment: &'static str) -> Result<T>
 where
     T: DeserializeOwned,
 {
@@ -142,8 +183,8 @@ where
     // prefixed with `;;!`.
     let config_lines: Vec<_> = wat
         .lines()
-        .take_while(|l| l.starts_with(";;!"))
-        .map(|l| &l[3..])
+        .take_while(|l| l.starts_with(comment))
+        .map(|l| &l[comment.len()..])
         .collect();
     let config_text = config_lines.join("\n");
 
@@ -175,6 +216,7 @@ macro_rules! foreach_config_option {
             custom_page_sizes
             multi_memory
             threads
+            shared_everything_threads
             gc
             function_references
             relaxed_simd
@@ -187,8 +229,16 @@ macro_rules! foreach_config_option {
             component_model_async
             component_model_async_builtins
             component_model_async_stackful
+            component_model_threading
+            component_model_error_context
+            component_model_gc
             simd
             gc_types
+            exceptions
+            legacy_exceptions
+            stack_switching
+            spec_test
+            custom_descriptors
         }
     };
 }
@@ -210,7 +260,6 @@ macro_rules! define_test_config {
                 }
             )*
         }
-
     }
 }
 
@@ -280,12 +329,8 @@ impl Compiler {
     /// `Config::compiler_panicking_wasm_features`.
     pub fn should_fail(&self, config: &TestConfig) -> bool {
         match self {
-            // Currently Cranelift supports all wasm proposals that wasmtime
-            // tests.
-            Compiler::CraneliftNative => {}
+            Compiler::CraneliftNative => config.legacy_exceptions(),
 
-            // Winch doesn't have quite the full breadth of support that
-            // Cranelift has quite yet.
             Compiler::Winch => {
                 if config.gc()
                     || config.tail_call()
@@ -293,22 +338,28 @@ impl Compiler {
                     || config.gc()
                     || config.relaxed_simd()
                     || config.gc_types()
+                    || config.exceptions()
+                    || config.legacy_exceptions()
+                    || config.stack_switching()
+                    || config.legacy_exceptions()
+                    || config.component_model_async()
                 {
                     return true;
                 }
+
+                if cfg!(target_arch = "aarch64") {
+                    return config.wide_arithmetic()
+                        || (config.simd() && !config.spec_test())
+                        || config.threads();
+                }
+
+                !cfg!(target_arch = "x86_64")
             }
 
             Compiler::CraneliftPulley => {
-                // Pulley at this time fundamentally does not support threads
-                // due to being unable to implement non-atomic loads/stores
-                // safely.
-                if config.threads() {
-                    return true;
-                }
+                config.threads() || config.legacy_exceptions() || config.stack_switching()
             }
         }
-
-        false
     }
 
     /// Returns whether this compiler configuration supports the current host
@@ -321,9 +372,7 @@ impl Compiler {
                     || cfg!(target_arch = "riscv64")
                     || cfg!(target_arch = "s390x")
             }
-            Compiler::Winch => {
-                cfg!(target_arch = "x86_64")
-            }
+            Compiler::Winch => cfg!(target_arch = "x86_64") || cfg!(target_arch = "aarch64"),
             Compiler::CraneliftPulley => true,
         }
     }
@@ -362,6 +411,7 @@ impl WastTest {
                 "misc_testsuite/memory64/more-than-4gb.wast",
                 // shared memories + pooling allocator aren't supported yet
                 "misc_testsuite/memory-combos.wast",
+                "misc_testsuite/threads/atomics-end-of-memory.wast",
                 "misc_testsuite/threads/LB.wast",
                 "misc_testsuite/threads/LB_atomic.wast",
                 "misc_testsuite/threads/MP.wast",
@@ -386,13 +436,13 @@ impl WastTest {
             return true;
         }
 
-        // Disable spec tests for proposals that Winch does not implement yet.
+        // Disable spec tests per target for proposals that Winch does not implement yet.
         if config.compiler == Compiler::Winch {
+            // Common list for tests that fail in all targets supported by Winch.
             let unsupported = [
-                // externref/reference-types related
-                "component-model/modules.wast",
                 "extended-const/elem.wast",
                 "extended-const/global.wast",
+                "misc_testsuite/component-model/modules.wast",
                 "misc_testsuite/externref-id-function.wast",
                 "misc_testsuite/externref-segment.wast",
                 "misc_testsuite/externref-segments.wast",
@@ -405,10 +455,7 @@ impl WastTest {
                 "misc_testsuite/simple_ref_is_null.wast",
                 "misc_testsuite/table_grow_with_funcref.wast",
                 "spec_testsuite/br_table.wast",
-                "spec_testsuite/data-invalid.wast",
-                "spec_testsuite/elem.wast",
                 "spec_testsuite/global.wast",
-                "spec_testsuite/linking.wast",
                 "spec_testsuite/ref_func.wast",
                 "spec_testsuite/ref_is_null.wast",
                 "spec_testsuite/ref_null.wast",
@@ -418,32 +465,49 @@ impl WastTest {
                 "spec_testsuite/table_grow.wast",
                 "spec_testsuite/table_set.wast",
                 "spec_testsuite/table_size.wast",
-                // simd-related failures
-                "misc_testsuite/simd/canonicalize-nan.wast",
+                "spec_testsuite/elem.wast",
+                "spec_testsuite/linking.wast",
             ];
 
             if unsupported.iter().any(|part| self.path.ends_with(part)) {
                 return true;
             }
 
-            // SIMD on Winch requires AVX instructions.
-            #[cfg(target_arch = "x86_64")]
-            if !(std::is_x86_feature_detected!("avx") && std::is_x86_feature_detected!("avx2")) {
+            #[cfg(target_arch = "aarch64")]
+            {
                 let unsupported = [
-                    "annotations/simd_lane.wast",
-                    "memory64/simd.wast",
                     "misc_testsuite/int-to-float-splat.wast",
                     "misc_testsuite/issue6562.wast",
+                    "misc_testsuite/memory64/simd.wast",
                     "misc_testsuite/simd/almost-extmul.wast",
+                    "misc_testsuite/simd/canonicalize-nan.wast",
                     "misc_testsuite/simd/cvt-from-uint.wast",
-                    "misc_testsuite/simd/issue_3327_bnot_lowering.wast",
+                    "misc_testsuite/simd/edge-of-memory.wast",
+                    "misc_testsuite/simd/interesting-float-splat.wast",
+                    "misc_testsuite/simd/issue4807.wast",
                     "misc_testsuite/simd/issue6725-no-egraph-panic.wast",
+                    "misc_testsuite/simd/issue_3173_select_v128.wast",
+                    "misc_testsuite/simd/issue_3327_bnot_lowering.wast",
+                    "misc_testsuite/simd/load_splat_out_of_bounds.wast",
                     "misc_testsuite/simd/replace-lane-preserve.wast",
                     "misc_testsuite/simd/spillslot-size-fuzzbug.wast",
+                    "misc_testsuite/simd/sse-cannot-fold-unaligned-loads.wast",
+                    "misc_testsuite/simd/unaligned-load.wast",
+                    "misc_testsuite/simd/v128-select.wast",
                     "misc_testsuite/winch/issue-10331.wast",
+                    "misc_testsuite/winch/issue-10357.wast",
+                    "misc_testsuite/winch/issue-10460.wast",
                     "misc_testsuite/winch/replace_lane.wast",
+                    "misc_testsuite/winch/simd_multivalue.wast",
+                    "misc_testsuite/winch/v128_load_lane_invalid_address.wast",
+                    "spec_testsuite/proposals/annotations/simd_lane.wast",
+                    "spec_testsuite/proposals/multi-memory/simd_memory-multi.wast",
+                    "spec_testsuite/simd_address.wast",
                     "spec_testsuite/simd_align.wast",
+                    "spec_testsuite/simd_bit_shift.wast",
+                    "spec_testsuite/simd_bitwise.wast",
                     "spec_testsuite/simd_boolean.wast",
+                    "spec_testsuite/simd_const.wast",
                     "spec_testsuite/simd_conversions.wast",
                     "spec_testsuite/simd_f32x4.wast",
                     "spec_testsuite/simd_f32x4_arith.wast",
@@ -455,84 +519,183 @@ impl WastTest {
                     "spec_testsuite/simd_f64x2_cmp.wast",
                     "spec_testsuite/simd_f64x2_pmin_pmax.wast",
                     "spec_testsuite/simd_f64x2_rounding.wast",
+                    "spec_testsuite/simd_i16x8_arith.wast",
+                    "spec_testsuite/simd_i16x8_arith2.wast",
                     "spec_testsuite/simd_i16x8_cmp.wast",
+                    "spec_testsuite/simd_i16x8_extadd_pairwise_i8x16.wast",
+                    "spec_testsuite/simd_i16x8_extmul_i8x16.wast",
+                    "spec_testsuite/simd_i16x8_q15mulr_sat_s.wast",
+                    "spec_testsuite/simd_i16x8_sat_arith.wast",
+                    "spec_testsuite/simd_i32x4_arith.wast",
+                    "spec_testsuite/simd_i32x4_arith2.wast",
                     "spec_testsuite/simd_i32x4_cmp.wast",
+                    "spec_testsuite/simd_i32x4_dot_i16x8.wast",
+                    "spec_testsuite/simd_i32x4_extadd_pairwise_i16x8.wast",
+                    "spec_testsuite/simd_i32x4_extmul_i16x8.wast",
+                    "spec_testsuite/simd_i32x4_trunc_sat_f32x4.wast",
+                    "spec_testsuite/simd_i32x4_trunc_sat_f64x2.wast",
+                    "spec_testsuite/simd_i64x2_arith.wast",
                     "spec_testsuite/simd_i64x2_arith2.wast",
                     "spec_testsuite/simd_i64x2_cmp.wast",
+                    "spec_testsuite/simd_i64x2_extmul_i32x4.wast",
+                    "spec_testsuite/simd_i8x16_arith.wast",
                     "spec_testsuite/simd_i8x16_arith2.wast",
                     "spec_testsuite/simd_i8x16_cmp.wast",
+                    "spec_testsuite/simd_i8x16_sat_arith.wast",
                     "spec_testsuite/simd_int_to_int_extend.wast",
+                    "spec_testsuite/simd_lane.wast",
                     "spec_testsuite/simd_load.wast",
-                    "spec_testsuite/simd_load_extend.wast",
-                    "spec_testsuite/simd_load_splat.wast",
-                    "spec_testsuite/simd_load_zero.wast",
-                    "spec_testsuite/simd_splat.wast",
-                    "spec_testsuite/simd_store16_lane.wast",
-                    "spec_testsuite/simd_store32_lane.wast",
-                    "spec_testsuite/simd_store64_lane.wast",
-                    "spec_testsuite/simd_store8_lane.wast",
                     "spec_testsuite/simd_load16_lane.wast",
                     "spec_testsuite/simd_load32_lane.wast",
                     "spec_testsuite/simd_load64_lane.wast",
                     "spec_testsuite/simd_load8_lane.wast",
-                    "spec_testsuite/simd_bitwise.wast",
-                    "misc_testsuite/simd/load_splat_out_of_bounds.wast",
-                    "misc_testsuite/simd/unaligned-load.wast",
-                    "multi-memory/simd_memory-multi.wast",
-                    "misc_testsuite/simd/issue4807.wast",
-                    "spec_testsuite/simd_const.wast",
-                    "spec_testsuite/simd_i8x16_sat_arith.wast",
-                    "spec_testsuite/simd_i64x2_arith.wast",
-                    "spec_testsuite/simd_i16x8_arith.wast",
-                    "spec_testsuite/simd_i16x8_arith2.wast",
-                    "spec_testsuite/simd_i16x8_q15mulr_sat_s.wast",
-                    "spec_testsuite/simd_i16x8_sat_arith.wast",
-                    "spec_testsuite/simd_i32x4_arith.wast",
-                    "spec_testsuite/simd_i32x4_dot_i16x8.wast",
-                    "spec_testsuite/simd_i32x4_trunc_sat_f32x4.wast",
-                    "spec_testsuite/simd_i32x4_trunc_sat_f64x2.wast",
-                    "spec_testsuite/simd_i8x16_arith.wast",
-                    "spec_testsuite/simd_bit_shift.wast",
-                    "spec_testsuite/simd_lane.wast",
-                    "spec_testsuite/simd_i16x8_extmul_i8x16.wast",
-                    "spec_testsuite/simd_i32x4_extmul_i16x8.wast",
-                    "spec_testsuite/simd_i64x2_extmul_i32x4.wast",
-                    "spec_testsuite/simd_i16x8_extadd_pairwise_i8x16.wast",
-                    "spec_testsuite/simd_i32x4_extadd_pairwise_i16x8.wast",
-                    "spec_testsuite/simd_i32x4_arith2.wast",
+                    "spec_testsuite/simd_load_extend.wast",
+                    "spec_testsuite/simd_load_splat.wast",
+                    "spec_testsuite/simd_load_zero.wast",
+                    "spec_testsuite/simd_select.wast",
+                    "spec_testsuite/simd_splat.wast",
+                    "spec_testsuite/simd_store.wast",
+                    "spec_testsuite/simd_store16_lane.wast",
+                    "spec_testsuite/simd_store32_lane.wast",
+                    "spec_testsuite/simd_store64_lane.wast",
+                    "spec_testsuite/simd_store8_lane.wast",
                 ];
 
                 if unsupported.iter().any(|part| self.path.ends_with(part)) {
                     return true;
                 }
             }
-        }
 
-        for part in self.path.iter() {
-            // Not implemented in Wasmtime yet
-            if part == "exception-handling" {
-                return !self.path.ends_with("binary.wast") && !self.path.ends_with("exports.wast");
-            }
+            #[cfg(target_arch = "x86_64")]
+            {
+                let unsupported = [
+                    // externref/reference-types related
+                    // simd-related failures
+                    "misc_testsuite/simd/canonicalize-nan.wast",
+                ];
 
-            if part == "memory64" {
-                if [
-                    // wasmtime doesn't implement exceptions yet
-                    "imports.wast",
-                    "ref_null.wast",
-                    "throw.wast",
-                    "throw_ref.wast",
-                    "try_table.wast",
-                    "tag.wast",
-                    "instance.wast",
-                ]
-                .iter()
-                .any(|i| self.path.ends_with(i))
-                {
+                if unsupported.iter().any(|part| self.path.ends_with(part)) {
                     return true;
                 }
+
+                // SIMD on Winch requires AVX instructions.
+                #[cfg(target_arch = "x86_64")]
+                if !(std::is_x86_feature_detected!("avx") && std::is_x86_feature_detected!("avx2"))
+                {
+                    let unsupported = [
+                        "annotations/simd_lane.wast",
+                        "memory64/simd.wast",
+                        "misc_testsuite/int-to-float-splat.wast",
+                        "misc_testsuite/issue6562.wast",
+                        "misc_testsuite/simd/almost-extmul.wast",
+                        "misc_testsuite/simd/cvt-from-uint.wast",
+                        "misc_testsuite/simd/edge-of-memory.wast",
+                        "misc_testsuite/simd/issue_3327_bnot_lowering.wast",
+                        "misc_testsuite/simd/issue6725-no-egraph-panic.wast",
+                        "misc_testsuite/simd/replace-lane-preserve.wast",
+                        "misc_testsuite/simd/spillslot-size-fuzzbug.wast",
+                        "misc_testsuite/simd/sse-cannot-fold-unaligned-loads.wast",
+                        "misc_testsuite/winch/issue-10331.wast",
+                        "misc_testsuite/winch/replace_lane.wast",
+                        "spec_testsuite/simd_align.wast",
+                        "spec_testsuite/simd_boolean.wast",
+                        "spec_testsuite/simd_conversions.wast",
+                        "spec_testsuite/simd_f32x4.wast",
+                        "spec_testsuite/simd_f32x4_arith.wast",
+                        "spec_testsuite/simd_f32x4_cmp.wast",
+                        "spec_testsuite/simd_f32x4_pmin_pmax.wast",
+                        "spec_testsuite/simd_f32x4_rounding.wast",
+                        "spec_testsuite/simd_f64x2.wast",
+                        "spec_testsuite/simd_f64x2_arith.wast",
+                        "spec_testsuite/simd_f64x2_cmp.wast",
+                        "spec_testsuite/simd_f64x2_pmin_pmax.wast",
+                        "spec_testsuite/simd_f64x2_rounding.wast",
+                        "spec_testsuite/simd_i16x8_cmp.wast",
+                        "spec_testsuite/simd_i32x4_cmp.wast",
+                        "spec_testsuite/simd_i64x2_arith2.wast",
+                        "spec_testsuite/simd_i64x2_cmp.wast",
+                        "spec_testsuite/simd_i8x16_arith2.wast",
+                        "spec_testsuite/simd_i8x16_cmp.wast",
+                        "spec_testsuite/simd_int_to_int_extend.wast",
+                        "spec_testsuite/simd_load.wast",
+                        "spec_testsuite/simd_load_extend.wast",
+                        "spec_testsuite/simd_load_splat.wast",
+                        "spec_testsuite/simd_load_zero.wast",
+                        "spec_testsuite/simd_splat.wast",
+                        "spec_testsuite/simd_store16_lane.wast",
+                        "spec_testsuite/simd_store32_lane.wast",
+                        "spec_testsuite/simd_store64_lane.wast",
+                        "spec_testsuite/simd_store8_lane.wast",
+                        "spec_testsuite/simd_load16_lane.wast",
+                        "spec_testsuite/simd_load32_lane.wast",
+                        "spec_testsuite/simd_load64_lane.wast",
+                        "spec_testsuite/simd_load8_lane.wast",
+                        "spec_testsuite/simd_bitwise.wast",
+                        "misc_testsuite/simd/load_splat_out_of_bounds.wast",
+                        "misc_testsuite/simd/unaligned-load.wast",
+                        "multi-memory/simd_memory-multi.wast",
+                        "misc_testsuite/simd/issue4807.wast",
+                        "spec_testsuite/simd_const.wast",
+                        "spec_testsuite/simd_i8x16_sat_arith.wast",
+                        "spec_testsuite/simd_i64x2_arith.wast",
+                        "spec_testsuite/simd_i16x8_arith.wast",
+                        "spec_testsuite/simd_i16x8_arith2.wast",
+                        "spec_testsuite/simd_i16x8_q15mulr_sat_s.wast",
+                        "spec_testsuite/simd_i16x8_sat_arith.wast",
+                        "spec_testsuite/simd_i32x4_arith.wast",
+                        "spec_testsuite/simd_i32x4_dot_i16x8.wast",
+                        "spec_testsuite/simd_i32x4_trunc_sat_f32x4.wast",
+                        "spec_testsuite/simd_i32x4_trunc_sat_f64x2.wast",
+                        "spec_testsuite/simd_i8x16_arith.wast",
+                        "spec_testsuite/simd_bit_shift.wast",
+                        "spec_testsuite/simd_lane.wast",
+                        "spec_testsuite/simd_i16x8_extmul_i8x16.wast",
+                        "spec_testsuite/simd_i32x4_extmul_i16x8.wast",
+                        "spec_testsuite/simd_i64x2_extmul_i32x4.wast",
+                        "spec_testsuite/simd_i16x8_extadd_pairwise_i8x16.wast",
+                        "spec_testsuite/simd_i32x4_extadd_pairwise_i16x8.wast",
+                        "spec_testsuite/simd_i32x4_arith2.wast",
+                    ];
+
+                    if unsupported.iter().any(|part| self.path.ends_with(part)) {
+                        return true;
+                    }
+                }
             }
         }
 
+        let failing_component_model_tests = [
+            // FIXME(#11683)
+            "component-model/test/values/trap-in-post-return.wast",
+            // Awaiting https://github.com/WebAssembly/component-model/pull/570
+            "component-model/test/resources/multiple-resources.wast",
+            "component-model/test/async/empty-wait.wast",
+            "component-model/test/async/drop-stream.wast",
+            "component-model/test/async/passing-resources.wast",
+            "component-model/test/async/async-calls-sync.wast",
+            "component-model/test/async/partial-stream-copies.wast",
+            "component-model/test/async/futures-must-write.wast",
+            "component-model/test/async/cancel-stream.wast",
+            "component-model/test/async/drop-waitable-set.wast",
+        ];
+        if failing_component_model_tests
+            .iter()
+            .any(|part| self.path.ends_with(part))
+        {
+            return true;
+        }
+
+        // Not implemented in Wasmtime anywhere yet.
+        if self.config.custom_descriptors() {
+            let happens_to_work =
+                ["spec_testsuite/proposals/custom-descriptors/binary-leb128.wast"];
+
+            if happens_to_work.iter().any(|part| self.path.ends_with(part)) {
+                return false;
+            }
+            return true;
+        }
+
         false
     }
 }
