@@ -4,17 +4,15 @@ use crate::{bindings::http::types, types::FieldMap};
 use anyhow::anyhow;
 use bytes::Bytes;
 use http_body::{Body, Frame};
-use http_body_util::combinators::BoxBody;
 use http_body_util::BodyExt;
+use http_body_util::combinators::BoxBody;
 use std::future::Future;
 use std::mem;
 use std::task::{Context, Poll};
 use std::{pin::Pin, sync::Arc, time::Duration};
 use tokio::sync::{mpsc, oneshot};
-use wasmtime_wasi::{
-    runtime::{poll_noop, AbortOnDropJoinHandle},
-    InputStream, OutputStream, Pollable, StreamError,
-};
+use wasmtime_wasi::p2::{InputStream, OutputStream, Pollable, StreamError};
+use wasmtime_wasi::runtime::{AbortOnDropJoinHandle, poll_noop};
 
 /// Common type for incoming bodies.
 pub type HyperIncomingBody = BoxBody<Bytes, types::ErrorCode>;
@@ -22,7 +20,7 @@ pub type HyperIncomingBody = BoxBody<Bytes, types::ErrorCode>;
 /// Common type for outgoing bodies.
 pub type HyperOutgoingBody = BoxBody<Bytes, types::ErrorCode>;
 
-/// The concrete type behind a `was:http/types/incoming-body` resource.
+/// The concrete type behind a `was:http/types.incoming-body` resource.
 #[derive(Debug)]
 pub struct HostIncomingBody {
     body: IncomingBodyState,
@@ -160,8 +158,8 @@ enum StreamEnd {
     Trailers(Option<FieldMap>),
 }
 
-/// The concrete type behind the `wasi:io/streams/input-stream` resource returned
-/// by `wasi:http/types/incoming-body`'s `stream` method.
+/// The concrete type behind the `wasi:io/streams.input-stream` resource returned
+/// by `wasi:http/types.incoming-body`'s `stream` method.
 #[derive(Debug)]
 pub struct HostIncomingBodyStream {
     state: IncomingBodyStreamState,
@@ -298,7 +296,7 @@ impl Drop for HostIncomingBodyStream {
     }
 }
 
-/// The concrete type behind a `wasi:http/types/future-trailers` resource.
+/// The concrete type behind a `wasi:http/types.future-trailers` resource.
 #[derive(Debug)]
 pub enum HostFutureTrailers {
     /// Trailers aren't here yet.
@@ -412,7 +410,7 @@ impl WrittenState {
     }
 }
 
-/// The concrete type behind a `wasi:http/types/outgoing-body` resource.
+/// The concrete type behind a `wasi:http/types.outgoing-body` resource.
 pub struct HostOutgoingBody {
     /// The output stream that the body is written to.
     body_output_stream: Option<Box<dyn OutputStream>>,
@@ -529,7 +527,7 @@ impl HostOutgoingBody {
         };
 
         // Ignoring failure: receiver died sending body, but we can't report that here.
-        let _ = sender.send(message.into());
+        let _ = sender.send(message);
 
         Ok(())
     }
@@ -615,9 +613,9 @@ impl OutputStream for BodyWriteStream {
                 if let Some(written) = self.written.as_ref() {
                     if !written.update(len) {
                         let total = written.written();
-                        return Err(StreamError::LastOperationFailed(anyhow!(self
-                            .context
-                            .as_body_size_error(total))));
+                        return Err(StreamError::LastOperationFailed(anyhow!(
+                            self.context.as_body_size_error(total)
+                        )));
                     }
                 }
 
