@@ -1,12 +1,9 @@
 //! Implements the base structure (i.e. [WasiHttpCtx]) that will provide the
 //! implementation of the wasi-http API.
 
-use crate::io::TokioIo;
 use crate::{
     bindings::http::types::{self, Method, Scheme},
     body::{HostIncomingBody, HyperIncomingBody, HyperOutgoingBody},
-    error::dns_error,
-    hyper_request_error,
 };
 use anyhow::bail;
 use bytes::Bytes;
@@ -15,10 +12,17 @@ use hyper::body::Body;
 use hyper::header::HeaderName;
 use std::any::Any;
 use std::time::Duration;
-use tokio::net::TcpStream;
-use tokio::time::timeout;
 use wasmtime::component::{Resource, ResourceTable};
-use wasmtime_wasi::{runtime::AbortOnDropJoinHandle, IoImpl, IoView, Pollable};
+use wasmtime_wasi::p2::Pollable;
+use wasmtime_wasi::runtime::AbortOnDropJoinHandle;
+
+#[cfg(feature = "default-send-request")]
+use {
+    crate::io::TokioIo,
+    crate::{error::dns_error, hyper_request_error},
+    tokio::net::TcpStream,
+    tokio::time::timeout,
+};
 
 /// Capture the state necessary for use in the wasi-http API implementation.
 #[derive(Debug)]
@@ -39,7 +43,7 @@ impl WasiHttpCtx {
 ///
 /// ```
 /// use wasmtime::component::ResourceTable;
-/// use wasmtime_wasi::{IoView, WasiCtx, WasiView, WasiCtxBuilder};
+/// use wasmtime_wasi::{WasiCtx, WasiCtxView, WasiView};
 /// use wasmtime_wasi_http::{WasiHttpCtx, WasiHttpView};
 ///
 /// struct MyState {
@@ -48,20 +52,20 @@ impl WasiHttpCtx {
 ///     table: ResourceTable,
 /// }
 ///
-/// impl IoView for MyState {
-///     fn table(&mut self) -> &mut ResourceTable { &mut self.table }
-/// }
 /// impl WasiHttpView for MyState {
 ///     fn ctx(&mut self) -> &mut WasiHttpCtx { &mut self.http_ctx }
+///     fn table(&mut self) -> &mut ResourceTable { &mut self.table }
 /// }
 ///
 /// impl WasiView for MyState {
-///     fn ctx(&mut self) -> &mut WasiCtx { &mut self.ctx }
+///     fn ctx(&mut self) -> WasiCtxView<'_> {
+///         WasiCtxView { ctx: &mut self.ctx, table: &mut self.table }
+///     }
 /// }
 ///
 /// impl MyState {
 ///     fn new() -> MyState {
-///         let mut wasi = WasiCtxBuilder::new();
+///         let mut wasi = WasiCtx::builder();
 ///         wasi.arg("./foo.wasm");
 ///         wasi.arg("--help");
 ///         wasi.env("FOO", "bar");
@@ -74,10 +78,13 @@ impl WasiHttpCtx {
 ///     }
 /// }
 /// ```
-pub trait WasiHttpView: IoView {
+pub trait WasiHttpView {
     /// Returns a mutable reference to the WASI HTTP context.
     fn ctx(&mut self) -> &mut WasiHttpCtx;
 
+    /// Returns the table used to manage resources.
+    fn table(&mut self) -> &mut ResourceTable;
+
     /// Create a new incoming request resource.
     fn new_incoming_request<B>(
         &mut self,
@@ -111,6 +118,7 @@ pub trait WasiHttpView: IoView {
     }
 
     /// Send an outgoing request.
+    #[cfg(feature = "default-send-request")]
     fn send_request(
         &mut self,
         request: hyper::Request<HyperOutgoingBody>,
@@ -119,9 +127,17 @@ pub trait WasiHttpView: IoView {
         Ok(default_send_request(request, config))
     }
 
+    /// Send an outgoing request.
+    #[cfg(not(feature = "default-send-request"))]
+    fn send_request(
+        &mut self,
+        request: hyper::Request<HyperOutgoingBody>,
+        config: OutgoingRequestConfig,
+    ) -> crate::HttpResult<HostFutureIncomingResponse>;
+
     /// Whether a given header should be considered forbidden and not allowed.
-    fn is_forbidden_header(&mut self, _name: &HeaderName) -> bool {
-        false
+    fn is_forbidden_header(&mut self, name: &HeaderName) -> bool {
+        DEFAULT_FORBIDDEN_HEADERS.contains(name)
     }
 
     /// Number of distinct write calls to the outgoing body's output-stream
@@ -148,6 +164,10 @@ impl<T: ?Sized + WasiHttpView> WasiHttpView for &mut T {
         T::ctx(self)
     }
 
+    fn table(&mut self) -> &mut ResourceTable {
+        T::table(self)
+    }
+
     fn new_response_outparam(
         &mut self,
         result: tokio::sync::oneshot::Sender<
@@ -183,6 +203,10 @@ impl<T: ?Sized + WasiHttpView> WasiHttpView for Box<T> {
         T::ctx(self)
     }
 
+    fn table(&mut self) -> &mut ResourceTable {
+        T::table(self)
+    }
+
     fn new_response_outparam(
         &mut self,
         result: tokio::sync::oneshot::Sender<
@@ -226,16 +250,15 @@ impl<T: ?Sized + WasiHttpView> WasiHttpView for Box<T> {
 /// [`add_to_linker_sync`](crate::add_to_linker_sync)
 /// and doesn't need to be manually configured.
 #[repr(transparent)]
-pub struct WasiHttpImpl<T>(pub IoImpl<T>);
+pub struct WasiHttpImpl<T>(pub T);
 
-impl<T: IoView> IoView for WasiHttpImpl<T> {
-    fn table(&mut self) -> &mut ResourceTable {
-        T::table(&mut self.0 .0)
-    }
-}
 impl<T: WasiHttpView> WasiHttpView for WasiHttpImpl<T> {
     fn ctx(&mut self) -> &mut WasiHttpCtx {
-        self.0 .0.ctx()
+        self.0.ctx()
+    }
+
+    fn table(&mut self) -> &mut ResourceTable {
+        self.0.table()
     }
 
     fn new_response_outparam(
@@ -244,7 +267,7 @@ impl<T: WasiHttpView> WasiHttpView for WasiHttpImpl<T> {
             Result<hyper::Response<HyperOutgoingBody>, types::ErrorCode>,
         >,
     ) -> wasmtime::Result<Resource<HostResponseOutparam>> {
-        self.0 .0.new_response_outparam(result)
+        self.0.new_response_outparam(result)
     }
 
     fn send_request(
@@ -252,39 +275,35 @@ impl<T: WasiHttpView> WasiHttpView for WasiHttpImpl<T> {
         request: hyper::Request<HyperOutgoingBody>,
         config: OutgoingRequestConfig,
     ) -> crate::HttpResult<HostFutureIncomingResponse> {
-        self.0 .0.send_request(request, config)
+        self.0.send_request(request, config)
     }
 
     fn is_forbidden_header(&mut self, name: &HeaderName) -> bool {
-        self.0 .0.is_forbidden_header(name)
+        self.0.is_forbidden_header(name)
     }
 
     fn outgoing_body_buffer_chunks(&mut self) -> usize {
-        self.0 .0.outgoing_body_buffer_chunks()
+        self.0.outgoing_body_buffer_chunks()
     }
 
     fn outgoing_body_chunk_size(&mut self) -> usize {
-        self.0 .0.outgoing_body_chunk_size()
+        self.0.outgoing_body_chunk_size()
     }
 }
 
-/// Returns `true` when the header is forbidden according to this [`WasiHttpView`] implementation.
-pub(crate) fn is_forbidden_header(view: &mut dyn WasiHttpView, name: &HeaderName) -> bool {
-    static FORBIDDEN_HEADERS: [HeaderName; 10] = [
-        hyper::header::CONNECTION,
-        HeaderName::from_static("keep-alive"),
-        hyper::header::PROXY_AUTHENTICATE,
-        hyper::header::PROXY_AUTHORIZATION,
-        HeaderName::from_static("proxy-connection"),
-        hyper::header::TE,
-        hyper::header::TRANSFER_ENCODING,
-        hyper::header::UPGRADE,
-        hyper::header::HOST,
-        HeaderName::from_static("http2-settings"),
-    ];
-
-    FORBIDDEN_HEADERS.contains(name) || view.is_forbidden_header(name)
-}
+/// Set of [http::header::HeaderName], that are forbidden by default
+/// for requests and responses originating in the guest.
+pub const DEFAULT_FORBIDDEN_HEADERS: [http::header::HeaderName; 9] = [
+    hyper::header::CONNECTION,
+    HeaderName::from_static("keep-alive"),
+    hyper::header::PROXY_AUTHENTICATE,
+    hyper::header::PROXY_AUTHORIZATION,
+    HeaderName::from_static("proxy-connection"),
+    hyper::header::TRANSFER_ENCODING,
+    hyper::header::UPGRADE,
+    hyper::header::HOST,
+    HeaderName::from_static("http2-settings"),
+];
 
 /// Removes forbidden headers from a [`hyper::HeaderMap`].
 pub(crate) fn remove_forbidden_headers(
@@ -292,7 +311,7 @@ pub(crate) fn remove_forbidden_headers(
     headers: &mut hyper::HeaderMap,
 ) {
     let forbidden_keys = Vec::from_iter(headers.keys().filter_map(|name| {
-        if is_forbidden_header(view, name) {
+        if view.is_forbidden_header(name) {
             Some(name.clone())
         } else {
             None
@@ -320,6 +339,7 @@ pub struct OutgoingRequestConfig {
 ///
 /// This implementation is used by the `wasi:http/outgoing-handler` interface
 /// default implementation.
+#[cfg(feature = "default-send-request")]
 pub fn default_send_request(
     request: hyper::Request<HyperOutgoingBody>,
     config: OutgoingRequestConfig,
@@ -334,6 +354,7 @@ pub fn default_send_request(
 /// in a task.
 ///
 /// This is called from [default_send_request] to actually send the request.
+#[cfg(feature = "default-send-request")]
 pub async fn default_send_request_handler(
     mut request: hyper::Request<HyperOutgoingBody>,
     OutgoingRequestConfig {
@@ -509,7 +530,7 @@ impl TryInto<http::Method> for types::Method {
     }
 }
 
-/// The concrete type behind a `wasi:http/types/incoming-request` resource.
+/// The concrete type behind a `wasi:http/types.incoming-request` resource.
 #[derive(Debug)]
 pub struct HostIncomingRequest {
     pub(crate) parts: http::request::Parts,
@@ -545,14 +566,14 @@ impl HostIncomingRequest {
     }
 }
 
-/// The concrete type behind a `wasi:http/types/response-outparam` resource.
+/// The concrete type behind a `wasi:http/types.response-outparam` resource.
 pub struct HostResponseOutparam {
     /// The sender for sending a response.
     pub result:
         tokio::sync::oneshot::Sender<Result<hyper::Response<HyperOutgoingBody>, types::ErrorCode>>,
 }
 
-/// The concrete type behind a `wasi:http/types/outgoing-response` resource.
+/// The concrete type behind a `wasi:http/types.outgoing-response` resource.
 pub struct HostOutgoingResponse {
     /// The status of the response.
     pub status: http::StatusCode,
@@ -585,7 +606,7 @@ impl TryFrom<HostOutgoingResponse> for hyper::Response<HyperOutgoingBody> {
     }
 }
 
-/// The concrete type behind a `wasi:http/types/outgoing-request` resource.
+/// The concrete type behind a `wasi:http/types.outgoing-request` resource.
 #[derive(Debug)]
 pub struct HostOutgoingRequest {
     /// The method of the request.
@@ -602,7 +623,7 @@ pub struct HostOutgoingRequest {
     pub body: Option<HyperOutgoingBody>,
 }
 
-/// The concrete type behind a `wasi:http/types/request-options` resource.
+/// The concrete type behind a `wasi:http/types.request-options` resource.
 #[derive(Debug, Default)]
 pub struct HostRequestOptions {
     /// How long to wait for a connection to be established.
@@ -613,7 +634,7 @@ pub struct HostRequestOptions {
     pub between_bytes_timeout: Option<std::time::Duration>,
 }
 
-/// The concrete type behind a `wasi:http/types/incoming-response` resource.
+/// The concrete type behind a `wasi:http/types.incoming-response` resource.
 #[derive(Debug)]
 pub struct HostIncomingResponse {
     /// The response status
@@ -624,7 +645,7 @@ pub struct HostIncomingResponse {
     pub body: Option<HostIncomingBody>,
 }
 
-/// The concrete type behind a `wasi:http/types/fields` resource.
+/// The concrete type behind a `wasi:http/types.fields` resource.
 #[derive(Debug)]
 pub enum HostFields {
     /// A reference to the fields of a parent entry.
@@ -664,7 +685,7 @@ pub struct IncomingResponse {
     pub between_bytes_timeout: std::time::Duration,
 }
 
-/// The concrete type behind a `wasi:http/types/future-incoming-response` resource.
+/// The concrete type behind a `wasi:http/types.future-incoming-response` resource.
 #[derive(Debug)]
 pub enum HostFutureIncomingResponse {
     /// A pending response
