@@ -1,3 +1,4 @@
+use crate::Config;
 use crate::prelude::*;
 #[cfg(feature = "runtime")]
 pub use crate::runtime::code_memory::CustomCodeMemory;
@@ -5,7 +6,6 @@ pub use crate::runtime::code_memory::CustomCodeMemory;
 use crate::runtime::type_registry::TypeRegistry;
 #[cfg(feature = "runtime")]
 use crate::runtime::vm::GcRuntime;
-use crate::Config;
 use alloc::sync::Arc;
 use core::ptr::NonNull;
 #[cfg(target_has_atomic = "64")]
@@ -65,7 +65,6 @@ struct EngineInner {
 
     /// One-time check of whether the compiler's settings, if present, are
     /// compatible with the native host.
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
     compatible_with_native_host: crate::sync::OnceLock<Result<(), String>>,
 }
 
@@ -93,12 +92,11 @@ impl Engine {
     /// configurations are incompatible.
     ///
     /// For example, feature `reference_types` will need to set
-    /// the compiler setting `enable_safepoints` and `unwind_info`
-    /// to `true`, but explicitly disable these two compiler settings
-    /// will cause errors.
+    /// the compiler setting `unwind_info` to `true`, but explicitly
+    /// disable these two compiler settings will cause errors.
     pub fn new(config: &Config) -> Result<Engine> {
         let config = config.clone();
-        let (tunables, features) = config.validate()?;
+        let (mut tunables, features) = config.validate()?;
 
         #[cfg(feature = "runtime")]
         if tunables.signals_based_traps {
@@ -115,14 +113,26 @@ impl Engine {
         }
 
         #[cfg(any(feature = "cranelift", feature = "winch"))]
-        let (config, compiler) = config.build_compiler(&tunables, features)?;
+        let (config, compiler) = config.build_compiler(&mut tunables, features)?;
+        #[cfg(not(any(feature = "cranelift", feature = "winch")))]
+        let _ = &mut tunables;
 
         Ok(Engine {
             inner: Arc::new(EngineInner {
                 #[cfg(any(feature = "cranelift", feature = "winch"))]
                 compiler,
                 #[cfg(feature = "runtime")]
-                allocator: config.build_allocator(&tunables)?,
+                allocator: {
+                    let allocator = config.build_allocator(&tunables)?;
+                    #[cfg(feature = "gc")]
+                    {
+                        let mem_ty = tunables.gc_heap_memory_type();
+                        allocator.validate_memory(&mem_ty).context(
+                            "instance allocator cannot support configured GC heap memory",
+                        )?;
+                    }
+                    allocator
+                },
                 #[cfg(feature = "runtime")]
                 gc_runtime: config.build_gc_runtime()?,
                 #[cfg(feature = "runtime")]
@@ -131,7 +141,6 @@ impl Engine {
                 signatures: TypeRegistry::new(),
                 #[cfg(all(feature = "runtime", target_has_atomic = "64"))]
                 epoch: AtomicU64::new(0),
-                #[cfg(any(feature = "cranelift", feature = "winch"))]
                 compatible_with_native_host: Default::default(),
                 config,
                 tunables,
@@ -185,6 +194,40 @@ impl Engine {
             .collect::<Result<Vec<B>, E>>()
     }
 
+    #[cfg(any(feature = "cranelift", feature = "winch"))]
+    pub(crate) fn run_maybe_parallel_mut<
+        T: Send,
+        E: Send,
+        F: Fn(&mut T) -> Result<(), E> + Send + Sync,
+    >(
+        &self,
+        input: &mut [T],
+        f: F,
+    ) -> Result<(), E> {
+        if self.config().parallel_compilation {
+            #[cfg(feature = "parallel-compilation")]
+            {
+                use rayon::prelude::*;
+                // If we collect into `Result<(), E>` directly, the returned
+                // error is not deterministic, because any error could be
+                // returned early. So we first materialize all results in order
+                // and then return the first error deterministically, or
+                // `Ok(_)`.
+                return input
+                    .into_par_iter()
+                    .map(|a| f(a))
+                    .collect::<Vec<Result<(), E>>>()
+                    .into_iter()
+                    .collect::<Result<(), E>>();
+            }
+        }
+
+        // In case the parallel-compilation feature is disabled or the
+        // parallel_compilation config was turned off dynamically fallback to
+        // the non-parallel version.
+        input.into_iter().map(|a| f(a)).collect::<Result<(), E>>()
+    }
+
     /// Take a weak reference to this engine.
     pub fn weak(&self) -> EngineWeak {
         EngineWeak {
@@ -249,7 +292,6 @@ impl Engine {
     /// engine can indeed load modules for the configured compiler (if any).
     /// Note that if cranelift is disabled this trivially returns `Ok` because
     /// loaded serialized modules are checked separately.
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
     pub(crate) fn check_compatible_with_native_host(&self) -> Result<()> {
         self.inner
             .compatible_with_native_host
@@ -258,18 +300,16 @@ impl Engine {
             .map_err(anyhow::Error::msg)
     }
 
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
     fn _check_compatible_with_native_host(&self) -> Result<(), String> {
         use target_lexicon::Triple;
 
-        let compiler = self.compiler();
-
-        let target = compiler.triple();
         let host = Triple::host();
+        let target = self.config().compiler_target();
+
         let target_matches_host = || {
             // If the host target and target triple match, then it's valid
             // to run results of compilation on this host.
-            if host == *target {
+            if host == target {
                 return true;
             }
 
@@ -293,12 +333,16 @@ impl Engine {
             ));
         }
 
-        // Also double-check all compiler settings
-        for (key, value) in compiler.flags().iter() {
-            self.check_compatible_with_shared_flag(key, value)?;
-        }
-        for (key, value) in compiler.isa_flags().iter() {
-            self.check_compatible_with_isa_flag(key, value)?;
+        #[cfg(any(feature = "cranelift", feature = "winch"))]
+        {
+            let compiler = self.compiler();
+            // Also double-check all compiler settings
+            for (key, value) in compiler.flags().iter() {
+                self.check_compatible_with_shared_flag(key, value)?;
+            }
+            for (key, value) in compiler.isa_flags().iter() {
+                self.check_compatible_with_isa_flag(key, value)?;
+            }
         }
 
         // Double-check that this configuration isn't requesting capabilities
@@ -312,6 +356,22 @@ impl Engine {
         if !cfg!(target_has_atomic = "64") && self.tunables().epoch_interruption {
             return Err("epochs currently require 64-bit atomics".into());
         }
+
+        // Double-check that the host's float ABI matches Cranelift's float ABI.
+        // See `Config::x86_float_abi_ok` for some more
+        // information.
+        if target == target_lexicon::triple!("x86_64-unknown-none")
+            && self.config().x86_float_abi_ok != Some(true)
+        {
+            return Err("\
+the x86_64-unknown-none target by default uses a soft-float ABI that is \
+incompatible with Cranelift and Wasmtime -- use \
+`Config::x86_float_abi_ok` to disable this check and see more \
+information about this check\
+"
+            .into());
+        }
+
         Ok(())
     }
 
@@ -343,10 +403,10 @@ impl Engine {
             // These settings must all have be enabled, since their value
             // can affect the way the generated code performs or behaves at
             // runtime.
-            "libcall_call_conv" => *value == FlagValue::Enum("isa_default".into()),
+            "libcall_call_conv" => *value == FlagValue::Enum("isa_default"),
             "preserve_frame_pointers" => *value == FlagValue::Bool(true),
             "enable_probestack" => *value == FlagValue::Bool(true),
-            "probestack_strategy" => *value == FlagValue::Enum("inline".into()),
+            "probestack_strategy" => *value == FlagValue::Enum("inline"),
             "enable_multi_ret_implicit_sret" => *value == FlagValue::Bool(true),
 
             // Features wasmtime doesn't use should all be disabled, since
@@ -357,21 +417,28 @@ impl Engine {
             "use_colocated_libcalls" => *value == FlagValue::Bool(false),
             "use_pinned_reg_as_heap_base" => *value == FlagValue::Bool(false),
 
-            // If reference types (or anything that depends on reference types,
-            // like typed function references and GC) are enabled this must be
-            // enabled, otherwise this setting can have any value.
-            "enable_safepoints" => {
-                if self.features().contains(WasmFeatures::REFERENCE_TYPES) {
+            // Windows requires unwind info as part of its ABI.
+            "unwind_info" => {
+                if target.operating_system == target_lexicon::OperatingSystem::Windows {
                     *value == FlagValue::Bool(true)
                 } else {
                     return Ok(())
                 }
             }
 
-            // Windows requires unwind info as part of its ABI.
-            "unwind_info" => {
-                if target.operating_system == target_lexicon::OperatingSystem::Windows {
-                    *value == FlagValue::Bool(true)
+            // stack switch model must match the current OS
+            "stack_switch_model" => {
+                if self.features().contains(WasmFeatures::STACK_SWITCHING) {
+                    use target_lexicon::OperatingSystem;
+                    let expected =
+                    match target.operating_system  {
+                        OperatingSystem::Windows => "update_windows_tib",
+                        OperatingSystem::Linux
+                        | OperatingSystem::MacOSX(_)
+                        | OperatingSystem::Darwin(_)  => "basic",
+                        _ => { return Err(String::from("stack-switching feature not supported on this platform")); }
+                    };
+                    *value == FlagValue::Enum(expected)
                 } else {
                     return Ok(())
                 }
@@ -383,7 +450,6 @@ impl Engine {
             "enable_heap_access_spectre_mitigation"
             | "enable_table_access_spectre_mitigation"
             | "enable_nan_canonicalization"
-            | "enable_jump_tables"
             | "enable_float"
             | "enable_verifier"
             | "enable_pcc"
@@ -395,7 +461,6 @@ impl Engine {
             | "log2_min_function_alignment"
             | "machine_code_cfg_info"
             | "tls_model" // wasmtime doesn't use tls right now
-            | "stack_switch_model" // wasmtime doesn't use stack switching right now
             | "opt_level" // opt level doesn't change semantics
             | "enable_alias_analysis" // alias analysis-based opts don't change semantics
             | "probestack_size_log2" // probestack above asserted disabled
@@ -441,14 +506,14 @@ impl Engine {
                     Ok(())
                 } else {
                     Err("wrong host pointer width".to_string())
-                }
+                };
             }
             FlagValue::Enum("pointer64") => {
                 return if cfg!(target_pointer_width = "64") {
                     Ok(())
                 } else {
                     Err("wrong host pointer width".to_string())
-                }
+                };
             }
 
             // Only `bool` values are supported right now, other settings would
@@ -456,7 +521,7 @@ impl Engine {
             _ => {
                 return Err(format!(
                     "isa-specific feature {flag:?} configured to unknown value {value:?}"
-                ))
+                ));
             }
         }
 
@@ -481,7 +546,9 @@ impl Engine {
 
             // s390x features to detect
             "has_vxrs_ext2" => "vxrs_ext2",
-            "has_mie2" => "mie2",
+            "has_vxrs_ext3" => "vxrs_ext3",
+            "has_mie3" => "mie3",
+            "has_mie4" => "mie4",
 
             // x64 features to detect
             "has_cmpxchg16b" => "cmpxchg16b",
@@ -505,7 +572,7 @@ impl Engine {
             // pulley features
             "big_endian" if cfg!(target_endian = "big") => return Ok(()),
             "big_endian" if cfg!(target_endian = "little") => {
-                return Err("wrong host endianness".to_string())
+                return Err("wrong host endianness".to_string());
             }
 
             _ => {
@@ -527,7 +594,7 @@ impl Engine {
                     "cannot determine if host feature {host_feature:?} is \
                      available at runtime, configure a probing function with \
                      `Config::detect_host_feature`"
-                ))
+                ));
             }
         };
 
@@ -538,8 +605,8 @@ impl Engine {
                  available on the host",
             )),
             None => Err(format!(
-                "failed to detect if target-specific flag {flag:?} is \
-                 available at runtime"
+                "failed to detect if target-specific flag {host_feature:?} is \
+                 available at runtime (compile setting {flag:?})"
             )),
         }
     }
@@ -657,16 +724,20 @@ impl Engine {
         crate::runtime::vm::tls_eager_initialize();
     }
 
+    /// Returns a [`PoolingAllocatorMetrics`](crate::PoolingAllocatorMetrics) if
+    /// this engine was configured with
+    /// [`InstanceAllocationStrategy::Pooling`](crate::InstanceAllocationStrategy::Pooling).
+    #[cfg(feature = "pooling-allocator")]
+    pub fn pooling_allocator_metrics(&self) -> Option<crate::vm::PoolingAllocatorMetrics> {
+        crate::runtime::vm::PoolingAllocatorMetrics::new(self)
+    }
+
     pub(crate) fn allocator(&self) -> &dyn crate::runtime::vm::InstanceAllocator {
         self.inner.allocator.as_ref()
     }
 
-    pub(crate) fn gc_runtime(&self) -> Result<&Arc<dyn GcRuntime>> {
-        if let Some(rt) = &self.inner.gc_runtime {
-            Ok(rt)
-        } else {
-            bail!("no GC runtime: GC disabled at compile time or configuration time")
-        }
+    pub(crate) fn gc_runtime(&self) -> Option<&Arc<dyn GcRuntime>> {
+        self.inner.gc_runtime.as_ref()
     }
 
     pub(crate) fn profiler(&self) -> &dyn crate::profiling_agent::ProfilingAgent {
@@ -674,8 +745,8 @@ impl Engine {
     }
 
     #[cfg(all(feature = "cache", any(feature = "cranelift", feature = "winch")))]
-    pub(crate) fn cache_config(&self) -> &wasmtime_cache::CacheConfig {
-        &self.config().cache_config
+    pub(crate) fn cache(&self) -> Option<&wasmtime_cache::Cache> {
+        self.config().cache.as_ref()
     }
 
     pub(crate) fn signatures(&self) -> &TypeRegistry {
@@ -738,25 +809,6 @@ impl Engine {
         crate::compile::HashedEngineCompileEnv(self)
     }
 
-    /// Executes `f1` and `f2` in parallel if parallel compilation is enabled at
-    /// both runtime and compile time, otherwise runs them synchronously.
-    #[allow(dead_code)] // only used for the component-model feature right now
-    pub(crate) fn join_maybe_parallel<T, U>(
-        &self,
-        f1: impl FnOnce() -> T + Send,
-        f2: impl FnOnce() -> U + Send,
-    ) -> (T, U)
-    where
-        T: Send,
-        U: Send,
-    {
-        if self.config().parallel_compilation {
-            #[cfg(feature = "parallel-compilation")]
-            return rayon::join(f1, f2);
-        }
-        (f1(), f2())
-    }
-
     /// Returns the required alignment for a code image, if we
     /// allocate in a way that is not a system `mmap()` that naturally
     /// aligns it.
@@ -805,7 +857,9 @@ impl Engine {
         memory: NonNull<[u8]>,
         expected: ObjectKind,
     ) -> Result<Arc<crate::CodeMemory>> {
-        self.load_code(crate::runtime::vm::MmapVec::from_raw(memory)?, expected)
+        // SAFETY: the contract of this function is the same as that of
+        // `from_raw`.
+        unsafe { self.load_code(crate::runtime::vm::MmapVec::from_raw(memory)?, expected) }
     }
 
     /// Like `load_code_bytes`, but creates a mmap from a file on disk.
@@ -827,6 +881,9 @@ impl Engine {
         mmap: crate::runtime::vm::MmapVec,
         expected: ObjectKind,
     ) -> Result<Arc<crate::CodeMemory>> {
+        self.check_compatible_with_native_host()
+            .context("compilation settings are not compatible with the native host")?;
+
         serialization::check_compatible(self, &mmap, expected)?;
         let mut code = crate::CodeMemory::new(self, mmap)?;
         code.publish()?;
@@ -885,8 +942,11 @@ impl Engine {
         assert_eq!(Arc::weak_count(&self.inner), 0);
         assert_eq!(Arc::strong_count(&self.inner), 1);
 
+        // SAFETY: the contract of this function is the same as `deinit_traps`.
         #[cfg(not(miri))]
-        crate::runtime::vm::deinit_traps();
+        unsafe {
+            crate::runtime::vm::deinit_traps();
+        }
     }
 }
 
