@@ -23,11 +23,12 @@
 
 use crate::prelude::*;
 use crate::{Engine, ModuleVersionStrategy, Precompiled};
+use core::fmt;
 use core::str::FromStr;
 use object::endian::Endianness;
 #[cfg(any(feature = "cranelift", feature = "winch"))]
 use object::write::{Object, StandardSegment};
-use object::{read::elf::ElfFile64, FileFlags, Object as _, ObjectSection};
+use object::{FileFlags, Object as _, ObjectSection, read::elf::ElfFile64};
 use serde_derive::{Deserialize, Serialize};
 use wasmtime_environ::obj;
 use wasmtime_environ::{FlagValue, ObjectKind, Tunables};
@@ -96,20 +97,14 @@ pub fn check_compatible(engine: &Engine, mmap: &[u8], expected: ObjectKind) -> R
     match &engine.config().module_version {
         ModuleVersionStrategy::WasmtimeVersion => {
             let version = core::str::from_utf8(version)?;
-            if version != env!("CARGO_PKG_VERSION") {
-                bail!(
-                    "Module was compiled with incompatible Wasmtime version '{}'",
-                    version
-                );
+            if version != env!("CARGO_PKG_VERSION_MAJOR") {
+                bail!("Module was compiled with incompatible Wasmtime version '{version}'");
             }
         }
         ModuleVersionStrategy::Custom(v) => {
             let version = core::str::from_utf8(&version)?;
             if version != v {
-                bail!(
-                    "Module was compiled with incompatible version '{}'",
-                    version
-                );
+                bail!("Module was compiled with incompatible version '{version}'");
             }
         }
         ModuleVersionStrategy::None => { /* ignore the version info, accept all */ }
@@ -127,7 +122,7 @@ pub fn append_compiler_info(engine: &Engine, obj: &mut Object<'_>, metadata: &Me
     let mut data = Vec::new();
     data.push(VERSION);
     let version = match &engine.config().module_version {
-        ModuleVersionStrategy::WasmtimeVersion => env!("CARGO_PKG_VERSION"),
+        ModuleVersionStrategy::WasmtimeVersion => env!("CARGO_PKG_VERSION_MAJOR"),
         ModuleVersionStrategy::Custom(c) => c,
         ModuleVersionStrategy::None => "",
     };
@@ -179,110 +174,18 @@ pub struct Metadata<'a> {
     #[serde(borrow)]
     isa_flags: Vec<(&'a str, FlagValue<'a>)>,
     tunables: Tunables,
-    features: WasmFeatures,
-}
-
-// This exists because `wasmparser::WasmFeatures` isn't serializable
-#[derive(Debug, Copy, Clone, Serialize, Deserialize)]
-struct WasmFeatures {
-    reference_types: bool,
-    multi_value: bool,
-    bulk_memory: bool,
-    component_model: bool,
-    simd: bool,
-    tail_call: bool,
-    threads: bool,
-    multi_memory: bool,
-    exceptions: bool,
-    memory64: bool,
-    relaxed_simd: bool,
-    extended_const: bool,
-    function_references: bool,
-    gc: bool,
-    custom_page_sizes: bool,
-    component_model_async: bool,
-    component_model_async_builtins: bool,
-    component_model_async_stackful: bool,
-    gc_types: bool,
-    wide_arithmetic: bool,
-    stack_switching: bool,
+    features: u64,
 }
 
 impl Metadata<'_> {
     #[cfg(any(feature = "cranelift", feature = "winch"))]
     pub fn new(engine: &Engine) -> Metadata<'static> {
-        let wasmparser::WasmFeaturesInflated {
-            reference_types,
-            multi_value,
-            bulk_memory,
-            component_model,
-            simd,
-            threads,
-            tail_call,
-            multi_memory,
-            exceptions,
-            memory64,
-            relaxed_simd,
-            extended_const,
-            memory_control,
-            function_references,
-            gc,
-            custom_page_sizes,
-            shared_everything_threads,
-            cm_async,
-            cm_async_builtins,
-            cm_async_stackful,
-            cm_nested_names,
-            cm_values,
-            legacy_exceptions,
-            gc_types,
-            stack_switching,
-            wide_arithmetic,
-
-            // Always on; we don't currently have knobs for these.
-            mutable_global: _,
-            saturating_float_to_int: _,
-            sign_extension: _,
-            floats: _,
-        } = engine.features().inflate();
-
-        // These features are not implemented in Wasmtime yet. We match on them
-        // above so that once we do implement support for them, we won't
-        // silently ignore them during serialization.
-        assert!(!memory_control);
-        assert!(!cm_nested_names);
-        assert!(!cm_values);
-        assert!(!shared_everything_threads);
-        assert!(!legacy_exceptions);
-
         Metadata {
             target: engine.compiler().triple().to_string(),
             shared_flags: engine.compiler().flags(),
             isa_flags: engine.compiler().isa_flags(),
             tunables: engine.tunables().clone(),
-            features: WasmFeatures {
-                reference_types,
-                multi_value,
-                bulk_memory,
-                component_model,
-                simd,
-                threads,
-                tail_call,
-                multi_memory,
-                exceptions,
-                memory64,
-                relaxed_simd,
-                extended_const,
-                function_references,
-                gc,
-                custom_page_sizes,
-                gc_types,
-                wide_arithmetic,
-                stack_switching,
-                component_model_async: cm_async,
-                component_model_async_builtins: cm_async_builtins,
-                component_model_async_stackful: cm_async_stackful,
-            },
+            features: engine.features().bits(),
         }
     }
 
@@ -337,20 +240,17 @@ impl Metadata<'_> {
         Ok(())
     }
 
-    fn check_int<T: Eq + core::fmt::Display>(found: T, expected: T, feature: &str) -> Result<()> {
+    fn check_int<T: Eq + fmt::Display>(found: T, expected: T, feature: &str) -> Result<()> {
         if found == expected {
             return Ok(());
         }
 
         bail!(
-            "Module was compiled with a {} of '{}' but '{}' is expected for the host",
-            feature,
-            found,
-            expected
+            "Module was compiled with a {feature} of '{found}' but '{expected}' is expected for the host"
         );
     }
 
-    fn check_bool(found: bool, expected: bool, feature: &str) -> Result<()> {
+    fn check_bool(found: bool, expected: bool, feature: impl fmt::Display) -> Result<()> {
         if found == expected {
             return Ok(());
         }
@@ -368,7 +268,8 @@ impl Metadata<'_> {
             collector,
             memory_reservation,
             memory_guard_size,
-            generate_native_debuginfo,
+            debug_native,
+            debug_guest,
             parse_wasm_debuginfo,
             consume_fuel,
             epoch_interruption,
@@ -379,6 +280,11 @@ impl Metadata<'_> {
             winch_callable,
             signals_based_traps,
             memory_init_cow,
+            inlining,
+            inlining_intra_module,
+            inlining_small_callee_size,
+            inlining_sum_size_threshold,
+
             // This doesn't affect compilation, it's just a runtime setting.
             memory_reservation_for_growth: _,
 
@@ -404,10 +310,11 @@ impl Metadata<'_> {
             "memory guard size",
         )?;
         Self::check_bool(
-            generate_native_debuginfo,
-            other.generate_native_debuginfo,
-            "debug information support",
+            debug_native,
+            other.debug_native,
+            "native debug information support",
         )?;
+        Self::check_bool(debug_guest, other.debug_guest, "guest debug")?;
         Self::check_bool(
             parse_wasm_debuginfo,
             other.parse_wasm_debuginfo,
@@ -446,160 +353,32 @@ impl Metadata<'_> {
             other.memory_init_cow,
             "memory initialization with CoW",
         )?;
+        Self::check_bool(inlining, other.inlining, "function inlining")?;
+        Self::check_int(
+            inlining_small_callee_size,
+            other.inlining_small_callee_size,
+            "function inlining small-callee size",
+        )?;
+        Self::check_int(
+            inlining_sum_size_threshold,
+            other.inlining_sum_size_threshold,
+            "function inlining sum-size threshold",
+        )?;
+        Self::check_intra_module_inlining(inlining_intra_module, other.inlining_intra_module)?;
 
         Ok(())
     }
 
-    fn check_cfg_bool(
-        cfg: bool,
-        cfg_str: &str,
-        found: bool,
-        expected: bool,
-        feature: &str,
-    ) -> Result<()> {
-        if cfg {
-            Self::check_bool(found, expected, feature)
-        } else {
-            assert!(!expected);
-            ensure!(
-                !found,
-                "Module was compiled with {feature} but support in the host \
-                 was disabled at compile time because the `{cfg_str}` Cargo \
-                 feature was not enabled",
-            );
-            Ok(())
-        }
-    }
-
     fn check_features(&mut self, other: &wasmparser::WasmFeatures) -> Result<()> {
-        let WasmFeatures {
-            reference_types,
-            multi_value,
-            bulk_memory,
-            component_model,
-            simd,
-            tail_call,
-            threads,
-            multi_memory,
-            exceptions,
-            memory64,
-            relaxed_simd,
-            extended_const,
-            function_references,
-            gc,
-            custom_page_sizes,
-            component_model_async,
-            component_model_async_builtins,
-            component_model_async_stackful,
-            gc_types,
-            wide_arithmetic,
-            stack_switching,
-        } = self.features;
-
-        use wasmparser::WasmFeatures as F;
-        Self::check_bool(
-            reference_types,
-            other.contains(F::REFERENCE_TYPES),
-            "WebAssembly reference types support",
-        )?;
-        Self::check_bool(
-            function_references,
-            other.contains(F::FUNCTION_REFERENCES),
-            "WebAssembly function-references support",
-        )?;
-        Self::check_bool(
-            gc,
-            other.contains(F::GC),
-            "WebAssembly garbage collection support",
-        )?;
-        Self::check_bool(
-            multi_value,
-            other.contains(F::MULTI_VALUE),
-            "WebAssembly multi-value support",
-        )?;
-        Self::check_bool(
-            bulk_memory,
-            other.contains(F::BULK_MEMORY),
-            "WebAssembly bulk memory support",
-        )?;
-        Self::check_bool(
-            component_model,
-            other.contains(F::COMPONENT_MODEL),
-            "WebAssembly component model support",
-        )?;
-        Self::check_bool(simd, other.contains(F::SIMD), "WebAssembly SIMD support")?;
-        Self::check_bool(
-            tail_call,
-            other.contains(F::TAIL_CALL),
-            "WebAssembly tail calls support",
-        )?;
-        Self::check_bool(
-            threads,
-            other.contains(F::THREADS),
-            "WebAssembly threads support",
-        )?;
-        Self::check_bool(
-            multi_memory,
-            other.contains(F::MULTI_MEMORY),
-            "WebAssembly multi-memory support",
-        )?;
-        Self::check_bool(
-            exceptions,
-            other.contains(F::EXCEPTIONS),
-            "WebAssembly exceptions support",
-        )?;
-        Self::check_bool(
-            memory64,
-            other.contains(F::MEMORY64),
-            "WebAssembly 64-bit memory support",
-        )?;
-        Self::check_bool(
-            extended_const,
-            other.contains(F::EXTENDED_CONST),
-            "WebAssembly extended-const support",
-        )?;
-        Self::check_bool(
-            relaxed_simd,
-            other.contains(F::RELAXED_SIMD),
-            "WebAssembly relaxed-simd support",
-        )?;
-        Self::check_bool(
-            custom_page_sizes,
-            other.contains(F::CUSTOM_PAGE_SIZES),
-            "WebAssembly custom-page-sizes support",
-        )?;
-        Self::check_bool(
-            component_model_async,
-            other.contains(F::CM_ASYNC),
-            "WebAssembly component model support for async lifts/lowers, futures, streams, and errors",
-        )?;
-        Self::check_bool(
-            component_model_async_builtins,
-            other.contains(F::CM_ASYNC_BUILTINS),
-            "WebAssembly component model support for async builtins",
-        )?;
-        Self::check_bool(
-            component_model_async_stackful,
-            other.contains(F::CM_ASYNC_STACKFUL),
-            "WebAssembly component model support for async stackful",
-        )?;
-        Self::check_cfg_bool(
-            cfg!(feature = "gc"),
-            "gc",
-            gc_types,
-            other.contains(F::GC_TYPES),
-            "support for WebAssembly gc types",
-        )?;
-        Self::check_bool(
-            wide_arithmetic,
-            other.contains(F::WIDE_ARITHMETIC),
-            "WebAssembly wide-arithmetic support",
-        )?;
-        Self::check_bool(
-            stack_switching,
-            other.contains(F::STACK_SWITCHING),
-            "WebAssembly stack switching support",
-        )?;
+        let module_features = wasmparser::WasmFeatures::from_bits_truncate(self.features);
+        let missing_features = (*other & module_features) ^ module_features;
+        for (name, _) in missing_features.iter_names() {
+            let name = name.to_ascii_lowercase();
+            bail!(
+                "Module was compiled with support for WebAssembly feature \
+                `{name}` but it is not enabled for the host",
+            );
+        }
         Ok(())
     }
 
@@ -608,12 +387,11 @@ impl Metadata<'_> {
         host: Option<wasmtime_environ::Collector>,
     ) -> Result<()> {
         match (module, host) {
-            (None, None) => Ok(()),
+            // If the module doesn't require GC support it doesn't matter
+            // whether the host has GC support enabled or not.
+            (None, _) => Ok(()),
             (Some(module), Some(host)) if module == host => Ok(()),
 
-            (None, Some(_)) => {
-                bail!("module was compiled without GC but GC is enabled in the host")
-            }
             (Some(_), None) => {
                 bail!("module was compiled with GC however GC is disabled in the host")
             }
@@ -626,12 +404,34 @@ impl Metadata<'_> {
             }
         }
     }
+
+    fn check_intra_module_inlining(
+        module: wasmtime_environ::IntraModuleInlining,
+        host: wasmtime_environ::IntraModuleInlining,
+    ) -> Result<()> {
+        if module == host {
+            return Ok(());
+        }
+
+        let desc = |cfg| match cfg {
+            wasmtime_environ::IntraModuleInlining::No => "without intra-module inlining",
+            wasmtime_environ::IntraModuleInlining::Yes => "with intra-module inlining",
+            wasmtime_environ::IntraModuleInlining::WhenUsingGc => {
+                "with intra-module inlining only when using GC"
+            }
+        };
+
+        let module = desc(module);
+        let host = desc(host);
+
+        bail!("module was compiled {module} however the host is configured {host}")
+    }
 }
 
 #[cfg(test)]
 mod test {
     use super::*;
-    use crate::{Config, Module, OptLevel};
+    use crate::{Cache, Config, Module, OptLevel};
     use std::{
         collections::hash_map::DefaultHasher,
         hash::{Hash, Hasher},
@@ -655,9 +455,9 @@ mod test {
         Ok(())
     }
 
+    // Note that this test runs on a platform that is known to use Cranelift
     #[test]
-    #[cfg(target_arch = "x86_64")] // test on a platform that is known to use
-                                   // Cranelift
+    #[cfg(all(target_arch = "x86_64", not(miri)))]
     fn test_os_mismatch() -> Result<()> {
         let engine = Engine::default();
         let mut metadata = Metadata::new(&engine);
@@ -738,7 +538,10 @@ Caused by:
 
         match metadata.check_compatible(&engine) {
             Ok(_) => unreachable!(),
-            Err(e) => assert_eq!(e.to_string(), "Module was compiled with a memory guard size of '0' but '33554432' is expected for the host"),
+            Err(e) => assert_eq!(
+                e.to_string(),
+                "Module was compiled with a memory guard size of '0' but '33554432' is expected for the host"
+            ),
         }
 
         Ok(())
@@ -779,32 +582,35 @@ Caused by:
         Ok(())
     }
 
+    /// This test is only run a platform that is known to implement threads
     #[test]
-    #[cfg(target_arch = "x86_64")] // test on a platform that is known to
-                                   // implement threads
+    #[cfg(all(target_arch = "x86_64", not(miri)))]
     fn test_feature_mismatch() -> Result<()> {
         let mut config = Config::new();
         config.wasm_threads(true);
 
         let engine = Engine::new(&config)?;
         let mut metadata = Metadata::new(&engine);
-        metadata.features.threads = false;
+        metadata.features &= !wasmparser::WasmFeatures::THREADS.bits();
 
-        match metadata.check_compatible(&engine) {
-            Ok(_) => unreachable!(),
-            Err(e) => assert_eq!(e.to_string(), "Module was compiled without WebAssembly threads support but it is enabled for the host"),
-        }
+        // If a feature is disabled in the module and enabled in the host,
+        // that's always ok.
+        metadata.check_compatible(&engine)?;
 
         let mut config = Config::new();
         config.wasm_threads(false);
 
         let engine = Engine::new(&config)?;
         let mut metadata = Metadata::new(&engine);
-        metadata.features.threads = true;
+        metadata.features |= wasmparser::WasmFeatures::THREADS.bits();
 
         match metadata.check_compatible(&engine) {
             Ok(_) => unreachable!(),
-            Err(e) => assert_eq!(e.to_string(), "Module was compiled with WebAssembly threads support but it is not enabled for the host"),
+            Err(e) => assert_eq!(
+                e.to_string(),
+                "Module was compiled with support for WebAssembly feature \
+                `threads` but it is not enabled for the host"
+            ),
         }
 
         Ok(())
@@ -826,6 +632,8 @@ Caused by:
     #[test]
     #[cfg_attr(miri, ignore)]
     fn cache_accounts_for_opt_level() -> Result<()> {
+        let _ = env_logger::try_init();
+
         let td = TempDir::new()?;
         let config_path = td.path().join("config.toml");
         std::fs::write(
@@ -833,7 +641,6 @@ Caused by:
             &format!(
                 "
                     [cache]
-                    enabled = true
                     directory = '{}'
                 ",
                 td.path().join("cache").display()
@@ -841,46 +648,67 @@ Caused by:
         )?;
         let mut cfg = Config::new();
         cfg.cranelift_opt_level(OptLevel::None)
-            .cache_config_load(&config_path)?;
+            .cache(Some(Cache::from_file(Some(&config_path))?));
         let engine = Engine::new(&cfg)?;
         Module::new(&engine, "(module (func))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 0);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        let cache_config = engine
+            .config()
+            .cache
+            .as_ref()
+            .expect("Missing cache config");
+        assert_eq!(cache_config.cache_hits(), 0);
+        assert_eq!(cache_config.cache_misses(), 1);
         Module::new(&engine, "(module (func))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 1);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        assert_eq!(cache_config.cache_hits(), 1);
+        assert_eq!(cache_config.cache_misses(), 1);
 
         let mut cfg = Config::new();
         cfg.cranelift_opt_level(OptLevel::Speed)
-            .cache_config_load(&config_path)?;
+            .cache(Some(Cache::from_file(Some(&config_path))?));
         let engine = Engine::new(&cfg)?;
+        let cache_config = engine
+            .config()
+            .cache
+            .as_ref()
+            .expect("Missing cache config");
         Module::new(&engine, "(module (func))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 0);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        assert_eq!(cache_config.cache_hits(), 0);
+        assert_eq!(cache_config.cache_misses(), 1);
         Module::new(&engine, "(module (func))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 1);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        assert_eq!(cache_config.cache_hits(), 1);
+        assert_eq!(cache_config.cache_misses(), 1);
 
         let mut cfg = Config::new();
         cfg.cranelift_opt_level(OptLevel::SpeedAndSize)
-            .cache_config_load(&config_path)?;
+            .cache(Some(Cache::from_file(Some(&config_path))?));
         let engine = Engine::new(&cfg)?;
+        let cache_config = engine
+            .config()
+            .cache
+            .as_ref()
+            .expect("Missing cache config");
         Module::new(&engine, "(module (func))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 0);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        assert_eq!(cache_config.cache_hits(), 0);
+        assert_eq!(cache_config.cache_misses(), 1);
         Module::new(&engine, "(module (func))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 1);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        assert_eq!(cache_config.cache_hits(), 1);
+        assert_eq!(cache_config.cache_misses(), 1);
 
         let mut cfg = Config::new();
-        cfg.debug_info(true).cache_config_load(&config_path)?;
+        cfg.debug_info(true)
+            .cache(Some(Cache::from_file(Some(&config_path))?));
         let engine = Engine::new(&cfg)?;
+        let cache_config = engine
+            .config()
+            .cache
+            .as_ref()
+            .expect("Missing cache config");
         Module::new(&engine, "(module (func))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 0);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        assert_eq!(cache_config.cache_hits(), 0);
+        assert_eq!(cache_config.cache_misses(), 1);
         Module::new(&engine, "(module (func))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 1);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        assert_eq!(cache_config.cache_hits(), 1);
+        assert_eq!(cache_config.cache_misses(), 1);
 
         Ok(())
     }
@@ -941,21 +769,25 @@ Caused by:
             &format!(
                 "
                     [cache]
-                    enabled = true
                     directory = '{}'
                 ",
                 td.path().join("cache").display()
             ),
         )?;
         let mut cfg = Config::new();
-        cfg.cache_config_load(&config_path)?;
+        cfg.cache(Some(Cache::from_file(Some(&config_path))?));
         let engine = Engine::new(&cfg)?;
+        let cache_config = engine
+            .config()
+            .cache
+            .as_ref()
+            .expect("Missing cache config");
         Component::new(&engine, "(component (core module (func)))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 0);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        assert_eq!(cache_config.cache_hits(), 0);
+        assert_eq!(cache_config.cache_misses(), 1);
         Component::new(&engine, "(component (core module (func)))")?;
-        assert_eq!(engine.config().cache_config.cache_hits(), 1);
-        assert_eq!(engine.config().cache_config.cache_misses(), 1);
+        assert_eq!(cache_config.cache_hits(), 1);
+        assert_eq!(cache_config.cache_misses(), 1);
 
         Ok(())
     }
