@@ -1,10 +1,12 @@
 use crate::linker::{Definition, DefinitionType};
 use crate::prelude::*;
 use crate::runtime::vm::{
-    Imports, InstanceAllocationRequest, ModuleRuntimeInfo, StorePtr, VMFuncRef, VMFunctionImport,
-    VMGlobalImport, VMMemoryImport, VMOpaqueContext, VMTableImport, VMTagImport,
+    self, Imports, ModuleRuntimeInfo, VMFuncRef, VMFunctionImport, VMGlobalImport, VMMemoryImport,
+    VMStore, VMTableImport, VMTagImport,
+};
+use crate::store::{
+    AllocateInstanceKind, InstanceId, StoreInstanceId, StoreOpaque, StoreResourceLimiter,
 };
-use crate::store::{InstanceId, StoreOpaque, Stored};
 use crate::types::matching;
 use crate::{
     AsContextMut, Engine, Export, Extern, Func, Global, Memory, Module, ModuleExport, SharedMemory,
@@ -32,27 +34,20 @@ use wasmtime_environ::{
 /// [`Linker`](crate::Linker) methods, but a more low-level constructor is also
 /// available as [`Instance::new`].
 #[derive(Copy, Clone, Debug)]
-#[repr(transparent)]
-pub struct Instance(Stored<InstanceData>);
-
-pub(crate) struct InstanceData {
-    /// The id of the instance within the store, used to find the original
-    /// `InstanceHandle`.
-    id: InstanceId,
-    /// A lazily-populated list of exports of this instance. The order of
-    /// exports here matches the order of the exports in the original
-    /// module.
-    exports: Vec<Option<Extern>>,
+#[repr(C)]
+pub struct Instance {
+    id: StoreInstanceId,
 }
 
-impl InstanceData {
-    pub fn from_id(id: InstanceId) -> InstanceData {
-        InstanceData {
-            id,
-            exports: vec![],
-        }
-    }
-}
+// Double-check that the C representation in `instance.h` matches our in-Rust
+// representation here in terms of size/alignment/etc.
+const _: () = {
+    #[repr(C)]
+    struct C(u64, usize);
+    assert!(core::mem::size_of::<C>() == core::mem::size_of::<Instance>());
+    assert!(core::mem::align_of::<C>() == core::mem::align_of::<Instance>());
+    assert!(core::mem::offset_of!(Instance, id) == 0);
+};
 
 impl Instance {
     /// Creates a new [`Instance`] from the previously compiled [`Module`] and
@@ -124,7 +119,8 @@ impl Instance {
         // Note that the unsafety here should be satisfied by the call to
         // `typecheck_externs` above which satisfies the condition that all
         // the imports are valid for this module.
-        unsafe { Instance::new_started(&mut store, module, imports.as_ref()) }
+        assert!(!store.0.async_support());
+        vm::assert_ready(unsafe { Instance::new_started(&mut store, module, imports.as_ref()) })
     }
 
     /// Same as [`Instance::new`], except for usage in [asynchronous stores].
@@ -144,19 +140,70 @@ impl Instance {
     ///
     /// This function will also panic, like [`Instance::new`], if any [`Extern`]
     /// specified does not belong to `store`.
+    ///
+    /// # Examples
+    ///
+    /// An example of using this function:
+    ///
+    /// ```
+    /// use wasmtime::{Result, Store, Engine, Config, Module, Instance};
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> Result<()> {
+    ///     let mut config = Config::new();
+    ///     config.async_support(true);
+    ///     let engine = Engine::new(&config)?;
+    ///
+    ///     // For this example, a module with no imports is being used hence
+    ///     // the empty array to `Instance::new_async`.
+    ///     let module = Module::new(&engine, "(module)")?;
+    ///     let mut store = Store::new(&engine, ());
+    ///     let instance = Instance::new_async(&mut store, &module, &[]).await?;
+    ///
+    ///     // ... use `instance` and exports and such ...
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
+    ///
+    /// Note, though, that the future returned from this function is only
+    /// `Send` if the store's own data is `Send` meaning that this does not
+    /// compile for example:
+    ///
+    /// ```compile_fail
+    /// use wasmtime::{Result, Store, Engine, Config, Module, Instance};
+    /// use std::rc::Rc;
+    ///
+    /// #[tokio::main]
+    /// async fn main() -> Result<()> {
+    ///     let mut config = Config::new();
+    ///     config.async_support(true);
+    ///     let engine = Engine::new(&config)?;
+    ///
+    ///     let module = Module::new(&engine, "(module)")?;
+    ///
+    ///     // Note that `Rc<()>` is NOT `Send`, which is what many future
+    ///     // runtimes require and below will cause a failure.
+    ///     let mut store = Store::new(&engine, Rc::new(()));
+    ///
+    ///     // Compile failure because `Store<Rc<()>>` is not `Send`
+    ///     assert_send(Instance::new_async(&mut store, &module, &[])).await?;
+    ///
+    ///     Ok(())
+    /// }
+    ///
+    /// fn assert_send<T: Send>(t: T) -> T { t }
+    /// ```
     #[cfg(feature = "async")]
-    pub async fn new_async<T>(
-        mut store: impl AsContextMut<Data = T>,
+    pub async fn new_async(
+        mut store: impl AsContextMut,
         module: &Module,
         imports: &[Extern],
-    ) -> Result<Instance>
-    where
-        T: Send,
-    {
+    ) -> Result<Instance> {
         let mut store = store.as_context_mut();
         let imports = Instance::typecheck_externs(store.0, module, imports)?;
         // See `new` for notes on this unsafety
-        unsafe { Instance::new_started_async(&mut store, module, imports.as_ref()).await }
+        unsafe { Instance::new_started(&mut store, module, imports.as_ref()).await }
     }
 
     fn typecheck_externs(
@@ -169,13 +216,28 @@ impl Instance {
                 bail!("cross-`Store` instantiation is not currently supported");
             }
         }
+
         typecheck(module, imports, |cx, ty, item| {
             let item = DefinitionType::from(store, item);
             cx.definition(ty, &item)
         })?;
+
+        // When pushing functions into `OwnedImports` it's required that their
+        // `wasm_call` fields are all filled out. This `module` is guaranteed
+        // to have any trampolines necessary for functions so register the
+        // module with the store and then attempt to fill out any outstanding
+        // holes.
+        //
+        // Note that under normal operation this shouldn't do much as the list
+        // of funcs-with-holes should generally be empty. As a result the
+        // process of filling this out is not super optimized at this point.
+        store.modules_mut().register_module(module);
+        let (funcrefs, modules) = store.func_refs_and_modules();
+        funcrefs.fill(modules);
+
         let mut owned_imports = OwnedImports::new(module);
         for import in imports {
-            owned_imports.push(import, store, module);
+            owned_imports.push(import, store);
         }
         Ok(owned_imports)
     }
@@ -183,56 +245,34 @@ impl Instance {
     /// Internal function to create an instance and run the start function.
     ///
     /// This function's unsafety is the same as `Instance::new_raw`.
-    pub(crate) unsafe fn new_started<T>(
+    pub(crate) async unsafe fn new_started<T>(
         store: &mut StoreContextMut<'_, T>,
         module: &Module,
         imports: Imports<'_>,
     ) -> Result<Instance> {
-        assert!(
-            !store.0.async_support(),
-            "must use async instantiation when async support is enabled",
-        );
-        Self::new_started_impl(store, module, imports)
-    }
-
-    /// Internal function to create an instance and run the start function.
-    ///
-    /// ONLY CALL THIS IF YOU HAVE ALREADY CHECKED FOR ASYNCNESS AND HANDLED
-    /// THE FIBER NONSENSE
-    pub(crate) unsafe fn new_started_impl<T>(
-        store: &mut StoreContextMut<'_, T>,
-        module: &Module,
-        imports: Imports<'_>,
-    ) -> Result<Instance> {
-        let (instance, start) = Instance::new_raw(store.0, module, imports)?;
+        let (instance, start) = {
+            let (mut limiter, store) = store.0.resource_limiter_and_store_opaque();
+            // SAFETY: the safety contract of `new_raw` is the same as this
+            // function.
+            unsafe { Instance::new_raw(store, limiter.as_mut(), module, imports).await? }
+        };
         if let Some(start) = start {
-            instance.start_raw(store, start)?;
+            if store.0.async_support() {
+                #[cfg(feature = "async")]
+                {
+                    store
+                        .on_fiber(|store| instance.start_raw(store, start))
+                        .await??;
+                }
+                #[cfg(not(feature = "async"))]
+                unreachable!();
+            } else {
+                instance.start_raw(store, start)?;
+            }
         }
         Ok(instance)
     }
 
-    /// Internal function to create an instance and run the start function.
-    ///
-    /// This function's unsafety is the same as `Instance::new_raw`.
-    #[cfg(feature = "async")]
-    async unsafe fn new_started_async<T>(
-        store: &mut StoreContextMut<'_, T>,
-        module: &Module,
-        imports: Imports<'_>,
-    ) -> Result<Instance>
-    where
-        T: Send,
-    {
-        assert!(
-            store.0.async_support(),
-            "must use sync instantiation when async support is disabled",
-        );
-
-        store
-            .on_fiber(|store| Self::new_started_impl(store, module, imports))
-            .await?
-    }
-
     /// Internal function to create an instance which doesn't have its `start`
     /// function run yet.
     ///
@@ -248,8 +288,9 @@ impl Instance {
     /// This method is unsafe because it does not type-check the `imports`
     /// provided. The `imports` provided must be suitable for the module
     /// provided as well.
-    unsafe fn new_raw(
+    async unsafe fn new_raw(
         store: &mut StoreOpaque,
+        mut limiter: Option<&mut StoreResourceLimiter<'_>>,
         module: &Module,
         imports: Imports<'_>,
     ) -> Result<(Instance, Option<FuncIndex>)> {
@@ -259,8 +300,8 @@ impl Instance {
         store.bump_resource_counts(module)?;
 
         // Allocate the GC heap, if necessary.
-        if store.engine().features().gc_types() {
-            let _ = store.gc_store_mut()?;
+        if module.env_module().needs_gc_heap {
+            store.ensure_gc_store(limiter.as_deref_mut()).await?;
         }
 
         let compiled_module = module.compiled_module();
@@ -268,46 +309,23 @@ impl Instance {
         // Register the module just before instantiation to ensure we keep the module
         // properly referenced while in use by the store.
         let module_id = store.modules_mut().register_module(module);
-        store.fill_func_refs();
 
         // The first thing we do is issue an instance allocation request
         // to the instance allocator. This, on success, will give us an
         // instance handle.
         //
-        // Note that the `host_state` here is a pointer back to the
-        // `Instance` we'll be returning from this function. This is a
-        // circular reference so we can't construct it before we construct
-        // this instance, so we determine what the ID is and then assert
-        // it's the same later when we do actually insert it.
-        let instance_to_be = store.store_data().next_id::<InstanceData>();
-
-        let mut instance_handle =
+        // SAFETY: this module, by construction, was already validated within
+        // the store.
+        let id = unsafe {
             store
-                .engine()
-                .allocator()
-                .allocate_module(InstanceAllocationRequest {
-                    runtime_info: &ModuleRuntimeInfo::Module(module.clone()),
+                .allocate_instance(
+                    limiter.as_deref_mut(),
+                    AllocateInstanceKind::Module(module_id),
+                    &ModuleRuntimeInfo::Module(module.clone()),
                     imports,
-                    host_state: Box::new(Instance(instance_to_be)),
-                    store: StorePtr::new(store.traitobj()),
-                    wmemcheck: store.engine().config().wmemcheck,
-                    pkey: store.get_pkey(),
-                    tunables: store.engine().tunables(),
-                })?;
-
-        // The instance still has lots of setup, for example
-        // data/elements/start/etc. This can all fail, but even on failure
-        // the instance may persist some state via previous successful
-        // initialization. For this reason once we have an instance handle
-        // we immediately insert it into the store to keep it alive.
-        //
-        // Note that we `clone` the instance handle just to make easier
-        // working the borrow checker here easier. Technically the `&mut
-        // instance` has somewhat of a borrow on `store` (which
-        // conflicts with the borrow on `store.engine`) but this doesn't
-        // matter in practice since initialization isn't even running any
-        // code here anyway.
-        let id = store.add_instance(instance_handle.clone(), module_id);
+                )
+                .await?
+        };
 
         // Additionally, before we start doing fallible instantiation, we
         // do one more step which is to insert an `InstanceData`
@@ -321,15 +339,7 @@ impl Instance {
         // For module/instance exports, though, those aren't actually
         // stored in the instance handle so we need to immediately handle
         // those here.
-        let instance = {
-            let exports = vec![None; compiled_module.module().exports.len()];
-            let data = InstanceData { id, exports };
-            Instance::from_wasmtime(data, store)
-        };
-
-        // double-check our guess of what the new instance's ID would be
-        // was actually correct.
-        assert_eq!(instance.0, instance_to_be);
+        let instance = Instance::from_wasmtime(id, store);
 
         // Now that we've recorded all information we need to about this
         // instance within a `Store` we can start performing fallible
@@ -345,42 +355,43 @@ impl Instance {
             .engine()
             .features()
             .contains(WasmFeatures::BULK_MEMORY);
-        instance_handle.initialize(store, compiled_module.module(), bulk_memory)?;
+
+        vm::initialize_instance(store, limiter, id, compiled_module.module(), bulk_memory).await?;
 
         Ok((instance, compiled_module.module().start_func))
     }
 
-    pub(crate) fn from_wasmtime(handle: InstanceData, store: &mut StoreOpaque) -> Instance {
-        Instance(store.store_data_mut().insert(handle))
+    pub(crate) fn from_wasmtime(id: InstanceId, store: &mut StoreOpaque) -> Instance {
+        Instance {
+            id: StoreInstanceId::new(store.id(), id),
+        }
     }
 
     fn start_raw<T>(&self, store: &mut StoreContextMut<'_, T>, start: FuncIndex) -> Result<()> {
-        let id = store.0.store_data()[self.0].id;
         // If a start function is present, invoke it. Make sure we use all the
         // trap-handling configuration in `store` as well.
-        let instance = store.0.instance_mut(id);
-        let f = instance.get_exported_func(start);
+        let store_id = store.0.id();
+        let mut instance = self.id.get_mut(store.0);
+        // SAFETY: the `store_id` is the id of the store that owns this
+        // instance and any function stored within the instance.
+        let f = unsafe { instance.as_mut().get_exported_func(store_id, start) };
         let caller_vmctx = instance.vmctx();
         unsafe {
+            let funcref = f.vm_func_ref(store.0);
             super::func::invoke_wasm_and_catch_traps(store, |_default_caller, vm| {
-                f.func_ref.as_ref().array_call(
-                    vm,
-                    VMOpaqueContext::from_vmcontext(caller_vmctx),
-                    NonNull::from(&mut []),
-                )
+                VMFuncRef::array_call(funcref, vm, caller_vmctx, NonNull::from(&mut []))
             })?;
         }
         Ok(())
     }
 
     /// Get this instance's module.
-    pub fn module<'a, T: 'a>(&self, store: impl Into<StoreContext<'a, T>>) -> &'a Module {
+    pub fn module<'a, T: 'static>(&self, store: impl Into<StoreContext<'a, T>>) -> &'a Module {
         self._module(store.into().0)
     }
 
     fn _module<'a>(&self, store: &'a StoreOpaque) -> &'a Module {
-        let InstanceData { id, .. } = store[self.0];
-        store.module_for_instance(id).unwrap()
+        store.module_for_instance(self.id).unwrap()
     }
 
     /// Returns the list of exported items from this [`Instance`].
@@ -388,7 +399,7 @@ impl Instance {
     /// # Panics
     ///
     /// Panics if `store` does not own this instance.
-    pub fn exports<'a, T: 'a>(
+    pub fn exports<'a, T: 'static>(
         &'a self,
         store: impl Into<StoreContextMut<'a, T>>,
     ) -> impl ExactSizeIterator<Item = Export<'a>> + 'a {
@@ -399,31 +410,17 @@ impl Instance {
         &'a self,
         store: &'a mut StoreOpaque,
     ) -> impl ExactSizeIterator<Item = Export<'a>> + 'a {
-        // If this is an `Instantiated` instance then all the `exports` may not
-        // be filled in. Fill them all in now if that's the case.
-        let InstanceData { exports, id, .. } = &store[self.0];
-        if exports.iter().any(|e| e.is_none()) {
-            let module = Arc::clone(store.instance(*id).module());
-            let data = &store[self.0];
-            let id = data.id;
-
-            for name in module.exports.keys() {
-                let instance = store.instance(id);
-                if let Some((export_name_index, _, &entity)) =
-                    instance.module().exports.get_full(name)
-                {
-                    self._get_export(store, entity, export_name_index);
-                }
-            }
+        let module = store[self.id].env_module().clone();
+        let mut items = Vec::new();
+        for (_name, entity) in module.exports.iter() {
+            items.push(self._get_export(store, *entity));
         }
-
-        let data = &store.store_data()[self.0];
-        let module = store.instance(data.id).module();
-        module
+        store[self.id]
+            .env_module()
             .exports
             .iter()
-            .zip(&data.exports)
-            .map(|((name, _), export)| Export::new(name, export.clone().unwrap()))
+            .zip(items)
+            .map(|((name, _), item)| Export::new(name, item))
     }
 
     /// Looks up an exported [`Extern`] value by name.
@@ -445,10 +442,8 @@ impl Instance {
     /// mutable context.
     pub fn get_export(&self, mut store: impl AsContextMut, name: &str) -> Option<Extern> {
         let store = store.as_context_mut().0;
-        let data = &store[self.0];
-        let instance = store.instance(data.id);
-        let (export_name_index, _, &entity) = instance.module().exports.get_full(name)?;
-        self._get_export(store, entity, export_name_index)
+        let entity = *store[self.id].env_module().exports.get(name)?;
+        Some(self._get_export(store, entity))
     }
 
     /// Looks up an exported [`Extern`] value by a [`ModuleExport`] value.
@@ -461,6 +456,7 @@ impl Instance {
     /// the value, if found.
     ///
     /// Returns `None` if there was no export with a matching entity index.
+    ///
     /// # Panics
     ///
     /// Panics if `store` does not own this instance.
@@ -476,29 +472,15 @@ impl Instance {
             return None;
         }
 
-        self._get_export(store, export.entity, export.export_name_index)
+        Some(self._get_export(store, export.entity))
     }
 
-    fn _get_export(
-        &self,
-        store: &mut StoreOpaque,
-        entity: EntityIndex,
-        export_name_index: usize,
-    ) -> Option<Extern> {
-        // Instantiated instances will lazily fill in exports, so we process
-        // all that lazy logic here.
-        let data = &store[self.0];
-
-        if let Some(export) = &data.exports[export_name_index] {
-            return Some(export.clone());
-        }
-
-        let instance = store.instance_mut(data.id); // Reborrow the &mut InstanceHandle
-        let item =
-            unsafe { Extern::from_wasmtime_export(instance.get_export_by_index(entity), store) };
-        let data = &mut store[self.0];
-        data.exports[export_name_index] = Some(item.clone());
-        Some(item)
+    fn _get_export(&self, store: &mut StoreOpaque, entity: EntityIndex) -> Extern {
+        let id = store.id();
+        // SAFETY: the store `id` owns this instance and all exports contained
+        // within.
+        let export = unsafe { self.id.get_mut(store).get_export_by_index_mut(id, entity) };
+        unsafe { Extern::from_wasmtime_export(export, store) }
     }
 
     /// Looks up an exported [`Func`] value by name.
@@ -536,7 +518,7 @@ impl Instance {
         let f = self
             .get_export(store.as_context_mut(), name)
             .and_then(|f| f.into_func())
-            .ok_or_else(|| anyhow!("failed to find function export `{}`", name))?;
+            .ok_or_else(|| anyhow!("failed to find function export `{name}`"))?;
         Ok(f.typed::<Params, Results>(store)
             .with_context(|| format!("failed to convert function `{name}` to given type"))?)
     }
@@ -606,9 +588,12 @@ impl Instance {
         self.get_export(store, name)?.into_tag()
     }
 
-    #[cfg(feature = "component-model")]
-    pub(crate) fn id(&self, store: &StoreOpaque) -> InstanceId {
-        store[self.0].id
+    #[allow(
+        dead_code,
+        reason = "c-api crate does not yet support exnrefs and causes this method to be dead."
+    )]
+    pub(crate) fn id(&self) -> InstanceId {
+        self.id.instance()
     }
 
     /// Get all globals within this instance.
@@ -623,13 +608,8 @@ impl Instance {
         &'a self,
         store: &'a mut StoreOpaque,
     ) -> impl ExactSizeIterator<Item = (GlobalIndex, Global)> + 'a {
-        let data = &store[self.0];
-        let instance = store.instance_mut(data.id);
-        instance
-            .all_globals()
-            .collect::<Vec<_>>()
-            .into_iter()
-            .map(|(i, g)| (i, unsafe { Global::from_wasmtime_global(g, store) }))
+        let store_id = store.id();
+        store[self.id].all_globals(store_id)
     }
 
     /// Get all memories within this instance.
@@ -642,15 +622,10 @@ impl Instance {
     #[cfg(feature = "coredump")]
     pub(crate) fn all_memories<'a>(
         &'a self,
-        store: &'a mut StoreOpaque,
+        store: &'a StoreOpaque,
     ) -> impl ExactSizeIterator<Item = (MemoryIndex, Memory)> + 'a {
-        let data = &store[self.0];
-        let instance = store.instance_mut(data.id);
-        instance
-            .all_memories()
-            .collect::<Vec<_>>()
-            .into_iter()
-            .map(|(i, m)| (i, unsafe { Memory::from_wasmtime_memory(m, store) }))
+        let store_id = store.id();
+        store[self.id].all_memories(store_id)
     }
 }
 
@@ -697,10 +672,10 @@ impl OwnedImports {
         self.tags.clear();
     }
 
-    fn push(&mut self, item: &Extern, store: &mut StoreOpaque, module: &Module) {
+    fn push(&mut self, item: &Extern, store: &mut StoreOpaque) {
         match item {
             Extern::Func(i) => {
-                self.functions.push(i.vmimport(store, module));
+                self.functions.push(i.vmimport(store));
             }
             Extern::Global(i) => {
                 self.globals.push(i.vmimport(store));
@@ -723,10 +698,12 @@ impl OwnedImports {
     /// Note that this is unsafe as the validity of `item` is not verified and
     /// it contains a bunch of raw pointers.
     #[cfg(feature = "component-model")]
-    pub(crate) unsafe fn push_export(&mut self, item: &crate::runtime::vm::Export) {
+    pub(crate) fn push_export(&mut self, store: &StoreOpaque, item: &crate::runtime::vm::Export) {
         match item {
             crate::runtime::vm::Export::Function(f) => {
-                let f = f.func_ref.as_ref();
+                // SAFETY: the funcref associated with a `Func` is valid to use
+                // under the `store` that owns the function.
+                let f = unsafe { f.vm_func_ref(store).as_ref() };
                 self.functions.push(VMFunctionImport {
                     wasm_call: f.wasm_call.unwrap(),
                     array_call: f.array_call,
@@ -734,27 +711,16 @@ impl OwnedImports {
                 });
             }
             crate::runtime::vm::Export::Global(g) => {
-                self.globals.push(VMGlobalImport {
-                    from: g.definition.into(),
-                });
+                self.globals.push(g.vmimport(store));
             }
             crate::runtime::vm::Export::Table(t) => {
-                self.tables.push(VMTableImport {
-                    from: t.definition.into(),
-                    vmctx: t.vmctx.into(),
-                });
+                self.tables.push(t.vmimport(store));
             }
-            crate::runtime::vm::Export::Memory(m) => {
-                self.memories.push(VMMemoryImport {
-                    from: m.definition.into(),
-                    vmctx: m.vmctx.into(),
-                    index: m.index,
-                });
+            crate::runtime::vm::Export::Memory { memory, .. } => {
+                self.memories.push(memory.vmimport(store));
             }
             crate::runtime::vm::Export::Tag(t) => {
-                self.tags.push(VMTagImport {
-                    from: t.definition.into(),
-                });
+                self.tags.push(t.vmimport(store));
             }
         }
     }
@@ -824,7 +790,7 @@ impl<T> Clone for InstancePre<T> {
     }
 }
 
-impl<T> InstancePre<T> {
+impl<T: 'static> InstancePre<T> {
     /// Creates a new `InstancePre` which type-checks the `items` provided and
     /// on success is ready to instantiate a new instance.
     ///
@@ -901,7 +867,10 @@ impl<T> InstancePre<T> {
         // This unsafety should be handled by the type-checking performed by the
         // constructor of `InstancePre` to assert that all the imports we're passing
         // in match the module we're instantiating.
-        unsafe { Instance::new_started(&mut store, &self.module, imports.as_ref()) }
+        assert!(!store.0.async_support());
+        vm::assert_ready(unsafe {
+            Instance::new_started(&mut store, &self.module, imports.as_ref())
+        })
     }
 
     /// Creates a new instance, running the start function asynchronously
@@ -918,10 +887,7 @@ impl<T> InstancePre<T> {
     pub async fn instantiate_async(
         &self,
         mut store: impl AsContextMut<Data = T>,
-    ) -> Result<Instance>
-    where
-        T: Send,
-    {
+    ) -> Result<Instance> {
         let mut store = store.as_context_mut();
         let imports = pre_instantiate_raw(
             &mut store.0,
@@ -934,7 +900,7 @@ impl<T> InstancePre<T> {
         // This unsafety should be handled by the type-checking performed by the
         // constructor of `InstancePre` to assert that all the imports we're passing
         // in match the module we're instantiating.
-        unsafe { Instance::new_started_async(&mut store, &self.module, imports.as_ref()).await }
+        unsafe { Instance::new_started(&mut store, &self.module, imports.as_ref()).await }
     }
 }
 
@@ -951,19 +917,25 @@ fn pre_instantiate_raw(
     host_funcs: usize,
     func_refs: &Arc<[VMFuncRef]>,
 ) -> Result<OwnedImports> {
+    // Register this module and use it to fill out any funcref wasm_call holes
+    // we can. For more comments on this see `typecheck_externs`.
+    store.modules_mut().register_module(module);
+    let (funcrefs, modules) = store.func_refs_and_modules();
+    funcrefs.fill(modules);
+
     if host_funcs > 0 {
         // Any linker-defined function of the `Definition::HostFunc` variant
         // will insert a function into the store automatically as part of
         // instantiation, so reserve space here to make insertion more efficient
         // as it won't have to realloc during the instantiation.
-        store.store_data_mut().reserve_funcs(host_funcs);
+        funcrefs.reserve_storage(host_funcs);
 
         // The usage of `to_extern_store_rooted` requires that the items are
         // rooted via another means, which happens here by cloning the list of
         // items into the store once. This avoids cloning each individual item
         // below.
-        store.push_rooted_funcs(items.clone());
-        store.push_instance_pre_func_refs(func_refs.clone());
+        funcrefs.push_instance_pre_definitions(items.clone());
+        funcrefs.push_instance_pre_func_refs(func_refs.clone());
     }
 
     let mut func_refs = func_refs.iter().map(|f| NonNull::from(f));
@@ -990,7 +962,7 @@ fn pre_instantiate_raw(
                 .into()
             },
         };
-        imports.push(&item, store, module);
+        imports.push(&item, store);
     }
 
     Ok(imports)
