@@ -1,15 +1,15 @@
 use crate::prelude::*;
 #[cfg(feature = "std")]
 use crate::runtime::vm::open_file_for_mmap;
-use crate::runtime::vm::{CompiledModuleId, ModuleMemoryImages, VMWasmCallFunction};
+use crate::runtime::vm::{CompiledModuleId, MmapVec, ModuleMemoryImages, VMWasmCallFunction};
 use crate::sync::OnceLock;
 use crate::{
+    Engine,
     code::CodeObject,
     code_memory::CodeMemory,
     instantiate::CompiledModule,
     resources::ResourcesRequired,
     types::{ExportType, ExternType, ImportType},
-    Engine,
 };
 use alloc::sync::Arc;
 use core::fmt;
@@ -18,10 +18,14 @@ use core::ptr::NonNull;
 #[cfg(feature = "std")]
 use std::{fs::File, path::Path};
 use wasmparser::{Parser, ValidPayload, Validator};
+#[cfg(feature = "debug")]
+use wasmtime_environ::FrameTable;
 use wasmtime_environ::{
-    CompiledModuleInfo, EntityIndex, HostPtr, ModuleTypes, ObjectKind, TypeTrace, VMOffsets,
-    VMSharedTypeIndex,
+    CompiledFunctionsTable, CompiledModuleInfo, EntityIndex, HostPtr, ModuleTypes, ObjectKind,
+    TypeTrace, VMOffsets, VMSharedTypeIndex,
 };
+#[cfg(feature = "gc")]
+use wasmtime_unwinder::ExceptionTable;
 mod registry;
 
 pub use registry::*;
@@ -402,6 +406,26 @@ impl Module {
         Module::from_parts(engine, code, None)
     }
 
+    /// In-place deserialization of an in-memory compiled module previously
+    /// created with [`Module::serialize`] or [`Engine::precompile_module`].
+    ///
+    /// See [`Self::deserialize`] for additional information; this method
+    /// works identically except that it will not create a copy of the provided
+    /// memory but will use it directly.
+    ///
+    /// # Unsafety
+    ///
+    /// All of the safety notes from [`Self::deserialize`] apply here as well
+    /// with the additional constraint that the code memory provide by `memory`
+    /// lives for as long as the module and is nevery externally modified for
+    /// the lifetime of the deserialized module.
+    pub unsafe fn deserialize_raw(engine: &Engine, memory: NonNull<[u8]>) -> Result<Module> {
+        // SAFETY: the contract required by `load_code_raw` is the same as this
+        // function.
+        let code = unsafe { engine.load_code_raw(memory, ObjectKind::Module)? };
+        Module::from_parts(engine, code, None)
+    }
+
     /// Same as [`deserialize`], except that the contents of `path` are read to
     /// deserialize into a [`Module`].
     ///
@@ -428,8 +452,12 @@ impl Module {
     #[cfg(feature = "std")]
     pub unsafe fn deserialize_file(engine: &Engine, path: impl AsRef<Path>) -> Result<Module> {
         let file = open_file_for_mmap(path.as_ref())?;
-        Self::deserialize_open_file(engine, file)
-            .with_context(|| format!("failed deserialization for: {}", path.as_ref().display()))
+        // SAFETY: the contract of `deserialize_open_file` is the samea s this
+        // function.
+        unsafe {
+            Self::deserialize_open_file(engine, file)
+                .with_context(|| format!("failed deserialization for: {}", path.as_ref().display()))
+        }
     }
 
     /// Same as [`deserialize_file`], except that it takes an open `File`
@@ -469,13 +497,13 @@ impl Module {
     pub(crate) fn from_parts(
         engine: &Engine,
         code_memory: Arc<CodeMemory>,
-        info_and_types: Option<(CompiledModuleInfo, ModuleTypes)>,
+        info_and_types: Option<(CompiledModuleInfo, CompiledFunctionsTable, ModuleTypes)>,
     ) -> Result<Self> {
         // Acquire this module's metadata and type information, deserializing
         // it from the provided artifact if it wasn't otherwise provided
         // already.
-        let (mut info, mut types) = match info_and_types {
-            Some((info, types)) => (info, types),
+        let (mut info, index, mut types) = match info_and_types {
+            Some((info, index, types)) => (info, index, types),
             None => postcard::from_bytes(code_memory.wasmtime_info())?,
         };
 
@@ -493,17 +521,23 @@ impl Module {
         // Package up all our data into a `CodeObject` and delegate to the final
         // step of module compilation.
         let code = Arc::new(CodeObject::new(code_memory, signatures, types.into()));
-        Module::from_parts_raw(engine, code, info, true)
+        let index = Arc::new(index);
+        Module::from_parts_raw(engine, code, info, index, true)
     }
 
     pub(crate) fn from_parts_raw(
         engine: &Engine,
         code: Arc<CodeObject>,
         info: CompiledModuleInfo,
+        index: Arc<CompiledFunctionsTable>,
         serializable: bool,
     ) -> Result<Self> {
-        let module =
-            CompiledModule::from_artifacts(code.code_memory().clone(), info, engine.profiler())?;
+        let module = CompiledModule::from_artifacts(
+            code.code_memory().clone(),
+            info,
+            index,
+            engine.profiler(),
+        )?;
 
         // Validate the module can be used with the current instance allocator.
         let offsets = VMOffsets::new(HostPtr, module.module());
@@ -844,14 +878,11 @@ impl Module {
     pub fn get_export_index(&self, name: &str) -> Option<ModuleExport> {
         let compiled_module = self.compiled_module();
         let module = compiled_module.module();
-        module
-            .exports
-            .get_full(name)
-            .map(|(export_name_index, _, &entity)| ModuleExport {
-                module: self.id(),
-                entity,
-                export_name_index,
-            })
+        let entity = *module.exports.get(name)?;
+        Some(ModuleExport {
+            module: self.id(),
+            entity,
+        })
     }
 
     /// Returns the [`Engine`] that this [`Module`] was compiled by.
@@ -1064,24 +1095,26 @@ impl Module {
             .code
             .signatures()
             .trampoline_type(trampoline_shared_ty)?;
-        debug_assert!(self
-            .inner
-            .engine
-            .signatures()
-            .borrow(
-                self.inner
-                    .code
-                    .signatures()
-                    .shared_type(trampoline_module_ty)
-                    .unwrap()
-            )
-            .unwrap()
-            .unwrap_func()
-            .is_trampoline_type());
+        debug_assert!(
+            self.inner
+                .engine
+                .signatures()
+                .borrow(
+                    self.inner
+                        .code
+                        .signatures()
+                        .shared_type(trampoline_module_ty)
+                        .unwrap()
+                )
+                .unwrap()
+                .unwrap_func()
+                .is_trampoline_type()
+        );
 
         let ptr = self
             .compiled_module()
             .wasm_to_array_trampoline(trampoline_module_ty)
+            .expect("always have a trampoline for the trampoline type")
             .as_ptr()
             .cast::<VMWasmCallFunction>()
             .cast_mut();
@@ -1092,35 +1125,43 @@ impl Module {
         let images = self
             .inner
             .memory_images
-            .get_or_try_init(|| memory_images(&self.inner.engine, &self.inner.module))?
+            .get_or_try_init(|| memory_images(&self.inner))?
             .as_ref();
         Ok(images)
     }
 
+    /// Get the text offset (relative PC) for a given absolute PC in
+    /// this module.
+    #[cfg(feature = "gc")]
+    pub(crate) fn text_offset(&self, pc: usize) -> u32 {
+        u32::try_from(pc - self.inner.module.text().as_ptr() as usize).unwrap()
+    }
+
     /// Lookup the stack map at a program counter value.
     #[cfg(feature = "gc")]
-    pub(crate) fn lookup_stack_map(&self, pc: usize) -> Option<&wasmtime_environ::StackMap> {
-        let text_offset = pc - self.inner.module.text().as_ptr() as usize;
-        let (index, func_offset) = self.inner.module.func_by_text_offset(text_offset)?;
-        let info = self.inner.module.wasm_func_info(index);
+    pub(crate) fn lookup_stack_map(&self, pc: usize) -> Option<wasmtime_environ::StackMap<'_>> {
+        let text_offset = self.text_offset(pc);
+        let info = self.inner.code.code_memory().stack_map_data();
+        wasmtime_environ::StackMap::lookup(text_offset, info)
+    }
 
-        // Do a binary search to find the stack map for the given offset.
-        let index = match info
-            .stack_maps
-            .binary_search_by_key(&func_offset, |i| i.code_offset)
-        {
-            // Found it.
-            Ok(i) => i,
+    /// Obtain an exception-table parser on this module's exception metadata.
+    #[cfg(feature = "gc")]
+    pub(crate) fn exception_table<'a>(&'a self) -> ExceptionTable<'a> {
+        ExceptionTable::parse(self.inner.code.code_memory().exception_tables())
+            .expect("Exception tables were validated on module load")
+    }
 
-            // No stack map associated with this PC.
-            //
-            // Because we know we are in Wasm code, and we must be at some kind
-            // of call/safepoint, then the Cranelift backend must have avoided
-            // emitting a stack map for this location because no refs were live.
-            Err(_) => return None,
-        };
-
-        Some(&info.stack_maps[index].stack_map)
+    /// Obtain a frame-table parser on this module's frame state slot
+    /// (debug instrumentation) metadata.
+    #[cfg(feature = "debug")]
+    pub(crate) fn frame_table<'a>(&'a self) -> Option<FrameTable<'a>> {
+        let data = self.inner.code.code_memory().frame_tables();
+        if data.is_empty() {
+            None
+        } else {
+            Some(FrameTable::parse(data).expect("Frame tables were validated on module load"))
+        }
     }
 }
 
@@ -1151,8 +1192,6 @@ pub struct ModuleExport {
     pub(crate) module: CompiledModuleId,
     /// A raw index into the wasm module.
     pub(crate) entity: EntityIndex,
-    /// The index of the export name.
-    pub(crate) export_name_index: usize,
 }
 
 fn _assert_send_sync() {
@@ -1162,21 +1201,30 @@ fn _assert_send_sync() {
 
 /// Helper method to construct a `ModuleMemoryImages` for an associated
 /// `CompiledModule`.
-fn memory_images(engine: &Engine, module: &CompiledModule) -> Result<Option<ModuleMemoryImages>> {
+fn memory_images(inner: &Arc<ModuleInner>) -> Result<Option<ModuleMemoryImages>> {
     // If initialization via copy-on-write is explicitly disabled in
     // configuration then this path is skipped entirely.
-    if !engine.tunables().memory_init_cow {
+    if !inner.engine.tunables().memory_init_cow {
         return Ok(None);
     }
 
     // ... otherwise logic is delegated to the `ModuleMemoryImages::new`
     // constructor.
-    let mmap = if engine.config().force_memory_init_memfd {
-        None
-    } else {
-        Some(module.mmap())
-    };
-    ModuleMemoryImages::new(module.module(), module.code_memory().wasm_data(), mmap)
+    ModuleMemoryImages::new(
+        &inner.engine,
+        inner.module.module(),
+        inner.code.code_memory(),
+    )
+}
+
+impl crate::vm::ModuleMemoryImageSource for CodeMemory {
+    fn wasm_data(&self) -> &[u8] {
+        <Self>::wasm_data(self)
+    }
+
+    fn mmap(&self) -> Option<&MmapVec> {
+        Some(<Self>::mmap(self))
+    }
 }
 
 #[cfg(test)]
