@@ -5,14 +5,45 @@
 // selectively enabled here.
 #![warn(clippy::cast_sign_loss)]
 
+// Polyfill `std::simd::i8x16` etc. until they're stable.
+#[cfg(all(target_arch = "x86_64", target_feature = "sse"))]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
+pub(crate) type i8x16 = core::arch::x86_64::__m128i;
+#[cfg(all(target_arch = "x86_64", target_feature = "sse"))]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
+pub(crate) type f32x4 = core::arch::x86_64::__m128;
+#[cfg(all(target_arch = "x86_64", target_feature = "sse"))]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
+pub(crate) type f64x2 = core::arch::x86_64::__m128d;
+
+// On platforms other than x86_64, define i8x16 to a non-constructible type;
+// we need a type because we have a lot of macros for defining builtin
+// functions that are awkward to make conditional on the target, but it
+// doesn't need to actually be constructible unless we're on x86_64.
+#[cfg(not(all(target_arch = "x86_64", target_feature = "sse")))]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
+#[derive(Copy, Clone)]
+pub(crate) struct i8x16(crate::uninhabited::Uninhabited);
+#[cfg(not(all(target_arch = "x86_64", target_feature = "sse")))]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
+#[derive(Copy, Clone)]
+pub(crate) struct f32x4(crate::uninhabited::Uninhabited);
+#[cfg(not(all(target_arch = "x86_64", target_feature = "sse")))]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
+#[derive(Copy, Clone)]
+pub(crate) struct f64x2(crate::uninhabited::Uninhabited);
+
+use crate::StoreContextMut;
 use crate::prelude::*;
-use crate::store::StoreOpaque;
+use crate::store::{StoreInner, StoreOpaque, StoreResourceLimiter};
+use crate::type_registry::RegisteredType;
 use alloc::sync::Arc;
 use core::fmt;
-use core::ops::Deref;
-use core::ops::DerefMut;
+use core::ops::{Deref, DerefMut};
+use core::pin::pin;
 use core::ptr::NonNull;
 use core::sync::atomic::{AtomicUsize, Ordering};
+use core::task::{Context, Poll, Waker};
 use wasmtime_environ::{
     DefinedFuncIndex, DefinedMemoryIndex, HostPtr, VMOffsets, VMSharedTypeIndex,
 };
@@ -20,8 +51,7 @@ use wasmtime_environ::{
 #[cfg(feature = "gc")]
 use wasmtime_environ::ModuleInternedTypeIndex;
 
-#[cfg(has_host_compiler_backend)]
-mod arch;
+mod always_mut;
 #[cfg(feature = "component-model")]
 pub mod component;
 mod const_expr;
@@ -31,13 +61,17 @@ mod imports;
 mod instance;
 mod memory;
 mod mmap_vec;
+#[cfg(has_virtual_memory)]
+mod pagemap_disabled;
 mod provenance;
 mod send_sync_ptr;
+mod stack_switching;
 mod store_box;
 mod sys;
 mod table;
+#[cfg(feature = "gc")]
+mod throw;
 mod traphandlers;
-mod unwind;
 mod vmcontext;
 
 #[cfg(feature = "threads")]
@@ -61,19 +95,17 @@ pub(crate) use interpreter_disabled as interpreter;
 #[cfg(feature = "debug-builtins")]
 pub use wasmtime_jit_debug::gdb_jit_int::GdbJitImageRegistration;
 
-#[cfg(has_host_compiler_backend)]
-pub use crate::runtime::vm::arch::get_stack_pointer;
+pub use crate::runtime::vm::always_mut::*;
 pub use crate::runtime::vm::export::*;
 pub use crate::runtime::vm::gc::*;
 pub use crate::runtime::vm::imports::Imports;
 pub use crate::runtime::vm::instance::{
-    GcHeapAllocationIndex, Instance, InstanceAllocationRequest, InstanceAllocator,
-    InstanceAllocatorImpl, InstanceAndStore, InstanceHandle, MemoryAllocationIndex,
-    OnDemandInstanceAllocator, StorePtr, TableAllocationIndex,
+    GcHeapAllocationIndex, Instance, InstanceAllocationRequest, InstanceAllocator, InstanceHandle,
+    MemoryAllocationIndex, OnDemandInstanceAllocator, TableAllocationIndex, initialize_instance,
 };
 #[cfg(feature = "pooling-allocator")]
 pub use crate::runtime::vm::instance::{
-    InstanceLimits, PoolConcurrencyLimitError, PoolingInstanceAllocator,
+    InstanceLimits, PoolConcurrencyLimitError, PoolingAllocatorMetrics, PoolingInstanceAllocator,
     PoolingInstanceAllocatorConfig,
 };
 pub use crate::runtime::vm::interpreter::*;
@@ -82,22 +114,29 @@ pub use crate::runtime::vm::memory::{
 };
 pub use crate::runtime::vm::mmap_vec::MmapVec;
 pub use crate::runtime::vm::provenance::*;
+pub use crate::runtime::vm::stack_switching::*;
 pub use crate::runtime::vm::store_box::*;
 #[cfg(feature = "std")]
 pub use crate::runtime::vm::sys::mmap::open_file_for_mmap;
 #[cfg(has_host_compiler_backend)]
 pub use crate::runtime::vm::sys::unwind::UnwindRegistration;
-pub use crate::runtime::vm::table::{Table, TableElement};
+pub use crate::runtime::vm::table::{Table, TableElementType};
+#[cfg(feature = "gc")]
+pub use crate::runtime::vm::throw::*;
 pub use crate::runtime::vm::traphandlers::*;
-pub use crate::runtime::vm::unwind::*;
-#[cfg(feature = "component-model")]
-pub use crate::runtime::vm::vmcontext::VMTableDefinition;
 pub use crate::runtime::vm::vmcontext::{
-    VMArrayCallFunction, VMArrayCallHostFuncContext, VMContext, VMFuncRef, VMFunctionBody,
-    VMFunctionImport, VMGlobalDefinition, VMGlobalImport, VMMemoryDefinition, VMMemoryImport,
-    VMOpaqueContext, VMStoreContext, VMTable, VMTagImport, VMWasmCallFunction, ValRaw,
+    VMArrayCallFunction, VMArrayCallHostFuncContext, VMContext, VMFuncRef, VMFunctionImport,
+    VMGlobalDefinition, VMGlobalImport, VMGlobalKind, VMMemoryDefinition, VMMemoryImport,
+    VMOpaqueContext, VMStoreContext, VMTableImport, VMTagImport, VMWasmCallFunction, ValRaw,
 };
+#[cfg(has_custom_sync)]
+pub(crate) use sys::capi;
+
 pub use send_sync_ptr::SendSyncPtr;
+pub use wasmtime_unwinder::Unwind;
+
+#[cfg(has_host_compiler_backend)]
+pub use wasmtime_unwinder::{UnwindHost, get_stack_pointer};
 
 mod module_id;
 pub use module_id::CompiledModuleId;
@@ -131,76 +170,55 @@ cfg_if::cfg_if! {
     }
 }
 
+/// Source of data used for [`MemoryImage`]
+pub trait ModuleMemoryImageSource: Send + Sync + 'static {
+    /// Returns this image's slice of all wasm data for a module which is then
+    /// further sub-sliced for a particular initialization segment.
+    fn wasm_data(&self) -> &[u8];
+
+    /// Optionally returns the backing mmap. Used for using the backing mmap's
+    /// file to perform other mmaps, for example.
+    fn mmap(&self) -> Option<&MmapVec>;
+}
+
 /// Dynamic runtime functionality needed by this crate throughout the execution
 /// of a wasm instance.
 ///
 /// This trait is used to store a raw pointer trait object within each
 /// `VMContext`. This raw pointer trait object points back to the
-/// `wasmtime::Store` internally but is type-erased so this `wasmtime-runtime`
-/// crate doesn't need the entire `wasmtime` crate to build.
+/// `wasmtime::Store` internally but is type-erased to avoid needing to
+/// monomorphize the entire runtime on the `T` in `Store<T>`
 ///
-/// Note that this is an extra-unsafe trait because no heed is paid to the
-/// lifetime of this store or the Send/Sync-ness of this store. All of that must
-/// be respected by embedders (e.g. the `wasmtime::Store` structure). The theory
-/// is that `wasmtime::Store` handles all this correctly.
-pub unsafe trait VMStore {
+/// # Safety
+///
+/// This trait should be implemented by nothing other than `StoreInner<T>` in
+/// this crate. It's not sound to implement it for anything else due to
+/// `unchecked_context_mut` below.
+///
+/// It's also worth nothing that there are various locations where a `*mut dyn
+/// VMStore` is asserted to be both `Send` and `Sync` which disregards the `T`
+/// that's actually stored in the store itself. It's assume that the high-level
+/// APIs using `Store<T>` are correctly inferring send/sync on the returned
+/// values (e.g. futures) and that internally in the runtime we aren't doing
+/// anything "weird" with threads for example.
+pub unsafe trait VMStore: 'static {
     /// Get a shared borrow of this store's `StoreOpaque`.
     fn store_opaque(&self) -> &StoreOpaque;
 
     /// Get an exclusive borrow of this store's `StoreOpaque`.
     fn store_opaque_mut(&mut self) -> &mut StoreOpaque;
 
-    /// Callback invoked to allow the store's resource limiter to reject a
-    /// memory grow operation.
-    fn memory_growing(
+    /// Returns a split borrow to the limiter plus `StoreOpaque` at the same
+    /// time.
+    fn resource_limiter_and_store_opaque(
         &mut self,
-        current: usize,
-        desired: usize,
-        maximum: Option<usize>,
-    ) -> Result<bool, Error>;
-
-    /// Callback invoked to notify the store's resource limiter that a memory
-    /// grow operation has failed.
-    ///
-    /// Note that this is not invoked if `memory_growing` returns an error.
-    fn memory_grow_failed(&mut self, error: Error) -> Result<()>;
-
-    /// Callback invoked to allow the store's resource limiter to reject a
-    /// table grow operation.
-    fn table_growing(
-        &mut self,
-        current: usize,
-        desired: usize,
-        maximum: Option<usize>,
-    ) -> Result<bool, Error>;
-
-    /// Callback invoked to notify the store's resource limiter that a table
-    /// grow operation has failed.
-    ///
-    /// Note that this is not invoked if `table_growing` returns an error.
-    fn table_grow_failed(&mut self, error: Error) -> Result<()>;
-
-    /// Callback invoked whenever fuel runs out by a wasm instance. If an error
-    /// is returned that's raised as a trap. Otherwise wasm execution will
-    /// continue as normal.
-    fn out_of_gas(&mut self) -> Result<(), Error>;
+    ) -> (Option<StoreResourceLimiter<'_>>, &mut StoreOpaque);
 
     /// Callback invoked whenever an instance observes a new epoch
     /// number. Cannot fail; cooperative epoch-based yielding is
     /// completely semantically transparent. Returns the new deadline.
     #[cfg(target_has_atomic = "64")]
-    fn new_epoch(&mut self) -> Result<u64, Error>;
-
-    /// Callback invoked whenever an instance needs to trigger a GC.
-    ///
-    /// Optionally given a GC reference that is rooted for the collection, and
-    /// then whose updated GC reference is returned.
-    ///
-    /// Cooperative, async-yielding (if configured) is completely transparent.
-    ///
-    /// If the async GC was cancelled, returns an error. This should be raised
-    /// as a trap to clean up Wasm execution.
-    fn maybe_async_gc(&mut self, root: Option<VMGcRef>) -> Result<Option<VMGcRef>>;
+    fn new_epoch_updated_deadline(&mut self) -> Result<crate::UpdateDeadline>;
 
     /// Metadata required for resources for the component model.
     #[cfg(feature = "component-model")]
@@ -210,6 +228,10 @@ pub unsafe trait VMStore {
     fn component_async_store(
         &mut self,
     ) -> &mut dyn crate::runtime::component::VMComponentAsyncStore;
+
+    /// Invoke a debug handler, if present, at a debug event.
+    #[cfg(feature = "debug")]
+    fn block_on_debug_handler(&mut self, event: crate::DebugEvent) -> anyhow::Result<()>;
 }
 
 impl Deref for dyn VMStore + '_ {
@@ -226,6 +248,19 @@ impl DerefMut for dyn VMStore + '_ {
     }
 }
 
+impl dyn VMStore + '_ {
+    /// Asserts that this `VMStore` was originally paired with `StoreInner<T>`
+    /// and then casts to the `StoreContextMut` type.
+    ///
+    /// # Unsafety
+    ///
+    /// This method is not safe as there's no static guarantee that `T` is
+    /// correct for this store.
+    pub(crate) unsafe fn unchecked_context_mut<T>(&mut self) -> StoreContextMut<'_, T> {
+        unsafe { StoreContextMut(&mut *(self as *mut dyn VMStore as *mut StoreInner<T>)) }
+    }
+}
+
 /// A newtype wrapper around `NonNull<dyn VMStore>` intended to be a
 /// self-pointer back to the `Store<T>` within raw data structures like
 /// `VMContext`.
@@ -275,23 +310,23 @@ pub enum ModuleRuntimeInfo {
 #[derive(Clone)]
 pub struct BareModuleInfo {
     module: Arc<wasmtime_environ::Module>,
-    one_signature: Option<VMSharedTypeIndex>,
     offsets: VMOffsets<HostPtr>,
+    _registered_type: Option<RegisteredType>,
 }
 
 impl ModuleRuntimeInfo {
     pub(crate) fn bare(module: Arc<wasmtime_environ::Module>) -> Self {
-        ModuleRuntimeInfo::bare_maybe_imported_func(module, None)
+        ModuleRuntimeInfo::bare_with_registered_type(module, None)
     }
 
-    pub(crate) fn bare_maybe_imported_func(
+    pub(crate) fn bare_with_registered_type(
         module: Arc<wasmtime_environ::Module>,
-        one_signature: Option<VMSharedTypeIndex>,
+        registered_type: Option<RegisteredType>,
     ) -> Self {
         ModuleRuntimeInfo::Bare(Box::new(BareModuleInfo {
             offsets: VMOffsets::new(HostPtr, &module),
             module,
-            one_signature,
+            _registered_type: registered_type,
         }))
     }
 
@@ -393,10 +428,7 @@ impl ModuleRuntimeInfo {
                 .as_module_map()
                 .values()
                 .as_slice(),
-            ModuleRuntimeInfo::Bare(b) => match &b.one_signature {
-                Some(s) => core::slice::from_ref(s),
-                None => &[],
-            },
+            ModuleRuntimeInfo::Bare(_) => &[],
         }
     }
 
@@ -412,6 +444,8 @@ impl ModuleRuntimeInfo {
 /// Returns the host OS page size, in bytes.
 #[cfg(has_virtual_memory)]
 pub fn host_page_size() -> usize {
+    // NB: this function is duplicated in `crates/fiber/src/unix.rs` so if this
+    // changes that should probably get updated as well.
     static PAGE_SIZE: AtomicUsize = AtomicUsize::new(0);
 
     return match PAGE_SIZE.load(Ordering::Relaxed) {
@@ -457,3 +491,38 @@ impl fmt::Display for WasmFault {
         )
     }
 }
+
+/// Asserts that the future `f` is ready and returns its output.
+///
+/// This function is intended to be used when `async_support` is verified as
+/// disabled. Internals of Wasmtime are generally `async` when they optionally
+/// can be, meaning that synchronous entrypoints will invoke this function
+/// after invoking the asynchronous internals. Due to `async_support` being
+/// disabled there should be no way to introduce a yield point meaning that all
+/// futures built from internal functions should always be ready.
+///
+/// # Panics
+///
+/// Panics if `f` is not yet ready.
+pub fn assert_ready<F: Future>(f: F) -> F::Output {
+    one_poll(f).unwrap()
+}
+
+/// Attempts one poll of `f` to see if its output is available.
+///
+/// This function is intended for a few minor entrypoints into the Wasmtime API
+/// where a synchronous function is documented to work even when `async_support`
+/// is enabled. For example growing a `Memory` can be done with a synchronous
+/// function, but it's documented to panic with an async resource limiter.
+///
+/// This function provides the opportunity to poll `f` once to see if its output
+/// is available. If it isn't then `None` is returned and an appropriate panic
+/// message should be generated recommending to use an async function (e.g.
+/// `grow_async` instead of `grow`).
+pub fn one_poll<F: Future>(f: F) -> Option<F::Output> {
+    let mut context = Context::from_waker(&Waker::noop());
+    match pin!(f).poll(&mut context) {
+        Poll::Ready(output) => Some(output),
+        Poll::Pending => None,
+    }
+}
