@@ -75,10 +75,11 @@
 //! they should be merged together.
 
 use crate::prelude::*;
+use crate::runtime::store::StoreResourceLimiter;
 use crate::runtime::vm::vmcontext::VMMemoryDefinition;
 #[cfg(has_virtual_memory)]
 use crate::runtime::vm::{HostAlignedByteCount, MmapOffset};
-use crate::runtime::vm::{MemoryImage, MemoryImageSlot, SendSyncPtr, VMStore};
+use crate::runtime::vm::{MemoryImage, MemoryImageSlot, SendSyncPtr};
 use alloc::sync::Arc;
 use core::{ops::Range, ptr::NonNull};
 use wasmtime_environ::Tunables;
@@ -169,6 +170,9 @@ pub trait RuntimeLinearMemory: Send + Sync {
     /// offset within it.
     fn base(&self) -> MemoryBase;
 
+    /// Get a `VMMemoryDefinition` for this linear memory.
+    fn vmmemory(&self) -> VMMemoryDefinition;
+
     /// Internal method for Wasmtime when used in conjunction with CoW images.
     /// This is used to inform the underlying memory that the size of memory has
     /// changed.
@@ -227,14 +231,14 @@ pub enum Memory {
 
 impl Memory {
     /// Create a new dynamic (movable) memory instance for the specified plan.
-    pub fn new_dynamic(
+    pub async fn new_dynamic(
         ty: &wasmtime_environ::Memory,
         tunables: &Tunables,
         creator: &dyn RuntimeMemoryCreator,
-        store: &mut dyn VMStore,
         memory_image: Option<&Arc<MemoryImage>>,
+        limiter: Option<&mut StoreResourceLimiter<'_>>,
     ) -> Result<Self> {
-        let (minimum, maximum) = Self::limit_new(ty, Some(store))?;
+        let (minimum, maximum) = Self::limit_new(ty, limiter).await?;
         let allocation = creator.new_memory(ty, tunables, minimum, maximum)?;
 
         let memory = LocalMemory::new(ty, tunables, allocation, memory_image)?;
@@ -247,15 +251,15 @@ impl Memory {
 
     /// Create a new static (immovable) memory instance for the specified plan.
     #[cfg(feature = "pooling-allocator")]
-    pub fn new_static(
+    pub async fn new_static(
         ty: &wasmtime_environ::Memory,
         tunables: &Tunables,
         base: MemoryBase,
         base_capacity: usize,
         memory_image: MemoryImageSlot,
-        store: &mut dyn VMStore,
+        limiter: Option<&mut StoreResourceLimiter<'_>>,
     ) -> Result<Self> {
-        let (minimum, maximum) = Self::limit_new(ty, Some(store))?;
+        let (minimum, maximum) = Self::limit_new(ty, limiter).await?;
         let pooled_memory = StaticMemory::new(base, base_capacity, minimum, maximum)?;
         let allocation = Box::new(pooled_memory);
 
@@ -282,9 +286,9 @@ impl Memory {
     ///
     /// Returns a tuple of the minimum size, optional maximum size, and log(page
     /// size) of the memory, all in bytes.
-    pub(crate) fn limit_new(
+    pub(crate) async fn limit_new(
         ty: &wasmtime_environ::Memory,
-        store: Option<&mut dyn VMStore>,
+        limiter: Option<&mut StoreResourceLimiter<'_>>,
     ) -> Result<(usize, Option<usize>)> {
         let page_size = usize::try_from(ty.page_size()).unwrap();
 
@@ -327,8 +331,11 @@ impl Memory {
         // `minimum` calculation overflowed. This means that the `minimum` we're
         // informing the limiter is lossy and may not be 100% accurate, but for
         // now the expected uses of limiter means that's ok.
-        if let Some(store) = store {
-            if !store.memory_growing(0, minimum.unwrap_or(absolute_max), maximum)? {
+        if let Some(limiter) = limiter {
+            if !limiter
+                .memory_growing(0, minimum.unwrap_or(absolute_max), maximum)
+                .await?
+            {
                 bail!(
                     "memory minimum size of {} pages exceeds memory limits",
                     ty.limits.min
@@ -356,7 +363,7 @@ impl Memory {
         }
     }
 
-    /// Returns the number of allocated wasm pages.
+    /// Returns the size of this memory, in bytes.
     pub fn byte_size(&self) -> usize {
         match self {
             Memory::Local(mem) => mem.byte_size(),
@@ -391,14 +398,14 @@ impl Memory {
     ///
     /// Ensure that the provided Store is not used to get access any Memory
     /// which lives inside it.
-    pub unsafe fn grow(
+    pub async unsafe fn grow(
         &mut self,
         delta_pages: u64,
-        store: Option<&mut dyn VMStore>,
+        limiter: Option<&mut StoreResourceLimiter<'_>>,
     ) -> Result<Option<usize>, Error> {
         let result = match self {
-            Memory::Local(mem) => mem.grow(delta_pages, store)?,
-            Memory::Shared(mem) => mem.grow(delta_pages, store)?,
+            Memory::Local(mem) => mem.grow(delta_pages, limiter).await?,
+            Memory::Shared(mem) => mem.grow(delta_pages)?,
         };
         match result {
             Some((old, _new)) => Ok(Some(old)),
@@ -407,7 +414,7 @@ impl Memory {
     }
 
     /// Return a `VMMemoryDefinition` for exposing the memory to compiled wasm code.
-    pub fn vmmemory(&mut self) -> VMMemoryDefinition {
+    pub fn vmmemory(&self) -> VMMemoryDefinition {
         match self {
             Memory::Local(mem) => mem.vmmemory(),
             // `vmmemory()` is used for writing the `VMMemoryDefinition` of a
@@ -429,9 +436,14 @@ impl Memory {
         }
     }
 
+    /// Is this a shared memory?
+    pub fn is_shared_memory(&self) -> bool {
+        matches!(self, Memory::Shared(_))
+    }
+
     /// If the [Memory] is a [SharedMemory], unwrap it and return a clone to
     /// that shared memory.
-    pub fn as_shared_memory(&mut self) -> Option<&mut SharedMemory> {
+    pub fn as_shared_memory(&self) -> Option<&SharedMemory> {
         match self {
             Memory::Local(_) => None,
             Memory::Shared(mem) => Some(mem),
@@ -539,29 +551,6 @@ impl LocalMemory {
 
                     let mut slot =
                         MemoryImageSlot::create(mmap_base, byte_size, alloc.byte_capacity());
-                    // On drop, we will unmap our mmap'd range that this slot
-                    // was mapped on top of, so there is no need for the slot to
-                    // wipe it with an anonymous mapping first.
-                    //
-                    // Note that this code would be incorrect if clear-on-drop
-                    // were enabled. That's because:
-                    //
-                    // * In the struct definition, `memory_image` above is listed
-                    //   after `alloc`.
-                    // * Rust drops fields in the order they're defined, so
-                    //   `memory_image` would be dropped after `alloc`.
-                    // * `alloc` can represent either owned memory (i.e. the mmap is
-                    //   freed on drop) or logically borrowed memory (something else
-                    //   manages the mmap).
-                    // * If `alloc` is borrowed memory, then this isn't an issue.
-                    // * But if `alloc` is owned memory, then it would first drop
-                    //   the mmap, and then `memory_image` would try to remap
-                    //   part of that same memory as part of clear-on-drop.
-                    //
-                    // A lot of this really suggests representing the ownership
-                    // via Rust lifetimes -- that would be a major refactor,
-                    // though.
-                    slot.no_clear_on_drop();
                     slot.instantiate(alloc.byte_size(), Some(image), ty, tunables)?;
                     Some(slot)
                 } else {
@@ -592,10 +581,10 @@ impl LocalMemory {
     /// the underlying `grow_to` implementation.
     ///
     /// The `store` is used only for error reporting.
-    pub fn grow(
+    pub async fn grow(
         &mut self,
         delta_pages: u64,
-        mut store: Option<&mut dyn VMStore>,
+        mut limiter: Option<&mut StoreResourceLimiter<'_>>,
     ) -> Result<Option<(usize, usize)>, Error> {
         let old_byte_size = self.alloc.byte_size();
 
@@ -626,8 +615,11 @@ impl LocalMemory {
             .and_then(|n| usize::try_from(n).ok());
 
         // Store limiter gets first chance to reject memory_growing.
-        if let Some(store) = &mut store {
-            if !store.memory_growing(old_byte_size, new_byte_size, maximum)? {
+        if let Some(limiter) = &mut limiter {
+            if !limiter
+                .memory_growing(old_byte_size, new_byte_size, maximum)
+                .await?
+            {
                 return Ok(None);
             }
         }
@@ -693,19 +685,16 @@ impl LocalMemory {
                 // report the growth failure to but the error should not be
                 // dropped
                 // (https://github.com/bytecodealliance/wasmtime/issues/4240).
-                if let Some(store) = store {
-                    store.memory_grow_failed(e)?;
+                if let Some(limiter) = limiter {
+                    limiter.memory_grow_failed(e)?;
                 }
                 Ok(None)
             }
         }
     }
 
-    pub fn vmmemory(&mut self) -> VMMemoryDefinition {
-        VMMemoryDefinition {
-            base: self.alloc.base().as_non_null().into(),
-            current_length: self.alloc.byte_size().into(),
-        }
+    pub fn vmmemory(&self) -> VMMemoryDefinition {
+        self.alloc.vmmemory()
     }
 
     pub fn byte_size(&self) -> usize {
@@ -758,7 +747,7 @@ pub fn validate_atomic_addr(
     }
 
     let length = u64::try_from(def.current_length()).unwrap();
-    if !(addr.saturating_add(access_size) < length) {
+    if !(addr.saturating_add(access_size) <= length) {
         return Err(Trap::MemoryOutOfBounds);
     }
 
