@@ -26,24 +26,24 @@ use alloc::collections::VecDeque;
 use alloc::rc::Rc;
 use alloc::string::{String, ToString};
 use alloc::vec::Vec;
-use anyhow::{bail, Result};
+use anyhow::{Result, bail};
 use core::cell::{Cell, RefCell};
 use core::fmt::Write as _;
 use core::future::Future;
 use core::pin::Pin;
 use core::task::{Context, Poll, Waker};
 use wasmtime::component::{Component, Linker, Resource, ResourceTable};
-use wasmtime::{Config, Engine, Store};
+use wasmtime::{Engine, Store};
 use wasmtime_wasi_io::{
-    bytes::Bytes,
-    poll::{subscribe, DynPollable, Pollable},
-    streams::{DynInputStream, DynOutputStream, InputStream, OutputStream},
     IoView,
+    bytes::Bytes,
+    poll::{DynPollable, Pollable, subscribe},
+    streams::{DynInputStream, DynOutputStream, InputStream, OutputStream},
 };
 
 /// Unlike super::run, its nice to provide some sort of output showing what the
 /// wasi program did while it executed, so this function reports in out_buf
-/// what stdout/stderr prints occured on success (returns 0), or the error
+/// what stdout/stderr prints occurred on success (returns 0), or the error
 /// message on failure (returns != 0).
 #[unsafe(no_mangle)]
 pub unsafe extern "C" fn run_wasi(
@@ -52,21 +52,23 @@ pub unsafe extern "C" fn run_wasi(
     wasi_component: *const u8,
     wasi_component_size: usize,
 ) -> usize {
-    let buf = core::slice::from_raw_parts_mut(out_buf, *out_size);
-    let wasi_component = core::slice::from_raw_parts(wasi_component, wasi_component_size);
-    match run(wasi_component) {
-        Ok(output) => {
-            let len = buf.len().min(output.len());
-            buf[..len].copy_from_slice(&output.as_bytes()[..len]);
-            *out_size = len;
-            return 0;
-        }
-        Err(e) => {
-            let msg = format!("{e:?}");
-            let len = buf.len().min(msg.len());
-            buf[..len].copy_from_slice(&msg.as_bytes()[..len]);
-            *out_size = len;
-            return 1;
+    unsafe {
+        let buf = core::slice::from_raw_parts_mut(out_buf, *out_size);
+        let wasi_component = core::slice::from_raw_parts(wasi_component, wasi_component_size);
+        match run(wasi_component) {
+            Ok(output) => {
+                let len = buf.len().min(output.len());
+                buf[..len].copy_from_slice(&output.as_bytes()[..len]);
+                *out_size = len;
+                return 0;
+            }
+            Err(e) => {
+                let msg = format!("{e:?}");
+                let len = buf.len().min(msg.len());
+                buf[..len].copy_from_slice(&msg.as_bytes()[..len]);
+                *out_size = len;
+                return 1;
+            }
         }
     }
 }
@@ -76,7 +78,7 @@ fn run(wasi_component: &[u8]) -> Result<String> {
     // interface will poll as Pending while execution is suspended and it is
     // waiting for a Pollable to become Ready. This example provides a very
     // small async executor which is entered below with `block_on`.
-    let mut config = Config::default();
+    let mut config = super::config();
     config.async_support(true);
     // For future: we could consider turning on fuel in the Config to meter
     // how long a wasm guest could execute for.
@@ -148,10 +150,11 @@ fn deserialize(engine: &Engine, component: &[u8]) -> Result<Option<Component>> {
 // Generate bindings for the entire wasi:cli command world. We won't impl and
 // link with all of these generated bindings for the sake of this example.
 wasmtime::component::bindgen!({
-    path: "../../../crates/wasi/wit",
+    path: "../../../crates/wasi/src/p2/wit",
     world: "wasi:cli/command",
-    async: { only_imports: [] },
-    trappable_imports: true,
+    imports: { default: trappable },
+    exports: { default: async },
+    require_store_data_send: true,
     // Important: tell bindgen that anywhere it encounters the wasi:io
     // package, refer to the bindings generated in the wasmtime_wasi_io crate.
     // This way, all uses of the streams and pollable in the bindings in this
@@ -210,16 +213,18 @@ impl ExampleCtx {
 // wasi:cli/command world. Many of these impls are bare-bones and some are
 // intentionally broken, see notes below.
 pub fn add_to_linker_async(linker: &mut Linker<ExampleCtx>) -> Result<()> {
-    wasi::clocks::monotonic_clock::add_to_linker(linker, |t| t)?;
-    wasi::clocks::wall_clock::add_to_linker(linker, |t| t)?;
-    wasi::cli::environment::add_to_linker(linker, |t| t)?;
-    wasi::cli::exit::add_to_linker(linker, &wasi::cli::exit::LinkOptions::default(), |t| t)?;
-    wasi::cli::stdin::add_to_linker(linker, |t| t)?;
-    wasi::cli::stdout::add_to_linker(linker, |t| t)?;
-    wasi::cli::stderr::add_to_linker(linker, |t| t)?;
-    wasi::random::random::add_to_linker(linker, |t| t)?;
-    wasi::filesystem::preopens::add_to_linker(linker, |t| t)?;
-    wasi::filesystem::types::add_to_linker(linker, |t| t)?;
+    type Data = wasmtime::component::HasSelf<ExampleCtx>;
+
+    wasi::clocks::monotonic_clock::add_to_linker::<_, Data>(linker, |t| t)?;
+    wasi::clocks::wall_clock::add_to_linker::<_, Data>(linker, |t| t)?;
+    wasi::cli::environment::add_to_linker::<_, Data>(linker, |t| t)?;
+    wasi::cli::exit::add_to_linker::<_, Data>(linker, &Default::default(), |t| t)?;
+    wasi::cli::stdin::add_to_linker::<_, Data>(linker, |t| t)?;
+    wasi::cli::stdout::add_to_linker::<_, Data>(linker, |t| t)?;
+    wasi::cli::stderr::add_to_linker::<_, Data>(linker, |t| t)?;
+    wasi::random::random::add_to_linker::<_, Data>(linker, |t| t)?;
+    wasi::filesystem::preopens::add_to_linker::<_, Data>(linker, |t| t)?;
+    wasi::filesystem::types::add_to_linker::<_, Data>(linker, |t| t)?;
     Ok(())
 }
 
@@ -296,7 +301,7 @@ impl InputStream for NeverReadable {
     }
 }
 
-// WriteLog is used implement stdout and stderr. Clonable because wasi:cli
+// WriteLog is used implement stdout and stderr. Cloneable because wasi:cli
 // requires, when calling get_stdout/get_stderr multiple times, to provide
 // distinct resources that point to the same underlying stream. RefCell
 // provides mutation, and VecDeque provides O(1) push_back operation.
@@ -412,24 +417,6 @@ impl ExecutorInner {
     }
 }
 
-// Yanked from core::task::wake, which is unfortunately still unstable :/
-fn noop_waker() -> Waker {
-    use core::task::{RawWaker, RawWakerVTable};
-    const VTABLE: RawWakerVTable = RawWakerVTable::new(
-        // Cloning just returns a new no-op raw waker
-        |_| RAW,
-        // `wake` does nothing
-        |_| {},
-        // `wake_by_ref` does nothing
-        |_| {},
-        // Dropping does nothing as we don't allocate anything
-        |_| {},
-    );
-    const RAW: RawWaker = RawWaker::new(core::ptr::null(), &VTABLE);
-
-    unsafe { Waker::from_raw(RAW) }
-}
-
 fn block_on<R>(clock: Clock, f: impl Future<Output = Result<R>> + Send + 'static) -> Result<R> {
     // Guard against nested invocations
     if EXECUTOR.0.borrow_mut().is_some() {
@@ -439,8 +426,7 @@ fn block_on<R>(clock: Clock, f: impl Future<Output = Result<R>> + Send + 'static
     *EXECUTOR.0.borrow_mut() = Some(Executor(executor.0.clone()));
 
     // No special waker needed for this executor.
-    let waker = noop_waker();
-    let mut cx = Context::from_waker(&waker);
+    let mut cx = Context::from_waker(Waker::noop());
     let mut f = core::pin::pin!(f);
 
     // Drive the Future to completion in the following loop
