@@ -6,22 +6,19 @@ You can execute this example with:
     cargo run --example wasip1
 */
 
-use wasi_common::sync::WasiCtxBuilder;
 use wasmtime::*;
+use wasmtime_wasi::WasiCtx;
 
 fn main() -> Result<()> {
     // Define the WASI functions globally on the `Config`.
     let engine = Engine::default();
     let mut linker = Linker::new(&engine);
-    wasi_common::sync::add_to_linker(&mut linker, |s| s)?;
+    wasmtime_wasi::p1::add_to_linker_sync(&mut linker, |s| s)?;
 
     // Create a WASI context and put it in a Store; all instances in the store
     // share this context. `WasiCtxBuilder` provides a number of ways to
     // configure what the target program will have access to.
-    let wasi = WasiCtxBuilder::new()
-        .inherit_stdio()
-        .inherit_args()?
-        .build();
+    let wasi = WasiCtx::builder().inherit_stdio().inherit_args().build_p1();
     let mut store = Store::new(&engine, wasi);
 
     // Instantiate our module with the imports we've created, and run it.
