@@ -3,8 +3,8 @@
 use core::ptr::NonNull;
 
 use alloc::vec::Vec;
-use cranelift_bitset::scalar::ScalarBitSetStorage;
 use cranelift_bitset::ScalarBitSet;
+use cranelift_bitset::scalar::ScalarBitSetStorage;
 
 use crate::imms::*;
 use crate::opcode::*;
@@ -223,7 +223,7 @@ impl UnsafeBytecodeStream {
     /// that the address at `self._as_ptr() + offset` contains valid Pulley
     /// bytecode.
     pub unsafe fn offset(&self, offset: isize) -> Self {
-        UnsafeBytecodeStream(NonNull::new_unchecked(self.0.as_ptr().offset(offset)))
+        UnsafeBytecodeStream(unsafe { NonNull::new_unchecked(self.0.as_ptr().offset(offset)) })
     }
 
     /// Get this stream's underlying raw pointer.
