@@ -6,34 +6,30 @@
 )]
 
 use crate::common::{Profile, RunCommon, RunTarget};
-
-use anyhow::{anyhow, bail, Context as _, Error, Result};
+use anyhow::{Context as _, Error, Result, anyhow, bail};
 use clap::Parser;
 use std::ffi::OsString;
 use std::path::{Path, PathBuf};
 use std::sync::{Arc, Mutex};
 use std::thread;
-use wasi_common::sync::{ambient_authority, Dir, TcpListener, WasiCtxBuilder};
+use wasi_common::sync::{Dir, TcpListener, WasiCtxBuilder, ambient_authority};
 use wasmtime::{Engine, Func, Module, Store, StoreLimits, Val, ValType};
-use wasmtime_wasi::{IoView, WasiView};
-
-#[cfg(feature = "wasi-nn")]
-use wasmtime_wasi_nn::wit::WasiNnView;
-
-#[cfg(feature = "wasi-threads")]
-use wasmtime_wasi_threads::WasiThreadsCtx;
+use wasmtime_wasi::{WasiCtxView, WasiView};
 
 #[cfg(feature = "wasi-config")]
 use wasmtime_wasi_config::{WasiConfig, WasiConfigVariables};
 #[cfg(feature = "wasi-http")]
 use wasmtime_wasi_http::{
-    WasiHttpCtx, DEFAULT_OUTGOING_BODY_BUFFER_CHUNKS, DEFAULT_OUTGOING_BODY_CHUNK_SIZE,
+    DEFAULT_OUTGOING_BODY_BUFFER_CHUNKS, DEFAULT_OUTGOING_BODY_CHUNK_SIZE, WasiHttpCtx,
 };
 #[cfg(feature = "wasi-keyvalue")]
 use wasmtime_wasi_keyvalue::{WasiKeyValue, WasiKeyValueCtx, WasiKeyValueCtxBuilder};
-
+#[cfg(feature = "wasi-nn")]
+use wasmtime_wasi_nn::wit::WasiNnView;
+#[cfg(feature = "wasi-threads")]
+use wasmtime_wasi_threads::WasiThreadsCtx;
 #[cfg(feature = "wasi-tls")]
-use wasmtime_wasi_tls::WasiTlsCtx;
+use wasmtime_wasi_tls::{WasiTls, WasiTlsCtx};
 
 fn parse_preloads(s: &str) -> Result<(String, PathBuf)> {
     let parts: Vec<&str> = s.splitn(2, '=').collect();
@@ -54,14 +50,9 @@ pub struct RunCommand {
     #[arg(long, value_name = "FUNCTION")]
     pub invoke: Option<String>,
 
-    /// Load the given WebAssembly module before the main module
-    #[arg(
-        long = "preload",
-        number_of_values = 1,
-        value_name = "NAME=MODULE_PATH",
-        value_parser = parse_preloads,
-    )]
-    pub preloads: Vec<(String, PathBuf)>,
+    #[command(flatten)]
+    #[expect(missing_docs, reason = "don't want to mess with clap doc-strings")]
+    pub preloads: Preloads,
 
     /// Override the value of `argv[0]`, typically the name of the executable of
     /// the application being run.
@@ -81,17 +72,61 @@ pub struct RunCommand {
     pub module_and_args: Vec<OsString>,
 }
 
-enum CliLinker {
+#[expect(missing_docs, reason = "don't want to mess with clap doc-strings")]
+#[derive(Parser, Default, Clone)]
+pub struct Preloads {
+    /// Load the given WebAssembly module before the main module
+    #[arg(
+        long = "preload",
+        number_of_values = 1,
+        value_name = "NAME=MODULE_PATH",
+        value_parser = parse_preloads,
+    )]
+    modules: Vec<(String, PathBuf)>,
+}
+
+/// Dispatch between either a core or component linker.
+#[expect(missing_docs, reason = "self-explanatory")]
+pub enum CliLinker {
     Core(wasmtime::Linker<Host>),
     #[cfg(feature = "component-model")]
     Component(wasmtime::component::Linker<Host>),
 }
 
+/// Dispatch between either a core or component instance.
+#[expect(missing_docs, reason = "self-explanatory")]
+pub enum CliInstance {
+    Core(wasmtime::Instance),
+    #[cfg(feature = "component-model")]
+    Component(wasmtime::component::Instance),
+}
+
 impl RunCommand {
     /// Executes the command.
+    #[cfg(feature = "run")]
     pub fn execute(mut self) -> Result<()> {
-        self.run.common.init_logging()?;
+        let runtime = tokio::runtime::Builder::new_multi_thread()
+            .enable_time()
+            .enable_io()
+            .build()?;
 
+        runtime.block_on(async {
+            self.run.common.init_logging()?;
+
+            let engine = self.new_engine()?;
+            let main = self
+                .run
+                .load_module(&engine, self.module_and_args[0].as_ref())?;
+            let (mut store, mut linker) = self.new_store_and_linker(&engine, &main)?;
+
+            self.instantiate_and_run(&engine, &mut linker, &main, &mut store)
+                .await?;
+            Ok(())
+        })
+    }
+
+    /// Creates a new `Engine` with the configuration for this command.
+    pub fn new_engine(&mut self) -> Result<Engine> {
         let mut config = self.run.common.config(None)?;
         config.async_support(true);
 
@@ -109,13 +144,19 @@ impl RunCommand {
             None => {}
         }
 
-        let engine = Engine::new(&config)?;
-
-        // Read the wasm module binary either as `*.wat` or a raw binary.
-        let main = self
-            .run
-            .load_module(&engine, self.module_and_args[0].as_ref())?;
+        Engine::new(&config)
+    }
 
+    /// Populatse a new `Store` and `CliLinker` with the configuration in this
+    /// command.
+    ///
+    /// The `engine` provided is used to for the store/linker and the `main`
+    /// provided is the module/component that is going to be run.
+    pub fn new_store_and_linker(
+        &mut self,
+        engine: &Engine,
+        main: &RunTarget,
+    ) -> Result<(Store<Host>, CliLinker)> {
         // Validate coredump-on-trap argument
         if let Some(path) = &self.run.common.debug.coredump {
             if path.contains("%") {
@@ -166,88 +207,94 @@ impl RunCommand {
             store.set_fuel(fuel)?;
         }
 
-        // Always run the module asynchronously to ensure that the module can be
-        // interrupted, even if it is blocking on I/O or a timeout or something.
-        let runtime = tokio::runtime::Builder::new_multi_thread()
-            .enable_time()
-            .enable_io()
-            .build()?;
+        Ok((store, linker))
+    }
 
+    /// Executes the `main` after instantiating it within `store`.
+    ///
+    /// This applies all configuration within `self`, such as timeouts and
+    /// profiling, and performs the execution. The resulting instance is
+    /// returned.
+    pub async fn instantiate_and_run(
+        &self,
+        engine: &Engine,
+        linker: &mut CliLinker,
+        main: &RunTarget,
+        store: &mut Store<Host>,
+    ) -> Result<CliInstance> {
         let dur = self
             .run
             .common
             .wasm
             .timeout
             .unwrap_or(std::time::Duration::MAX);
-        let result = runtime.block_on(async {
-            tokio::time::timeout(dur, async {
-                let mut profiled_modules: Vec<(String, Module)> = Vec::new();
-                if let RunTarget::Core(m) = &main {
-                    profiled_modules.push(("".to_string(), m.clone()));
-                }
+        let result = tokio::time::timeout(dur, async {
+            let mut profiled_modules: Vec<(String, Module)> = Vec::new();
+            if let RunTarget::Core(m) = &main {
+                profiled_modules.push(("".to_string(), m.clone()));
+            }
 
-                // Load the preload wasm modules.
-                for (name, path) in self.preloads.iter() {
-                    // Read the wasm module binary either as `*.wat` or a raw binary
-                    let preload_target = self.run.load_module(&engine, path)?;
-                    let preload_module = match preload_target {
-                        RunTarget::Core(m) => m,
-                        #[cfg(feature = "component-model")]
-                        RunTarget::Component(_) => {
-                            bail!("components cannot be loaded with `--preload`")
-                        }
-                    };
-                    profiled_modules.push((name.to_string(), preload_module.clone()));
+            // Load the preload wasm modules.
+            for (name, path) in self.preloads.modules.iter() {
+                // Read the wasm module binary either as `*.wat` or a raw binary
+                let preload_target = self.run.load_module(&engine, path)?;
+                let preload_module = match preload_target {
+                    RunTarget::Core(m) => m,
+                    #[cfg(feature = "component-model")]
+                    RunTarget::Component(_) => {
+                        bail!("components cannot be loaded with `--preload`")
+                    }
+                };
+                profiled_modules.push((name.to_string(), preload_module.clone()));
 
-                    // Add the module's functions to the linker.
-                    match &mut linker {
-                        #[cfg(feature = "cranelift")]
-                        CliLinker::Core(linker) => {
-                            linker
-                                .module_async(&mut store, name, &preload_module)
-                                .await
-                                .context(format!(
-                                    "failed to process preload `{}` at `{}`",
-                                    name,
-                                    path.display()
-                                ))?;
-                        }
-                        #[cfg(not(feature = "cranelift"))]
-                        CliLinker::Core(_) => {
-                            bail!("support for --preload disabled at compile time");
-                        }
-                        #[cfg(feature = "component-model")]
-                        CliLinker::Component(_) => {
-                            bail!("--preload cannot be used with components");
-                        }
+                // Add the module's functions to the linker.
+                match linker {
+                    #[cfg(feature = "cranelift")]
+                    CliLinker::Core(linker) => {
+                        linker
+                            .module_async(&mut *store, name, &preload_module)
+                            .await
+                            .context(format!(
+                                "failed to process preload `{}` at `{}`",
+                                name,
+                                path.display()
+                            ))?;
+                    }
+                    #[cfg(not(feature = "cranelift"))]
+                    CliLinker::Core(_) => {
+                        bail!("support for --preload disabled at compile time");
+                    }
+                    #[cfg(feature = "component-model")]
+                    CliLinker::Component(_) => {
+                        bail!("--preload cannot be used with components");
                     }
                 }
+            }
 
-                self.load_main_module(&mut store, &mut linker, &main, profiled_modules)
-                    .await
-                    .with_context(|| {
-                        format!(
-                            "failed to run main module `{}`",
-                            self.module_and_args[0].to_string_lossy()
-                        )
-                    })
-            })
-            .await
-        });
+            self.load_main_module(store, linker, &main, profiled_modules)
+                .await
+                .with_context(|| {
+                    format!(
+                        "failed to run main module `{}`",
+                        self.module_and_args[0].to_string_lossy()
+                    )
+                })
+        })
+        .await;
 
         // Load the main wasm module.
-        match result.unwrap_or_else(|elapsed| {
+        let instance = match result.unwrap_or_else(|elapsed| {
             Err(anyhow::Error::from(wasmtime::Trap::Interrupt))
                 .with_context(|| format!("timed out after {elapsed}"))
         }) {
-            Ok(()) => (),
+            Ok(instance) => instance,
             Err(e) => {
                 // Exit the process if Wasmtime understands the error;
                 // otherwise, fall back on Rust's default error printing/return
                 // code.
-                if store.data().preview1_ctx.is_some() {
+                if store.data().legacy_p1_ctx.is_some() {
                     return Err(wasi_common::maybe_exit_on_error(e));
-                } else if store.data().preview2_ctx.is_some() {
+                } else if store.data().wasip1_ctx.is_some() {
                     if let Some(exit) = e.downcast_ref::<wasmtime_wasi::I32Exit>() {
                         std::process::exit(exit.0);
                     }
@@ -265,9 +312,9 @@ impl RunCommand {
                 }
                 return Err(e);
             }
-        }
+        };
 
-        Ok(())
+        Ok(instance)
     }
 
     fn compute_argv(&self) -> Result<Vec<String>> {
@@ -395,9 +442,11 @@ impl RunCommand {
 
         store.set_epoch_deadline(1);
         let engine = store.engine().clone();
-        thread::spawn(move || loop {
-            thread::sleep(interval);
-            engine.increment_epoch();
+        thread::spawn(move || {
+            loop {
+                thread::sleep(interval);
+                engine.increment_epoch();
+            }
         });
 
         let path = path.to_string();
@@ -423,7 +472,7 @@ impl RunCommand {
         linker: &mut CliLinker,
         main_target: &RunTarget,
         profiled_modules: Vec<(String, Module)>,
-    ) -> Result<()> {
+    ) -> Result<CliInstance> {
         // The main module might be allowed to have unknown imports, which
         // should be defined as traps:
         if self.run.common.wasm.unknown_imports_trap == Some(true) {
@@ -443,7 +492,7 @@ impl RunCommand {
             match linker {
                 CliLinker::Core(linker) => {
                     linker.define_unknown_imports_as_default_values(
-                        store,
+                        &mut *store,
                         main_target.unwrap_core(),
                     )?;
                 }
@@ -479,7 +528,7 @@ impl RunCommand {
                     Some(
                         instance
                             .get_func(&mut *store, name)
-                            .ok_or_else(|| anyhow!("no func export named `{}` found", name))?,
+                            .ok_or_else(|| anyhow!("no func export named `{name}` found"))?,
                     )
                 } else {
                     instance
@@ -487,38 +536,23 @@ impl RunCommand {
                         .or_else(|| instance.get_func(&mut *store, "_start"))
                 };
 
-                match func {
-                    Some(func) => self.invoke_func(store, func).await,
-                    None => Ok(()),
+                if let Some(func) = func {
+                    self.invoke_func(store, func).await?;
                 }
+                Ok(CliInstance::Core(instance))
             }
             #[cfg(feature = "component-model")]
             CliLinker::Component(linker) => {
-                if self.invoke.is_some() {
-                    bail!("using `--invoke` with components is not supported");
-                }
-
                 let component = main_target.unwrap_component();
-
-                let command = wasmtime_wasi::bindings::Command::instantiate_async(
-                    &mut *store,
-                    component,
-                    linker,
-                )
-                .await?;
-                let result = command
-                    .wasi_cli_run()
-                    .call_run(&mut *store)
-                    .await
-                    .context("failed to invoke `run` function")
-                    .map_err(|e| self.handle_core_dump(&mut *store, e));
-
-                // Translate the `Result<(),()>` produced by wasm into a feigned
-                // explicit exit here with status 1 if `Err(())` is returned.
-                result.and_then(|wasm_result| match wasm_result {
-                    Ok(()) => Ok(()),
-                    Err(()) => Err(wasmtime_wasi::I32Exit(1).into()),
-                })
+                let result = if self.invoke.is_some() {
+                    self.invoke_component(&mut *store, component, linker).await
+                } else {
+                    self.run_command_component(&mut *store, component, linker)
+                        .await
+                };
+                result
+                    .map(CliInstance::Component)
+                    .map_err(|e| self.handle_core_dump(&mut *store, e))
             }
         };
         finish_epoch_handler(store);
@@ -526,6 +560,162 @@ impl RunCommand {
         result
     }
 
+    #[cfg(feature = "component-model")]
+    async fn invoke_component(
+        &self,
+        store: &mut Store<Host>,
+        component: &wasmtime::component::Component,
+        linker: &mut wasmtime::component::Linker<Host>,
+    ) -> Result<wasmtime::component::Instance> {
+        use wasmtime::component::{
+            Val,
+            wasm_wave::{
+                untyped::UntypedFuncCall,
+                wasm::{DisplayFuncResults, WasmFunc},
+            },
+        };
+
+        // Check if the invoke string is present
+        let invoke: &String = self.invoke.as_ref().unwrap();
+
+        let untyped_call = UntypedFuncCall::parse(invoke).with_context(|| {
+                format!(
+                    "Failed to parse invoke '{invoke}': See https://docs.wasmtime.dev/cli-options.html#run for syntax",
+                )
+        })?;
+
+        let name = untyped_call.name();
+        let matches =
+            Self::search_component_funcs(store.engine(), component.component_type(), name);
+        let (names, func_type) = match matches.len() {
+            0 => bail!("No exported func named `{name}` in component."),
+            1 => &matches[0],
+            _ => bail!(
+                "Multiple exports named `{name}`: {matches:?}. FIXME: support some way to disambiguate names"
+            ),
+        };
+
+        let param_types = WasmFunc::params(func_type).collect::<Vec<_>>();
+        let params = untyped_call
+            .to_wasm_params(&param_types)
+            .with_context(|| format!("while interpreting parameters in invoke \"{invoke}\""))?;
+
+        let export = names
+            .iter()
+            .fold(None, |instance, name| {
+                component.get_export_index(instance.as_ref(), name)
+            })
+            .expect("export has at least one name");
+
+        let instance = linker.instantiate_async(&mut *store, component).await?;
+
+        let func = instance
+            .get_func(&mut *store, export)
+            .expect("found export index");
+
+        let mut results = vec![Val::Bool(false); func_type.results().len()];
+        func.call_async(&mut *store, &params, &mut results).await?;
+        func.post_return_async(&mut *store).await?;
+
+        println!("{}", DisplayFuncResults(&results));
+
+        Ok(instance)
+    }
+
+    /// Execute the default behavior for components on the CLI, looking for
+    /// `wasi:cli`-based commands and running their exported `run` function.
+    #[cfg(feature = "component-model")]
+    async fn run_command_component(
+        &self,
+        store: &mut Store<Host>,
+        component: &wasmtime::component::Component,
+        linker: &wasmtime::component::Linker<Host>,
+    ) -> Result<wasmtime::component::Instance> {
+        let instance = linker.instantiate_async(&mut *store, component).await?;
+
+        let mut result = None;
+        let _ = &mut result;
+
+        // If WASIp3 is enabled at compile time, enabled at runtime, and found
+        // in this component then use that to generate the result.
+        #[cfg(feature = "component-model-async")]
+        if self.run.common.wasi.p3.unwrap_or(crate::common::P3_DEFAULT) {
+            if let Ok(command) = wasmtime_wasi::p3::bindings::Command::new(&mut *store, &instance) {
+                result = Some(
+                    store
+                        .run_concurrent(async |store| command.wasi_cli_run().call_run(store).await)
+                        .await?,
+                );
+            }
+        }
+
+        let result = match result {
+            Some(result) => result,
+            // If WASIp3 wasn't found then fall back to requiring WASIp2 and
+            // this'll report an error if the right export doesn't exist.
+            None => {
+                wasmtime_wasi::p2::bindings::Command::new(&mut *store, &instance)?
+                    .wasi_cli_run()
+                    .call_run(&mut *store)
+                    .await
+            }
+        };
+        let wasm_result = result.context("failed to invoke `run` function")?;
+
+        // Translate the `Result<(),()>` produced by wasm into a feigned
+        // explicit exit here with status 1 if `Err(())` is returned.
+        match wasm_result {
+            Ok(()) => Ok(instance),
+            Err(()) => Err(wasmtime_wasi::I32Exit(1).into()),
+        }
+    }
+
+    #[cfg(feature = "component-model")]
+    fn search_component_funcs(
+        engine: &Engine,
+        component: wasmtime::component::types::Component,
+        name: &str,
+    ) -> Vec<(Vec<String>, wasmtime::component::types::ComponentFunc)> {
+        use wasmtime::component::types::ComponentItem as CItem;
+        fn collect_exports(
+            engine: &Engine,
+            item: CItem,
+            basename: Vec<String>,
+        ) -> Vec<(Vec<String>, CItem)> {
+            match item {
+                CItem::Component(c) => c
+                    .exports(engine)
+                    .flat_map(move |(name, item)| {
+                        let mut names = basename.clone();
+                        names.push(name.to_string());
+                        collect_exports(engine, item, names)
+                    })
+                    .collect::<Vec<_>>(),
+                CItem::ComponentInstance(c) => c
+                    .exports(engine)
+                    .flat_map(move |(name, item)| {
+                        let mut names = basename.clone();
+                        names.push(name.to_string());
+                        collect_exports(engine, item, names)
+                    })
+                    .collect::<Vec<_>>(),
+                _ => vec![(basename, item)],
+            }
+        }
+
+        collect_exports(engine, CItem::Component(component), Vec::new())
+            .into_iter()
+            .filter_map(|(names, item)| {
+                let CItem::ComponentFunc(func) = item else {
+                    return None;
+                };
+                let func_name = names.last().expect("at least one name");
+                let base_func_name = func_name.strip_prefix("[async]").unwrap_or(func_name);
+                (base_func_name == name).then_some((names, func))
+            })
+            .collect()
+    }
+
     async fn invoke_func(&self, store: &mut Store<Host>, func: Func) -> Result<()> {
         let ty = func.ty(&store);
         if ty.params().len() > 0 {
@@ -541,7 +731,7 @@ impl RunCommand {
                 Some(s) => s,
                 None => {
                     if let Some(name) = &self.invoke {
-                        bail!("not enough arguments for `{}`", name)
+                        bail!("not enough arguments for `{name}`")
                     } else {
                         bail!("not enough arguments for command default")
                     }
@@ -564,7 +754,7 @@ impl RunCommand {
                 }),
                 ValType::F32 => Val::F32(val.parse::<f32>()?.to_bits()),
                 ValType::F64 => Val::F64(val.parse::<f64>()?.to_bits()),
-                t => bail!("unsupported argument type {:?}", t),
+                t => bail!("unsupported argument type {t:?}"),
             });
         }
 
@@ -606,6 +796,10 @@ impl RunCommand {
                 Val::FuncRef(Some(_)) => println!("<funcref>"),
                 Val::AnyRef(None) => println!("<null anyref>"),
                 Val::AnyRef(Some(_)) => println!("<anyref>"),
+                Val::ExnRef(None) => println!("<null exnref>"),
+                Val::ExnRef(Some(_)) => println!("<exnref>"),
+                Val::ContRef(None) => println!("<null contref>"),
+                Val::ContRef(Some(_)) => println!("<contref>"),
             }
         }
 
@@ -645,20 +839,8 @@ impl RunCommand {
         store: &mut Store<Host>,
         module: &RunTarget,
     ) -> Result<()> {
-        let mut cli = self.run.common.wasi.cli;
-
-        // Accept -Scommon as a deprecated alias for -Scli.
-        if let Some(common) = self.run.common.wasi.common {
-            if cli.is_some() {
-                bail!(
-                    "The -Scommon option should not be use with -Scli as it is a deprecated alias"
-                );
-            } else {
-                // In the future, we may add a warning here to tell users to use
-                // `-S cli` instead of `-S common`.
-                cli = Some(common);
-            }
-        }
+        self.run.validate_p3_option()?;
+        let cli = self.run.validate_cli_enabled()?;
 
         if cli != Some(false) {
             match linker {
@@ -669,34 +851,29 @@ impl RunCommand {
                         // implementation.
                         (Some(false), _) | (None, Some(true)) => {
                             wasi_common::tokio::add_to_linker(linker, |host| {
-                                host.preview1_ctx.as_mut().unwrap()
+                                host.legacy_p1_ctx.as_mut().unwrap()
                             })?;
-                            self.set_preview1_ctx(store)?;
+                            self.set_legacy_p1_ctx(store)?;
                         }
                         // If preview2 was explicitly requested, always use it.
                         // Otherwise use it so long as threads are disabled.
                         //
-                        // Note that for now `preview0` is currently
+                        // Note that for now `p0` is currently
                         // default-enabled but this may turn into
                         // default-disabled in the future.
                         (Some(true), _) | (None, Some(false) | None) => {
                             if self.run.common.wasi.preview0 != Some(false) {
-                                wasmtime_wasi::preview0::add_to_linker_async(linker, |t| {
-                                    t.preview2_ctx()
-                                })?;
+                                wasmtime_wasi::p0::add_to_linker_async(linker, |t| t.wasip1_ctx())?;
                             }
-                            wasmtime_wasi::preview1::add_to_linker_async(linker, |t| {
-                                t.preview2_ctx()
-                            })?;
-                            self.set_preview2_ctx(store)?;
+                            wasmtime_wasi::p1::add_to_linker_async(linker, |t| t.wasip1_ctx())?;
+                            self.set_wasi_ctx(store)?;
                         }
                     }
                 }
                 #[cfg(feature = "component-model")]
                 CliLinker::Component(linker) => {
-                    let link_options = self.run.compute_wasi_features();
-                    wasmtime_wasi::add_to_linker_with_options_async(linker, &link_options)?;
-                    self.set_preview2_ctx(store)?;
+                    self.run.add_wasmtime_wasi_to_linker(linker)?;
+                    self.set_wasi_ctx(store)?;
                 }
             }
         }
@@ -722,15 +899,14 @@ impl RunCommand {
                     #[cfg(feature = "component-model")]
                     CliLinker::Component(linker) => {
                         wasmtime_wasi_nn::wit::add_to_linker(linker, |h: &mut Host| {
-                            let preview2_ctx =
-                                h.preview2_ctx.as_mut().expect("wasip2 is not configured");
-                            let preview2_ctx = Arc::get_mut(preview2_ctx)
+                            let ctx = h.wasip1_ctx.as_mut().expect("wasi is not configured");
+                            let ctx = Arc::get_mut(ctx)
                                 .expect("wasmtime_wasi is not compatible with threads")
                                 .get_mut()
                                 .unwrap();
                             let nn_ctx = Arc::get_mut(h.wasi_nn_wit.as_mut().unwrap())
                                 .expect("wasi-nn is not implemented with multi-threading support");
-                            WasiNnView::new(preview2_ctx.table(), nn_ctx)
+                            WasiNnView::new(ctx.ctx().table, nn_ctx)
                         })?;
                         store.data_mut().wasi_nn_wit = Some(Arc::new(
                             wasmtime_wasi_nn::wit::WasiNnCtx::new(backends, registry),
@@ -775,7 +951,9 @@ impl RunCommand {
         if self.run.common.wasi.keyvalue == Some(true) {
             #[cfg(not(feature = "wasi-keyvalue"))]
             {
-                bail!("Cannot enable wasi-keyvalue when the binary is not compiled with this feature.");
+                bail!(
+                    "Cannot enable wasi-keyvalue when the binary is not compiled with this feature."
+                );
             }
             #[cfg(all(feature = "wasi-keyvalue", feature = "component-model"))]
             {
@@ -796,13 +974,11 @@ impl RunCommand {
                             .build();
 
                         wasmtime_wasi_keyvalue::add_to_linker(linker, |h| {
-                            let preview2_ctx =
-                                h.preview2_ctx.as_mut().expect("wasip2 is not configured");
-                            let preview2_ctx =
-                                Arc::get_mut(preview2_ctx).unwrap().get_mut().unwrap();
+                            let ctx = h.wasip1_ctx.as_mut().expect("wasip2 is not configured");
+                            let ctx = Arc::get_mut(ctx).unwrap().get_mut().unwrap();
                             WasiKeyValue::new(
                                 Arc::get_mut(h.wasi_keyvalue.as_mut().unwrap()).unwrap(),
-                                preview2_ctx.table(),
+                                ctx.ctx().table,
                             )
                         })?;
                         store.data_mut().wasi_keyvalue = Some(Arc::new(ctx));
@@ -852,6 +1028,10 @@ impl RunCommand {
                     }
                     CliLinker::Component(linker) => {
                         wasmtime_wasi_http::add_only_http_to_linker_sync(linker)?;
+                        #[cfg(feature = "component-model-async")]
+                        if self.run.common.wasi.p3.unwrap_or(crate::common::P3_DEFAULT) {
+                            wasmtime_wasi_http::p3::add_to_linker(linker)?;
+                        }
                     }
                 }
 
@@ -874,12 +1054,16 @@ impl RunCommand {
                         let mut opts = wasmtime_wasi_tls::LinkOptions::default();
                         opts.tls(true);
                         wasmtime_wasi_tls::add_to_linker(linker, &mut opts, |h| {
-                            let preview2_ctx =
-                                h.preview2_ctx.as_mut().expect("wasip2 is not configured");
-                            let preview2_ctx =
-                                Arc::get_mut(preview2_ctx).unwrap().get_mut().unwrap();
-                            WasiTlsCtx::new(preview2_ctx.table())
+                            let ctx = h.wasip1_ctx.as_mut().expect("wasi is not configured");
+                            let ctx = Arc::get_mut(ctx).unwrap().get_mut().unwrap();
+                            WasiTls::new(
+                                Arc::get_mut(h.wasi_tls.as_mut().unwrap()).unwrap(),
+                                ctx.ctx().table,
+                            )
                         })?;
+
+                        let ctx = wasmtime_wasi_tls::WasiTlsCtxBuilder::new().build();
+                        store.data_mut().wasi_tls = Some(Arc::new(ctx));
                     }
                 }
             }
@@ -888,7 +1072,7 @@ impl RunCommand {
         Ok(())
     }
 
-    fn set_preview1_ctx(&self, store: &mut Store<Host>) -> Result<()> {
+    fn set_legacy_p1_ctx(&self, store: &mut Store<Host>) -> Result<()> {
         let mut builder = WasiCtxBuilder::new();
         builder.inherit_stdio().args(&self.compute_argv()?)?;
 
@@ -931,16 +1115,23 @@ impl RunCommand {
             builder.preopened_dir(dir, guest)?;
         }
 
-        store.data_mut().preview1_ctx = Some(builder.build());
+        store.data_mut().legacy_p1_ctx = Some(builder.build());
         Ok(())
     }
 
-    fn set_preview2_ctx(&self, store: &mut Store<Host>) -> Result<()> {
+    /// Note the naming here is subtle, but this is effectively setting up a
+    /// `wasmtime_wasi::WasiCtx` structure.
+    ///
+    /// This is stored in `Host` as `WasiP1Ctx` which internally contains the
+    /// `WasiCtx` and `ResourceTable` used for WASI implementations. Exactly
+    /// which "p" for WASIpN is more a reference to
+    /// `wasmtime-wasi`-vs-`wasi-common` here more than anything else.
+    fn set_wasi_ctx(&self, store: &mut Store<Host>) -> Result<()> {
         let mut builder = wasmtime_wasi::WasiCtxBuilder::new();
         builder.inherit_stdio().args(&self.compute_argv()?);
         self.run.configure_wasip2(&mut builder)?;
         let ctx = builder.build_p1();
-        store.data_mut().preview2_ctx = Some(Arc::new(Mutex::new(ctx)));
+        store.data_mut().wasip1_ctx = Some(Arc::new(Mutex::new(ctx)));
         Ok(())
     }
 
@@ -960,14 +1151,35 @@ impl RunCommand {
     }
 }
 
+/// The `T` in `Store<T>` for what the CLI is running.
+///
+/// This structures has a number of contexts used for various WASI proposals.
+/// Note that all of them are optional meaning that they're `None` by default
+/// and enabled with various CLI flags (some CLI flags are on-by-default). Note
+/// additionally that this structure is `Clone` to implement the `wasi-threads`
+/// proposal. Many WASI proposals are not compatible with `wasi-threads` so to
+/// model this `Arc` and `Arc<Mutex<T>>` is used for many configurations. If a
+/// WASI proposal is inherently threadsafe it's protected with just an `Arc` to
+/// share its configuration across many threads.
+///
+/// If mutation is required then `Mutex` is used. Note though that the mutex is
+/// not actually locked as access always goes through `Arc::get_mut` which
+/// effectively asserts that there's only one thread. In short much of this is
+/// not compatible with `wasi-threads`.
 #[derive(Default, Clone)]
-struct Host {
-    preview1_ctx: Option<wasi_common::WasiCtx>,
+pub struct Host {
+    // Legacy wasip1 context using `wasi_common`, not set unless opted-in-to
+    // with the CLI.
+    legacy_p1_ctx: Option<wasi_common::WasiCtx>,
 
+    // Context for both WASIp1 and WASIp2 (and beyond) for the `wasmtime_wasi`
+    // crate. This has both `wasmtime_wasi::WasiCtx` as well as a
+    // `ResourceTable` internally to be used.
+    //
     // The Mutex is only needed to satisfy the Sync constraint but we never
     // actually perform any locking on it as we use Mutex::get_mut for every
     // access.
-    preview2_ctx: Option<Arc<Mutex<wasmtime_wasi::preview1::WasiP1Ctx>>>,
+    wasip1_ctx: Option<Arc<Mutex<wasmtime_wasi::p1::WasiP1Ctx>>>,
 
     #[cfg(feature = "wasi-nn")]
     wasi_nn_wit: Option<Arc<wasmtime_wasi_nn::wit::WasiNnCtx>>,
@@ -982,6 +1194,8 @@ struct Host {
     wasi_http_outgoing_body_buffer_chunks: Option<usize>,
     #[cfg(feature = "wasi-http")]
     wasi_http_outgoing_body_chunk_size: Option<usize>,
+    #[cfg(all(feature = "wasi-http", feature = "component-model-async"))]
+    p3_http: crate::common::DefaultP3Ctx,
     limits: StoreLimits,
     #[cfg(feature = "profiling")]
     guest_profiler: Option<Arc<wasmtime::GuestProfiler>>,
@@ -990,29 +1204,27 @@ struct Host {
     wasi_config: Option<Arc<WasiConfigVariables>>,
     #[cfg(feature = "wasi-keyvalue")]
     wasi_keyvalue: Option<Arc<WasiKeyValueCtx>>,
+    #[cfg(feature = "wasi-tls")]
+    wasi_tls: Option<Arc<WasiTlsCtx>>,
 }
 
 impl Host {
-    fn preview2_ctx(&mut self) -> &mut wasmtime_wasi::preview1::WasiP1Ctx {
-        let ctx = self
-            .preview2_ctx
-            .as_mut()
-            .expect("wasip2 is not configured");
-        Arc::get_mut(ctx)
-            .expect("wasmtime_wasi is not compatible with threads")
-            .get_mut()
-            .unwrap()
+    fn wasip1_ctx(&mut self) -> &mut wasmtime_wasi::p1::WasiP1Ctx {
+        unwrap_singlethread_context(&mut self.wasip1_ctx)
     }
 }
 
-impl IoView for Host {
-    fn table(&mut self) -> &mut wasmtime::component::ResourceTable {
-        self.preview2_ctx().table()
-    }
+fn unwrap_singlethread_context<T>(ctx: &mut Option<Arc<Mutex<T>>>) -> &mut T {
+    let ctx = ctx.as_mut().expect("context not configured");
+    Arc::get_mut(ctx)
+        .expect("context is not compatible with threads")
+        .get_mut()
+        .unwrap()
 }
+
 impl WasiView for Host {
-    fn ctx(&mut self) -> &mut wasmtime_wasi::WasiCtx {
-        self.preview2_ctx().ctx()
+    fn ctx(&mut self) -> WasiCtxView<'_> {
+        WasiView::ctx(self.wasip1_ctx())
     }
 }
 
@@ -1023,6 +1235,10 @@ impl wasmtime_wasi_http::types::WasiHttpView for Host {
         Arc::get_mut(ctx).expect("wasmtime_wasi is not compatible with threads")
     }
 
+    fn table(&mut self) -> &mut wasmtime::component::ResourceTable {
+        WasiView::ctx(self).table
+    }
+
     fn outgoing_body_buffer_chunks(&mut self) -> usize {
         self.wasi_http_outgoing_body_buffer_chunks
             .unwrap_or_else(|| DEFAULT_OUTGOING_BODY_BUFFER_CHUNKS)
@@ -1034,29 +1250,39 @@ impl wasmtime_wasi_http::types::WasiHttpView for Host {
     }
 }
 
-#[cfg(not(unix))]
-fn ctx_set_listenfd(num_fd: usize, _builder: &mut WasiCtxBuilder) -> Result<usize> {
-    Ok(num_fd)
-}
-
-#[cfg(unix)]
-fn ctx_set_listenfd(mut num_fd: usize, builder: &mut WasiCtxBuilder) -> Result<usize> {
-    use listenfd::ListenFd;
-
-    for env in ["LISTEN_FDS", "LISTEN_FDNAMES"] {
-        if let Ok(val) = std::env::var(env) {
-            builder.env(env, &val)?;
+#[cfg(all(feature = "wasi-http", feature = "component-model-async"))]
+impl wasmtime_wasi_http::p3::WasiHttpView for Host {
+    fn http(&mut self) -> wasmtime_wasi_http::p3::WasiHttpCtxView<'_> {
+        wasmtime_wasi_http::p3::WasiHttpCtxView {
+            table: WasiView::ctx(unwrap_singlethread_context(&mut self.wasip1_ctx)).table,
+            ctx: &mut self.p3_http,
         }
     }
+}
 
-    let mut listenfd = ListenFd::from_env();
+fn ctx_set_listenfd(mut num_fd: usize, builder: &mut WasiCtxBuilder) -> Result<usize> {
+    let _ = &mut num_fd;
+    let _ = &mut *builder;
 
-    for i in 0..listenfd.len() {
-        if let Some(stdlistener) = listenfd.take_tcp_listener(i)? {
-            let _ = stdlistener.set_nonblocking(true)?;
-            let listener = TcpListener::from_std(stdlistener);
-            builder.preopened_socket((3 + i) as _, listener)?;
-            num_fd = 3 + i;
+    #[cfg(all(unix, feature = "run"))]
+    {
+        use listenfd::ListenFd;
+
+        for env in ["LISTEN_FDS", "LISTEN_FDNAMES"] {
+            if let Ok(val) = std::env::var(env) {
+                builder.env(env, &val)?;
+            }
+        }
+
+        let mut listenfd = ListenFd::from_env();
+
+        for i in 0..listenfd.len() {
+            if let Some(stdlistener) = listenfd.take_tcp_listener(i)? {
+                let _ = stdlistener.set_nonblocking(true)?;
+                let listener = TcpListener::from_std(stdlistener);
+                builder.preopened_socket((3 + i) as _, listener)?;
+                num_fd = 3 + i;
+            }
         }
     }
 
