@@ -1,26 +1,37 @@
 use crate::common::{Profile, RunCommon, RunTarget};
-use anyhow::{anyhow, bail, Result};
+use anyhow::{Context as _, Result, bail};
+use bytes::Bytes;
 use clap::Parser;
+use futures::future::FutureExt;
+use http::{Response, StatusCode};
+use http_body_util::BodyExt as _;
+use http_body_util::combinators::UnsyncBoxBody;
+use std::convert::Infallible;
 use std::net::SocketAddr;
-use std::time::Instant;
+use std::pin::Pin;
+use std::task::{Context, Poll};
 use std::{
     path::PathBuf,
     sync::{
-        atomic::{AtomicBool, AtomicU64, Ordering},
         Arc, Mutex,
+        atomic::{AtomicBool, Ordering},
     },
     time::Duration,
 };
+use tokio::io::{self, AsyncWrite};
 use tokio::sync::Notify;
-use wasmtime::component::{Component, Linker};
-use wasmtime::{Engine, Store, StoreLimits, UpdateDeadline};
-use wasmtime_wasi::{IoView, StreamError, StreamResult, WasiCtx, WasiCtxBuilder, WasiView};
-use wasmtime_wasi_http::bindings::http::types::Scheme;
-use wasmtime_wasi_http::bindings::ProxyPre;
+use wasmtime::component::{Component, Linker, ResourceTable};
+use wasmtime::{Engine, Store, StoreContextMut, StoreLimits, UpdateDeadline};
+use wasmtime_cli_flags::opt::WasmtimeOptionValue;
+use wasmtime_wasi::p2::{StreamError, StreamResult};
+use wasmtime_wasi::{WasiCtx, WasiCtxBuilder, WasiCtxView, WasiView};
+#[cfg(feature = "component-model-async")]
+use wasmtime_wasi_http::handler::p2::bindings as p2;
+use wasmtime_wasi_http::handler::{HandlerState, Proxy, ProxyHandler, ProxyPre, StoreBundle};
 use wasmtime_wasi_http::io::TokioIo;
 use wasmtime_wasi_http::{
-    body::HyperOutgoingBody, WasiHttpCtx, WasiHttpView, DEFAULT_OUTGOING_BODY_BUFFER_CHUNKS,
-    DEFAULT_OUTGOING_BODY_CHUNK_SIZE,
+    DEFAULT_OUTGOING_BODY_BUFFER_CHUNKS, DEFAULT_OUTGOING_BODY_CHUNK_SIZE, WasiHttpCtx,
+    WasiHttpView,
 };
 
 #[cfg(feature = "wasi-config")]
@@ -30,6 +41,10 @@ use wasmtime_wasi_keyvalue::{WasiKeyValue, WasiKeyValueCtx, WasiKeyValueCtxBuild
 #[cfg(feature = "wasi-nn")]
 use wasmtime_wasi_nn::wit::WasiNnCtx;
 
+const DEFAULT_WASIP3_MAX_INSTANCE_REUSE_COUNT: usize = 128;
+const DEFAULT_WASIP2_MAX_INSTANCE_REUSE_COUNT: usize = 1;
+const DEFAULT_WASIP3_MAX_INSTANCE_CONCURRENT_REUSE_COUNT: usize = 16;
+
 struct Host {
     table: wasmtime::component::ResourceTable,
     ctx: WasiCtx,
@@ -37,6 +52,9 @@ struct Host {
     http_outgoing_body_buffer_chunks: Option<usize>,
     http_outgoing_body_chunk_size: Option<usize>,
 
+    #[cfg(feature = "component-model-async")]
+    p3_http: crate::common::DefaultP3Ctx,
+
     limits: StoreLimits,
 
     #[cfg(feature = "wasi-nn")]
@@ -52,14 +70,12 @@ struct Host {
     guest_profiler: Option<Arc<wasmtime::GuestProfiler>>,
 }
 
-impl IoView for Host {
-    fn table(&mut self) -> &mut wasmtime::component::ResourceTable {
-        &mut self.table
-    }
-}
 impl WasiView for Host {
-    fn ctx(&mut self) -> &mut WasiCtx {
-        &mut self.ctx
+    fn ctx(&mut self) -> WasiCtxView<'_> {
+        WasiCtxView {
+            ctx: &mut self.ctx,
+            table: &mut self.table,
+        }
     }
 }
 
@@ -67,6 +83,9 @@ impl WasiHttpView for Host {
     fn ctx(&mut self) -> &mut WasiHttpCtx {
         &mut self.http
     }
+    fn table(&mut self) -> &mut ResourceTable {
+        &mut self.table
+    }
 
     fn outgoing_body_buffer_chunks(&mut self) -> usize {
         self.http_outgoing_body_buffer_chunks
@@ -79,11 +98,25 @@ impl WasiHttpView for Host {
     }
 }
 
+#[cfg(feature = "component-model-async")]
+impl wasmtime_wasi_http::p3::WasiHttpView for Host {
+    fn http(&mut self) -> wasmtime_wasi_http::p3::WasiHttpCtxView<'_> {
+        wasmtime_wasi_http::p3::WasiHttpCtxView {
+            table: &mut self.table,
+            ctx: &mut self.p3_http,
+        }
+    }
+}
+
 const DEFAULT_ADDR: std::net::SocketAddr = std::net::SocketAddr::new(
     std::net::IpAddr::V4(std::net::Ipv4Addr::new(0, 0, 0, 0)),
     8080,
 );
 
+fn parse_duration(s: &str) -> Result<Duration, String> {
+    Duration::parse(Some(s)).map_err(|e| e.to_string())
+}
+
 /// Runs a WebAssembly module
 #[derive(Parser)]
 pub struct ServeCommand {
@@ -109,6 +142,31 @@ pub struct ServeCommand {
     /// The WebAssembly component to run.
     #[arg(value_name = "WASM", required = true)]
     component: PathBuf,
+
+    /// Maximum number of requests to send to a single component instance before
+    /// dropping it.
+    ///
+    /// This defaults to 1 for WASIp2 components and 128 for WASIp3 components.
+    #[arg(long)]
+    max_instance_reuse_count: Option<usize>,
+
+    /// Maximum number of concurrent requests to send to a single component
+    /// instance.
+    ///
+    /// This defaults to 1 for WASIp2 components and 16 for WASIp3 components.
+    /// Note that setting it to more than 1 will have no effect for WASIp2
+    /// components since they cannot be called concurrently.
+    #[arg(long)]
+    max_instance_concurrent_reuse_count: Option<usize>,
+
+    /// Time to hold an idle component instance for possible reuse before
+    /// dropping it.
+    ///
+    /// A number with no suffix or with an `s` suffix is interpreted as seconds;
+    /// other accepted suffixes include `ms` (milliseconds), `us` or `μs`
+    /// (microseconds), and `ns` (nanoseconds).
+    #[arg(long, default_value = "1s", value_parser = parse_duration)]
+    idle_instance_timeout: Duration,
 }
 
 impl ServeCommand {
@@ -149,20 +207,25 @@ impl ServeCommand {
         Ok(())
     }
 
-    fn new_store(&self, engine: &Engine, req_id: u64) -> Result<Store<Host>> {
+    fn new_store(&self, engine: &Engine, req_id: Option<u64>) -> Result<Store<Host>> {
         let mut builder = WasiCtxBuilder::new();
         self.run.configure_wasip2(&mut builder)?;
 
-        builder.env("REQUEST_ID", req_id.to_string());
+        if let Some(req_id) = req_id {
+            builder.env("REQUEST_ID", req_id.to_string());
+        }
 
         let stdout_prefix: String;
         let stderr_prefix: String;
-        if self.no_logging_prefix {
-            stdout_prefix = "".to_string();
-            stderr_prefix = "".to_string();
-        } else {
-            stdout_prefix = format!("stdout [{req_id}] :: ");
-            stderr_prefix = format!("stderr [{req_id}] :: ");
+        match req_id {
+            Some(req_id) if !self.no_logging_prefix => {
+                stdout_prefix = format!("stdout [{req_id}] :: ");
+                stderr_prefix = format!("stderr [{req_id}] :: ");
+            }
+            _ => {
+                stdout_prefix = "".to_string();
+                stderr_prefix = "".to_string();
+            }
         }
         builder.stdout(LogStream::new(stdout_prefix, Output::Stdout));
         builder.stderr(LogStream::new(stderr_prefix, Output::Stderr));
@@ -184,6 +247,8 @@ impl ServeCommand {
             wasi_keyvalue: None,
             #[cfg(feature = "profiling")]
             guest_profiler: None,
+            #[cfg(feature = "component-model-async")]
+            p3_http: crate::common::DefaultP3Ctx,
         };
 
         if self.run.common.wasi.nn == Some(true) {
@@ -249,20 +314,8 @@ impl ServeCommand {
     }
 
     fn add_to_linker(&self, linker: &mut Linker<Host>) -> Result<()> {
-        let mut cli = self.run.common.wasi.cli;
-
-        // Accept -Scommon as a deprecated alias for -Scli.
-        if let Some(common) = self.run.common.wasi.common {
-            if cli.is_some() {
-                bail!(
-                    "The -Scommon option should not be use with -Scli as it is a deprecated alias"
-                );
-            } else {
-                // In the future, we may add a warning here to tell users to use
-                // `-S cli` instead of `-S common`.
-                cli = Some(common);
-            }
-        }
+        self.run.validate_p3_option()?;
+        let cli = self.run.validate_cli_enabled()?;
 
         // Repurpose the `-Scli` flag of `wasmtime run` for `wasmtime serve`
         // to serve as a signal to enable all WASI interfaces instead of just
@@ -273,11 +326,21 @@ impl ServeCommand {
         // bindings which adds just those interfaces that the proxy interface
         // uses.
         if cli == Some(true) {
-            let link_options = self.run.compute_wasi_features();
-            wasmtime_wasi::add_to_linker_with_options_async(linker, &link_options)?;
+            self.run.add_wasmtime_wasi_to_linker(linker)?;
             wasmtime_wasi_http::add_only_http_to_linker_async(linker)?;
+            #[cfg(feature = "component-model-async")]
+            if self.run.common.wasi.p3.unwrap_or(crate::common::P3_DEFAULT) {
+                wasmtime_wasi_http::p3::add_to_linker(linker)?;
+            }
         } else {
             wasmtime_wasi_http::add_to_linker_async(linker)?;
+            #[cfg(feature = "component-model-async")]
+            if self.run.common.wasi.p3.unwrap_or(crate::common::P3_DEFAULT) {
+                wasmtime_wasi_http::p3::add_to_linker(linker)?;
+                wasmtime_wasi::p3::clocks::add_to_linker(linker)?;
+                wasmtime_wasi::p3::random::add_to_linker(linker)?;
+                wasmtime_wasi::p3::cli::add_to_linker(linker)?;
+            }
         }
 
         if self.run.common.wasi.nn == Some(true) {
@@ -366,7 +429,13 @@ impl ServeCommand {
         };
 
         let instance = linker.instantiate_pre(&component)?;
-        let instance = ProxyPre::new(instance)?;
+        #[cfg(feature = "component-model-async")]
+        let instance = match wasmtime_wasi_http::p3::bindings::ProxyPre::new(instance.clone()) {
+            Ok(pre) => ProxyPre::P3(pre),
+            Err(_) => ProxyPre::P2(p2::ProxyPre::new(instance)?),
+        };
+        #[cfg(not(feature = "component-model-async"))]
+        let instance = ProxyPre::P2(p2::ProxyPre::new(instance)?);
 
         // Spawn background task(s) waiting for graceful shutdown signals. This
         // always listens for ctrl-c but additionally can listen for a TCP
@@ -412,7 +481,40 @@ impl ServeCommand {
 
         log::info!("Listening on {}", self.addr);
 
-        let handler = ProxyHandler::new(self, engine, instance);
+        let epoch_interval = if let Some(Profile::Guest { interval, .. }) = self.run.profile {
+            Some(interval)
+        } else if let Some(t) = self.run.common.wasm.timeout {
+            Some(EPOCH_INTERRUPT_PERIOD.min(t))
+        } else {
+            None
+        };
+        let _epoch_thread = epoch_interval.map(|t| EpochThread::spawn(t, engine.clone()));
+
+        let max_instance_reuse_count = self.max_instance_reuse_count.unwrap_or_else(|| {
+            if let ProxyPre::P3(_) = &instance {
+                DEFAULT_WASIP3_MAX_INSTANCE_REUSE_COUNT
+            } else {
+                DEFAULT_WASIP2_MAX_INSTANCE_REUSE_COUNT
+            }
+        });
+
+        let max_instance_concurrent_reuse_count = if let ProxyPre::P3(_) = &instance {
+            self.max_instance_concurrent_reuse_count
+                .unwrap_or(DEFAULT_WASIP3_MAX_INSTANCE_CONCURRENT_REUSE_COUNT)
+        } else {
+            1
+        };
+
+        let handler = ProxyHandler::new(
+            HostHandlerState {
+                cmd: self,
+                engine,
+                component,
+                max_instance_reuse_count,
+                max_instance_concurrent_reuse_count,
+            },
+            instance,
+        );
 
         loop {
             // Wait for a socket, but also "race" against shutdown to break out
@@ -422,7 +524,14 @@ impl ServeCommand {
                 _ = shutdown.requested.notified() => break,
                 v = listener.accept() => v?,
             };
-            let comp = component.clone();
+
+            // The Nagle algorithm can impose a significant latency penalty
+            // (e.g. 40ms on Linux) on guests which write small, intermittent
+            // response body chunks (e.g. SSE streams).  Here we disable that
+            // algorithm and rely on the guest to buffer if appropriate to avoid
+            // TCP fragmentation.
+            stream.set_nodelay(true)?;
+
             let stream = TokioIo::new(stream);
             let h = handler.clone();
             let shutdown_guard = shutdown.clone().increment();
@@ -432,7 +541,39 @@ impl ServeCommand {
                     .serve_connection(
                         stream,
                         hyper::service::service_fn(move |req| {
-                            handle_request(h.clone(), req, comp.clone())
+                            let h = h.clone();
+                            async move {
+                                use http_body_util::{BodyExt, Full};
+                                match handle_request(h, req).await {
+                                    Ok(r) => Ok::<_, Infallible>(r),
+                                    Err(e) => {
+                                        eprintln!("error: {e:?}");
+                                        let error_html = "\
+<!doctype html>
+<html>
+<head>
+    <title>500 Internal Server Error</title>
+</head>
+<body>
+    <center>
+        <h1>500 Internal Server Error</h1>
+        <hr>
+        wasmtime
+    </center>
+</body>
+</html>";
+                                        Ok(Response::builder()
+                                            .status(StatusCode::INTERNAL_SERVER_ERROR)
+                                            .header("Content-Type", "text/html; charset=UTF-8")
+                                            .body(
+                                                Full::new(bytes::Bytes::from(error_html))
+                                                    .map_err(|_| unreachable!())
+                                                    .boxed_unsync(),
+                                            )
+                                            .unwrap())
+                                    }
+                                }
+                            }
                         }),
                     )
                     .await
@@ -461,6 +602,48 @@ impl ServeCommand {
     }
 }
 
+struct HostHandlerState {
+    cmd: ServeCommand,
+    engine: Engine,
+    component: Component,
+    max_instance_reuse_count: usize,
+    max_instance_concurrent_reuse_count: usize,
+}
+
+impl HandlerState for HostHandlerState {
+    type StoreData = Host;
+
+    fn new_store(&self, req_id: Option<u64>) -> Result<StoreBundle<Host>> {
+        let mut store = self.cmd.new_store(&self.engine, req_id)?;
+        let write_profile = setup_epoch_handler(&self.cmd, &mut store, self.component.clone())?;
+
+        Ok(StoreBundle {
+            store,
+            write_profile,
+        })
+    }
+
+    fn request_timeout(&self) -> Duration {
+        self.cmd.run.common.wasm.timeout.unwrap_or(Duration::MAX)
+    }
+
+    fn idle_instance_timeout(&self) -> Duration {
+        self.cmd.idle_instance_timeout
+    }
+
+    fn max_instance_reuse_count(&self) -> usize {
+        self.max_instance_reuse_count
+    }
+
+    fn max_instance_concurrent_reuse_count(&self) -> usize {
+        self.max_instance_concurrent_reuse_count
+    }
+
+    fn handle_worker_error(&self, error: anyhow::Error) {
+        eprintln!("worker error: {error}");
+    }
+}
+
 /// Helper structure to manage graceful shutdown int he accept loop above.
 #[derive(Default)]
 struct GracefulShutdown {
@@ -548,17 +731,17 @@ impl Drop for EpochThread {
     }
 }
 
-type WriteProfile = Box<dyn FnOnce(&mut Store<Host>) + Send>;
+type WriteProfile = Box<dyn FnOnce(StoreContextMut<Host>) + Send>;
 
 fn setup_epoch_handler(
     cmd: &ServeCommand,
     store: &mut Store<Host>,
     component: Component,
-) -> Result<(WriteProfile, Option<EpochThread>)> {
+) -> Result<WriteProfile> {
     // Profiling Enabled
     if let Some(Profile::Guest { interval, path }) = &cmd.run.profile {
         #[cfg(feature = "profiling")]
-        return setup_guest_profiler(cmd, store, path.clone(), *interval, component.clone());
+        return setup_guest_profiler(store, path.clone(), *interval, component.clone());
         #[cfg(not(feature = "profiling"))]
         {
             let _ = (path, interval);
@@ -567,32 +750,20 @@ fn setup_epoch_handler(
     }
 
     // Profiling disabled but there's a global request timeout
-    let epoch_thread = if let Some(timeout) = cmd.run.common.wasm.timeout {
-        let start = Instant::now();
-        store.epoch_deadline_callback(move |_store| {
-            if start.elapsed() > timeout {
-                bail!("Timeout expired");
-            }
-            Ok(UpdateDeadline::Continue(1))
-        });
-        store.set_epoch_deadline(1);
-        let engine = store.engine().clone();
-        Some(EpochThread::spawn(EPOCH_INTERRUPT_PERIOD, engine))
-    } else {
-        None
-    };
+    if cmd.run.common.wasm.timeout.is_some() {
+        store.epoch_deadline_async_yield_and_update(1);
+    }
 
-    Ok((Box::new(|_store| {}), epoch_thread))
+    Ok(Box::new(|_store| {}))
 }
 
 #[cfg(feature = "profiling")]
 fn setup_guest_profiler(
-    cmd: &ServeCommand,
     store: &mut Store<Host>,
     path: String,
     interval: Duration,
     component: Component,
-) -> Result<(WriteProfile, Option<EpochThread>)> {
+) -> Result<WriteProfile> {
     use wasmtime::{AsContext, GuestProfiler, StoreContext, StoreContextMut};
 
     let module_name = "<main>";
@@ -622,30 +793,17 @@ fn setup_guest_profiler(
         Ok(())
     });
 
-    let start = Instant::now();
-    let timeout = cmd.run.common.wasm.timeout;
     store.epoch_deadline_callback(move |store| {
         sample(store, |profiler, store| {
             profiler.sample(store, std::time::Duration::ZERO)
         });
 
-        // Originally epoch counting was used here; this is problematic in
-        // a lot of cases due to there being a lot of time (e.g. in hostcalls)
-        // when we are not expected to get sample hits.
-        if let Some(timeout) = timeout {
-            if start.elapsed() > timeout {
-                bail!("Timeout expired");
-            }
-        }
-
         Ok(UpdateDeadline::Continue(1))
     });
 
     store.set_epoch_deadline(1);
-    let engine = store.engine().clone();
-    let epoch_thread = Some(EpochThread::spawn(interval, engine));
 
-    let write_profile = Box::new(move |store: &mut Store<Host>| {
+    let write_profile = Box::new(move |mut store: StoreContextMut<Host>| {
         let profiler = Arc::try_unwrap(store.data_mut().guest_profiler.take().unwrap())
             .expect("profiling doesn't support threads yet");
         if let Err(e) = std::fs::File::create(&path)
@@ -660,46 +818,18 @@ fn setup_guest_profiler(
         }
     });
 
-    Ok((write_profile, epoch_thread))
-}
-
-struct ProxyHandlerInner {
-    cmd: ServeCommand,
-    engine: Engine,
-    instance_pre: ProxyPre<Host>,
-    next_id: AtomicU64,
-}
-
-impl ProxyHandlerInner {
-    fn next_req_id(&self) -> u64 {
-        self.next_id.fetch_add(1, Ordering::Relaxed)
-    }
-}
-
-#[derive(Clone)]
-struct ProxyHandler(Arc<ProxyHandlerInner>);
-
-impl ProxyHandler {
-    fn new(cmd: ServeCommand, engine: Engine, instance_pre: ProxyPre<Host>) -> Self {
-        Self(Arc::new(ProxyHandlerInner {
-            cmd,
-            engine,
-            instance_pre,
-            next_id: AtomicU64::from(0),
-        }))
-    }
+    Ok(write_profile)
 }
 
 type Request = hyper::Request<hyper::body::Incoming>;
 
 async fn handle_request(
-    ProxyHandler(inner): ProxyHandler,
+    handler: ProxyHandler<HostHandlerState>,
     req: Request,
-    component: Component,
-) -> Result<hyper::Response<HyperOutgoingBody>> {
-    let (sender, receiver) = tokio::sync::oneshot::channel();
+) -> Result<hyper::Response<UnsyncBoxBody<Bytes, anyhow::Error>>> {
+    use tokio::sync::oneshot;
 
-    let req_id = inner.next_req_id();
+    let req_id = handler.next_req_id();
 
     log::info!(
         "Request {req_id} handling {} to {}",
@@ -707,54 +837,102 @@ async fn handle_request(
         req.uri()
     );
 
-    let mut store = inner.cmd.new_store(&inner.engine, req_id)?;
+    // Here we must declare different channel types for p2 and p3 since p2's
+    // `WasiHttpView::new_response_outparam` expects a specific kind of sender
+    // that uses `p2::http::types::ErrorCode`, and we don't want to have to
+    // convert from the p3 `ErrorCode` to the p2 one, only to convert again to
+    // `anyhow::Error`.
 
-    let req = store.data_mut().new_incoming_request(Scheme::Http, req)?;
-    let out = store.data_mut().new_response_outparam(sender)?;
-    let proxy = inner.instance_pre.instantiate_async(&mut store).await?;
+    type P2Response = Result<
+        hyper::Response<wasmtime_wasi_http::body::HyperOutgoingBody>,
+        p2::http::types::ErrorCode,
+    >;
+    type P3Response = hyper::Response<UnsyncBoxBody<Bytes, anyhow::Error>>;
 
-    let comp = component.clone();
-    let task = tokio::task::spawn(async move {
-        let (write_profile, epoch_thread) = setup_epoch_handler(&inner.cmd, &mut store, comp)?;
+    enum Sender {
+        P2(oneshot::Sender<P2Response>),
+        P3(oneshot::Sender<P3Response>),
+    }
 
-        if let Err(e) = proxy
-            .wasi_http_incoming_handler()
-            .call_handle(&mut store, req, out)
-            .await
-        {
-            log::error!("[{req_id}] :: {:?}", e);
-            return Err(e);
+    enum Receiver {
+        P2(oneshot::Receiver<P2Response>),
+        P3(oneshot::Receiver<P3Response>),
+    }
+
+    let (tx, rx) = match handler.instance_pre() {
+        ProxyPre::P2(_) => {
+            let (tx, rx) = oneshot::channel();
+            (Sender::P2(tx), Receiver::P2(rx))
         }
-
-        write_profile(&mut store);
-        drop(epoch_thread);
-
-        Ok(())
-    });
-
-    let result = match receiver.await {
-        Ok(Ok(resp)) => Ok(resp),
-        Ok(Err(e)) => Err(e.into()),
-        Err(_) => {
-            // An error in the receiver (`RecvError`) only indicates that the
-            // task exited before a response was sent (i.e., the sender was
-            // dropped); it does not describe the underlying cause of failure.
-            // Instead we retrieve and propagate the error from inside the task
-            // which should more clearly tell the user what went wrong. Note
-            // that we assume the task has already exited at this point so the
-            // `await` should resolve immediately.
-            let e = match task.await {
-                Ok(Ok(())) => {
-                    bail!("guest never invoked `response-outparam::set` method")
-                }
-                Ok(Err(e)) => e,
-                Err(e) => e.into(),
-            };
-            Err(e.context("guest never invoked `response-outparam::set` method"))
+        ProxyPre::P3(_) => {
+            let (tx, rx) = oneshot::channel();
+            (Sender::P3(tx), Receiver::P3(rx))
         }
     };
 
-    result
+    handler.spawn(
+        if handler.state().max_instance_reuse_count() == 1 {
+            Some(req_id)
+        } else {
+            None
+        },
+        Box::new(move |store, proxy| {
+            Box::pin(
+                async move {
+                    match proxy {
+                        Proxy::P2(proxy) => {
+                            let Sender::P2(tx) = tx else { unreachable!() };
+                            let (req, out) = store.with(move |mut store| {
+                                let req = store
+                                    .data_mut()
+                                    .new_incoming_request(p2::http::types::Scheme::Http, req)?;
+                                let out = store.data_mut().new_response_outparam(tx)?;
+                                anyhow::Ok((req, out))
+                            })?;
+
+                            proxy
+                                .wasi_http_incoming_handler()
+                                .call_handle(store, req, out)
+                                .await
+                        }
+                        Proxy::P3(proxy) => {
+                            use wasmtime_wasi_http::p3::bindings::http::types::{
+                                ErrorCode, Request,
+                            };
+
+                            let Sender::P3(tx) = tx else { unreachable!() };
+                            let (req, body) = req.into_parts();
+                            let body = body.map_err(ErrorCode::from_hyper_request_error);
+                            let req = http::Request::from_parts(req, body);
+                            let (request, request_io_result) = Request::from_http(req);
+                            let (res, task) = proxy.handle(store, request).await??;
+                            let res = store
+                                .with(|mut store| res.into_http(&mut store, request_io_result))?;
+                            _ = tx.send(res.map(|body| body.map_err(|e| e.into()).boxed_unsync()));
+
+                            // Wait for the task to finish.
+                            task.block(store).await;
+                            Ok(())
+                        }
+                    }
+                }
+                .map(move |result| {
+                    if let Err(error) = result {
+                        eprintln!("[{req_id}] :: {error:?}");
+                    }
+                }),
+            )
+        }),
+    );
+
+    Ok(match rx {
+        Receiver::P2(rx) => rx
+            .await
+            .context("guest never invoked `response-outparam::set` method")?
+            .map_err(|e| anyhow::Error::from(e))?
+            .map(|body| body.map_err(|e| e.into()).boxed_unsync()),
+        Receiver::P3(rx) => rx.await?,
+    })
 }
 
 #[derive(Clone)]
@@ -764,42 +942,81 @@ enum Output {
 }
 
 impl Output {
-    fn write_all(&self, buf: &[u8]) -> anyhow::Result<()> {
+    fn write_all(&self, buf: &[u8]) -> io::Result<()> {
         use std::io::Write;
 
         match self {
             Output::Stdout => std::io::stdout().write_all(buf),
             Output::Stderr => std::io::stderr().write_all(buf),
         }
-        .map_err(|e| anyhow!(e))
     }
 }
 
 #[derive(Clone)]
 struct LogStream {
-    prefix: String,
     output: Output,
-    needs_prefix_on_next_write: bool,
+    state: Arc<LogStreamState>,
+}
+
+struct LogStreamState {
+    prefix: String,
+    needs_prefix_on_next_write: AtomicBool,
 }
 
 impl LogStream {
     fn new(prefix: String, output: Output) -> LogStream {
         LogStream {
-            prefix,
             output,
-            needs_prefix_on_next_write: true,
+            state: Arc::new(LogStreamState {
+                prefix,
+                needs_prefix_on_next_write: AtomicBool::new(true),
+            }),
         }
     }
+
+    fn write_all(&mut self, mut bytes: &[u8]) -> io::Result<()> {
+        while !bytes.is_empty() {
+            if self
+                .state
+                .needs_prefix_on_next_write
+                .load(Ordering::Relaxed)
+            {
+                self.output.write_all(self.state.prefix.as_bytes())?;
+                self.state
+                    .needs_prefix_on_next_write
+                    .store(false, Ordering::Relaxed);
+            }
+            match bytes.iter().position(|b| *b == b'\n') {
+                Some(i) => {
+                    let (a, b) = bytes.split_at(i + 1);
+                    bytes = b;
+                    self.output.write_all(a)?;
+                    self.state
+                        .needs_prefix_on_next_write
+                        .store(true, Ordering::Relaxed);
+                }
+                None => {
+                    self.output.write_all(bytes)?;
+                    break;
+                }
+            }
+        }
+
+        Ok(())
+    }
 }
 
-impl wasmtime_wasi::StdoutStream for LogStream {
-    fn stream(&self) -> Box<dyn wasmtime_wasi::OutputStream> {
+impl wasmtime_wasi::cli::StdoutStream for LogStream {
+    fn p2_stream(&self) -> Box<dyn wasmtime_wasi::p2::OutputStream> {
         Box::new(self.clone())
     }
+    fn async_stream(&self) -> Box<dyn AsyncWrite + Send + Sync> {
+        Box::new(self.clone())
+    }
+}
 
-    fn isatty(&self) -> bool {
-        use std::io::IsTerminal;
-
+impl wasmtime_wasi::cli::IsTerminal for LogStream {
+    fn is_terminal(&self) -> bool {
         match &self.output {
             Output::Stdout => std::io::stdout().is_terminal(),
             Output::Stderr => std::io::stderr().is_terminal(),
@@ -807,35 +1024,10 @@ impl wasmtime_wasi::StdoutStream for LogStream {
     }
 }
 
-impl wasmtime_wasi::OutputStream for LogStream {
+impl wasmtime_wasi::p2::OutputStream for LogStream {
     fn write(&mut self, bytes: bytes::Bytes) -> StreamResult<()> {
-        let mut bytes = &bytes[..];
-
-        while !bytes.is_empty() {
-            if self.needs_prefix_on_next_write {
-                self.output
-                    .write_all(self.prefix.as_bytes())
-                    .map_err(StreamError::LastOperationFailed)?;
-                self.needs_prefix_on_next_write = false;
-            }
-            match bytes.iter().position(|b| *b == b'\n') {
-                Some(i) => {
-                    let (a, b) = bytes.split_at(i + 1);
-                    bytes = b;
-                    self.output
-                        .write_all(a)
-                        .map_err(StreamError::LastOperationFailed)?;
-                    self.needs_prefix_on_next_write = true;
-                }
-                None => {
-                    self.output
-                        .write_all(bytes)
-                        .map_err(StreamError::LastOperationFailed)?;
-                    break;
-                }
-            }
-        }
-
+        self.write_all(&bytes)
+            .map_err(|e| StreamError::LastOperationFailed(e.into()))?;
         Ok(())
     }
 
@@ -849,10 +1041,26 @@ impl wasmtime_wasi::OutputStream for LogStream {
 }
 
 #[async_trait::async_trait]
-impl wasmtime_wasi::Pollable for LogStream {
+impl wasmtime_wasi::p2::Pollable for LogStream {
     async fn ready(&mut self) {}
 }
 
+impl AsyncWrite for LogStream {
+    fn poll_write(
+        mut self: Pin<&mut Self>,
+        _cx: &mut Context<'_>,
+        buf: &[u8],
+    ) -> Poll<io::Result<usize>> {
+        Poll::Ready(self.write_all(buf).map(|_| buf.len()))
+    }
+    fn poll_flush(self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<io::Result<()>> {
+        Poll::Ready(Ok(()))
+    }
+    fn poll_shutdown(self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<io::Result<()>> {
+        Poll::Ready(Ok(()))
+    }
+}
+
 /// The pooling allocator is tailor made for the `wasmtime serve` use case, so
 /// try to use it when we can. The main cost of the pooling allocator, however,
 /// is the virtual memory required to run it. Not all systems support the same
