@@ -1,11 +1,10 @@
 #![cfg(not(miri))]
-#![allow(dead_code)]
 
 use super::engine;
 use anyhow::Result;
 use wasmtime::{
+    Config, Engine, Store,
     component::{Component, Linker},
-    Store,
 };
 
 mod ownership;
@@ -13,6 +12,7 @@ mod results;
 
 mod no_imports {
     use super::*;
+    use std::rc::Rc;
 
     wasmtime::component::bindgen!({
         inline: "
@@ -54,12 +54,90 @@ mod no_imports {
         let no_imports = NoImports::instantiate(&mut store, &component, &linker)?;
         no_imports.call_bar(&mut store)?;
         no_imports.foo().call_foo(&mut store)?;
+
+        let linker = Linker::new(&engine);
+        let mut non_send_store = Store::new(&engine, Rc::new(()));
+        let no_imports = NoImports::instantiate(&mut non_send_store, &component, &linker)?;
+        no_imports.call_bar(&mut non_send_store)?;
+        no_imports.foo().call_foo(&mut non_send_store)?;
         Ok(())
     }
 }
 
+mod no_imports_concurrent {
+    use super::*;
+    use futures::{
+        FutureExt,
+        stream::{FuturesUnordered, TryStreamExt},
+    };
+
+    wasmtime::component::bindgen!({
+        inline: "
+            package foo:foo;
+
+            world no-imports {
+                export foo: interface {
+                    foo: async func();
+                }
+
+                export bar: async func();
+            }
+        ",
+    });
+
+    #[tokio::test]
+    async fn run() -> Result<()> {
+        let mut config = Config::new();
+        config.wasm_component_model_async(true);
+        config.async_support(true);
+        let engine = &Engine::new(&config)?;
+
+        let component = Component::new(
+            &engine,
+            r#"
+                (component
+                    (core module $m
+                        (import "" "task.return" (func $task-return))
+                        (func (export "bar") (result i32)
+                            call $task-return
+                            i32.const 0
+                        )
+                        (func (export "callback") (param i32 i32 i32) (result i32) unreachable)
+                    )
+                    (core func $task-return (canon task.return))
+                    (core instance $i (instantiate $m
+                        (with "" (instance (export "task.return" (func $task-return))))
+                    ))
+
+                    (func $f (export "[async]bar")
+                        (canon lift (core func $i "bar") async (callback (func $i "callback")))
+                    )
+
+                    (instance $i (export "[async]foo" (func $f)))
+                    (export "foo" (instance $i))
+                )
+            "#,
+        )?;
+
+        let linker = Linker::new(&engine);
+        let mut store = Store::new(&engine, ());
+        let no_imports = NoImports::instantiate_async(&mut store, &component, &linker).await?;
+        store
+            .run_concurrent(async move |accessor| {
+                let mut futures = FuturesUnordered::new();
+                futures.push(no_imports.call_bar(accessor).boxed());
+                futures.push(no_imports.foo().call_foo(accessor).boxed());
+                assert!(futures.try_next().await?.is_some());
+                assert!(futures.try_next().await?.is_some());
+                Ok(())
+            })
+            .await?
+    }
+}
+
 mod one_import {
     use super::*;
+    use wasmtime::component::HasSelf;
 
     wasmtime::component::bindgen!({
         inline: "
@@ -112,7 +190,7 @@ mod one_import {
         }
 
         let mut linker = Linker::new(&engine);
-        foo::add_to_linker(&mut linker, |f: &mut MyImports| f)?;
+        foo::add_to_linker::<_, HasSelf<_>>(&mut linker, |f| f)?;
         let mut store = Store::new(&engine, MyImports::default());
         let one_import = OneImport::instantiate(&mut store, &component, &linker)?;
         one_import.call_bar(&mut store)?;
@@ -121,9 +199,104 @@ mod one_import {
     }
 }
 
+mod one_import_concurrent {
+    use super::*;
+    use wasmtime::component::{Accessor, HasData};
+
+    wasmtime::component::bindgen!({
+        inline: "
+            package foo:foo;
+
+            world no-imports {
+                import foo: interface {
+                    foo: async func();
+                }
+
+                export bar: async func();
+            }
+        ",
+    });
+
+    #[tokio::test]
+    async fn run() -> Result<()> {
+        let mut config = Config::new();
+        config.wasm_component_model_async(true);
+        config.async_support(true);
+        let engine = &Engine::new(&config)?;
+
+        let component = Component::new(
+            &engine,
+            r#"
+                (component
+                    (import "foo" (instance $foo-instance
+                        (export "[async]foo" (func))
+                    ))
+                    (core module $libc
+                        (memory (export "memory") 1)
+                    )
+                    (core instance $libc-instance (instantiate $libc))
+                    (core module $m
+                        (import "" "foo" (func $foo (param) (result i32)))
+                        (import "" "task.return" (func $task-return))
+                        (func (export "bar") (result i32)
+                            call $foo
+                            drop
+                            call $task-return
+                            i32.const 0
+                        )
+                        (func (export "callback") (param i32 i32 i32) (result i32) unreachable)
+                    )
+                    (core func $foo (canon lower (func $foo-instance "[async]foo") async (memory $libc-instance "memory")))
+                    (core func $task-return (canon task.return))
+                    (core instance $i (instantiate $m
+                        (with "" (instance
+                            (export "task.return" (func $task-return))
+                            (export "foo" (func $foo))
+                        ))
+                    ))
+
+                    (func $f (export "[async]bar")
+                        (canon lift (core func $i "bar") async (callback (func $i "callback")))
+                    )
+
+                    (instance $i (export "[async]foo" (func $f)))
+                    (export "foo" (instance $i))
+                )
+            "#,
+        )?;
+
+        #[derive(Default)]
+        struct MyImports {
+            hit: bool,
+        }
+
+        impl HasData for MyImports {
+            type Data<'a> = &'a mut MyImports;
+        }
+
+        impl foo::HostWithStore for MyImports {
+            async fn foo<T>(accessor: &Accessor<T, Self>) {
+                accessor.with(|mut view| view.get().hit = true);
+            }
+        }
+
+        impl foo::Host for MyImports {}
+
+        let mut linker = Linker::new(&engine);
+        foo::add_to_linker::<_, MyImports>(&mut linker, |x| x)?;
+        let mut store = Store::new(&engine, MyImports::default());
+        let no_imports = NoImports::instantiate_async(&mut store, &component, &linker).await?;
+        store
+            .run_concurrent(async move |accessor| no_imports.call_bar(accessor).await)
+            .await??;
+        assert!(store.data().hit);
+        Ok(())
+    }
+}
+
 mod resources_at_world_level {
     use super::*;
-    use wasmtime::component::Resource;
+    use wasmtime::component::{HasSelf, Resource};
 
     wasmtime::component::bindgen!({
         inline: "
@@ -200,7 +373,7 @@ mod resources_at_world_level {
         impl ResourcesImports for MyImports {}
 
         let mut linker = Linker::new(&engine);
-        Resources::add_to_linker(&mut linker, |f: &mut MyImports| f)?;
+        Resources::add_to_linker::<_, HasSelf<_>>(&mut linker, |f| f)?;
         let mut store = Store::new(&engine, MyImports::default());
         let one_import = Resources::instantiate(&mut store, &component, &linker)?;
         one_import.call_y(&mut store, Resource::new_own(40))?;
@@ -212,7 +385,7 @@ mod resources_at_world_level {
 
 mod resources_at_interface_level {
     use super::*;
-    use wasmtime::component::Resource;
+    use wasmtime::component::{HasSelf, Resource};
 
     wasmtime::component::bindgen!({
         inline: "
@@ -245,7 +418,7 @@ mod resources_at_interface_level {
             r#"
                 (component
                     (import (interface "foo:foo/def") (instance $i
-                        (export $x "x" (type (sub resource)))
+                        (export "x" (type $x (sub resource)))
                         (export "[constructor]x" (func (result (own $x))))
                     ))
                     (alias export $i "x" (type $x))
@@ -305,7 +478,7 @@ mod resources_at_interface_level {
         impl foo::foo::def::Host for MyImports {}
 
         let mut linker = Linker::new(&engine);
-        Resources::add_to_linker(&mut linker, |f: &mut MyImports| f)?;
+        Resources::add_to_linker::<_, HasSelf<_>>(&mut linker, |f| f)?;
         let mut store = Store::new(&engine, MyImports::default());
         let one_import = Resources::instantiate(&mut store, &component, &linker)?;
         one_import
@@ -333,9 +506,11 @@ mod async_config {
                 export z: func();
             }
         ",
-        async: true,
+        imports: { default: async },
+        exports: { default: async },
     });
 
+    #[expect(dead_code, reason = "just here for bindings")]
     struct T;
 
     impl T1Imports for T {
@@ -358,9 +533,11 @@ mod async_config {
                 export z: func();
             }
         ",
-        async: {
-            except_imports: ["x"],
+        imports: {
+            "x": tracing,
+            default: async,
         },
+        exports: { default: async },
     });
 
     impl T2Imports for T {
@@ -383,9 +560,8 @@ mod async_config {
                 export z: func();
             }
         ",
-        async: {
-            only_imports: ["x"],
-        },
+        imports: { "x": async },
+        exports: { default: async },
     });
 
     impl T3Imports for T {
@@ -402,7 +578,7 @@ mod async_config {
 mod exported_resources {
     use super::*;
     use std::mem;
-    use wasmtime::component::Resource;
+    use wasmtime::component::{HasSelf, Resource};
 
     wasmtime::component::bindgen!({
         inline: "
@@ -481,7 +657,7 @@ mod exported_resources {
 (component
   ;; setup the `foo:foo/a` import
   (import (interface "foo:foo/a") (instance $a
-    (export $x "x" (type (sub resource)))
+    (export "x" (type $x (sub resource)))
     (export "[constructor]x" (func (result (own $x))))
   ))
   (alias export $a "x" (type $a-x))
@@ -591,7 +767,7 @@ mod exported_resources {
         )?;
 
         let mut linker = Linker::new(&engine);
-        Resources::add_to_linker(&mut linker, |f: &mut MyImports| f)?;
+        Resources::add_to_linker::<_, HasSelf<_>>(&mut linker, |f| f)?;
         let mut store = Store::new(&engine, MyImports::default());
         let i = Resources::instantiate(&mut store, &component, &linker)?;
 
@@ -629,6 +805,7 @@ mod exported_resources {
 
 mod unstable_import {
     use super::*;
+    use wasmtime::component::HasSelf;
 
     wasmtime::component::bindgen!({
         inline: "
@@ -688,7 +865,10 @@ mod unstable_import {
     }
     fn assert_failure(link_options: &LinkOptions) {
         let err = run_with_options(link_options).unwrap_err().to_string();
-        assert_eq!(err, "component imports instance `foo:foo/my-interface`, but a matching implementation was not found in the linker");
+        assert_eq!(
+            err,
+            "component imports instance `foo:foo/my-interface`, but a matching implementation was not found in the linker"
+        );
     }
 
     fn run_with_options(link_options: &LinkOptions) -> Result<()> {
@@ -723,7 +903,7 @@ mod unstable_import {
         }
 
         let mut linker = Linker::new(&engine);
-        MyWorld::add_to_linker(&mut linker, link_options, |h: &mut MyHost| h)?;
+        MyWorld::add_to_linker::<_, HasSelf<_>>(&mut linker, link_options, |h| h)?;
         let mut store = Store::new(&engine, MyHost::default());
         let one_import = MyWorld::instantiate(&mut store, &component, &linker)?;
         one_import.call_bar(&mut store)?;
