@@ -1,6 +1,6 @@
 #![cfg(not(miri))]
 
-use super::{make_echo_component, make_echo_component_with_params, Param, Type};
+use super::{Param, Type, make_echo_component, make_echo_component_with_params};
 use anyhow::Result;
 use wasmtime::component::types::{self, Case, ComponentItem, Field};
 use wasmtime::component::{Component, Linker, ResourceType, Val};
@@ -87,7 +87,7 @@ fn primitives() -> Result<()> {
         .call_and_post_return(&mut store, &output, &mut [])
         .unwrap_err();
     assert!(
-        err.to_string().contains("expected 1 results(s), got 0"),
+        err.to_string().contains("expected 1 result(s), got 0"),
         "{err}"
     );
 
