@@ -1,6 +1,6 @@
 use super::ref_types_module;
-use std::sync::atomic::{AtomicBool, AtomicUsize, Ordering::SeqCst};
 use std::sync::Arc;
+use std::sync::atomic::{AtomicBool, AtomicUsize, Ordering::SeqCst};
 use wasmtime::*;
 
 struct SetFlagOnDrop(Arc<AtomicBool>);
@@ -49,7 +49,7 @@ fn smoke_test_gc_impl(use_epochs: bool) -> Result<()> {
 
     let do_gc = Func::wrap(&mut store, |mut caller: Caller<'_, _>| {
         // Do a GC with `externref`s on the stack in Wasm frames.
-        caller.gc();
+        caller.gc(None);
     });
     let instance = Instance::new(&mut store, &module, &[do_gc.into()])?;
     let func = instance.get_func(&mut store, "func").unwrap();
@@ -67,7 +67,7 @@ fn smoke_test_gc_impl(use_epochs: bool) -> Result<()> {
 
         // Doing a GC should see that there aren't any `externref`s on the stack in
         // Wasm frames anymore.
-        scope.as_context_mut().gc();
+        scope.as_context_mut().gc(None);
 
         // But the scope should still be rooting `r`.
         assert!(!inner_dropped.load(SeqCst));
@@ -107,8 +107,6 @@ fn wasm_dropping_refs() -> Result<()> {
 
     let num_refs_dropped = Arc::new(AtomicUsize::new(0));
 
-    // NB: 4096 is greater than the initial `VMExternRefActivationsTable`
-    // capacity, so this will trigger at least one GC.
     for _ in 0..4096 {
         let mut scope = RootScope::new(&mut store);
         let r = ExternRef::new(&mut scope, CountDrops(num_refs_dropped.clone()))?;
@@ -116,10 +114,8 @@ fn wasm_dropping_refs() -> Result<()> {
         drop_ref.call(&mut scope, &args, &mut [])?;
     }
 
-    assert!(num_refs_dropped.load(SeqCst) > 0);
-
-    // And after doing a final GC, all the refs should have been dropped.
-    store.gc();
+    // After doing a GC, all the refs should have been dropped.
+    store.gc(None);
     assert_eq!(num_refs_dropped.load(SeqCst), 4096);
 
     return Ok(());
@@ -196,7 +192,7 @@ fn many_live_refs() -> Result<()> {
 
     many_live_refs.call(&mut store, &[], &mut [])?;
 
-    store.as_context_mut().gc();
+    store.as_context_mut().gc(None);
     assert_eq!(live_refs.load(SeqCst), 0);
 
     return Ok(());
@@ -252,7 +248,7 @@ fn drop_externref_via_table_set() -> Result<()> {
             table_set.call(&mut scope, &args, &mut [])?;
         }
 
-        scope.as_context_mut().gc();
+        scope.as_context_mut().gc(None);
         assert!(!foo_is_dropped.load(SeqCst));
         assert!(!bar_is_dropped.load(SeqCst));
 
@@ -262,7 +258,7 @@ fn drop_externref_via_table_set() -> Result<()> {
         }
     }
 
-    store.gc();
+    store.gc(None);
     assert!(foo_is_dropped.load(SeqCst));
     assert!(!bar_is_dropped.load(SeqCst));
 
@@ -641,7 +637,7 @@ fn gc_and_tail_calls_and_stack_arguments() -> Result<()> {
         },
     )?;
     linker.func_wrap("", "gc", |mut caller: Caller<()>| {
-        caller.gc();
+        caller.gc(None);
     })?;
 
     let instance = linker.instantiate(&mut store, &module)?;
@@ -740,28 +736,31 @@ fn rooted_gets_collected_after_scope_exit() -> Result<()> {
         let mut scope = RootScope::new(&mut store);
         let _externref = ExternRef::new(&mut scope, SetFlagOnDrop(flag.clone()))?;
 
-        scope.as_context_mut().gc();
+        scope.as_context_mut().gc(None);
         assert!(!flag.load(SeqCst), "not dropped when still rooted");
     }
 
-    store.as_context_mut().gc();
+    store.as_context_mut().gc(None);
     assert!(flag.load(SeqCst), "dropped after being unrooted");
 
     Ok(())
 }
 
 #[test]
-fn manually_rooted_gets_collected_after_unrooting() -> Result<()> {
+fn owned_rooted_gets_collected_after_unrooting() -> Result<()> {
     let mut store = Store::<()>::default();
     let flag = Arc::new(AtomicBool::new(false));
 
-    let externref = ExternRef::new_manually_rooted(&mut store, SetFlagOnDrop(flag.clone()))?;
+    let externref = {
+        let mut scope = RootScope::new(&mut store);
+        ExternRef::new(&mut scope, SetFlagOnDrop(flag.clone()))?.to_owned_rooted(&mut scope)?
+    };
 
-    store.gc();
+    store.gc(None);
     assert!(!flag.load(SeqCst), "not dropped when still rooted");
 
-    externref.unroot(&mut store);
-    store.gc();
+    drop(externref);
+    store.gc(None);
     assert!(flag.load(SeqCst), "dropped after being unrooted");
 
     Ok(())
@@ -783,7 +782,7 @@ fn round_trip_gc_ref_through_typed_wasm_func() -> Result<()> {
             )
         "#,
     )?;
-    let gc = Func::wrap(&mut store, |mut caller: Caller<'_, _>| caller.gc());
+    let gc = Func::wrap(&mut store, |mut caller: Caller<'_, _>| caller.gc(None));
     let instance = Instance::new(&mut store, &module, &[gc.into()])?;
     let f = instance
         .get_typed_func::<Option<Rooted<ExternRef>>, Option<Rooted<ExternRef>>>(&mut store, "f")?;
@@ -800,7 +799,7 @@ fn round_trip_gc_ref_through_func_wrap() -> Result<()> {
     let f = Func::wrap(
         &mut store,
         |mut caller: Caller<'_, _>, x: Rooted<ExternRef>| {
-            caller.gc();
+            caller.gc(None);
             x
         },
     );
@@ -819,11 +818,11 @@ fn to_raw_from_raw_doesnt_leak() -> Result<()> {
     {
         let mut scope = RootScope::new(&mut store);
         let x = ExternRef::new(&mut scope, SetFlagOnDrop(flag.clone()))?;
-        let raw = unsafe { x.to_raw(&mut scope)? };
-        let _x = unsafe { ExternRef::from_raw(&mut scope, raw) };
+        let raw = x.to_raw(&mut scope)?;
+        let _x = ExternRef::from_raw(&mut scope, raw);
     }
 
-    store.gc();
+    store.gc(None);
     assert!(flag.load(SeqCst));
     Ok(())
 }
@@ -846,7 +845,7 @@ fn table_fill_doesnt_leak() -> Result<()> {
         table.fill(&mut scope, 0, Ref::Extern(None), 10)?;
     }
 
-    store.gc();
+    store.gc(None);
     assert!(flag.load(SeqCst));
     Ok(())
 }
@@ -873,11 +872,112 @@ fn table_copy_doesnt_leak() -> Result<()> {
         Table::copy(&mut scope, &table, 0, &table, 5, 5)?;
     }
 
-    store.gc();
+    store.gc(None);
     assert!(flag.load(SeqCst));
     Ok(())
 }
 
+#[test]
+#[cfg_attr(miri, ignore)]
+fn table_set_doesnt_leak() -> Result<()> {
+    let _ = env_logger::try_init();
+
+    let mut store = Store::<()>::default();
+    let flag = Arc::new(AtomicBool::new(false));
+
+    {
+        let mut scope = RootScope::new(&mut store);
+        let table = Table::new(
+            &mut scope,
+            TableType::new(RefType::EXTERNREF, 10, Some(10)),
+            Ref::Extern(None),
+        )?;
+
+        let x = ExternRef::new(&mut scope, SetFlagOnDrop(flag.clone()))?;
+        table.set(&mut scope, 2, x.into())?;
+        table.set(&mut scope, 2, x.into())?;
+        table.set(&mut scope, 2, Ref::Extern(None))?;
+    }
+
+    store.gc(None);
+    assert!(flag.load(SeqCst));
+    Ok(())
+}
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn table_grow_doesnt_leak() -> Result<()> {
+    let _ = env_logger::try_init();
+
+    let mut store = Store::<()>::default();
+    let flag = Arc::new(AtomicBool::new(false));
+
+    {
+        let mut scope = RootScope::new(&mut store);
+        let table = Table::new(
+            &mut scope,
+            TableType::new(RefType::EXTERNREF, 10, Some(10)),
+            Ref::Extern(None),
+        )?;
+
+        let x = ExternRef::new(&mut scope, SetFlagOnDrop(flag.clone()))?;
+        table.grow(&mut scope, 0, x.into())?;
+    }
+
+    store.gc(None);
+    assert!(flag.load(SeqCst));
+    Ok(())
+}
+
+// This is a test that the argument to `table.init` is properly handled w.r.t.
+// write barriers and such. This doesn't use `SetFlagOnDrop` because that would
+// require initializing a table with an element initialized from a global and
+// the global keeps the externref alive regardless. Instead this has a small GC
+// heap and we continuously make more garbage than is in the heap and expect
+// this to work as when it triggers a GC everything prior should get cleaned up.
+#[test]
+#[cfg_attr(miri, ignore)]
+fn table_init_doesnt_leak() -> Result<()> {
+    const SIZE: u64 = 64 << 10;
+
+    let _ = env_logger::try_init();
+
+    let mut config = Config::new();
+    config.wasm_gc(true);
+    config.wasm_function_references(true);
+    config.memory_may_move(false);
+    config.memory_reservation(SIZE);
+    config.memory_reservation_for_growth(0);
+    let engine = Engine::new(&config)?;
+    let mut store = Store::new(&engine, ());
+
+    let module = Module::new(
+        store.engine(),
+        r#"
+            (module
+                (table 1 arrayref)
+
+                (type $a (array i31ref))
+
+                (func (export "run")
+                    i32.const 0
+                    i32.const 0
+                    i32.const 1
+                    table.init $e)
+                (elem $e arrayref (array.new $a (ref.i31 (i32.const 0)) (i32.const 200)))
+            )
+        "#,
+    )?;
+
+    let instance = Instance::new(&mut store, &module, &[])?;
+    let func = instance.get_typed_func::<(), ()>(&mut store, "run")?;
+    for _ in 0..200 {
+        func.call(&mut store, ())?;
+    }
+
+    Ok(())
+}
+
 #[test]
 fn ref_matches() -> Result<()> {
     let mut store = Store::<()>::default();
@@ -1106,7 +1206,7 @@ fn drc_transitive_drop_cons_list() -> Result<()> {
 
     // Not holding the cons list alive anymore; should transitively drop
     // everything we created.
-    store.gc();
+    store.gc(None);
     assert_eq!(num_refs_dropped.load(SeqCst), len);
 
     Ok(())
@@ -1166,7 +1266,7 @@ fn drc_transitive_drop_nested_arrays_tree() -> Result<()> {
 
     // Not holding the tree alive anymore; should transitively drop everything
     // we created.
-    store.gc();
+    store.gc(None);
     assert_eq!(num_refs_dropped.load(SeqCst), expected);
 
     Ok(())
@@ -1208,7 +1308,7 @@ fn drc_traces_the_correct_number_of_gc_refs_in_arrays() -> Result<()> {
     }
 
     // Make sure the poison array is collected.
-    store.gc();
+    store.gc(None);
 
     // Allocate and then collect an array of GC refs from Wasm. This should not
     // trick the collector into tracing any poison and panicking.
@@ -1225,7 +1325,480 @@ fn drc_traces_the_correct_number_of_gc_refs_in_arrays() -> Result<()> {
         "#,
     )?;
     let _instance = Instance::new(&mut store, &module, &[])?;
-    store.gc();
+    store.gc(None);
+
+    Ok(())
+}
+
+// Test that we can completely fill the GC heap until we get an OOM. This
+// exercises growing the GC heap and that we configure compilation tunables and
+// runtime memories backing GC heaps correctly.
+#[test]
+#[cfg_attr(any(miri, not(target_pointer_width = "64")), ignore)]
+fn gc_heap_oom() -> Result<()> {
+    if std::env::var("WASMTIME_TEST_NO_HOG_MEMORY").is_ok() {
+        return Ok(());
+    }
+
+    let _ = env_logger::try_init();
+
+    for heap_size in [
+        // Very small heap.
+        1 << 16,
+        // Bigger heap: 4 GiB
+        1 << 32,
+    ] {
+        for pooling in [true, false] {
+            let mut config = Config::new();
+            config.wasm_function_references(true);
+            config.wasm_gc(true);
+            config.collector(Collector::Null);
+            config.memory_reservation(heap_size);
+            config.memory_reservation_for_growth(0);
+            config.memory_guard_size(0);
+            config.memory_may_move(false);
+
+            if pooling {
+                let mut pooling = crate::small_pool_config();
+                pooling.max_memory_size(heap_size.try_into().unwrap());
+                config.allocation_strategy(InstanceAllocationStrategy::Pooling(pooling));
+            }
+
+            let engine = Engine::new(&config)?;
+
+            let module = Module::new(
+                &engine,
+                r#"
+                (module
+                    (type $s (struct))
+                    (global $g (export "g") (mut i32) (i32.const 0))
+                    (func (export "run")
+                      loop
+                        struct.new $s
+
+                        global.get $g
+                        i32.const 1
+                        i32.add
+                        global.set $g
+
+                        br 0
+                      end
+                    )
+                )
+            "#,
+            )?;
+
+            let mut store = Store::new(&engine, ());
+            let instance = Instance::new(&mut store, &module, &[])?;
+
+            let run = instance.get_typed_func::<(), ()>(&mut store, "run")?;
+            let err = run.call(&mut store, ()).expect_err("should oom");
+            assert!(err.is::<Trap>(), "should get trap, got: {err:?}");
+            let trap = err.downcast::<Trap>().unwrap();
+            assert_eq!(trap, Trap::AllocationTooLarge);
+
+            let g = instance.get_global(&mut store, "g").unwrap();
+            const SIZE_OF_NULL_GC_HEADER: u64 = 8;
+            const FUDGE: u64 = 2;
+            let actual = g.get(&mut store).unwrap_i32() as u64;
+            let expected = heap_size / SIZE_OF_NULL_GC_HEADER;
+            assert!(
+                actual.abs_diff(expected) <= FUDGE,
+                "actual approx= expected failed: \
+                 actual = {actual}, expected = {expected}, FUDGE={FUDGE}"
+            );
+        }
+    }
+    Ok(())
+}
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn issue_10772() -> Result<()> {
+    let mut store = crate::gc_store()?;
+    let engine = store.engine().clone();
+
+    let module = Module::new(
+        &engine,
+        r#"
+            (module
+                (type $empty (struct))
+                (type $tuple-concrete (struct (field (ref $empty))))
+                (type $tuple-abstract (struct (field (ref struct))))
+                (func (export "abstract") (param $t (ref $tuple-abstract))
+                    (drop (ref.cast (ref $tuple-concrete) (local.get $t)))
+                )
+            )
+        "#,
+    )?;
+
+    let linker = Linker::new(&engine);
+
+    let instance = linker.instantiate(&mut store, &module)?;
+    let abstract_ = instance.get_func(&mut store, "abstract").unwrap();
+    let empty_pre = StructRefPre::new(&mut store, StructType::new(&engine, [])?);
+    let empty_struct = StructRef::new(&mut store, &empty_pre, &[])?;
+    let tuple_pre = StructRefPre::new(
+        &mut store,
+        StructType::new(
+            &engine,
+            [FieldType::new(
+                Mutability::Const,
+                StorageType::ValType(ValType::Ref(RefType::new(false, HeapType::Struct))),
+            )],
+        )?,
+    );
+    let tuple_struct = StructRef::new(&mut store, &tuple_pre, &[empty_struct.into()])?;
+    let tuple_any = Val::from(tuple_struct);
+
+    match abstract_.call(store, &[tuple_any], &mut []) {
+        Ok(()) => panic!("should have trapped on cast failure"),
+        Err(e) => {
+            let trap = e.downcast::<Trap>().expect("should fail with a trap");
+            assert_eq!(trap, Trap::CastFailure);
+        }
+    }
+
+    Ok(())
+}
+
+#[test]
+fn drc_gc_inbetween_host_calls() -> Result<()> {
+    let _ = env_logger::try_init();
+
+    let mut config = Config::new();
+    config.wasm_function_references(true);
+    config.wasm_gc(true);
+    config.collector(Collector::DeferredReferenceCounting);
+
+    let engine = Engine::new(&config)?;
+
+    let mut store = Store::new(&engine, ());
+    let func = Func::wrap(&mut store, |_: Option<Rooted<ExternRef>>| {});
+
+    let mut invoke_func = || {
+        let inner_dropped = Arc::new(AtomicBool::new(false));
+        {
+            let mut scope = RootScope::new(&mut store);
+            let r = ExternRef::new(&mut scope, SetFlagOnDrop(inner_dropped.clone()))?;
+            func.call(&mut scope, &[r.into()], &mut [])?;
+        }
+
+        assert!(!inner_dropped.load(SeqCst));
+        store.gc(None);
+        assert!(inner_dropped.load(SeqCst));
+        anyhow::Ok(())
+    };
+
+    invoke_func()?;
+    invoke_func()?;
+
+    Ok(())
+}
+
+#[test]
+fn owned_rooted() -> Result<()> {
+    let _ = env_logger::try_init();
+
+    let mut config = Config::new();
+    config.wasm_function_references(true);
+    config.wasm_gc(true);
+    config.collector(Collector::DeferredReferenceCounting);
+
+    let engine = Engine::new(&config)?;
+
+    let mut store = Store::new(&engine, ());
+    let inner_dropped = Arc::new(AtomicBool::new(false));
+    let r = {
+        let mut scope = RootScope::new(&mut store);
+        let r = ExternRef::new(&mut scope, SetFlagOnDrop(inner_dropped.clone()))?;
+        r.to_owned_rooted(&mut scope)?
+    };
+    assert!(!inner_dropped.load(SeqCst));
+    store.gc(None);
+    assert!(!inner_dropped.load(SeqCst));
+    let r2 = r.clone();
+    store.gc(None);
+    assert!(!inner_dropped.load(SeqCst));
+    drop(r);
+    store.gc(None);
+    assert!(!inner_dropped.load(SeqCst));
+    drop(r2);
+    store.gc(None);
+    assert!(inner_dropped.load(SeqCst));
+
+    Ok(())
+}
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn owned_rooted_lots_of_root_creation() -> Result<()> {
+    let mut config = Config::new();
+    config.wasm_function_references(true);
+    config.wasm_gc(true);
+    config.collector(Collector::DeferredReferenceCounting);
+
+    let engine = Engine::new(&config)?;
+
+    let mut store = Store::new(&engine, ());
+    let inner_dropped = Arc::new(AtomicBool::new(false));
+    let r = {
+        let mut scope = RootScope::new(&mut store);
+        let r = ExternRef::new(&mut scope, SetFlagOnDrop(inner_dropped.clone()))?;
+        r.to_owned_rooted(&mut scope)?
+    };
+    assert!(!inner_dropped.load(SeqCst));
+    store.gc(None);
+
+    for _ in 0..100_000 {
+        let mut scope = RootScope::new(&mut store);
+        // Go through a LIFO root then back to an owned root to create
+        // a distinct root.
+        let r2 = r.to_rooted(&mut scope);
+        let r3 = r2.to_owned_rooted(&mut scope);
+        drop(r3);
+    }
+
+    store.gc(None);
+    assert!(!inner_dropped.load(SeqCst));
+
+    drop(r);
+    store.gc(None);
+    assert!(inner_dropped.load(SeqCst));
+
+    Ok(())
+}
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn runtime_table_init_oom() -> Result<()> {
+    let mut config = Config::new();
+    config.wasm_gc(true);
+    config.wasm_function_references(true);
+    config.memory_may_move(false);
+    config.memory_reservation(64 << 10);
+    config.memory_reservation_for_growth(0);
+    let engine = Engine::new(&config)?;
+    let mut store = Store::new(&engine, ());
+
+    let module = Module::new(
+        store.engine(),
+        r#"
+            (module
+                (table 100 arrayref)
+
+                (type $a (array i31ref))
+
+                (func (export "run")
+                    i32.const 0
+                    i32.const 0
+                    i32.const 5
+                    table.init $e)
+                (elem $e arrayref
+                    (array.new_default $a (i32.const 100))
+                    (array.new_default $a (i32.const 10000))
+                    (array.new_default $a (i32.const 10000))
+                    (array.new_default $a (i32.const 10000))
+                    (array.new_default $a (i32.const 1000000))
+                )
+            )
+        "#,
+    )?;
+
+    let instance = Instance::new(&mut store, &module, &[])?;
+    let func = instance.get_typed_func::<(), ()>(&mut store, "run")?;
+    func.call(&mut store, ())
+        .unwrap_err()
+        .downcast::<GcHeapOutOfMemory<()>>()?;
+
+    Ok(())
+}
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn instantiate_table_init_oom() -> Result<()> {
+    let mut config = Config::new();
+    config.wasm_gc(true);
+    config.wasm_function_references(true);
+    config.memory_may_move(false);
+    config.memory_reservation(64 << 10);
+    config.memory_reservation_for_growth(0);
+    let engine = Engine::new(&config)?;
+    let mut store = Store::new(&engine, ());
+
+    let module = Module::new(
+        store.engine(),
+        r#"
+            (module
+                (table 100 arrayref)
+
+                (type $a (array i31ref))
+
+                (elem (i32.const 0) arrayref
+                    (array.new_default $a (i32.const 100))
+                    (array.new_default $a (i32.const 10000))
+                    (array.new_default $a (i32.const 10000))
+                    (array.new_default $a (i32.const 10000))
+                    (array.new_default $a (i32.const 1000000))
+                )
+            )
+        "#,
+    )?;
+
+    Instance::new(&mut store, &module, &[])
+        .unwrap_err()
+        .downcast::<GcHeapOutOfMemory<()>>()?;
+
+    Ok(())
+}
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn instantiate_table_init_expr_oom() -> Result<()> {
+    let mut config = Config::new();
+    config.wasm_gc(true);
+    config.wasm_function_references(true);
+    config.memory_may_move(false);
+    config.memory_reservation(64 << 10);
+    config.memory_reservation_for_growth(0);
+    let engine = Engine::new(&config)?;
+    let mut store = Store::new(&engine, ());
+
+    let module = Module::new(
+        store.engine(),
+        r#"
+            (module
+                (type $a (array i31ref))
+                (table 100 (ref $a) (array.new_default $a (i32.const 100000)))
+            )
+        "#,
+    )?;
+
+    Instance::new(&mut store, &module, &[])
+        .unwrap_err()
+        .downcast::<GcHeapOutOfMemory<()>>()?;
+
+    Ok(())
+}
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn instantiate_global_init_oom() -> Result<()> {
+    let mut config = Config::new();
+    config.wasm_gc(true);
+    config.wasm_function_references(true);
+    config.memory_may_move(false);
+    config.memory_reservation(64 << 10);
+    config.memory_reservation_for_growth(0);
+    let engine = Engine::new(&config)?;
+    let mut store = Store::new(&engine, ());
+
+    let module = Module::new(
+        store.engine(),
+        r#"
+            (module
+                (table 100 arrayref)
+                (type $a (array i31ref))
+                (global (ref $a) (array.new_default $a (i32.const 10000000)))
+            )
+        "#,
+    )?;
+
+    Instance::new(&mut store, &module, &[])
+        .unwrap_err()
+        .downcast::<GcHeapOutOfMemory<()>>()?;
+
+    Ok(())
+}
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn array_new_elem_oom() -> Result<()> {
+    let mut config = Config::new();
+    config.wasm_gc(true);
+    config.wasm_function_references(true);
+    config.memory_may_move(false);
+    config.memory_reservation(64 << 10);
+    config.memory_reservation_for_growth(0);
+    let engine = Engine::new(&config)?;
+    let mut store = Store::new(&engine, ());
+
+    let module = Module::new(
+        store.engine(),
+        r#"
+            (module
+                (type $a (array (mut arrayref)))
+                (type $i (array i31ref))
+
+                (func (export "run")
+                    i32.const 0
+                    i32.const 5
+                    array.new_elem $a $e
+                    drop)
+
+                (elem $e arrayref
+                    (array.new_default $i (i32.const 100))
+                    (array.new_default $i (i32.const 10000))
+                    (array.new_default $i (i32.const 10000))
+                    (array.new_default $i (i32.const 10000))
+                    (array.new_default $i (i32.const 1000000))
+                )
+            )
+        "#,
+    )?;
+
+    let instance = Instance::new(&mut store, &module, &[])?;
+    let func = instance.get_typed_func::<(), ()>(&mut store, "run")?;
+    func.call(&mut store, ())
+        .unwrap_err()
+        .downcast::<GcHeapOutOfMemory<()>>()?;
+
+    Ok(())
+}
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn array_init_elem_oom() -> Result<()> {
+    let mut config = Config::new();
+    config.wasm_gc(true);
+    config.wasm_function_references(true);
+    config.memory_may_move(false);
+    config.memory_reservation(64 << 10);
+    config.memory_reservation_for_growth(0);
+    let engine = Engine::new(&config)?;
+    let mut store = Store::new(&engine, ());
+
+    let module = Module::new(
+        store.engine(),
+        r#"
+            (module
+                (type $a (array (mut arrayref)))
+                (type $i (array i31ref))
+
+                (func (export "run")
+                    i32.const 5
+                    array.new_default $a
+                    i32.const 0
+                    i32.const 0
+                    i32.const 5
+                    array.init_elem $a $e)
+
+                (elem $e arrayref
+                    (array.new_default $i (i32.const 100))
+                    (array.new_default $i (i32.const 10000))
+                    (array.new_default $i (i32.const 10000))
+                    (array.new_default $i (i32.const 10000))
+                    (array.new_default $i (i32.const 1000000))
+                )
+            )
+        "#,
+    )?;
+
+    let instance = Instance::new(&mut store, &module, &[])?;
+    let func = instance.get_typed_func::<(), ()>(&mut store, "run")?;
+    func.call(&mut store, ())
+        .unwrap_err()
+        .downcast::<GcHeapOutOfMemory<()>>()?;
 
     Ok(())
 }
