@@ -1,17 +1,22 @@
-#![expect(clippy::allow_attributes_without_reason, reason = "crate not migrated")]
 #![cfg_attr(miri, allow(dead_code, unused_imports))]
 
+use wasmtime::Result;
+
 mod arrays;
 mod async_functions;
 mod call_hook;
 mod cli_tests;
 mod code_too_large;
+mod compile_time_builtins;
 mod component_model;
 mod coredump;
 mod custom_code_memory;
 mod debug;
 mod defaults;
 mod epoch_interruption;
+mod eqref;
+mod exceptions;
+mod exnrefs;
 mod externals;
 mod fuel;
 mod func;
@@ -24,6 +29,7 @@ mod iloop;
 mod import_calling_export;
 mod import_indexes;
 mod instance;
+mod intrinsics;
 mod invoke_func_via_table;
 mod limits;
 mod linker;
@@ -32,6 +38,7 @@ mod memory_creator;
 mod module;
 mod module_serialize;
 mod name;
+mod native_debug;
 mod noextern;
 mod piped_tests;
 mod pooling_allocator;
@@ -42,12 +49,12 @@ mod stack_overflow;
 mod store;
 mod structs;
 mod table;
+#[cfg(all(feature = "stack-switching", unix, target_arch = "x86_64"))]
 mod tags;
 mod threads;
 mod traps;
 mod types;
 mod wait_notify;
-mod wasi_testsuite;
 mod winch_engine_features;
 
 /// A helper to compile a module in a new store with reference types enabled.
@@ -100,7 +107,7 @@ pub(crate) fn small_pool_config() -> wasmtime::PoolingAllocationConfig {
     // When testing, we may choose to start with MPK force-enabled to ensure
     // we use that functionality.
     if std::env::var("WASMTIME_TEST_FORCE_MPK").is_ok() {
-        config.memory_protection_keys(wasmtime::MpkEnabled::Enable);
+        config.memory_protection_keys(wasmtime::Enabled::No);
     }
 
     config.total_stacks(1);
@@ -108,7 +115,7 @@ pub(crate) fn small_pool_config() -> wasmtime::PoolingAllocationConfig {
     config
 }
 
-pub(crate) fn gc_store() -> wasmtime::Result<wasmtime::Store<()>> {
+pub(crate) fn gc_store() -> Result<wasmtime::Store<()>> {
     let _ = env_logger::try_init();
 
     let mut config = wasmtime::Config::new();
