@@ -114,8 +114,8 @@ fn offsets_static_dynamic_oh_my(config: &mut Config) -> Result<()> {
         }
     }
 
-    engines.par_iter().for_each(|engine| {
-        let module = module(&engine).unwrap();
+    engines.par_iter().try_for_each(|engine| {
+        let module = module(&engine)?;
 
         for (min, max) in [(1, Some(2)), (1, None)].iter() {
             let mut store = Store::new(&engine, ());
@@ -133,13 +133,15 @@ fn offsets_static_dynamic_oh_my(config: &mut Config) -> Result<()> {
             test_traps(&mut store, &funcs, 65536, &mem);
             test_traps(&mut store, &funcs, u32::MAX, &mem);
         }
-    });
+        Ok::<_, wasmtime::Error>(())
+    })?;
 
     Ok(())
 }
 
 #[test]
 #[cfg_attr(miri, ignore)]
+#[cfg_attr(asan, ignore)]
 fn guards_present() -> Result<()> {
     const GUARD_SIZE: u64 = 65536;
 
@@ -188,13 +190,14 @@ fn guards_present() -> Result<()> {
 
 #[wasmtime_test]
 #[cfg_attr(miri, ignore)]
+#[cfg_attr(asan, ignore)]
 fn guards_present_pooling(config: &mut Config) -> Result<()> {
     const GUARD_SIZE: u64 = 65536;
 
     let mut pool = crate::small_pool_config();
     pool.total_memories(2)
         .max_memory_size(10 << 16)
-        .memory_protection_keys(MpkEnabled::Disable);
+        .memory_protection_keys(Enabled::No);
     config.memory_reservation(1 << 20);
     config.memory_guard_size(GUARD_SIZE);
     config.guard_before_linear_memory(true);
@@ -217,17 +220,19 @@ fn guards_present_pooling(config: &mut Config) -> Result<()> {
     };
 
     unsafe fn assert_guards(store: &Store<()>, mem: &Memory) {
-        // guards before
-        println!("check pre-mem");
-        assert_faults(mem.data_ptr(&store).offset(-(GUARD_SIZE as isize)));
+        unsafe {
+            // guards before
+            println!("check pre-mem");
+            assert_faults(mem.data_ptr(&store).offset(-(GUARD_SIZE as isize)));
 
-        // unmapped just after memory
-        println!("check mem");
-        assert_faults(mem.data_ptr(&store).add(mem.data_size(&store)));
+            // unmapped just after memory
+            println!("check mem");
+            assert_faults(mem.data_ptr(&store).add(mem.data_size(&store)));
 
-        // guards after memory
-        println!("check post-mem");
-        assert_faults(mem.data_ptr(&store).add(1 << 20));
+            // guards after memory
+            println!("check post-mem");
+            assert_faults(mem.data_ptr(&store).add(1 << 20));
+        }
     }
     unsafe {
         assert_guards(&store, &mem1);
@@ -244,6 +249,8 @@ fn guards_present_pooling(config: &mut Config) -> Result<()> {
 
 #[wasmtime_test]
 #[cfg_attr(miri, ignore)]
+#[cfg_attr(asan, ignore)]
+#[cfg(target_arch = "x86_64")] // only platform with mpk
 fn guards_present_pooling_mpk(config: &mut Config) -> Result<()> {
     if !wasmtime::PoolingAllocationConfig::are_memory_protection_keys_available() {
         println!("skipping `guards_present_pooling_mpk` test; mpk is not supported");
@@ -254,7 +261,7 @@ fn guards_present_pooling_mpk(config: &mut Config) -> Result<()> {
     let mut pool = crate::small_pool_config();
     pool.total_memories(4)
         .max_memory_size(10 << 16)
-        .memory_protection_keys(MpkEnabled::Enable)
+        .memory_protection_keys(Enabled::Yes)
         .max_memory_protection_keys(2);
     config.memory_reservation(1 << 20);
     config.memory_guard_size(GUARD_SIZE);
@@ -278,17 +285,19 @@ fn guards_present_pooling_mpk(config: &mut Config) -> Result<()> {
     };
 
     unsafe fn assert_guards(store: &Store<()>, mem: &Memory) {
-        // guards before
-        println!("check pre-mem");
-        assert_faults(mem.data_ptr(&store).offset(-(GUARD_SIZE as isize)));
+        unsafe {
+            // guards before
+            println!("check pre-mem");
+            assert_faults(mem.data_ptr(&store).offset(-(GUARD_SIZE as isize)));
 
-        // unmapped just after memory
-        println!("check mem");
-        assert_faults(mem.data_ptr(&store).add(mem.data_size(&store)));
+            // unmapped just after memory
+            println!("check mem");
+            assert_faults(mem.data_ptr(&store).add(mem.data_size(&store)));
 
-        // guards after memory
-        println!("check post-mem");
-        assert_faults(mem.data_ptr(&store).add(1 << 20));
+            // guards after memory
+            println!("check post-mem");
+            assert_faults(mem.data_ptr(&store).add(1 << 20));
+        }
     }
     unsafe {
         assert_guards(&store, &mem1);
@@ -306,11 +315,12 @@ fn guards_present_pooling_mpk(config: &mut Config) -> Result<()> {
 unsafe fn assert_faults(ptr: *mut u8) {
     use std::io::Error;
     #[cfg(unix)]
-    {
+    unsafe {
         // There's probably a faster way to do this here, but, uh, when in rome?
         match libc::fork() {
             0 => {
                 *ptr = 4;
+
                 std::process::exit(0);
             }
             -1 => panic!("failed to fork: {}", Error::last_os_error()),
@@ -326,7 +336,7 @@ unsafe fn assert_faults(ptr: *mut u8) {
         }
     }
     #[cfg(windows)]
-    {
+    unsafe {
         use windows_sys::Win32::System::Memory::*;
 
         let mut info = std::mem::MaybeUninit::uninit();
@@ -343,7 +353,10 @@ unsafe fn assert_faults(ptr: *mut u8) {
     }
 }
 
+// Disable test on s390x because the large allocation may actually succeed;
+// the whole 64-bit address space is available on this platform.
 #[test]
+#[cfg(not(target_arch = "s390x"))]
 fn massive_64_bit_still_limited() -> Result<()> {
     // Creating a 64-bit memory which exceeds the limits of the address space
     // should still send a request to the `ResourceLimiter` to ensure that it
@@ -500,11 +513,13 @@ fn memory64_maximum_minimum(config: &mut Config) -> Result<()> {
     let engine = Engine::new(&config)?;
     let mut store = Store::new(&engine, ());
 
-    assert!(MemoryTypeBuilder::default()
-        .memory64(true)
-        .min(1 << 48)
-        .build()
-        .is_err());
+    assert!(
+        MemoryTypeBuilder::default()
+            .memory64(true)
+            .min(1 << 48)
+            .build()
+            .is_err()
+    );
 
     let module = Module::new(
         &engine,
@@ -556,12 +571,14 @@ fn shared_memory_basics() -> Result<()> {
     assert!(SharedMemory::new(&engine, MemoryType::new(1, Some(1))).is_err());
     assert!(SharedMemory::new(&engine, MemoryType::new64(1, None)).is_err());
     assert!(SharedMemory::new(&engine, MemoryType::new64(1, Some(1))).is_err());
-    assert!(MemoryTypeBuilder::default()
-        .shared(true)
-        .min(1)
-        .max(Some(0))
-        .build()
-        .is_err());
+    assert!(
+        MemoryTypeBuilder::default()
+            .shared(true)
+            .min(1)
+            .max(Some(0))
+            .build()
+            .is_err()
+    );
 
     let memory = SharedMemory::new(&engine, MemoryType::shared(1, 1))?;
     assert!(memory.ty().is_shared());
@@ -660,8 +677,8 @@ fn shared_memory_wait_notify() -> Result<()> {
 #[wasmtime_test]
 #[cfg_attr(miri, ignore)]
 #[cfg(target_pointer_width = "64")] // requires large VM reservation
-fn init_with_negative_segment(_: &mut Config) -> Result<()> {
-    let engine = Engine::default();
+fn init_with_negative_segment(cfg: &mut Config) -> Result<()> {
+    let engine = Engine::new(cfg)?;
     let module = Module::new(
         &engine,
         r#"
