@@ -1,76 +1,102 @@
 use super::{
+    ABI, RegAlloc,
     abi::Aarch64ABI,
     address::Address,
-    asm::Assembler,
-    regs::{self, scratch},
-    ABI,
+    asm::{Assembler, PatchableAddToReg},
+    regs::{self, scratch_fpr_bitset, scratch_gpr_bitset},
 };
 use crate::{
     abi::{self, align_to, calculate_frame_adjustment, local::LocalSlot, vmctx},
-    codegen::{ptr_type_from_ptr_size, CodeGenContext, CodeGenError, Emission, FuncEnv},
+    codegen::{CodeGenContext, CodeGenError, Emission, FuncEnv, ptr_type_from_ptr_size},
     isa::{
-        aarch64::abi::SHADOW_STACK_POINTER_SLOT_SIZE,
-        reg::{writable, Reg, WritableReg},
         CallingConvention,
+        aarch64::abi::SHADOW_STACK_POINTER_SLOT_SIZE,
+        reg::{Reg, WritableReg, writable},
     },
     masm::{
-        CalleeKind, DivKind, Extend, ExtendKind, ExtractLaneKind, FloatCmpKind, Imm as I,
-        IntCmpKind, LoadKind, MacroAssembler as Masm, MulWideKind, OperandSize, RegImm, RemKind,
-        ReplaceLaneKind, RmwOp, RoundingMode, SPOffset, ShiftKind, SplatKind, StackSlot, StoreKind,
-        TrapCode, TruncKind, V128AbsKind, V128AddKind, V128ConvertKind, V128ExtAddKind,
+        CalleeKind, DivKind, Extend, ExtendKind, ExtractLaneKind, FloatCmpKind, FloatScratch,
+        Imm as I, IntCmpKind, IntScratch, LoadKind, MacroAssembler as Masm, MulWideKind,
+        OperandSize, RegImm, RemKind, ReplaceLaneKind, RmwOp, RoundingMode, SPOffset, Scratch,
+        ScratchType, ShiftKind, SplatKind, StackSlot, StoreKind, TRUSTED_FLAGS, TrapCode,
+        TruncKind, UNTRUSTED_FLAGS, V128AbsKind, V128AddKind, V128ConvertKind, V128ExtAddKind,
         V128ExtMulKind, V128ExtendKind, V128MaxKind, V128MinKind, V128MulKind, V128NarrowKind,
         V128NegKind, V128SubKind, V128TruncKind, VectorCompareKind, VectorEqualityKind, Zero,
-        TRUSTED_FLAGS, UNTRUSTED_FLAGS,
     },
     stack::TypedReg,
 };
-use anyhow::{anyhow, bail, Result};
+use anyhow::{Result, anyhow, bail};
 use cranelift_codegen::{
+    Final, MachBufferFinalized, MachLabel,
     binemit::CodeOffset,
     ir::{MemFlags, RelSourceLoc, SourceLoc},
-    isa::aarch64::inst::{Cond, VectorSize},
-    settings, Final, MachBufferFinalized, MachLabel,
+    isa::aarch64::inst::{self, Cond, Imm12, ImmLogic, ImmShift, VectorSize},
+    settings,
 };
 use regalloc2::RegClass;
 use wasmtime_environ::{PtrSize, WasmValType};
 
 /// Aarch64 MacroAssembler.
 pub(crate) struct MacroAssembler {
+    /// This value represents the maximum stack size seen while compiling the
+    /// function. While the function is still being compiled its value will not
+    /// be valid (the stack will grow and shrink as space is reserved and freed
+    /// during compilation), but once all instructions have been seen this value
+    /// will be the maximum stack usage seen.
+    sp_max: u32,
+
+    /// Add-with-immediate patchable instruction sequence used to add the
+    /// constant stack max to a register.
+    stack_max_use_add: Option<PatchableAddToReg>,
+
     /// Low level assembler.
     asm: Assembler,
     /// Stack pointer offset.
     sp_offset: u32,
     /// The target pointer size.
     ptr_size: OperandSize,
+    /// Scratch register scope.
+    scratch_scope: RegAlloc,
 }
 
 impl MacroAssembler {
     /// Create an Aarch64 MacroAssembler.
     pub fn new(ptr_size: impl PtrSize, shared_flags: settings::Flags) -> Result<Self> {
         Ok(Self {
+            sp_max: 0,
+            stack_max_use_add: None,
             asm: Assembler::new(shared_flags),
             sp_offset: 0u32,
             ptr_size: ptr_type_from_ptr_size(ptr_size.size()).try_into()?,
+            scratch_scope: RegAlloc::from(scratch_gpr_bitset(), scratch_fpr_bitset()),
         })
     }
 
+    /// Add the maximum stack used to a register, recording an obligation to update the
+    /// add-with-immediate instruction emitted to use the real stack max when the masm is being
+    /// finalized.
+    fn add_stack_max(&mut self, reg: WritableReg, tmp: WritableReg) {
+        assert!(self.stack_max_use_add.is_none());
+        let patch = PatchableAddToReg::new(reg, tmp, self.asm.buffer_mut());
+        self.stack_max_use_add.replace(patch);
+    }
+
     /// Ensures that the stack pointer remains 16-byte aligned for the duration
     /// of the provided function. This alignment is necessary for AArch64
     /// compliance, particularly for signal handlers that may be invoked
     /// during execution. While the compiler doesn't directly use the stack
     /// pointer for memory addressing, maintaining this alignment is crucial
     /// to prevent issues when handling signals.
-    pub fn with_aligned_sp<F, T>(&mut self, mut f: F) -> Result<T>
+    pub fn with_aligned_sp<F, T>(&mut self, f: F) -> Result<T>
     where
-        F: FnMut(&mut Self) -> Result<T>,
+        F: FnOnce(&mut Self) -> Result<T>,
     {
         let mut aligned = false;
         let alignment: u32 = <Aarch64ABI as ABI>::call_stack_align().into();
         let addend: u32 = <Aarch64ABI as ABI>::initial_frame_size().into();
         let delta = calculate_frame_adjustment(self.sp_offset()?.as_u32(), addend, alignment);
         if delta != 0 {
-            self.asm.sub_ir(
-                u64::from(delta),
+            self.sub(
+                writable!(regs::sp()),
                 // Since we don't need to synchronize the shadow stack pointer
                 // when freeing stack space [^1], the stack pointer may become
                 // out of sync with the primary shadow stack pointer. Therefore,
@@ -82,9 +108,10 @@ impl MacroAssembler {
                 // current stack pointer, preventing issues with signal handlers
                 // and interrupts.
                 regs::shadow_sp(),
-                writable!(regs::sp()),
+                RegImm::i32(delta as i32),
                 OperandSize::S64,
-            );
+            )?;
+
             aligned = true;
         }
 
@@ -120,9 +147,52 @@ impl Masm for MacroAssembler {
         Ok(())
     }
 
-    fn check_stack(&mut self, _vmctx: Reg) -> Result<()> {
-        // TODO: Implement when we have more complete assembler support.
-        Ok(())
+    fn check_stack(&mut self, vmctx: Reg) -> Result<()> {
+        let ptr_size_u8: u8 = self.ptr_size.bytes().try_into().unwrap();
+
+        // The PatchableAddToReg construct on aarch64 is not a single
+        // add-immediate instruction, but a 3-instruction sequence that loads an
+        // immediate using 2 mov-immediate instructions into _another_ scratch
+        // register before adding it into the target scratch register.
+        //
+        // In other words, to make this work we use _two_ scratch registers, one
+        // to hold the limit we're calculating and one helper that's just used
+        // to load the immediate.
+        //
+        // Luckily on aarch64 we have 2 available scratch registers, ip0 and
+        // ip1.
+        // NB that this in this case, we manually allocate the scratch registers
+        // as precision when it comes to its usage is
+
+        let ptr_size = self.ptr_size;
+        self.with_aligned_sp(|masm| {
+            masm.with_scratch::<IntScratch, _>(|masm, scratch_stk_limit| {
+                masm.with_scratch::<IntScratch, _>(|masm, scratch_tmp| {
+                    masm.load_ptr(
+                        masm.address_at_reg(vmctx, ptr_size_u8.vmcontext_store_context().into())?,
+                        scratch_stk_limit.writable(),
+                    )?;
+
+                    masm.load_ptr(
+                        Address::offset(
+                            scratch_stk_limit.inner(),
+                            ptr_size_u8.vmstore_context_stack_limit().into(),
+                        ),
+                        scratch_stk_limit.writable(),
+                    )?;
+
+                    masm.add_stack_max(scratch_stk_limit.writable(), scratch_tmp.writable());
+
+                    // Aarch can only do a cmp with sp in the first operand, which means we
+                    // use a less-than comparison, not a greater-than (stack grows down).
+                    masm.cmp(regs::sp(), scratch_stk_limit.inner().into(), ptr_size)?;
+                    masm.asm
+                        .trapif(IntCmpKind::LtU.into(), TrapCode::STACK_OVERFLOW);
+
+                    Ok(())
+                })
+            })
+        })
     }
 
     fn frame_restore(&mut self) -> Result<()> {
@@ -159,8 +229,18 @@ impl Masm for MacroAssembler {
         }
 
         let ssp = regs::shadow_sp();
-        self.asm
-            .sub_ir(bytes as u64, ssp, writable!(ssp), OperandSize::S64);
+
+        match Imm12::maybe_from_u64(bytes as u64) {
+            Some(v) => self.asm.sub_ir(v, ssp, writable!(ssp), OperandSize::S64),
+            None => {
+                self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                    masm.asm
+                        .mov_ir(scratch.writable(), I::I64(bytes as u64), OperandSize::S64);
+                    masm.asm
+                        .sub_rrr(scratch.inner(), ssp, writable!(ssp), OperandSize::S64);
+                });
+            }
+        }
 
         // Even though we're using the shadow stack pointer to reserve stack, we
         // must ensure that the real stack pointer reflects the stack claimed so
@@ -178,8 +258,17 @@ impl Masm for MacroAssembler {
         }
 
         let ssp = regs::shadow_sp();
-        self.asm
-            .add_ir(bytes as u64, ssp, writable!(ssp), OperandSize::S64);
+        match Imm12::maybe_from_u64(bytes as u64) {
+            Some(v) => self.asm.add_ir(v, ssp, writable!(ssp), OperandSize::S64),
+            None => {
+                self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                    masm.asm
+                        .mov_ir(scratch.writable(), I::I64(bytes as u64), OperandSize::S64);
+                    masm.asm
+                        .add_rrr(ssp, scratch.inner(), writable!(ssp), OperandSize::S64);
+                });
+            }
+        }
 
         // We must ensure that the real stack pointer reflects the offset
         // tracked by `self.sp_offset`, we use such value to calculate
@@ -235,29 +324,30 @@ impl Masm for MacroAssembler {
     }
 
     fn store(&mut self, src: RegImm, dst: Address, size: OperandSize) -> Result<()> {
-        let src = match src {
+        match src {
             RegImm::Imm(v) => {
-                let imm = match v {
-                    I::I32(v) | I::F32(v) => v as u64,
-                    I::F64(v) | I::I64(v) => v,
-                    I::V128(_) => unreachable!(),
+                match v {
+                    I::I32(_) | I::I64(_) => {
+                        self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                            masm.asm.mov_ir(scratch.writable(), v, v.size());
+                            masm.asm.str(scratch.inner(), dst, size, TRUSTED_FLAGS);
+                        });
+                    }
+                    imm @ (I::F32(_) | I::F64(_)) => {
+                        self.with_scratch::<FloatScratch, _>(|masm, scratch| {
+                            masm.asm.mov_ir(scratch.writable(), imm, imm.size());
+                            masm.asm.str(scratch.inner(), dst, size, TRUSTED_FLAGS);
+                        });
+                    }
+                    _ => bail!(CodeGenError::unsupported_wasm_type()),
                 };
-                let scratch = regs::scratch();
-                self.asm.load_constant(imm, writable!(scratch));
-                if v.is_float() {
-                    let float_scratch = regs::float_scratch();
-                    self.asm
-                        .mov_to_fpu(scratch, writable!(float_scratch), v.size());
-                    float_scratch
-                } else {
-                    scratch
-                }
+                Ok(())
             }
-            RegImm::Reg(reg) => reg,
-        };
-
-        self.asm.str(src, dst, size, TRUSTED_FLAGS);
-        Ok(())
+            RegImm::Reg(r) => {
+                self.asm.str(r, dst, size, TRUSTED_FLAGS);
+                Ok(())
+            }
+        }
     }
 
     fn wasm_store(&mut self, src: Reg, dst: Self::Address, op_kind: StoreKind) -> Result<()> {
@@ -275,6 +365,18 @@ impl Masm for MacroAssembler {
         })
     }
 
+    fn with_scratch<T: ScratchType, R>(&mut self, f: impl FnOnce(&mut Self, Scratch) -> R) -> R {
+        let r = self
+            .scratch_scope
+            .reg_for_class(T::reg_class(), &mut |_| Ok(()))
+            .expect("Scratch register to be available");
+
+        let ret = f(self, Scratch::new(r));
+
+        self.scratch_scope.free(r);
+        ret
+    }
+
     fn call(
         &mut self,
         stack_args_size: u32,
@@ -290,7 +392,6 @@ impl Masm for MacroAssembler {
         match callee {
             CalleeKind::Indirect(reg) => self.asm.call_with_reg(reg, call_conv),
             CalleeKind::Direct(idx) => self.asm.call_with_name(idx, call_conv),
-            CalleeKind::LibCall(lib) => self.asm.call_with_lib(lib, scratch(), call_conv),
         }
 
         Ok(total_stack)
@@ -308,7 +409,13 @@ impl Masm for MacroAssembler {
     fn wasm_load(&mut self, src: Self::Address, dst: WritableReg, kind: LoadKind) -> Result<()> {
         let size = kind.derive_operand_size();
         self.with_aligned_sp(|masm| match &kind {
-            LoadKind::Operand(_) => Ok(masm.asm.uload(src, dst, size, UNTRUSTED_FLAGS)),
+            LoadKind::Operand(_) => {
+                if size == OperandSize::S128 {
+                    bail!(CodeGenError::UnimplementedWasmLoadKind)
+                } else {
+                    Ok(masm.asm.uload(src, dst, size, UNTRUSTED_FLAGS))
+                }
+            }
             LoadKind::Splat(_) => bail!(CodeGenError::UnimplementedWasmLoadKind),
             LoadKind::ScalarExtend(extend_kind) => {
                 if extend_kind.signed() {
@@ -340,9 +447,7 @@ impl Masm for MacroAssembler {
         size: OperandSize,
     ) -> Result<()> {
         let (base, offset) = src.unwrap_offset();
-        self.asm
-            .add_ir(u64::try_from(offset).unwrap(), base, dst, size);
-        Ok(())
+        self.add_ir(dst, base, I::i64(offset), size)
     }
 
     fn pop(&mut self, dst: WritableReg, size: OperandSize) -> Result<()> {
@@ -355,29 +460,27 @@ impl Masm for MacroAssembler {
         Ok(SPOffset::from_u32(self.sp_offset))
     }
 
-    fn finalize(self, base: Option<SourceLoc>) -> Result<MachBufferFinalized<Final>> {
+    fn finalize(mut self, base: Option<SourceLoc>) -> Result<MachBufferFinalized<Final>> {
+        if let Some(patch) = self.stack_max_use_add {
+            patch.finalize(i32::try_from(self.sp_max).unwrap(), self.asm.buffer_mut());
+        }
+
         Ok(self.asm.finalize(base))
     }
 
     fn mov(&mut self, dst: WritableReg, src: RegImm, size: OperandSize) -> Result<()> {
         match (src, dst) {
-            (RegImm::Imm(v), rd) => {
-                let imm = match v {
-                    I::I32(v) => v as u64,
-                    I::I64(v) => v,
-                    I::F32(v) => v as u64,
-                    I::F64(v) => v,
-                    I::V128(_) => bail!(CodeGenError::unsupported_imm()),
-                };
-
-                let scratch = regs::scratch();
-                self.asm.load_constant(imm, writable!(scratch));
-                match rd.to_reg().class() {
-                    RegClass::Int => Ok(self.asm.mov_rr(scratch, rd, size)),
-                    RegClass::Float => Ok(self.asm.mov_to_fpu(scratch, rd, size)),
-                    _ => bail!(CodeGenError::invalid_operand_combination()),
+            (RegImm::Imm(v), _) => match v {
+                I::I32(_) | I::I64(_) => {
+                    self.asm.mov_ir(dst, v, v.size());
+                    Ok(())
                 }
-            }
+                imm @ (I::F32(_) | I::F64(_)) => {
+                    self.asm.mov_ir(dst, imm, imm.size());
+                    Ok(())
+                }
+                I::V128(_) => bail!(CodeGenError::unsupported_imm()),
+            },
             (RegImm::Reg(rs), rd) => match (rs.class(), rd.to_reg().class()) {
                 (RegClass::Int, RegClass::Int) => Ok(self.asm.mov_rr(rs, rd, size)),
                 (RegClass::Float, RegClass::Float) => Ok(self.asm.fmov_rr(rs, rd, size)),
@@ -392,24 +495,23 @@ impl Masm for MacroAssembler {
         dst: WritableReg,
         src: Reg,
         cc: IntCmpKind,
-        _size: OperandSize,
+        size: OperandSize,
     ) -> Result<()> {
-        self.asm.csel(src, src, dst, Cond::from(cc));
+        match (src.class(), dst.to_reg().class()) {
+            (RegClass::Int, RegClass::Int) => self.asm.csel(src, dst.to_reg(), dst, Cond::from(cc)),
+            (RegClass::Float, RegClass::Float) => {
+                self.asm
+                    .fpu_csel(src, dst.to_reg(), dst, Cond::from(cc), size)
+            }
+            _ => return Err(anyhow!(CodeGenError::invalid_operand_combination())),
+        }
+
         Ok(())
     }
 
     fn add(&mut self, dst: WritableReg, lhs: Reg, rhs: RegImm, size: OperandSize) -> Result<()> {
         match (rhs, lhs, dst) {
-            (RegImm::Imm(v), rn, rd) => {
-                let imm = match v {
-                    I::I32(v) => v as u64,
-                    I::I64(v) => v,
-                    _ => bail!(CodeGenError::unsupported_imm()),
-                };
-
-                self.asm.add_ir(imm, rn, rd, size);
-                Ok(())
-            }
+            (RegImm::Imm(v), rn, rd) => self.add_ir(rd, rn, v, size),
 
             (RegImm::Reg(rm), rn, rd) => {
                 self.asm.add_rrr(rm, rn, rd, size);
@@ -430,7 +532,27 @@ impl Masm for MacroAssembler {
         // ensure that the real SP is 16-byte aligned in case control flow is
         // transferred to a signal handler.
         self.with_aligned_sp(|masm| {
-            masm.add(dst, lhs, rhs, size)?;
+            match (rhs, lhs, dst) {
+                // NB: we don't use `Self::add_ir` since we explicitly
+                // want to emit the add variant which sets overflow
+                // flags.
+                (RegImm::Imm(i), rn, rd) => {
+                    let imm = i.unwrap_as_u64();
+                    match Imm12::maybe_from_u64(imm) {
+                        Some(imm12) => masm.asm.adds_ir(imm12, rn, rd, size),
+                        None => {
+                            masm.with_scratch::<IntScratch, _>(|masm, scratch| {
+                                masm.asm.mov_ir(scratch.writable(), i, i.size());
+                                masm.asm.adds_rrr(scratch.inner(), rn, rd, size);
+                            });
+                        }
+                    }
+                }
+
+                (RegImm::Reg(rm), rn, rd) => {
+                    masm.asm.adds_rrr(rm, rn, rd, size);
+                }
+            }
             masm.asm.trapif(Cond::Hs, trap);
             Ok(())
         })
@@ -439,13 +561,17 @@ impl Masm for MacroAssembler {
     fn sub(&mut self, dst: WritableReg, lhs: Reg, rhs: RegImm, size: OperandSize) -> Result<()> {
         match (rhs, lhs, dst) {
             (RegImm::Imm(v), rn, rd) => {
-                let imm = match v {
-                    I::I32(v) => v as u64,
-                    I::I64(v) => v,
-                    _ => bail!(CodeGenError::unsupported_imm()),
+                let imm = v.unwrap_as_u64();
+                match Imm12::maybe_from_u64(imm) {
+                    Some(imm12) => self.asm.sub_ir(imm12, rn, rd, size),
+                    None => {
+                        self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                            masm.asm.mov_ir(scratch.writable(), v, v.size());
+                            masm.asm.sub_rrr(scratch.inner(), rn, rd, size);
+                        });
+                    }
                 };
 
-                self.asm.sub_ir(imm, rn, rd, size);
                 Ok(())
             }
 
@@ -458,16 +584,11 @@ impl Masm for MacroAssembler {
 
     fn mul(&mut self, dst: WritableReg, lhs: Reg, rhs: RegImm, size: OperandSize) -> Result<()> {
         match (rhs, lhs, dst) {
-            (RegImm::Imm(v), rn, rd) => {
-                let imm = match v {
-                    I::I32(v) => v as u64,
-                    I::I64(v) => v,
-                    _ => bail!(CodeGenError::unsupported_imm()),
-                };
-
-                self.asm.mul_ir(imm, rn, rd, size);
+            (RegImm::Imm(v), rn, rd) => self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                masm.asm.mov_ir(scratch.writable(), v, v.size());
+                masm.asm.mul_rrr(scratch.inner(), rn, rd, size);
                 Ok(())
-            }
+            }),
 
             (RegImm::Reg(rm), rn, rd) => {
                 self.asm.mul_rrr(rm, rn, rd, size);
@@ -558,13 +679,19 @@ impl Masm for MacroAssembler {
     fn and(&mut self, dst: WritableReg, lhs: Reg, rhs: RegImm, size: OperandSize) -> Result<()> {
         match (rhs, lhs, dst) {
             (RegImm::Imm(v), rn, rd) => {
-                let imm = match v {
-                    I::I32(v) => v as u64,
-                    I::I64(v) => v,
-                    _ => bail!(CodeGenError::unsupported_imm()),
+                let imm = v.unwrap_as_u64();
+                let csize: inst::OperandSize = size.into();
+
+                match ImmLogic::maybe_from_u64(imm, csize.to_ty()) {
+                    Some(imml) => self.asm.and_ir(imml, rn, rd, size),
+                    None => {
+                        self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                            masm.asm.mov_ir(scratch.writable(), v, v.size());
+                            masm.asm.and_rrr(scratch.inner(), rn, rd, size);
+                        });
+                    }
                 };
 
-                self.asm.and_ir(imm, rn, rd, size);
                 Ok(())
             }
 
@@ -578,13 +705,19 @@ impl Masm for MacroAssembler {
     fn or(&mut self, dst: WritableReg, lhs: Reg, rhs: RegImm, size: OperandSize) -> Result<()> {
         match (rhs, lhs, dst) {
             (RegImm::Imm(v), rn, rd) => {
-                let imm = match v {
-                    I::I32(v) => v as u64,
-                    I::I64(v) => v,
-                    _ => bail!(CodeGenError::unsupported_imm()),
+                let imm = v.unwrap_as_u64();
+                let csize: inst::OperandSize = size.into();
+
+                match ImmLogic::maybe_from_u64(imm, csize.to_ty()) {
+                    Some(imml) => self.asm.or_ir(imml, rn, rd, size),
+                    None => {
+                        self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                            masm.asm.mov_ir(scratch.writable(), v, v.size());
+                            masm.asm.or_rrr(scratch.inner(), rn, rd, size);
+                        });
+                    }
                 };
 
-                self.asm.or_ir(imm, rn, rd, size);
                 Ok(())
             }
 
@@ -598,13 +731,18 @@ impl Masm for MacroAssembler {
     fn xor(&mut self, dst: WritableReg, lhs: Reg, rhs: RegImm, size: OperandSize) -> Result<()> {
         match (rhs, lhs, dst) {
             (RegImm::Imm(v), rn, rd) => {
-                let imm = match v {
-                    I::I32(v) => v as u64,
-                    I::I64(v) => v,
-                    _ => bail!(CodeGenError::unsupported_imm()),
-                };
+                let imm = v.unwrap_as_u64();
+                let csize: inst::OperandSize = size.into();
 
-                self.asm.xor_ir(imm, rn, rd, size);
+                match ImmLogic::maybe_from_u64(imm, csize.to_ty()) {
+                    Some(imml) => self.asm.xor_ir(imml, rn, rd, size),
+                    None => {
+                        self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                            masm.asm.mov_ir(scratch.writable(), v, v.size());
+                            masm.asm.xor_rrr(scratch.inner(), rn, rd, size);
+                        });
+                    }
+                };
                 Ok(())
             }
 
@@ -618,12 +756,20 @@ impl Masm for MacroAssembler {
     fn shift_ir(
         &mut self,
         dst: WritableReg,
-        imm: u64,
+        imm: I,
         lhs: Reg,
         kind: ShiftKind,
         size: OperandSize,
     ) -> Result<()> {
-        self.asm.shift_ir(imm, lhs, dst, kind, size);
+        match ImmShift::maybe_from_u64(imm.unwrap_as_u64()) {
+            Some(imml) => self.asm.shift_ir(imml, lhs, dst, kind, size),
+            None => {
+                self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                    masm.asm.mov_ir(scratch.writable(), imm, imm.size());
+                    masm.asm.shift_rrr(scratch.inner(), lhs, dst, kind, size);
+                });
+            }
+        };
         Ok(())
     }
 
@@ -652,8 +798,11 @@ impl Masm for MacroAssembler {
         size: OperandSize,
     ) -> Result<()> {
         context.binop(self, size, |this, dividend, divisor, size| {
-            this.asm
-                .div_rrr(divisor, dividend, writable!(dividend), kind, size);
+            this.with_aligned_sp(|this| {
+                this.asm
+                    .div_rrr(divisor, dividend, writable!(dividend), kind, size);
+                Ok(())
+            })?;
             match size {
                 OperandSize::S32 => Ok(TypedReg::new(WasmValType::I32, dividend)),
                 OperandSize::S64 => Ok(TypedReg::new(WasmValType::I64, dividend)),
@@ -669,8 +818,19 @@ impl Masm for MacroAssembler {
         size: OperandSize,
     ) -> Result<()> {
         context.binop(self, size, |this, dividend, divisor, size| {
-            this.asm
-                .rem_rrr(divisor, dividend, writable!(dividend), kind, size);
+            this.with_aligned_sp(|this| {
+                this.with_scratch::<IntScratch, _>(|masm, scratch| {
+                    masm.asm.rem_rrr(
+                        divisor,
+                        dividend,
+                        writable!(dividend),
+                        scratch.writable(),
+                        kind,
+                        size,
+                    );
+                });
+                Ok(())
+            })?;
             match size {
                 OperandSize::S32 => Ok(TypedReg::new(WasmValType::I32, dividend)),
                 OperandSize::S64 => Ok(TypedReg::new(WasmValType::I64, dividend)),
@@ -680,18 +840,20 @@ impl Masm for MacroAssembler {
     }
 
     fn zero(&mut self, reg: WritableReg) -> Result<()> {
-        self.asm.load_constant(0, reg);
+        self.asm.mov_ir(reg, I::i64(0), OperandSize::S64);
         Ok(())
     }
 
     fn popcnt(&mut self, context: &mut CodeGenContext<Emission>, size: OperandSize) -> Result<()> {
         let src = context.pop_to_reg(self, None)?;
-        let tmp = regs::float_scratch();
-        self.asm.mov_to_fpu(src.into(), writable!(tmp), size);
-        self.asm.cnt(writable!(tmp));
-        self.asm.addv(tmp, writable!(tmp), VectorSize::Size8x8);
-        self.asm
-            .mov_from_vec(tmp, writable!(src.into()), 0, OperandSize::S8);
+        self.with_scratch::<FloatScratch, _>(|masm, tmp| {
+            masm.asm.mov_to_fpu(src.into(), tmp.writable(), size);
+            masm.asm.cnt(tmp.writable());
+            masm.asm
+                .addv(tmp.inner(), tmp.writable(), VectorSize::Size8x8);
+            masm.asm
+                .mov_from_vec(tmp.inner(), writable!(src.into()), 0, OperandSize::S8);
+        });
         context.stack.push(src.into());
         Ok(())
     }
@@ -704,10 +866,13 @@ impl Masm for MacroAssembler {
         dst_size: OperandSize,
         kind: TruncKind,
     ) -> Result<()> {
-        self.asm
-            .fpu_to_int(dst, src, src_size, dst_size, kind, true);
-
-        Ok(())
+        self.with_aligned_sp(|masm| {
+            masm.with_scratch::<FloatScratch, _>(|masm, scratch| {
+                masm.asm
+                    .fpu_to_int(dst, src, scratch.writable(), src_size, dst_size, kind, true);
+            });
+            Ok(())
+        })
     }
 
     fn unsigned_truncate(
@@ -724,10 +889,20 @@ impl Masm for MacroAssembler {
         };
 
         ctx.convert_op(self, dst_ty, |masm, dst, src, dst_size| {
-            masm.asm
-                .fpu_to_int(writable!(dst), src, src_size, dst_size, kind, false);
-
-            Ok(())
+            masm.with_aligned_sp(|masm| {
+                masm.with_scratch::<FloatScratch, _>(|masm, scratch| {
+                    masm.asm.fpu_to_int(
+                        writable!(dst),
+                        src,
+                        scratch.writable(),
+                        src_size,
+                        dst_size,
+                        kind,
+                        false,
+                    );
+                    Ok(())
+                })
+            })
         })
     }
 
@@ -776,13 +951,13 @@ impl Masm for MacroAssembler {
 
     fn demote(&mut self, dst: WritableReg, src: Reg) -> Result<()> {
         self.asm
-            .cvt_float_to_float(src.into(), dst, OperandSize::S64, OperandSize::S32);
+            .cvt_float_to_float(src, dst, OperandSize::S64, OperandSize::S32);
         Ok(())
     }
 
     fn promote(&mut self, dst: WritableReg, src: Reg) -> Result<()> {
         self.asm
-            .cvt_float_to_float(src.into(), dst, OperandSize::S32, OperandSize::S64);
+            .cvt_float_to_float(src, dst, OperandSize::S32, OperandSize::S64);
         Ok(())
     }
 
@@ -820,12 +995,16 @@ impl Masm for MacroAssembler {
                 Ok(())
             }
             RegImm::Imm(v) => {
-                let imm = match v {
-                    I::I32(v) => v as u64,
-                    I::I64(v) => v,
-                    _ => bail!(CodeGenError::unsupported_imm()),
+                let val = v.unwrap_as_u64();
+                match Imm12::maybe_from_u64(val) {
+                    Some(imm12) => self.asm.subs_ir(imm12, src1, size),
+                    None => {
+                        self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                            masm.asm.mov_ir(scratch.writable(), v, v.size());
+                            masm.asm.subs_rrr(scratch.inner(), src1, size);
+                        });
+                    }
                 };
-                self.asm.subs_ir(imm, src1, size);
                 Ok(())
             }
         }
@@ -850,10 +1029,11 @@ impl Masm for MacroAssembler {
     }
 
     fn ctz(&mut self, dst: WritableReg, src: Reg, size: OperandSize) -> Result<()> {
-        let scratch = regs::scratch();
-        self.asm.rbit(src, writable!(scratch), size);
-        self.asm.clz(scratch, dst, size);
-        Ok(())
+        self.with_scratch::<IntScratch, _>(|masm, scratch| {
+            masm.asm.rbit(src, scratch.writable(), size);
+            masm.asm.clz(scratch.inner(), dst, size);
+            Ok(())
+        })
     }
 
     fn wrap(&mut self, dst: WritableReg, src: Reg) -> Result<()> {
@@ -889,7 +1069,7 @@ impl Masm for MacroAssembler {
         match &(lhs, rhs) {
             (rlhs, RegImm::Reg(rrhs)) => {
                 // If the comparison kind is zero or not zero and both operands
-                // are the same register, emit a ands instruction. Else we emit
+                // are the same register, emit an ands instruction. Else we emit
                 // a normal comparison.
                 if (kind == Eq || kind == Ne) && (rlhs == rrhs) {
                     self.asm.ands_rr(*rlhs, *rrhs, size);
@@ -909,21 +1089,34 @@ impl Masm for MacroAssembler {
     }
 
     fn unreachable(&mut self) -> Result<()> {
-        self.asm.udf(wasmtime_cranelift::TRAP_UNREACHABLE);
-        Ok(())
+        self.with_aligned_sp(|masm| {
+            masm.asm.udf(wasmtime_cranelift::TRAP_UNREACHABLE);
+            Ok(())
+        })
     }
 
     fn jmp_table(&mut self, targets: &[MachLabel], index: Reg, tmp: Reg) -> Result<()> {
         // At least one default target.
         debug_assert!(targets.len() >= 1);
-        let max = targets.len() as u64 - 1;
-        self.asm.subs_ir(max, index, OperandSize::S64);
-        let default_index = max as usize;
+        let default_index = targets.len() - 1;
+        let max = default_index;
+        self.asm.mov_ir(
+            writable!(tmp),
+            I::i32(i32::try_from(max).unwrap()),
+            OperandSize::S32,
+        );
+        // NB: We only emit the comparison instruction, since
+        // `Assembler::jmp_table` (and the underlying Cranelift
+        // instruction) will emit spectre mitigation and bounds
+        // checks.
+        self.asm.subs_rrr(tmp, index, OperandSize::S32);
         let default = targets[default_index];
-        let rest = &targets[..default_index];
-        let tmp1 = regs::scratch();
-        self.asm.jmp_table(rest, default, index, tmp1, tmp);
-        Ok(())
+        let rest = &targets[0..default_index];
+        self.with_scratch::<IntScratch, _>(|masm, scratch| {
+            masm.asm
+                .jmp_table(rest, default, index, scratch.inner(), tmp);
+            Ok(())
+        })
     }
 
     fn trap(&mut self, code: TrapCode) -> Result<()> {
@@ -1350,6 +1543,11 @@ impl Masm for MacroAssembler {
 impl MacroAssembler {
     fn increment_sp(&mut self, bytes: u32) {
         self.sp_offset += bytes;
+
+        // NOTE: we use `max` here to track the largest stack allocation in `sp_max`. Once we have
+        // seen the entire function, this value will represent the maximum size for the stack
+        // frame.
+        self.sp_max = self.sp_max.max(self.sp_offset);
     }
 
     fn decrement_sp(&mut self, bytes: u32) {
@@ -1366,6 +1564,21 @@ impl MacroAssembler {
         self.asm.mov_rr(sp, writable!(shadow_sp), OperandSize::S64);
     }
 
+    /// Heloper to add an immediate to a register.
+    fn add_ir(&mut self, dst: WritableReg, lhs: Reg, rhs: I, size: OperandSize) -> Result<()> {
+        let imm = rhs.unwrap_as_u64();
+        match Imm12::maybe_from_u64(imm) {
+            Some(imm12) => self.asm.add_ir(imm12, lhs, dst, size),
+            None => {
+                self.with_scratch::<IntScratch, _>(|masm, scratch| {
+                    masm.asm.mov_ir(scratch.writable(), rhs, rhs.size());
+                    masm.asm.add_rrr(scratch.inner(), lhs, dst, size);
+                });
+            }
+        };
+        Ok(())
+    }
+
     // Copies the value of the shadow stack pointer to the stack pointer: mov
     // sp, x28.
     //
@@ -1383,6 +1596,7 @@ impl MacroAssembler {
     fn move_shadow_sp_to_sp(&mut self) {
         let shadow_sp = regs::shadow_sp();
         let sp = writable!(regs::sp());
-        self.asm.add_ir(0, shadow_sp, sp, OperandSize::S64);
+        let imm = Imm12::maybe_from_u64(0).unwrap();
+        self.asm.add_ir(imm, shadow_sp, sp, OperandSize::S64);
     }
 }
